/**********************************************************************************************************************************/
/* Project      : TAPER                                                                                                           */
/* Module       : TPBASIC.C                                                                                                       */
/* Description  : Convert ASCII to BASIC                                                                                          */
/* Version type : Program module                                                                                                  */
/* Last changed : 24-06-1999  7:00                                                                                                */
/* Update count : 5                                                                                                               */
/* OS type      : PC (DOS)                                                                                                        */
/* Compiler     : Watcom C32 Optimizing Compiler Version 10.x                                                                     */
/* Linker       : Watcom C Linker                                                                                                 */
/*                                                                                                                                */
/*                Copyleft (C) 1997-2001 ThunderWare Research Center, written by Martijn van der Heide.                           */
/*                                                                                                                                */
/*                This program is free software; you can redistribute it and/or                                                   */
/*                modify it under the terms of the GNU General Public License                                                     */
/*                as published by the Free Software Foundation; either version 2                                                  */
/*                of the License, or (at your option) any later version.                                                          */
/*                                                                                                                                */
/*                This program is distributed in the hope that it will be useful,                                                 */
/*                but WITHOUT ANY WARRANTY; without even the implied warranty of                                                  */
/*                MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                                                   */
/*                GNU General Public License for more details.                                                                    */
/*                                                                                                                                */
/*                You should have received a copy of the GNU General Public License                                               */
/*                along with this program; if not, write to the Free Software                                                     */
/*                Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                                     */
/*                                                                                                                                */
/**********************************************************************************************************************************/

#define __TPBASIC_MAIN__

#include "wbench32/wbench32.h"
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <fcntl.h>
#include <io.h>
#include "tpconfig.h"
#include "tptape.h"
#include "tpfile.h"
#include "tpaudio.h"
#include "tpdisass.h"
#include "tpcust.h"
#include "tpedit.h"
#include "tpscreen.h"
#include "tpmain.h"
#include "tpbasic.h"

#pragma pack          ()

/**********************************************************************************************************************************/
/* Define the global variables                                                                                                    */
/**********************************************************************************************************************************/

static char          _TempFileName[MAXDIR + 1];
static char          _Path[MAXDIR + 1]  = "\0";
static char          _File[13]          = "\0";
static char          _Mask[13]          = "*.BAS";

struct BTokenMap_s
{
  char *Token;
  byte  TokenType;
        /* Type   0 = No special meaning                                                   */
        /* Type   1 = Always keyword                                                       */
        /* Type   2 = Can be both keyword and non-keyword (colour parameters)              */
        /* Type   3 = Numeric expression token                                             */
        /* Type   4 = String expression token                                              */
        /* Type   5 = May only appear in (L)PRINT statements (AT and TAB)                  */
        /* Type   6 = Type-less (normal ASCII or expression token)                         */
  byte  KeywordClass[8];                                     /* The class this keyword belongs to, as defined in the Spectrum ROM */
        /* This table is used by expression tokens as well. Class 12 was added for this purpose */
        /* Class  0 = No further operands                                                  */
        /* Class  1 = Used in LET. A variable is required                                  */
        /* Class  2 = Used in LET. An expression, numeric or string, must follow           */
        /* Class  3 = A numeric expression may follow. Zero to be used in case of default  */
        /* Class  4 = A single character variable must follow                              */
        /* Class  5 = A set of items may be given                                          */
        /* Class  6 = A numeric expression must follow                                     */
        /* Class  7 = Handles colour items                                                 */
        /* Class  8 = Two numeric expressions, separated by a comma, must follow           */
        /* Class  9 = As for class 8 but colour items may precede the expression           */
        /* Class 10 = A string expression must follow                                      */
        /* Class 11 = Handles cassette routines                                            */
        /* The following classes are not available in the ROM but were needed              */
        /* Class 12 = One or more string expressions, separated by commas, must follow     */
        /* Class 13 = One or more expressions, separated by commas, must follow            */
        /* Class 14 = One or more variables, separated by commas, must follow (READ)       */
        /* Class 15 = DEF FN                                                               */
} BTokenMap[256] = {

   /* Everything below ASCII 32 */
   {NULL, 6, { 0 }}, {NULL, 6, { 0 }}, {NULL, 6, { 0 }}, {NULL, 6, { 0 }}, {NULL, 6, { 0 }},
   {NULL, 6, { 0 }},                                                                                                   /* Print ' */
   {NULL, 6, { 0 }}, {NULL, 6, { 0 }}, {NULL, 6, { 0 }}, {NULL, 6, { 0 }}, {NULL, 6, { 0 }}, {NULL, 6, { 0 }}, {NULL, 6, { 0 }},
   {"(eoln)", 6, { 0 }},                                                                                                    /* CR */
   {NULL, 6, { 0 }},                                                                                                    /* Number */
   {NULL, 6, { 0 }},
   {NULL, 6, { 0 }},                                                                                                       /* INK */
   {NULL, 6, { 0 }},                                                                                                     /* PAPER */
   {NULL, 6, { 0 }},                                                                                                     /* FLASH */
   {NULL, 6, { 0 }},                                                                                                    /* BRIGHT */
   {NULL, 6, { 0 }},                                                                                                   /* INVERSE */
   {NULL, 6, { 0 }},                                                                                                      /* OVER */
   {NULL, 6, { 0 }},                                                                                                        /* AT */
   {NULL, 6, { 0 }},                                                                                                       /* TAB */
   {NULL, 6, { 0 }}, {NULL, 6, { 0 }}, {NULL, 6, { 0 }}, {NULL, 6, { 0 }}, {NULL, 6, { 0 }}, {NULL, 6, { 0 }}, {NULL, 6, { 0 }},
   {NULL, 6, { 0 }},

   /* Normal ASCII set */
   {"\x20", 6, { 0 }}, {"\x21", 6, { 0 }}, {"\x22", 6, { 0 }}, {"\x23", 6, { 0 }}, {"\x24", 6, { 0 }}, {"\x25", 6, { 0 }},
   {"\x26", 6, { 0 }}, {"\x27", 6, { 0 }}, {"\x28", 6, { 0 }}, {"\x29", 6, { 0 }}, {"\x2A", 6, { 0 }}, {"\x2B", 6, { 0 }},
   {"\x2C", 6, { 0 }}, {"\x2D", 6, { 0 }}, {"\x2E", 6, { 0 }}, {"\x2F", 6, { 0 }}, {"\x30", 6, { 0 }}, {"\x31", 6, { 0 }},
   {"\x32", 6, { 0 }}, {"\x33", 6, { 0 }}, {"\x34", 6, { 0 }}, {"\x35", 6, { 0 }}, {"\x36", 6, { 0 }}, {"\x37", 6, { 0 }},
   {"\x38", 6, { 0 }}, {"\x39", 6, { 0 }}, {"\x3A", 2, { 0 }}, {"\x3B", 6, { 0 }}, {"\x3C", 6, { 0 }}, {"\x3D", 6, { 0 }},
   {"\x3E", 6, { 0 }}, {"\x3F", 6, { 0 }}, {"\x40", 6, { 0 }}, {"\x41", 6, { 0 }}, {"\x42", 6, { 0 }}, {"\x43", 6, { 0 }},
   {"\x44", 6, { 0 }}, {"\x45", 6, { 0 }}, {"\x46", 6, { 0 }}, {"\x47", 6, { 0 }}, {"\x48", 6, { 0 }}, {"\x49", 6, { 0 }},
   {"\x4A", 6, { 0 }}, {"\x4B", 6, { 0 }}, {"\x4C", 6, { 0 }}, {"\x4D", 6, { 0 }}, {"\x4E", 6, { 0 }}, {"\x4F", 6, { 0 }},
   {"\x50", 6, { 0 }}, {"\x51", 6, { 0 }}, {"\x52", 6, { 0 }}, {"\x53", 6, { 0 }}, {"\x54", 6, { 0 }}, {"\x55", 6, { 0 }},
   {"\x56", 6, { 0 }}, {"\x57", 6, { 0 }}, {"\x58", 6, { 0 }}, {"\x59", 6, { 0 }}, {"\x5A", 6, { 0 }}, {"\x5B", 6, { 0 }},
   {"\x5C", 6, { 0 }}, {"\x5D", 6, { 0 }}, {"\x5E", 6, { 0 }}, {"\x5F", 6, { 0 }}, {"\x60", 6, { 0 }}, {"\x61", 6, { 0 }},
   {"\x62", 6, { 0 }}, {"\x63", 6, { 0 }}, {"\x64", 6, { 0 }}, {"\x65", 6, { 0 }}, {"\x66", 6, { 0 }}, {"\x67", 6, { 0 }},
   {"\x68", 6, { 0 }}, {"\x69", 6, { 0 }}, {"\x6A", 6, { 0 }}, {"\x6B", 6, { 0 }}, {"\x6C", 6, { 0 }}, {"\x6D", 6, { 0 }},
   {"\x6E", 6, { 0 }}, {"\x6F", 6, { 0 }}, {"\x70", 6, { 0 }}, {"\x71", 6, { 0 }}, {"\x72", 6, { 0 }}, {"\x73", 6, { 0 }},
   {"\x74", 6, { 0 }}, {"\x75", 6, { 0 }}, {"\x76", 6, { 0 }}, {"\x77", 6, { 0 }}, {"\x78", 6, { 0 }}, {"\x79", 6, { 0 }},
   {"\x7A", 6, { 0 }}, {"\x7B", 6, { 0 }}, {"\x7C", 6, { 0 }}, {"\x7D", 6, { 0 }}, {"\x7E", 6, { 0 }}, {"\x7F", 6, { 0 }},

   /* Block graphics without shift */
   {"\x80", 6, { 0 }}, {"\x81", 6, { 0 }}, {"\x82", 6, { 0 }}, {"\x83", 6, { 0 }}, {"\x84", 6, { 0 }}, {"\x85", 6, { 0 }},
   {"\x86", 6, { 0 }}, {"\x87", 6, { 0 }},

   /* Block graphics with shift */
   {"\x88", 6, { 0 }}, {"\x89", 6, { 0 }}, {"\x8A", 6, { 0 }}, {"\x8B", 6, { 0 }}, {"\x8C", 6, { 0 }}, {"\x8D", 6, { 0 }},
   {"\x8E", 6, { 0 }}, {"\x8F", 6, { 0 }},

   /* UDGs */
   {"\x90", 6, { 0 }}, {"\x91", 6, { 0 }}, {"\x92", 6, { 0 }}, {"\x93", 6, { 0 }}, {"\x94", 6, { 0 }}, {"\x95", 6, { 0 }},
   {"\x96", 6, { 0 }}, {"\x97", 6, { 0 }}, {"\x98", 6, { 0 }}, {"\x99", 6, { 0 }}, {"\x9A", 6, { 0 }}, {"\x9B", 6, { 0 }},
   {"\x9C", 6, { 0 }}, {"\x9D", 6, { 0 }}, {"\x9E", 6, { 0 }}, {"\x9F", 6, { 0 }}, {"\xA0", 6, { 0 }}, {"\xA1", 6, { 0 }},
   {"\xA2", 6, { 0 }},

   {"SPECTRUM",  1, { 0 }},                                                                                   /* For Spectrum 128 */
   {"PLAY",      1, { 12 }},

   /* BASIC tokens - expression */
   {"RND",       3, { 0 }},
   {"INKEY$",    4, { 0 }},
   {"PI",        3, { 0 }},
   {"FN",        3, { 1, '(', 13, ')', 0 }},
   {"POINT",     3, { '(', 8, ')', 0 }},
   {"SCREEN$",   4, { '(', 8, ')', 0 }},
   {"ATTR",      3, { '(', 8, ')', 0 }},
   {"AT",        5, { 8, 0 }},
   {"TAB",       5, { 6, 0 }},
   {"VAL$",      4, { 10, 0 }},
   {"CODE",      3, { 10, 0 }},
   {"VAL",       3, { 10, 0 }},
   {"LEN",       3, { 10, 0 }},
   {"SIN",       3, { 6, 0 }},
   {"COS",       3, { 6, 0 }},
   {"TAN",       3, { 6, 0 }},
   {"ASN",       3, { 6, 0 }},
   {"ACS",       3, { 6, 0 }},
   {"ATN",       3, { 6, 0 }},
   {"LN",        3, { 6, 0 }},
   {"EXP",       3, { 6, 0 }},
   {"INT",       3, { 6, 0 }},
   {"SQR",       3, { 6, 0 }},
   {"SGN",       3, { 6, 0 }},
   {"ABS",       3, { 6, 0 }},
   {"PEEK",      3, { 6, 0 }},
   {"IN",        3, { 6, 0 }},
   {"USR",       3, { 6, 0 }},
   {"STR$",      4, { 6, 0 }},
   {"CHR$",      4, { 6, 0 }},
   {"NOT",       3, { 6, 0 }},
   {"BIN",       3, { 3, 0 }},
   {"OR",        6, { 5, 0 }},   /*  -\                                                  */
   {"AND",       6, { 5, 0 }},   /*   |                                                  */
   {"<=",        6, { 5, 0 }},   /*   | These are handled directly within ScanExpression */
   {">=",        6, { 5, 0 }},   /*   |                                                  */
   {"<>",        6, { 5, 0 }},   /*  -/                                                  */
   {"LINE",      6, { 0 }},
   {"THEN",      6, { 0 }},
   {"TO",        6, { 0 }},
   {"STEP",      6, { 0 }},

   /* BASIC tokens - keywords */
   {"DEF FN",    1, { 15, 0 }},
   {"CAT",       1, { 11, 0 }},
   {"FORMAT",    1, { 11, 0 }},
   {"MOVE",      1, { 11, 0 }},
   {"ERASE",     1, { 11, 0 }},
   {"OPEN #",    1, { 11, 0 }},
   {"CLOSE #",   1, { 11, 0 }},
   {"MERGE",     1, { 11, 0 }},
   {"VERIFY",    1, { 11, 0 }},
   {"BEEP",      1, { 8, 0 }},
   {"CIRCLE",    1, { 9, ',', 6, 0 }},
   {"INK",       2, { 7, 0 }},
   {"PAPER",     2, { 7, 0 }},
   {"FLASH",     2, { 7, 0 }},
   {"BRIGHT",    2, { 7, 0 }},
   {"INVERSE",   2, { 7, 0 }},
   {"OVER",      2, { 7, 0 }},
   {"OUT",       1, { 8, 0 }},
   {"LPRINT",    1, { 5, 0 }},
   {"LLIST",     1, { 3, 0 }},
   {"STOP",      1, { 0 }},
   {"READ",      1, { 14, 0 }},
   {"DATA",      1, { 13, 0 }},
   {"RESTORE",   1, { 3, 0 }},
   {"NEW",       1, { 0 }},
   {"BORDER",    1, { 6, 0 }},
   {"CONTINUE",  1, { 0 }},
   {"DIM",       1, { 1, '(', 13, ')', 0 }},
   {"REM",       1, { 5, 0 }},                                                                 /* (Special: taken out separately) */
   {"FOR",       1, { 4, '=', 6, 0xCC, 6, 0xCD, 6, 0 }},                                /* (Special: STEP (0xCD) is not required) */
   {"GO TO",     1, { 6, 0 }},
   {"GO SUB",    1, { 6, 0 }},
   {"INPUT",     1, { 5, 0 }},
   {"LOAD",      1, { 11, 0 }},
   {"LIST",      1, { 3, 0 }},
   {"LET",       1, { 1, '=', 2, 0 }},
   {"PAUSE",     1, { 6, 0 }},
   {"NEXT",      1, { 4, 0 }},
   {"POKE",      1, { 8, 0 }},
   {"PRINT",     1, { 5, 0 }},
   {"PLOT",      1, { 9, 0 }},
   {"RUN",       1, { 3, 0 }},
   {"SAVE",      1, { 11, 0 }},
   {"RANDOMIZE", 1, { 3, 0 }},
   {"IF",        1, { 6, 0xCB, 0 }},
   {"CLS",       1, { 0 }},
   {"DRAW",      1, { 9, ',', 6, 0 }},
   {"CLEAR",     1, { 3, 0 }},
   {"RETURN",    1, { 0 }},
   {"COPY",      1, { 0 }}};

#define MAXLINELENGTH 1024

char ConvertedSpectrumLine[MAXLINELENGTH + 1];
byte ResultingLine[MAXLINELENGTH + 1];

int   Is48KProgram    = -1;                                                                                       /* -1 = unknown */
                                                                                                                  /*  1 = 48K     */
                                                                                                                  /*  0 = 128K    */
int   UsesInterface1  = -1;                                                           /* -1 = unknown                             */
                                                                                      /*  0 = either Interface1 or Opus Discovery */
                                                                                      /*  1 = Interface1                          */
                                                                                      /*  2 = Opus Discovery                      */
bool  CaseSensitive   = TRUE;
bool  DoCheckSyntax   = TRUE;
bool  TokenBracket    = FALSE;
char  BASICLineIn[MAXLINELENGTH + 1];                                                        /* One line read from the ASCII file */
int   PreviousBasicLineNo = -1;

struct ComboInfo  ConvertWindow;
struct ComboEntry ConvertLine;
short NumFatals       = 0;
short NumErrors       = 0;
short NumWarnings     = 0;

/**********************************************************************************************************************************/
/* Prototype all functions                                                                                                        */
/**********************************************************************************************************************************/

void InsertTextString  (void);
void ReportInfo        (char *Format, ...);
void ReportFatal       (char *Format, ...);
void ReportError       (char *Format, ...);
void ReportWarning     (char *Format, ...);
int  GetLineNumber     (char **FirstAfter);
int  MatchToken        (int BasicLineNo, bool WantKeyword, char **LineIndex, byte *Token);
int  HandleNumbers     (int BasicLineNo, char **BasicLine, byte **SpectrumLine);
int  ExpandSequences   (int BasicLineNo, char **BasicLine, byte **SpectrumLine, bool StripSpaces);
int  PrepareLine       (char *LineIn, int FileLineNo, char **FirstToken);
bool ScanVariable      (int BasicLineNo, int StatementNo, int Keyword, byte **Index, bool *Type, int *NameLen, int AllowSlicing);
bool SliceDirectString (int BasicLineNo, int StatementNo, int Keyword, byte **Index);
bool ScanStream        (int BasicLineNo, int StatementNo, int Keyword, byte **Index);
bool ScanChannel       (int BasicLineNo, int StatementNo, int Keyword, byte **Index, byte *WhichChannel);
bool SignalInterface1  (int BasicLineNo, int StatementNo, int NewMode);
bool CheckEnd          (int BasicLineNo, int StatementNo, byte **Index);
bool ScanExpression    (int BasicLineNo, int StatementNo, int Keyword, byte **Index, bool *Type, int Level);
bool HandleClass01     (int BasicLineNo, int StatementNo, int Keyword, byte **Index, bool *Type);
bool HandleClass02     (int BasicLineNo, int StatementNo, int Keyword, byte **Index, bool Type);
bool HandleClass03     (int BasicLineNo, int StatementNo, int Keyword, byte **Index);
bool HandleClass04     (int BasicLineNo, int StatementNo, int Keyword, byte **Index);
bool HandleClass05     (int BasicLineNo, int StatementNo, int Keyword, byte **Index);
bool HandleClass06     (int BasicLineNo, int StatementNo, int Keyword, byte **Index);
bool HandleClass07     (int BasicLineNo, int StatementNo, int Keyword, byte **Index);
bool HandleClass08     (int BasicLineNo, int StatementNo, int Keyword, byte **Index);
bool HandleClass09     (int BasicLineNo, int StatementNo, int Keyword, byte **Index);
bool HandleClass10     (int BasicLineNo, int StatementNo, int Keyword, byte **Index);
bool HandleClass11     (int BasicLineNo, int StatementNo, int Keyword, byte **Index);
bool HandleClass12     (int BasicLineNo, int StatementNo, int Keyword, byte **Index);
bool HandleClass13     (int BasicLineNo, int StatementNo, int Keyword, byte **Index);
bool HandleClass14     (int BasicLineNo, int StatementNo, int Keyword, byte **Index);
bool HandleClass15     (int BasicLineNo, int StatementNo, int Keyword, byte **Index);
bool CheckSyntax       (int BasicLineNo, byte *Line);

/**********************************************************************************************************************************/
/* Let's be lazy and define a very commonly used error message....                                                                */
/**********************************************************************************************************************************/

#define BADTOKEN(_Exp,_Got)        ReportError ("{BERROR in line %d, statement %d}B\n- Expected %s, but got \"%s\"", \
                                                BasicLineNo, StatementNo, _Exp, _Got)

/**********************************************************************************************************************************/
/* Start of the program                                                                                                           */
/**********************************************************************************************************************************/

void InsertTextString (void)
{
  FitComboString (&ConvertWindow, NULL, ConvertLine.Text, ConvertWindow.TextWidth, 0, 0, FALSE);
  if (ConvertWindow.NumEntries >= ConvertWindow.TextHeight)
    ConvertWindow.DFirstLine = ConvertWindow.NumEntries - ConvertWindow.TextHeight + 1;
  else
    ConvertWindow.DFirstLine = 0;
  ConvertWindow.VSlider->SliderStart = ConvertWindow.VSlider->StepRate * ConvertWindow.DFirstLine;
  PrintComboEntries (&ConvertWindow);
}

void ReportInfo (char *Format, ...)
{
  va_list Args;

  va_start (Args, Format);
  vsprintf (ConvertLine.Text, Format, Args);
  va_end (Args);
  InsertTextString ();
}

void ReportFatal (char *Format, ...)
{
  va_list Args;

  ConvertLine.Text[0] = '\0';
  InsertComboEntry (&ConvertWindow, &ConvertLine, NULL, -1);
  va_start (Args, Format);
  vsprintf (ConvertLine.Text, Format, Args);
  va_end (Args);
  InsertTextString ();
  NumFatals ++;
  NumErrors ++;
}

void ReportError (char *Format, ...)
{
  va_list Args;

  ConvertLine.Text[0] = '\0';
  InsertComboEntry (&ConvertWindow, &ConvertLine, NULL, -1);
  va_start (Args, Format);
  vsprintf (ConvertLine.Text, Format, Args);
  va_end (Args);
  FitComboString (&ConvertWindow, NULL, BASICLineIn, ConvertWindow.TextWidth, 0, 0, TRUE);        /* (First print the BASIC line) */
  InsertTextString ();
  if (++ NumErrors >= 50)
    ReportFatal ("{BFATAL}B - Too many errors - aborting");
}

void ReportWarning (char *Format, ...)
{
  va_list Args;

  va_start (Args, Format);
  vsprintf (ConvertLine.Text, Format, Args);
  va_end (Args);
  InsertTextString ();
  NumWarnings ++;
}

int GetLineNumber (char **FirstAfter)

/**********************************************************************************************************************************/
/* Pre   : The line must have been prepared into (global) `ConvertedSpectrumLine'.                                                */
/* Post  : The BASIC line number has been returned, or -1 if there was none.                                                      */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  int   LineNo     = 0;
  char *LineIndex;
  bool  SkipSpaces = TRUE;
  bool  Continue   = TRUE;

  LineIndex = ConvertedSpectrumLine;
  while (*LineIndex && Continue)
    if (*LineIndex == ' ')                                                                                 /* Skip leading spaces */
    {
      if (SkipSpaces)
        LineIndex ++;
      else
        Continue = FALSE;
    }
    else if (isdigit (*LineIndex))                                                                              /* Process number */
    {
      LineNo = LineNo * 10 + *(LineIndex ++) - '0';
      SkipSpaces = FALSE;
    }
    else
      Continue = FALSE;
  *FirstAfter = LineIndex;
  if (SkipSpaces)                                                                                          /* Nothing found yet ? */
    return (-1);
  else
    while ((**FirstAfter) == ' ')                                                                         /* Skip trailing spaces */
      (*FirstAfter) ++;
  return (LineNo);
}

int MatchToken (int BasicLineNo, bool WantKeyword, char **LineIndex, byte *Token)

/**********************************************************************************************************************************/
/* Pre   : `WantKeyword' is TRUE if we need in keyword match, `LineIndex' holds the position to match.                            */
/* Post  : If there was a match, the token value is returned in `Token' and `LineIndex' is pointing after the string plus any     */
/*         any trailing space.                                                                                                    */
/*         The return value is 0 for no match, -2 for an error, -1 for a match of the wrong type, 1 for a good match.             */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  int    Cnt;
  size_t Length;
  size_t LongestMatch = 0;
  bool   Match        = FALSE;
  bool   Match2;

  if ((**LineIndex) == ':')                                                                                  /* Special exception */
  {
    LongestMatch = 1;
    Match = TRUE;
    *Token = ':';
  }
  else for (Cnt = 0xA3 ; Cnt <= 0xFF ; Cnt ++)                                                 /* (Keywords start after the UDGs) */
  {
    Length = strlen (BTokenMap[Cnt].Token);
    if (CaseSensitive)
      Match2 = !strncmp (*LineIndex, BTokenMap[Cnt].Token, Length);
    else
      Match2 = !strnicmp (*LineIndex, BTokenMap[Cnt].Token, Length);
    if (Match2)
      if (Length > LongestMatch)
      {
        LongestMatch = Length;
        Match = TRUE;
        *Token = Cnt;
      }
  }
  if (!Match)
    return (0);                                                                                               /* Signal: no match */
  if (isalpha (*(*LineIndex + LongestMatch - 1)) && isalpha (*(*LineIndex + LongestMatch)))         /* Continueing alpha string ? */
    return (0);                                            /* Then there's no match after all! (eg. 'INT' must not match 'INTER') */
  *LineIndex += LongestMatch;                                                                                /* Go past the token */
  while ((**LineIndex) == ' ')                                                                            /* Skip trailing spaces */
    (*LineIndex) ++;
  if (*Token == 0xA3 || *Token == 0xA4)                                                                 /* 'SPECTRUM' or 'PLAY' ? */
    switch (Is48KProgram)                                                                        /* Then the program must be 128K */
    {
      case -1 : Is48KProgram = 0; break;                                                                          /* Set the flag */
      case  1 : ReportError ("{BERROR}B - Line %d contains a 128K keyword, but the program "
                             "also uses UDGs \'T\' and/or \'U\'", BasicLineNo);
                return (-2);
      case  0 : break;
    }
  if ((WantKeyword && BTokenMap[*Token].TokenType == 0) ||                               /* Wanted keyword but got something else */
      (!WantKeyword && BTokenMap[*Token].TokenType == 1))                       /* Did not want a keyword but got one nonetheless */
    return (-1);                                                                              /* Signal: match, but of wrong type */
  else
    return (1);                                                                                                 /* Signal: match! */
}

int HandleNumbers (int BasicLineNo, char **BasicLine, byte **SpectrumLine)

/**********************************************************************************************************************************/
/* Pre   : `BasicLineNo' holds the current BASIC line number, `BasicLine' points into the line, `SpectrumLine' points to the      */
/*         TAPped Spectrum line.                                                                                                  */
/* Post  : If there was a (floating point) number at this position, it has been processed into `SpectrumLine' and `LineIndex' is  */
/*         pointing after the number.                                                                                             */
/*         The return value is: 0 = no number, 1 = number done, -1 = number error (already reported).                             */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
#define SHIFT31BITS   (double)2147483648.0                                                                            /* (= 2^31) */
  char   *StartOfNumber;
  double  Value   = 0.0;
  double  Divider = 1.0;
  double  Exp     = 0.0;
  int     IntValue;
  byte    Sign    = 0x00;
  unsigned long Mantissa;

  if (!isdigit (**BasicLine) &&                                                             /* Current character is not a digit ? */
      (**BasicLine) != '.')                                                               /* And not a decimal point (eg. '.5') ? */
    return (0);                                                                                 /* Then it can hardly be a number */
  StartOfNumber = *BasicLine;
  while (isdigit (**BasicLine))                                                                    /* First read the integer part */
    Value = Value * 10 + *((*BasicLine) ++) - '0';
  if ((**BasicLine) == '.')                                                                                    /* Decimal point ? */
  {                                                                                                      /* Read the decimal part */
    (*BasicLine) ++;
    while (isdigit (**BasicLine))
      Value = Value + (Divider /= 10) * (*((*BasicLine) ++) - '0');
  }
  if ((**BasicLine) == 'e' || (**BasicLine) == 'E')                                                                 /* Exponent ? */
  {
    (*BasicLine) ++;
    if ((**BasicLine) == '+')                                                            /* Both "Ex" and "E+x" do the same thing */
      (*BasicLine) ++;
    else if ((**BasicLine) == '-')                                                                           /* Negative exponent */
    {
      Sign = 0xFF;
      (*BasicLine) ++;
    }
    while (isdigit (**BasicLine))                                                                      /* Read the exponent value */
      Exp = Exp * 10 + *((*BasicLine) ++) - '0';
    if (Sign == 0x00)                                                           /* Raise the resulting value to the read exponent */
      Value = Value * pow (10.0, Exp);
    else
      Value = Value / pow (10.0, Exp);
  }
  strncpy ((char *)*SpectrumLine, StartOfNumber, *BasicLine - StartOfNumber);                     /* Insert the ASCII value first */
  (*SpectrumLine) += (*BasicLine - StartOfNumber);
  IntValue = (int)floor (Value);
  if (Value == IntValue && Value >= -65536 && Value < 65536)                                                   /* Small integer ? */
  {
    *((*SpectrumLine) ++) = 0x0E;                                                                         /* Insert number marker */
    *((*SpectrumLine) ++) = 0x00;
    if (IntValue >= 0)                                                                                             /* Insert sign */
      *((*SpectrumLine) ++) = 0x00;
    else
    {
      *((*SpectrumLine) ++) = 0xFF;
      IntValue += 65536;                                          /* Maintain bug in Spectrum ROM - INT(-65536) will result in -1 */
    }
    *((*SpectrumLine) ++) = (byte)(IntValue & 0xFF);
    *((*SpectrumLine) ++) = (byte)(IntValue >> 8);
    *((*SpectrumLine) ++) = 0x00;
  }
  else                                                                             /* Need to store in full floating point format */
  {
    if (Value < 0)
    {
      Sign = 0x80;                                                                              /* Sign bit is high bit of byte 2 */
      Value = -Value;
    }
    else
      Sign = 0x00;
    Exp = floor (log2 (Value));
    if (Exp < -129 || Exp > 126)
    {
      ReportError ("{BERROR}B - Number too big in line %d", BasicLineNo);
      return (-1);
    }
    Mantissa = (unsigned long)floor ((Value / pow (2.0, Exp) - 1.0) * SHIFT31BITS + 0.5);                   /* Calculate mantissa */
    *((*SpectrumLine) ++) = 0x0E;                                                                         /* Insert number marker */
    *((*SpectrumLine) ++) = (byte)Exp + 0x81;                                                                  /* Insert exponent */
    *((*SpectrumLine) ++) = (byte)((Mantissa >> 24) & 0x7F) | Sign;                                            /* Insert mantissa */
    *((*SpectrumLine) ++) = (byte)((Mantissa >> 16) & 0xFF);                                                     /* (Big endian!) */
    *((*SpectrumLine) ++) = (byte)((Mantissa >> 8) & 0xFF);
    *((*SpectrumLine) ++) = (byte)(Mantissa & 0xFF);
  }
  return (1);
}

int ExpandSequences (int BasicLineNo, char **BasicLine, byte **SpectrumLine, bool StripSpaces)

/**********************************************************************************************************************************/
/* Pre   : `BasicLineNo' holds the current BASIC line number, `BasicLine' points into the line, `SpectrumLine' points to the      */
/*         TAPped Spectrum line.                                                                                                  */
/* Post  : If there was an expandable '{...}' sequence at this position, it has been processed into `SpectrumLine', `LineIndex'   */
/*         is pointing after the sequence. Returned is -1 for error, 0 for no expansion, 1 for expansion.                         */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  char *StartOfSequence;
  byte  Attribute       = 0;
  byte  AttributeLength = 0;
  byte  AttributeVal1   = 0;
  byte  AttributeVal2   = 0;
  byte  OldCharacter;
  int   Cnt;

  if (**BasicLine != '{')
    return (0);
  StartOfSequence = (*BasicLine) + 1;
  /* 'CODE' and 'CAT' were added for the sole purpuse of allowing them to be OPEN #'ed as channels! */
  if (!strnicmp (StartOfSequence, "CODE}", 5))                                                                 /* Special: 'CODE' */
  {
    *((*SpectrumLine) ++) = 0xAF;
    (*BasicLine) += 6;
    return (1);
  }
  if (!strnicmp (StartOfSequence, "CAT}", 4))                                                                   /* Special: 'CAT' */
  {
    *((*SpectrumLine) ++) = 0xCF;
    (*BasicLine) += 5;
    return (1);
  }
  if (!strnicmp (StartOfSequence, "(C)}", 4))
  {                                                                                             /* Form "{(C)}" -> copyright sign */
    *((*SpectrumLine) ++) = 0x7F;
    (*BasicLine) += 5;
    if (StripSpaces)
      while ((**BasicLine) == ' ')                                                                        /* Skip trailing spaces */
        (*BasicLine) ++;
    return (1);
  }
  if (*StartOfSequence == '+' && *(StartOfSequence + 1) >= '1' && *(StartOfSequence + 1) <= '8' && *(StartOfSequence + 2) == '}')
  {                                                                                   /* Form "{+X}" -> block graphics with shift */
    *((*SpectrumLine) ++) = 0x88 + (((*(StartOfSequence + 1) - '0') % 8) ^ 7);
    (*BasicLine) += 4;
    if (StripSpaces)
      while ((**BasicLine) == ' ')
        (*BasicLine) ++;
    return (1);
  }
  if (*StartOfSequence == '-' && *(StartOfSequence + 1) >= '1' && *(StartOfSequence + 1) <= '8' && *(StartOfSequence + 2) == '}')
  {                                                                                /* Form "{-X}" -> block graphics without shift */
    *((*SpectrumLine) ++) = 0x80 + (*(StartOfSequence + 1) - '0') % 8;
    (*BasicLine) += 4;
    if (StripSpaces)
      while ((**BasicLine) == ' ')
        (*BasicLine) ++;
    return (1);
  }
  if (toupper (*StartOfSequence) >= 'A' && toupper (*StartOfSequence) <= 'U' && *(StartOfSequence + 1) == '}')
  {                                                                                                          /* Form "{X}" -> UDG */
    if (toupper (*StartOfSequence) == 'T' || toupper (*StartOfSequence) == 'U')                                   /* 'T' or 'U' ? */
      switch (Is48KProgram)                                                                       /* Then the program must be 48K */
      {
        case -1 : Is48KProgram = 1; break;                                                                        /* Set the flag */
        case  0 : ReportError ("{BERROR}B - Line %d contains UDGs \'T\' and/or \'U\' "
                               "but the program was already marked 128K", BasicLineNo);
                  return (-1);
        case  1 : break;
      }
    *((*SpectrumLine) ++) = 0x90 + toupper (*StartOfSequence) - 'A';
    (*BasicLine) += 3;
    if (StripSpaces)
      while ((**BasicLine) == ' ')
        (*BasicLine) ++;
    return (1);
  }
  if (isxdigit (*StartOfSequence) && isxdigit (*(StartOfSequence + 1)) && *(StartOfSequence + 2) == '}')
  {                                                                                                    /* Form "{XX}" -> below 32 */
    if (*StartOfSequence <= '9')
      (**SpectrumLine) = *StartOfSequence - '0';
    else
      (**SpectrumLine) = toupper (*StartOfSequence) - 'A' + 10;
    if (*(StartOfSequence + 1) <= '9')
      (**SpectrumLine) = (**SpectrumLine) * 16 + *(StartOfSequence + 1) - '0';
    else
      (**SpectrumLine) = (**SpectrumLine) * 16 + toupper (*(StartOfSequence + 1)) - 'A' + 10;
    (*BasicLine) += 4;
    if (StripSpaces)
      while ((**BasicLine) == ' ')
        (*BasicLine) ++;
    return (1);
  }
  if (!strnicmp (StartOfSequence, "PAPER", 5))
  {
    Attribute = 0x10;
    AttributeLength = 5;
  }
  else if (!strnicmp (StartOfSequence, "INK", 3))
  {
    Attribute = 0x11;
    AttributeLength = 3;
  }
  else if (!strnicmp (StartOfSequence, "BRIGHT", 6))
  {
    Attribute = 0x12;
    AttributeLength = 6;
  }
  else if (!strnicmp (StartOfSequence, "FLASH", 5))
  {
    Attribute = 0x13;
    AttributeLength = 5;
  }
  else if (!strnicmp (StartOfSequence, "OVER", 4))
  {
    Attribute = 0x14;
    AttributeLength = 4;
  }
  else if (!strnicmp (StartOfSequence, "INVERSE", 7))
  {
    Attribute = 0x15;
    AttributeLength = 7;
  }
  else if (!strnicmp (StartOfSequence, "AT", 2))
  {
    Attribute = 0x16;
    AttributeLength = 2;
  }
  else if (!strnicmp (StartOfSequence, "TAB", 3))
  {
    Attribute = 0x17;
    AttributeLength = 3;
  }
  if (Attribute > 0)
  {
    StartOfSequence += AttributeLength;
    while (*StartOfSequence == ' ')
      StartOfSequence ++;
    while (isdigit (*StartOfSequence))
      AttributeVal1 = AttributeVal1 * 10 + *(StartOfSequence ++) - '0';
    if (Attribute == 0x16 || Attribute == 0x17)
    {
      if (*StartOfSequence != ',')
        Attribute = 0;
      else
      {
        while (*StartOfSequence == ' ')
          StartOfSequence ++;
        while (isdigit (*StartOfSequence))
          AttributeVal2 = AttributeVal2 * 10 + *(StartOfSequence ++) - '0';
      }
    }
    if (*StartOfSequence != '}')                                                                          /* Need closing bracket */
      Attribute = 0;
    if (Attribute > 0)
    {
      *((*SpectrumLine) ++) = Attribute;
      *((*SpectrumLine) ++) = AttributeVal1;
      if (Attribute == 0x16 || Attribute == 0x17)
        *((*SpectrumLine) ++) = AttributeVal2;
      (*BasicLine) = StartOfSequence + 1;
      if (StripSpaces)
        while ((**BasicLine) == ' ')
          (*BasicLine) ++;
      return (1);
    }
  }
  for (Cnt = 0 ; *((*BasicLine) + Cnt) && *((*BasicLine) + Cnt) != '}' ; Cnt ++)
    ;
  if (*((*BasicLine) + Cnt) == '}')
  {
    OldCharacter = *((*BasicLine) + Cnt + 1);
    *((*BasicLine) + Cnt + 1) = '\0';
    ReportWarning ("{BWARNING}N - Unexpandable sequence \"%s\" in line %d", (*BasicLine), BasicLineNo);
    *((*BasicLine) + Cnt + 1) = OldCharacter;
    return (0);
  }
  return (0);
}

int PrepareLine (char *LineIn, int FileLineNo, char **FirstToken)

/**********************************************************************************************************************************/
/* Pre   : `LineIn' points to the read line, `FileLineNo' holds the real line number.                                             */
/* Post  : Multiple spaces have been removed (unless within a string), the BASIC line number has been found and `FirstToken' is   */
/*         pointing at the first non-whitespace character after the line number.                                                  */
/*         Bad characters are reported, as well as any other error. The return value is the BASIC line number, -1 if error, or    */
/*         -2 if the (empty!) line should be skipped.                                                                             */
/* Import: GetLineNumber.                                                                                                         */
/**********************************************************************************************************************************/

{
  char  *IndexIn;
  char  *IndexOut;
  bool   InString        = FALSE;
  bool   SingleSeparator = FALSE;
  bool   StillOk         = TRUE;
  int    BasicLineNo     = -1;

  IndexIn = LineIn;
  IndexOut = ConvertedSpectrumLine;
  while (*IndexIn && StillOk)
  {
    if (*IndexIn == '\t')                                                                                   /* EXCEPTION: Print ' */
    {
      *(IndexOut ++) = 0x06;
      IndexIn ++;
    }
    else if (*IndexIn < 32 || *IndexIn >= 127)                                                /* (Exclude copyright sign as well) */
      StillOk = FALSE;
    else
    {
      if (InString)
        *(IndexOut ++) = *IndexIn;
      else
      {
        if (*IndexIn == ' ')
        {
          if (!SingleSeparator)                                                                         /* Remove multiple spaces */
          {
            SingleSeparator = TRUE;
            *(IndexOut ++) = *IndexIn;
          }
        }
        else
        {
          SingleSeparator = FALSE;
          *(IndexOut ++) = *IndexIn;
        }
      }
      if (*IndexIn == '\"')
        InString = !InString;
      IndexIn ++;
    }
  }
  *IndexOut = '\0';
  if (!StillOk)
    if (*IndexIn == 0x0D || *IndexIn == 0x0A)                                                        /* 'Correct' for end-of-line */
      StillOk = TRUE;                                                                     /* (Accept CR and/or LF as end-of-line) */
  BasicLineNo = GetLineNumber (FirstToken);
  if (InString)
    ReportError ("{BERROR}B - %s line %d misses terminating quote",
                 BasicLineNo < 0 ? "ASCII" : "BASIC", BasicLineNo < 0 ? FileLineNo : BasicLineNo);
  else if (!StillOk)
    ReportError ("{BERROR}B - %s line %d contains a bad character (code %02Xh)",
                 BasicLineNo < 0 ? "ASCII" : "BASIC", BasicLineNo < 0 ? FileLineNo : BasicLineNo, *IndexIn);
  else if (BasicLineNo < 0)                                                                         /* Could not read line number */
  {
    if (!(**FirstToken))                                                                            /* Line is completely empty ? */
    {
      ReportWarning ("{BWARNING}B - Skipping empty ASCII line %d", FileLineNo);
      return (-2);                                                                                    /* Signal: skip entire line */
    }
    else
    {
      ReportError ("{BERROR}B - Missing line number in ASCII line %d", FileLineNo);
      StillOk = FALSE;
    }
  }
  else if (PreviousBasicLineNo >= 0)                                                                      /* Not the first line ? */
  {
    if (BasicLineNo < PreviousBasicLineNo)                                            /* This line number smaller than previous ? */
    {
      ReportError ("{BERROR}B - Line number %d is smaller than previous line number %d", BasicLineNo, PreviousBasicLineNo);
      StillOk = FALSE;
    }
    else if (BasicLineNo == PreviousBasicLineNo)                                                /* Same line number as previous ? */
      ReportWarning ("{BWARNING}B - Duplicate use of line number %d", BasicLineNo);         /* (BASIC can handle it after all...) */
  }
  else if (!(**FirstToken))                                                                 /* Line contains only a line number ? */
  {
    ReportError ("{BERROR}B - Line %d contains no statements!", BasicLineNo);
    StillOk = FALSE;
  }
  PreviousBasicLineNo = BasicLineNo;                                                                 /* Remember this line number */
  if (!InString && StillOk)
    return (BasicLineNo);
  else
    return (-1);
}

bool CheckEnd (int BasicLineNo, int StatementNo, byte **Index)

/**********************************************************************************************************************************/
/* Pre   : `BasicLineNo' holds the line number, `StatementNo' the statement number, `Index' the current position in the line.     */
/* Post  : A check is made whether the end of the current statement has been reached.                                             */
/*         If so, an error is reported and TRUE is returned (so FALSE indicates that everything is still fine and dandy).         */
/* Import: none.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if (**Index == ':' || **Index == 0x0D)                                                     /* End of statement or end of line ? */
  {
    ReportError ("{BERROR in line %d, statement %d}B\n- Unexpected end of statement", BasicLineNo, StatementNo);
    return (TRUE);
  }
  return (FALSE);
}

bool ScanVariable (int BasicLineNo, int StatementNo, int Keyword, byte **Index, bool *Type, int *NameLen, int AllowSlicing)

/**********************************************************************************************************************************/
/* Pre   : `BasicLineNo' holds the line number, `StatementNo' the statement number, `Keyword' the keyword to which this operand   */
/*         belongs, `Index' the current position in the line.                                                                     */
/*         `AllocSlicing' is one of the following values:                                                                         */
/*         -1 = Don't check for slicing/indexing (used by DEF FN, FN and DIM)                                                     */
/*         0 = No slicing/indexing allowed                                                                                        */
/*         1 = Either slicing or indexing may follow (indices being numeric)                                                      */
/*         2 = Only numeric indexing may follow (used by LET and READ)                                                            */
/* Post  : A check has been made whether there's a variable at the current position. If so, it has been skipped.                  */
/*         Slicing is handled here as well, but notice that this is not necessarily correct!                                      */
/*         Single letter string variables can be either flat or array and both possibilities are considered here.                 */
/*         Both "a$(1 TO 10)" and "a$(1, 2)" are correct to BASIC, but depend on whether a "DIM" statement was used.              */
/*         The length of the found string (without any '$') is returned in `NameLen', its type is returned in `Type' (TRUE for    */
/*         numeric and FALSE for string variables). The return value is TRUE is all went well. Errors have already been reported. */
/*         The return value is FALSE either when no variable is at this point or an error was found.                              */
/*         `NameLen' is returned 0 if no variable was detected here, or > 0 if in error.                                          */
/* Import: ScanExpression.                                                                                                        */
/**********************************************************************************************************************************/

{
  bool SubType;
  bool IsArray = FALSE;
  bool SetTokenBracket = FALSE;

  Keyword = Keyword;                                                                                     /* (Keep compiler happy) */
  *Type = TRUE;                                                                                      /* Assume it will be numeric */
  *NameLen = 0;
  if (!isalpha (**Index))                                                /* The first character must be alphabetic for a variable */
    return (FALSE);
  *NameLen = 1;
  while (isalnum (*(++ (*Index))))                                                              /* Read on, until end of the word */
    (*NameLen) ++;
  if (**Index == '$')                                                                                 /* It's a string variable ? */
  {
    if (*NameLen > 1)                                                   /* String variables can only have a single character name */
    {
      ReportError ("{BERROR in line %d, statement %d}B\n- String variables can only have single character names",
                   BasicLineNo, StatementNo);
      return (FALSE);
    }
    (*Index) ++;
    *Type = FALSE;
  }
  if (AllowSlicing >= 0 && **Index == '(')                                                                  /* Slice the string ? */
  {
    if (*NameLen > 1)                                                             /* Arrays can only have a single character name */
    {
      ReportError ("{BERROR in line %d, statement %d}B\n- Arrays can only have single character names",
                   BasicLineNo, StatementNo);
      return (FALSE);
    }
    if (AllowSlicing == 0)                                                                      /* Slicing/Indexing not allowed ? */
    {
      ReportError ("{BERROR in line %d, statement %d}B\n- Slicing/Indexing not allowed", BasicLineNo, StatementNo);
      return (FALSE);
    }
    (*Index) ++;                                                                                            /* (Skip the bracket) */
    if (**Index == ')')                                                                           /* Empty slice "a$()" is not ok */
    {
      ReportError ("{BERROR in line %d, statement %d}B\n- Empty array index not allowed", BasicLineNo, StatementNo);
      return (FALSE);
    }
    if (**Index == 0xCC)                                                                           /* "a$( TO num)" or "a$( TO )" */
    {
      if (AllowSlicing == 2)
      {
        ReportError ("{BERROR in line %d, statement %d}B\n- Slicing token \"TO\" inappropriate for arrays",
                     BasicLineNo, StatementNo);
        return (FALSE);
      }
    }
    if (**Index != 0xCC)                                                                      /* Not "a$( TO num)" nor "a$( TO )" */
    {
      if (!TokenBracket)
      {
        TokenBracket = TRUE;                                                                          /* Allow complex expression */
        SetTokenBracket = TRUE;
      }
      if (!ScanExpression (BasicLineNo, StatementNo, '(', Index, &SubType, 0))                                 /* First parameter */
        return (FALSE);
      if (SetTokenBracket)
        TokenBracket = FALSE;
      if (!SubType)                                                                                            /* Must be numeric */
      {
        ReportError ("{BERROR in line %d, statement %d}B\n- Variables indices must be numeric", BasicLineNo, StatementNo);
        return (FALSE);
      }
      if (**Index == ')')                                                                                      /* "a$(num)" is ok */
      {
        (*Index) ++;
        return (TRUE);
      }
    }
    if (**Index != 0xCC && **Index != ',')                                                          /* Either an array or a slice */
    {
      ReportError ("{BERROR in line %d, statement %d}B\n- Unexpected index character \"%c\"", BasicLineNo, StatementNo, **Index);
      return (FALSE);
    }
    if (**Index == ',')
      IsArray = TRUE;
    else
    {
      if (AllowSlicing == 2)
      {
        ReportError ("{BERROR in line %d, statement %d}B\n- Slicing token \"TO\" inappropriate for arrays",
                     BasicLineNo, StatementNo);
        return (FALSE);
      }
      if (*Type)                                                                          /* Only character strings can be sliced */
      {
        ReportError ("{BERROR in line %d, statement %d}B\n- Only character strings can be sliced", BasicLineNo, StatementNo);
        return (FALSE);
      }
    }
    do
    {
      (*Index) ++;                                                                  /* Skip each "," (or the "TO" for non-arrays) */
      if (!TokenBracket)
      {
        TokenBracket = TRUE;                                                                          /* Allow complex expression */
        SetTokenBracket = TRUE;
      }
      if (!ScanExpression (BasicLineNo, StatementNo, '(', Index, &SubType, 0))                      /* Second or further paramter */
        return (FALSE);
      if (SetTokenBracket)
        TokenBracket = FALSE;
      if (!SubType)                                                                                            /* Must be numeric */
      {
        ReportError ("{BERROR in line %d, statement %d}B\n- Variables indices must be numeric", BasicLineNo, StatementNo);
        return (FALSE);
      }
      if (!IsArray && **Index != ')')
      {
        BADTOKEN ("\")\"", BTokenMap[**Index].Token);
        return (FALSE);
      }
      else if (IsArray && **Index != ',' && **Index != ')')
      {
        BADTOKEN ("\",\"", BTokenMap[**Index].Token);
        return (FALSE);
      }
    }
    while (**Index != ')');
    (*Index) ++;                                                                                   /* (Step past closing bracket) */
  }
  return (TRUE);
}

bool SliceDirectString (int BasicLineNo, int StatementNo, int Keyword, byte **Index)

/**********************************************************************************************************************************/
/* Pre   : `BasicLineNo' holds the line number, `StatementNo' the statement number, `Keyword' the keyword to which this operand   */
/*         belongs, `Index' the current position in the line.                                                                     */
/*         A direct string has just been read and a '(' character is currently under the cursor.                                  */
/* Post  : Slicing is handled here.                                                                                               */
/*         Possible are "string"(), "string"(num), "string"( TO ), "string"(num TO ), "string"( TO num) and "string"(num TO num). */
/*         The return value is FALSE if an error was found (which has already been reported here).                                */
/* Import: ScanExpression.                                                                                                        */
/**********************************************************************************************************************************/

{
  bool SubType;
  bool SetTokenBracket = FALSE;

  Keyword = Keyword;                                                                                     /* (Keep compiler happy) */
  (*Index) ++;                                                                                   /* Step past the opening bracket */
  if (**Index == ')')                                                                                /* Empty slice "abc"() is ok */
  {
    (*Index) ++;
    return (TRUE);
  }
  if (**Index != 0xCC)                                                                      /* Not "abc"( TO num) nor "abc"( TO ) */
  {
    if (!TokenBracket)
    {
      TokenBracket = TRUE;                                                                            /* Allow complex expression */
      SetTokenBracket = TRUE;
    }
    if (!ScanExpression (BasicLineNo, StatementNo, '(', Index, &SubType, 0))                                    /* First paramter */
      return (FALSE);
    if (SetTokenBracket)
      TokenBracket = FALSE;
    if (!SubType)                                                                                              /* Must be numeric */
    {
      ReportError ("{BERROR in line %d, statement %d}B\n- Slice values must be numeric", BasicLineNo, StatementNo);
      return (FALSE);
    }
  }
  if (**Index == ')')                                                                                         /* "abc"(num) is ok */
  {
    (*Index) ++;
    return (TRUE);
  }
  if (**Index != 0xCC)                                                                                                  /* ('TO') */
  {
    ReportError ("{BERROR in line %d, statement %d}B\n- Unexpected index character", BasicLineNo, StatementNo);
    return (FALSE);
  }
  (*Index) ++;
  if (**Index == ')')                                                                                     /* "abc"(num TO ) is ok */
  {
    (*Index) ++;
    return (TRUE);
  }
  if (!TokenBracket)
  {
    TokenBracket = TRUE;                                                                              /* Allow complex expression */
    SetTokenBracket = TRUE;
  }
  if (!ScanExpression (BasicLineNo, StatementNo, '(', Index, &SubType, 0))                                     /* Second paramter */
    return (FALSE);
  if (SetTokenBracket)
    TokenBracket = FALSE;
  if (!SubType)                                                                                                /* Must be numeric */
  {
    ReportError ("{BERROR in line %d, statement %d}B\n- Slice values must be numeric", BasicLineNo, StatementNo);
    return (FALSE);
  }
  if (**Index != ')')
  {
    BADTOKEN ("\")\"", BTokenMap[**Index].Token);
    return (FALSE);
  }
  (*Index) ++;                                                                                     /* (Step past closing bracket) */
  return (TRUE);
}

bool ScanStream (int BasicLineNo, int StatementNo, int Keyword, byte **Index)

/**********************************************************************************************************************************/
/* Pre   : `BasicLineNo' holds the line number, `StatementNo' the statement number, `Keyword' the keyword to which this operand   */
/*         belongs, `Index' the current position in the line.                                                                     */
/*         A stream hash mark (`#') has just been read.                                                                           */
/* Post  : The following stream number is checked to be a numeric expression.                                                     */
/*         The return value is FALSE if an error was found (which has already been reported here).                                */
/* Import: HandleClass06.                                                                                                         */
/**********************************************************************************************************************************/

{
  if (!SignalInterface1 (BasicLineNo, StatementNo, 0))
    return (FALSE);
  return (HandleClass06 (BasicLineNo, StatementNo, Keyword, Index));                                   /* Find numeric expression */
}

bool SignalInterface1 (int BasicLineNo, int StatementNo, int NewMode)

/**********************************************************************************************************************************/
/* Pre   : `BasicLineNo' holds the line number, `StatementNo' the statement number, `NewMode' holds the required hardware mode.   */
/* Post  : The required hardware is tested for conflicts.                                                                         */
/*         The return value is FALSE if there was a conflict (which has already been reported here).                              */
/* Import: none.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if ((NewMode == 1 && UsesInterface1 == 2) ||                                 /* Interface1 required, but already flagged Opus ? */
      (NewMode == 2 && UsesInterface1 == 1))                                   /* Opus required, but already flagged Interface1 ? */
  {
    ReportError ("{BERROR in line %d, statement %d}B\n- The program uses commands that are specific "
                 "for Interface 1 and Opus Discovery, but don't exist on both devices", BasicLineNo, StatementNo);
    return (FALSE);
  }
  UsesInterface1 = NewMode;
  return (TRUE);
}

bool ScanChannel (int BasicLineNo, int StatementNo, int Keyword, byte **Index, byte *WhichChannel)

/**********************************************************************************************************************************/
/* Pre   : `BasicLineNo' holds the line number, `StatementNo' the statement number, `Keyword' the keyword to which this operand   */
/*         belongs, `Index' the current position in the line.                                                                     */
/* Post  : A channel identifier of the form   "x";n;   must follow. `x' is a single alphanumeric character, `n' is a numeric      */
/*         expression, the rest are required characters.                                                                          */
/*         The found channel identifier ('x') is returned (in lowercase) in `WhichChannel'.                                       */
/*         The return value is FALSE if an error was found (which has already been reported here).                                */
/* Import: HandleClass06, CheckEnd.                                                                                               */
/**********************************************************************************************************************************/

{
  int  NeededHardware = 0;                                                                            /* (Default to Interface 1) */

  *WhichChannel = '\0';
  if (CheckEnd (BasicLineNo, StatementNo, Index))
    return (FALSE);
  if (**Index != '\"')
  {
    if (!HandleClass06 (BasicLineNo, StatementNo, Keyword, Index))/* EXCEPTION: The Opus allows '<num>' to abbreviate '"m";<num>' */
    {
      ReportError ("Expected to find a channel identifier");
      return (FALSE);
    }
    *WhichChannel = 'm';
    if (!SignalInterface1 (BasicLineNo, StatementNo, 2))                                          /* Signal the Opus specificness */
      return (FALSE);
    if (CheckEnd (BasicLineNo, StatementNo, Index))
      return (FALSE);
    if (**Index != ';')
    {
      BADTOKEN ("\";\"", BTokenMap[**Index].Token);
      return (FALSE);
    }
    (*Index) ++;
    if (CheckEnd (BasicLineNo, StatementNo, Index))
      return (FALSE);
  }
  else
  {
    (*Index) ++;
    if (CheckEnd (BasicLineNo, StatementNo, Index))
      return (FALSE);
    if (!isalpha (**Index) &&                                                                               /* (Ordinary channel) */
        **Index != '#' &&                                                                        /* (Linked channel, OPEN # only) */
        **Index != 0xAF &&                                                                       /* ('CODE' channel, OPEN # only) */
        **Index != 0xCF)                                                                          /* ('CAT' channel, OPEN # only) */
    {
      ReportError ("{BERROR in line %d, statement %d}B\n- Channel name must be alphanumeric", BasicLineNo, StatementNo);
      return (FALSE);
    }
    *WhichChannel = tolower (**Index);
    (*Index) ++;
    if (CheckEnd (BasicLineNo, StatementNo, Index))
      return (FALSE);
    if (**Index != '\"')
    {
      ReportError ("{BERROR in line %d, statement %d}B\n- Channel name must be single character", BasicLineNo, StatementNo);
      return (FALSE);
    }
    (*Index) ++;
    if (*WhichChannel == 'k' || *WhichChannel == 's' || *WhichChannel == 'p' ||                     /* (Normal Spectrum channels) */
        *WhichChannel == 'm' || *WhichChannel == 't' || *WhichChannel == 'b' ||
        *WhichChannel == '#' || *WhichChannel == 0xCF)                                                         /* ('CAT' channel) */
      NeededHardware = 0;
    else if (*WhichChannel == 'n')                                 /* Network channel is available on Interface 1 but not on Opus */
      NeededHardware = 1;
    else if (*WhichChannel == 'j' ||                                                                  /* (Opus: Joystick channel) */
             *WhichChannel == 'd' ||                                                                      /* (Opus: disk channel) */
             *WhichChannel == 0xAF)                                                                     /* (Opus: 'CODE' channel) */
      NeededHardware = 2;
    if (!SignalInterface1 (BasicLineNo, StatementNo, NeededHardware))
      return (FALSE);
    if (*WhichChannel == 'm' || *WhichChannel == 'd' || *WhichChannel == 'n' ||     /* Continue checking with these channels only */
        *WhichChannel == '#' || *WhichChannel == 0xCF)
    {
      if (CheckEnd (BasicLineNo, StatementNo, Index))
        return (FALSE);
      if (**Index != ';')
      {
        BADTOKEN ("\";\"", BTokenMap[**Index].Token);
        return (FALSE);
      }
      (*Index) ++;
      if (!HandleClass06 (BasicLineNo, StatementNo, Keyword, Index))                                   /* Find numeric expression */
        return (FALSE);
      if (*WhichChannel == 'm')                                        /* Omly the 'm' channel requires a ';' character following */
      {
        if (CheckEnd (BasicLineNo, StatementNo, Index))
          return (FALSE);
        if (**Index != ';')
        {
          BADTOKEN ("\";\"", BTokenMap[**Index].Token);
          return (FALSE);
        }
        (*Index) ++;
        if (CheckEnd (BasicLineNo, StatementNo, Index))
          return (FALSE);
      }
    }
  }
  return (TRUE);
}

bool ScanExpression (int BasicLineNo, int StatementNo, int Keyword, byte **Index, bool *Type, int Level)

/**********************************************************************************************************************************/
/* Pre   : `BasicLineNo' holds the line number, `StatementNo' the statement number, `Keyword' the keyword to which this operand   */
/*         belongs, `Index' the current position in the line.                                                                     */
/*         `Level' is used for recursion and must be 0 when called, unless when cassed from ScanVariable (then it must be 1).     */
/* Post  : An expression must be found, either numerical or string. Its type is returned in `Type' (TRUE for numerical).          */
/*         All subexpressions, between brackets, are dealt with using recursion.                                                  */
/*         The return value is FALSE if an error was found (which has already been reported here).                                */
/* Import: ScanExpression (recursive), SliceDirectString, ScanVariable, HandleClassXX.                                            */
/**********************************************************************************************************************************/

{
  bool More       = TRUE;
  bool SubType    = TRUE;                                                                          /* (Assume numeric expression) */
  bool SubSubType;
  bool TypeKnown  = FALSE;
  bool TotalTypeKnown = FALSE;
  bool Dummy;
  int  VarNameLen;
  int  ClassIndex = -1;
  byte ThisToken;

  if (**Index == '+' || **Index == '-')                                                                   /* Unary plus and minus */
  {
    *Type = TRUE;                                                                          /* Then we expect a numeric expression */
    TypeKnown = TRUE;
    (*Index) ++;                                                                                                 /* Skip the sign */
  }
  while (More)
  {
    if (**Index == '(')                                                                                      /* Opening bracket ? */
    {
      (*Index) ++;                                                                 /* The 'parent' steps past the opening bracket */
      if (!ScanExpression (BasicLineNo, StatementNo, '(', Index, &SubSubType, Level + 1))                              /* Recurse */
        return (FALSE);
      if (TypeKnown && SubSubType != SubType)                                                         /* Bad subexpression type ? */
      {
        ReportError ("{BERROR in line %d, statement %d}B\n- Type conflict in expression", BasicLineNo, StatementNo);
        return (FALSE);
      }
      else if (!TypeKnown)                                                               /* We didn't have an expected type yet ? */
      {
        SubType = SubSubType;
        TypeKnown = TRUE;
      }
      (*Index) ++;                                                             /* The 'parent' steps past the closing bracket too */
    }
    else if (**Index == ')')                                                                                 /* Closing bracket ? */
    {                                                  /* Leave the bracket for the parent, to allow functions (eg. "ATTR (...)") */
      if (!TotalTypeKnown)                                                                               /* 'Simple' expression ? */
        *Type = SubType;                                                                                       /* Set return type */
      return (TRUE);                                                                                 /* Step out of the recursion */
    }
    else if (**Index == ':' || **Index == 0x0D)                                              /* End of statement or end of line ? */
    {
      if (!TotalTypeKnown)                                                                               /* 'Simple' expression ? */
        *Type = SubType;                                                                                       /* Set return type */
      if (Level)                                                                                         /* Not on lowest level ? */
      {
        ReportError ("{BERROR in line %d, statement %d}B\n- Too few closing brackets", BasicLineNo, StatementNo);
        return (FALSE);
      }
      More = FALSE;
    }
    else if (isdigit (**Index) || **Index == '.')                                                                     /* Number ? */
    {
      if (!TypeKnown)                                                                            /* Unknown expression type yet ? */
      {
        TypeKnown = TRUE;                                                                                /* Signal: it is numeric */
        SubType = TRUE;
      }
      else if (!SubType)                                                                         /* Type was known to be string ? */
      {
        ReportError ("{BERROR in line %d, statement %d}B\n- Type conflict in expression", BasicLineNo, StatementNo);
        return (FALSE);
      }
      while (*(++ (*Index)) != 0x0E)                                                              /* Skip until the number marker */
        ;
      (*Index) ++;
    }
    else if (**Index == '\"')                                                                                  /* Direct string ? */
    {
      if (!TypeKnown)                                                                            /* Unknown expression type yet ? */
      {
        TypeKnown = TRUE;                                                                               /* Signal: it is a string */
        SubType = FALSE;
      }
      else if (SubType)                                                                         /* Type was known to be numeric ? */
      {
        ReportError ("{BERROR in line %d, statement %d}B\n- Type conflict in expression", BasicLineNo, StatementNo);
        return (FALSE);
      }
      while (**Index == '\"')                         /* Concatenated strings are ok, since they allow the use of the " character */
      {
        while (*(++ (*Index)) != '\"')                                                                      /* Find closing quote */
          ;
        (*Index) ++;                                                                                              /* Step past it */
      }
      if (**Index == '(')                                                                                   /* String is sliced ? */
        if (!SliceDirectString (BasicLineNo, StatementNo, Keyword, Index))
          return (FALSE);
    }
    else if (ScanVariable (BasicLineNo, StatementNo, Keyword, Index, &SubSubType, &VarNameLen, 1))          /* Is it a variable ? */
    {
      if (!TypeKnown)                                                                            /* Unknown expression type yet ? */
      {
        TypeKnown = TRUE;                                                                                 /* Signal: it is string */
        SubType = SubSubType;
      }
      else if (SubType != SubSubType)                                                                /* Different type variable ? */
      {
        ReportError ("{BERROR in line %d, statement %d}B\n- Type conflict in expression", BasicLineNo, StatementNo);
        return (FALSE);
      }
    }
    else if (VarNameLen != 0)                                                                                 /* (Not a variable) */
      return (FALSE);                                                                 /* (But an error that was already reported) */
    /* It's none of the above. Go check tokens */
    else switch (BTokenMap[**Index].TokenType)
    {
      case 0 : ReportError ("{BERROR in line %d, statement %d}B\n- Unexpected token \"%s\"", BasicLineNo, StatementNo,
                            BTokenMap[**Index].Token);
               return (FALSE);
      case 1 :
      case 2 : ReportError ("{BERROR in line %d, statement %d}B\n- Unexpected keyword \"%s\"",
                            BasicLineNo, StatementNo, BTokenMap[**Index].Token);
               return (FALSE);
      case 3 :
      case 4 :
      case 5 : ThisToken = *((*Index) ++);
               if (BTokenMap[ThisToken].TokenType == 5)
               {
                 if (Keyword != 0xF5 && Keyword != 0xE0)                                      /* Not handling a PRINT or LPRINT ? */
                 {
                   ReportError ("{BERROR in line %d, statement %d}B\n- Unexpected token \"%s\"",
                                BasicLineNo, StatementNo, BTokenMap[**Index].Token);
                   return (FALSE);
                 }
               }
               else if (ThisToken == 0xC0 && **Index == '\"')                                                 /* Special: USR "x" */
               {
                 (*Index) ++;                                                                    /* (Step past the opening quote) */
                 if (CheckEnd (BasicLineNo, StatementNo, Index))
                   return (FALSE);
                 if (toupper (**Index) < 'A' || toupper (**Index) > 'U')                                   /* Bad UDG character ? */
                 {
                   ReportError ("{BERROR in line %d, statement %d}B\n- Bad UDG \"%s\"",
                                BasicLineNo, StatementNo, BTokenMap[**Index].Token);
                   return (FALSE);
                 }
                 (*Index) ++;
                 if (CheckEnd (BasicLineNo, StatementNo, Index))
                   return (FALSE);
                 if (**Index != '\"')                                                                   /* More than one letter ? */
                 {
                   ReportError ("{BERROR in line %d, statement %d}B\n- An UDG name may be only 1 letter", BasicLineNo, StatementNo);
                   return (FALSE);
                 }
                 (*Index) --;
                 if (toupper (**Index) == 'T' || toupper (**Index) == 'U')                        /* One of the UDGs 'T' or 'U' ? */
                   switch (Is48KProgram)                                                          /* Then the program must be 48K */
                   {
                     case -1 : Is48KProgram = 1; break;                                                           /* Set the flag */
                     case  0 : ReportError ("{BERROR}B - Line %d contains UDGs \'T\' and/or \'U\' "
                                            "but the program was already marked 128K", BasicLineNo);
                               return (FALSE);
                     case  1 : break;
                   }
                 (*Index) += 2;                                                       /* Step past the UDG name and closing quote */
                 break;                                                                                         /* Done, step out */
               }
               else
               {
                 if (!TypeKnown)                                                                 /* Unknown expression type yet ? */
                 {
                   TypeKnown = TRUE;                                                                         /* Set expected type */
                   SubType = (BTokenMap[ThisToken].TokenType == 3);
                 }
                 else if ((SubType && BTokenMap[ThisToken].TokenType == 4) ||
                          (!SubType && BTokenMap[ThisToken].TokenType == 3))
                 {
                   ReportError ("{BERROR in line %d, statement %d}B\n- Type conflict in expression", BasicLineNo, StatementNo);
                   return (FALSE);
                 }
               }
               ClassIndex = -1;
               while (BTokenMap[ThisToken].KeywordClass[++ ClassIndex])                            /* Handle all class parameters */
               {
                 if (CheckEnd (BasicLineNo, StatementNo, Index))
                   return (FALSE);
                 else if (BTokenMap[ThisToken].KeywordClass[ClassIndex] >= 32)                       /* Required token or class ? */
                 {
                   if (**Index != BTokenMap[ThisToken].KeywordClass[ClassIndex])                              /* (Required token) */
                   {                                                                                         /* (Token not there) */
                     ReportError ("{BERROR in line %d, statement %d}B\n- Expected \"%c\", but got \"%s\"",
                                  BasicLineNo, StatementNo, BTokenMap[ThisToken].KeywordClass[ClassIndex],
                                  BTokenMap[**Index].Token);
                     return (FALSE);
                   }
                   else
                   {
                     if (**Index == '(')
                       TokenBracket = TRUE;
                     else if (**Index == ')')
                       TokenBracket = FALSE;
                     (*Index) ++;
                   }
                 }
                 else                                                                                          /* (Command class) */
                   switch (BTokenMap[ThisToken].KeywordClass[ClassIndex])
                   {
                     case  1 : if (!HandleClass01 (BasicLineNo, StatementNo, ThisToken, Index, &Dummy))  /* (Special: FN and DIM) */
                                 return (FALSE);
                               break;
                     case  3 : if (!HandleClass03 (BasicLineNo, StatementNo, ThisToken, Index))
                                 return (FALSE);
                               break;
                     case  5 : if (!HandleClass05 (BasicLineNo, StatementNo, ThisToken, Index))
                                 return (FALSE);
                               break;
                     case  6 : if (!HandleClass06 (BasicLineNo, StatementNo, ThisToken, Index))
                                 return (FALSE);
                               break;
                     case  8 : if (!HandleClass08 (BasicLineNo, StatementNo, ThisToken, Index))
                                 return (FALSE);
                               break;
                     case 10 : if (!HandleClass10 (BasicLineNo, StatementNo, ThisToken, Index))
                                 return (FALSE);
                               break;
                     case 12 : if (!HandleClass12 (BasicLineNo, StatementNo, ThisToken, Index))
                                 return (FALSE);
                               break;
                     case 13 : if (!HandleClass13 (BasicLineNo, StatementNo, ThisToken, Index))
                                 return (FALSE);
                               break;
                     case 14 : if (!HandleClass14 (BasicLineNo, StatementNo, ThisToken, Index))
                                 return (FALSE);
                               break;
                   }
               }
               if (ThisToken == 0xA6)                                                                                 /* INKEY$ ? */
                 if (**Index == '#')                                                                  /* Type 'INKEY$#<stream>' ? */
                 {
                   (*Index) ++;
                   if (!ScanStream (BasicLineNo, StatementNo, ThisToken, Index))
                     return (FALSE);
                 }
               break;
    }
    /* Piece done, continue */
    if (More)
    {
      if (**Index == 0xC5 || **Index == 0xC6)                                                                 /* ('OR' and 'AND') */
      {
        if (!TotalTypeKnown)                                                           /* 'Simple' expression before the AND/OR ? */
          *Type = SubType;
        if (**Index == 0xC5 && !*Type)
        {
          ReportError ("{BERROR in line %d, statement %d}B\n- \"OR\" requires a numeric left value", BasicLineNo, StatementNo);
          return (FALSE);
        }
        ThisToken = *((*Index) ++);                                                   /* Step over the operator - but remember it */
        if (!ScanExpression (BasicLineNo, StatementNo, ThisToken, Index, &SubSubType, 0))                /* Recurse - at level 0! */
          return (FALSE);
        if (!SubSubType)                                       /* The expression at the right must be numeric for both AND and OR */
        {
          ReportError ("{BERROR in line %d, statement %d}B\n- \"%s\" requires a numeric right value",
                       BasicLineNo, StatementNo, BTokenMap[ThisToken].Token);
          return (FALSE);
        }
        if (!TypeKnown)                                                                  /* We didn't have an expected type yet ? */
        {
          TypeKnown = TRUE;
          TotalTypeKnown = TRUE;
          SubType = *Type = (bool)(ThisToken == 0xC6 && !*Type ? FALSE : TRUE);                          /* Signal resulting type */
                                                                                                  /* x$ AND y -> result is string */
                                                                                                  /* x AND y -> result is numeric */
                                                                                                   /* x OR y -> result is numeric */
        }
        More = FALSE;              /* (Because the recursing causes the expression to be evaluated right to left, we're done now) */
      }
      else if ((**Index == '=' || **Index == '<' || **Index == '>' ||     /* EXCEPTION: equations between brackets (side effects) */
                **Index == 0xC7 || **Index == 0xC8 || **Index == 0xC9) &&                                /* ("<=", ">=" and "<>") */
               Level)                                                                   /* Not on level 0: that is handled below! */
      {                                             /* Expressions like 'LET A=(INKEY$="A")'; we're now in between these brackets */
        SubType = *Type = TRUE;                                                          /* Signal: result is going to be numeric */
        TotalTypeKnown = TRUE;
        TypeKnown = FALSE;                                                               /* Start with a fresh subexpression type */
        (*Index) ++;
      }
      else if ((BTokenMap[Keyword].TokenType != 4 && BTokenMap[Keyword].TokenType != 3) ||/* Not evaluating an expression token ? */
               TokenBracket)                                                             /* Or evaluating an operand of a token ? */
      {
        if (**Index == '+')                                                 /* (Can apply to both string and numeric expressions) */
          (*Index) ++;
        else if (**Index == '-' || **Index == '*' || **Index == '/' || **Index == '^')                          /* (Numeric only) */
        {
          if (!SubType)                                                                          /* Type was known to be string ? */
          {
            ReportError ("{BERROR in line %d, statement %d}B\n- Type conflict in expression", BasicLineNo, StatementNo);
            return (FALSE);
          }
          (*Index) ++;
        }
        /* Equations and logical operators turn the total result numeric, but each subexpression may be of any type */
        else if ((**Index == '=' || **Index == '<' || **Index == '>' ||
                  **Index == 0xC7 || **Index == 0xC8 || **Index == 0xC9) &&                              /* ("<=", ">=" and "<>") */
                 !Level)                                                                       /* Only evaluate these on level 0! */
        {
          TotalTypeKnown = TRUE;
          *Type = TRUE;                                                                  /* Signal: result is going to be numeric */
          TypeKnown = FALSE;                                                             /* Start with a fresh subexpression type */
          (*Index) ++;
        }
        else
          More = FALSE;
      }
      else
        More = FALSE;
    }
  }
  if (!TotalTypeKnown)                                                                                   /* 'Simple' expression ? */
    *Type = SubType;                                                                                           /* Set return type */
  return (TRUE);
}

bool HandleClass01 (int BasicLineNo, int StatementNo, int Keyword, byte **Index, bool *Type)

/**********************************************************************************************************************************/
/* Class 1 = Used in LET. A variable is required.                                                                                 */
/* `Type' is returned to handle the rest of this special statement (HandleClass02)                                                */
/* This function is also used to parse the variable name for FN and DIM.                                                          */
/**********************************************************************************************************************************/

{
  int VarNameLen;
  int ParseArray;

  ParseArray = ((Keyword == 0xA8 || Keyword == 0xE9) ? -1 : 2);              /* Do not parse any bracketing if checking FN or DIM */
  if (!ScanVariable (BasicLineNo, StatementNo, Keyword, Index, Type, &VarNameLen, ParseArray))
  {
    if (VarNameLen == 0)
      BADTOKEN ("variable", BTokenMap[**Index].Token);
    return (FALSE);
  }
  return (TRUE);
}

bool HandleClass02 (int BasicLineNo, int StatementNo, int Keyword, byte **Index, bool Type)

/**********************************************************************************************************************************/
/* Class 2 = Used in LET. An expression, numeric or string, must follow.                                                          */
/* `Type' is the type as returned previously by the HandleClass01 call                                                            */
/**********************************************************************************************************************************/

{
  bool SubType;

  if (!ScanExpression (BasicLineNo, StatementNo, Keyword, Index, &SubType, 0))
    return (FALSE);
  if (SubType != Type)                                                                                              /* Must match */
  {
    ReportError ("{BERROR in line %d, statement %d}B\n- Bad assignment expression type", BasicLineNo, StatementNo);
    return (FALSE);
  }
  return (TRUE);
}

bool HandleClass03 (int BasicLineNo, int StatementNo, int Keyword, byte **Index)

/**********************************************************************************************************************************/
/* Class 3 = A numeric expression may follow. Zero to be used in case of default.                                                 */
/**********************************************************************************************************************************/

{
  if (**Index == ':' || **Index == 0x0D)                                                             /* No expression following ? */
    return (TRUE);                                                                                     /* Then we're done already */
  if (Keyword == 0xFD && **Index == '#')                   /* EXCEPTION: CLEAR may take a stream rather than a numeric expression */
  {
    (*Index) ++;
    if (!SignalInterface1 (BasicLineNo, StatementNo, 0))                                   /* (Which is Interface1/Opus specific) */
      return (FALSE);
    if (**Index == ':' || **Index == 0x0D)                                                           /* No expression following ? */
      return (TRUE);                                      /* (An empty stream is allowed as well - it clears all streams at once) */
  }
  return (HandleClass06 (BasicLineNo, StatementNo, Keyword, Index));                                   /* Find numeric expression */
}

bool HandleClass04 (int BasicLineNo, int StatementNo, int Keyword, byte **Index)

/**********************************************************************************************************************************/
/* Class 4 = A single character variable must follow.                                                                             */
/**********************************************************************************************************************************/

{
  bool Type;
  int  VarNameLen;

  if (!ScanVariable (BasicLineNo, StatementNo, Keyword, Index, &Type, &VarNameLen, 0))
  {
    if (VarNameLen == 0)
      BADTOKEN ("variable", BTokenMap[**Index].Token);
    return (FALSE);
  }
  if (VarNameLen != 1 || !Type)                                                  /* Not single letter or not a numeric variable ? */
  {
    ReportError ("{BERROR in line %d, statement %d}B\n- Wrong variable type", BasicLineNo, StatementNo);
    return (FALSE);
  }
  return (TRUE);
}

bool HandleClass05 (int BasicLineNo, int StatementNo, int Keyword, byte **Index)

/**********************************************************************************************************************************/
/* Class 5 = A set of items may be given.                                                                                         */
/**********************************************************************************************************************************/

{
  bool Type;
  bool More = TRUE;
  int  VarNameLen;
  
  while (More)
  {
    while (**Index == ';' || **Index == ',' || **Index == '\'')                              /* One of the separator characters ? */
      (*Index) ++;                                                                                  /* (More than one may follow) */
    if (**Index == ':' || **Index == 0x0D)                                                   /* End of statement or end of line ? */
      More = FALSE;
    else if (**Index == '#')                                                                                        /* A stream ? */
    {
      (*Index) ++;                                                                                    /* (Step past the '#' mark) */
      if (!ScanStream (BasicLineNo, StatementNo, Keyword, Index))
        return (FALSE);
    }
    else if (BTokenMap[**Index].TokenType == 2 ||                                                         /* A colour parameter ? */
             **Index == 0xAD)                                                                                            /* TAB ? */
    {
      (*Index) ++;                                                                                            /* (Skip the token) */
      if (!HandleClass06 (BasicLineNo, StatementNo, Keyword, Index))                       /* Find parameter (numeric expression) */
        return (FALSE);
    }
    else if (**Index == 0xAC)                                                                                             /* AT ? */
    {
      (*Index) ++;                                                                                            /* (Skip the token) */
      if (!HandleClass06 (BasicLineNo, StatementNo, Keyword, Index))                 /* Find first parameter (numeric expression) */
        return (FALSE);
      if (CheckEnd (BasicLineNo, StatementNo, Index))
        return (FALSE);
      if (**Index != ',')                                                                           /* (Required separator token) */
      {
        BADTOKEN ("\",\"", BTokenMap[**Index].Token);
        return (FALSE);
      }
      (*Index) ++;                                                                                            /* (Skip the token) */
      if (!HandleClass06 (BasicLineNo, StatementNo, Keyword, Index))                /* Find second parameter (numeric expression) */
        return (FALSE);
    }
    else if (Keyword == 0xEE && **Index == 0xCA)                                                            /* INPUT may use LINE */
    {
      (*Index) ++;                                                                                            /* (Skip the token) */
      if (!ScanVariable (BasicLineNo, StatementNo, Keyword, Index, &Type, &VarNameLen, 0))
      {
        if (VarNameLen == 0)
          BADTOKEN ("variable", BTokenMap[**Index].Token);
        return (FALSE);
      }
      if (Type)                                                                                  /* Not a alphanumeric variable ? */
      {
        ReportError ("{BERROR in line %d, statement %d}B\n- INPUT LINE requires an alphanumeric variable",
                     BasicLineNo, StatementNo);
        return (FALSE);
      }
    }
    else if (!ScanExpression (BasicLineNo, StatementNo, Keyword, Index, &Type, 0))                              /* Get expression */
      return (FALSE);
  }
  return (TRUE);
}

bool HandleClass06 (int BasicLineNo, int StatementNo, int Keyword, byte **Index)

/**********************************************************************************************************************************/
/* Class 6 = A numeric expression must follow.                                                                                    */
/**********************************************************************************************************************************/

{
  bool Type;

  if (!ScanExpression (BasicLineNo, StatementNo, Keyword, Index, &Type, 0))                                     /* Get expression */
    return (FALSE);
  if (!Type && Keyword != 0xC0)                                                                                /* Must be numeric */
  {                                                   /* (Exception for USR, which can take both a numerical and string operand!) */
    ReportError ("{BERROR in line %d, statement %d}B\n- Expected numeric expression", BasicLineNo, StatementNo);
    return (FALSE);
  }
  return (TRUE);
}

bool HandleClass07 (int BasicLineNo, int StatementNo, int Keyword, byte **Index)

/**********************************************************************************************************************************/
/* Class 7 = Handles colour items.                                                                                                */
/* Effectively the same as Class 6                                                                                                */
/**********************************************************************************************************************************/

{
  return (HandleClass06 (BasicLineNo, StatementNo, Keyword, Index));                                   /* Find numeric expression */
}

bool HandleClass08 (int BasicLineNo, int StatementNo, int Keyword, byte **Index)

/**********************************************************************************************************************************/
/* Class 8 = Two numeric expressions, separated by a comma, must follow.                                                          */
/**********************************************************************************************************************************/

{
  if (!HandleClass06 (BasicLineNo, StatementNo, Keyword, Index))                                 /* Find first numeric expression */
    return (FALSE);
  if (**Index != ',')
  {
    BADTOKEN ("\",\"", BTokenMap[**Index].Token);
    return (FALSE);
  }
  (*Index) ++;
  return (HandleClass06 (BasicLineNo, StatementNo, Keyword, Index));                            /* Find second numeric expression */
}

bool HandleClass09 (int BasicLineNo, int StatementNo, int Keyword, byte **Index)

/**********************************************************************************************************************************/
/* Class 9 = As for class 8 but colour items may precede the expression.                                                          */
/* Used only by PLOT and DRAW. Colour items are TokenType 2                                                                       */
/**********************************************************************************************************************************/

{
  bool CheckColour = TRUE;

  while (CheckColour)
  {
    if (CheckEnd (BasicLineNo, StatementNo, Index))
      return (FALSE);
    if (BTokenMap[**Index].TokenType == 2)                                                                /* A colour parameter ? */
    {
      (*Index) ++;                                                                                              /* Skip the token */
      if (!HandleClass06 (BasicLineNo, StatementNo, Keyword, Index))                       /* Find parameter (numeric expression) */
        return (FALSE);
      if (CheckEnd (BasicLineNo, StatementNo, Index))
        return (FALSE);
      if (**Index != ';')                                              /* All colour parameters must be separated with semicolons */
      {
        BADTOKEN ("\";\"", BTokenMap[**Index].Token);
        return (FALSE);
      }
    }
    else
      CheckColour = FALSE;
  }
  if (CheckEnd (BasicLineNo, StatementNo, Index))
    return (FALSE);
  return (HandleClass08 (BasicLineNo, StatementNo, Keyword, Index));
}

bool HandleClass10 (int BasicLineNo, int StatementNo, int Keyword, byte **Index)

/**********************************************************************************************************************************/
/* Class 10 = A string expression must follow.                                                                                    */
/**********************************************************************************************************************************/

{
  bool Type;

  if (!ScanExpression (BasicLineNo, StatementNo, Keyword, Index, &Type, 0))                                     /* Get expression */
    return (FALSE);
  if (Type)                                                                                                     /* Must be string */
  {
    ReportError ("{BERROR in line %d, statement %d}B\n- Expected string expression", BasicLineNo, StatementNo);
    return (FALSE);
  }
  return (TRUE);
}

bool HandleClass11 (int BasicLineNo, int StatementNo, int Keyword, byte **Index)

/**********************************************************************************************************************************/
/* Class 11 = Handles cassette routines.                                                                                          */
/**********************************************************************************************************************************/

{
  bool Type;
  int  VarNameLen;
  int  MoveLoop;
  byte WhichChannel = '\0';                                                                  /* (Default is no channel; for tape) */

  switch (Keyword)
  {
    case 0xEF:                                                                                                          /* (LOAD) */
    case 0xD6:                                                                                                        /* (VERIFY) */
    case 0xD5: if (**Index == '*')                                                                                     /* (MERGE) */
               {
                 (*Index) ++;
                 if (!ScanChannel (BasicLineNo, StatementNo, Keyword, Index, &WhichChannel))
                   return (FALSE);
                 if (WhichChannel != 'm' && WhichChannel != 'b' && WhichChannel != 'n')
                 {
                   ReportError ("{BERROR in line %d, statement %d}B\n- You cannot LOAD/VERIFY/MERGE from the \"%s\" channel",
                                BasicLineNo, StatementNo, BTokenMap[WhichChannel].Token);
                   return (FALSE);
                 }
               }
               else if (**Index == '!')                                                                        /* 128K RAM-bank ? */
               {
                 (*Index) ++;
                 switch (Is48KProgram)                                                           /* Then the program must be 128K */
                 {
                   case -1 : Is48KProgram = 0; break;                                                             /* Set the flag */
                   case  1 : ReportError ("{BERROR}B - Line %d contains 128K file I/O, but the program "
                                          "also uses UDGs \'T\' and/or \'U\'", BasicLineNo);
                             return (FALSE);
                   case  0 : break;
                 }
               }
               if (WhichChannel != '\0' && WhichChannel != 'm')                         /* Not tape nor microdrive/disk channel ? */
               {
                 if (**Index != ':' && **Index != 0x0D &&                                                   /* (End of statement) */
                     **Index != 0xAF &&                                                                                 /* (CODE) */
                     **Index != 0xE4 &&                                                                                 /* (DATA) */
                     **Index != 0xC4 &&                                                                                 /* (LINE) */
                     **Index != 0xAA)                                                                                /* (SCREEN$) */
                 {
                   ReportError ("{BERROR in line %d, statement %d}B\n- The \"%s\" channel does not use filenames",
                                BasicLineNo, StatementNo, BTokenMap[WhichChannel].Token);
                   return (FALSE);
                 }
               }
               else
               {
                 if (**Index == '\"')                                                                      /* Look for a filename */
                 {
                   while (**Index == '\"')            /* Concatenated strings are ok, since they allow the use of the " character */
                   {                                                             /* (And an empty string is allowed here as well) */
                     while (*(++ (*Index)) != '\"')                                                         /* Find closing quote */
                       if (**Index == 0x0D)                                                                      /* End of line ? */
                       {
                         ReportError ("{BERROR in line %d, statement %d}B\n- Unexpected end of line", BasicLineNo, StatementNo);
                         return (FALSE);
                       }
                     (*Index) ++;                                                                                 /* Step past it */
                   }
                 }
                 else if (**Index == ':' || **Index == 0x0D ||                                              /* (End of statement) */
                          **Index == 0xAF ||                                                                            /* (CODE) */
                          **Index == 0xE4 ||                                                                            /* (DATA) */
                          **Index == 0xC4 ||                                                                            /* (LINE) */
                          **Index == 0xAA)                                                                           /* (SCREEN$) */
                 {
                   BADTOKEN ("filename", BTokenMap[**Index].Token);
                   return (FALSE);
                 }
                 else if (!HandleClass10 (BasicLineNo, StatementNo, Keyword, Index))              /* Look for a string expression */
                   return (FALSE);
               }
               if (**Index != ':' && **Index != 0x0D)                                       /* (Continue unless end of statement) */
               {
                 if (**Index == 0xAF)                                                                                     /* CODE */
                 {
                   if (Keyword == 0xD5)                                                                /* (We were doing MERGE ?) */
                   {
                     ReportError ("{BERROR in line %d, statement %d}B\n- Cannot MERGE CODE", BasicLineNo, StatementNo);
                     return (FALSE);
                   }
                   (*Index) ++;
                   if (**Index != ':' && **Index != 0x0D)                                                   /* Optional address ? */
                   {
                     if (!HandleClass06 (BasicLineNo, StatementNo, Keyword, Index))          /* Find address (numeric expression) */
                       return (FALSE);
                     if (**Index == ',')                                                                /* Also optional length ? */
                     {
                       (*Index) ++;
                       if (!HandleClass06 (BasicLineNo, StatementNo, Keyword, Index))         /* Find length (numeric expression) */
                         return (FALSE);
                     }
                     else if (**Index != ':' && **Index != 0x0D)
                     {
                       BADTOKEN ("\",\"", BTokenMap[**Index].Token);
                       return (FALSE);
                     }
                   }
                 }
                 else if (**Index == 0xAA)                                                                             /* SCREEN$ */
                   (*Index) ++;
                 else if (**Index == 0xE4)                                                                                /* DATA */
                 {
                   (*Index) ++;
                   if (CheckEnd (BasicLineNo, StatementNo, Index))
                     return (FALSE);
                   if (!ScanVariable (BasicLineNo, StatementNo, Keyword, Index, &Type, &VarNameLen, -1))
                   {
                     if (VarNameLen == 0)
                       BADTOKEN ("variable", BTokenMap[**Index].Token);
                     return (FALSE);
                   }
                   if (VarNameLen != 1)                                                                    /* Not single letter ? */
                   {
                     ReportError ("{BERROR in line %d, statement %d}B\n- Wrong variable type; must be single character",
                                  BasicLineNo, StatementNo);
                     return (FALSE);
                   }
                   if (**Index != '(')                                         /* The variable must be followed by an empty index */
                   {
                     ReportError ("{BERROR in line %d, statement %d}B\n- DATA requires an array", BasicLineNo, StatementNo);
                     return (FALSE);
                   }
                   (*Index) ++;
                   if (**Index != ')')
                   {
                     ReportError ("{BERROR in line %d, statement %d}B\n- DATA requires an empty array index",
                                  BasicLineNo, StatementNo);
                     return (FALSE);
                   }
                 }
                 else
                 {
                   ReportError ("{BERROR in line %d, statement %d}B\n- Unknown file-type \"%s\"",
                                BasicLineNo, StatementNo, BTokenMap[**Index].Token);
                   return (FALSE);
                 }
               }
               break;
    case 0xF8: if (**Index == '*')                                                                                      /* (SAVE) */
               {
                 (*Index) ++;
                 if (!ScanChannel (BasicLineNo, StatementNo, Keyword, Index, &WhichChannel))
                   return (FALSE);
                 if (WhichChannel != 'm' && WhichChannel != 'b' && WhichChannel != 'n')
                 {
                   ReportError ("{BERROR in line %d, statement %d}B\n- You cannot SAVE to the \"%s\" channel",
                                BasicLineNo, StatementNo, BTokenMap[WhichChannel].Token);
                   return (FALSE);
                 }
               }
               else if (**Index == '!')                                                                        /* 128K RAM-bank ? */
               {
                 (*Index) ++;
                 switch (Is48KProgram)                                                           /* Then the program must be 128K */
                 {
                   case -1 : Is48KProgram = 0; break;                                                             /* Set the flag */
                   case  1 : ReportError ("{BERROR}B - Line %d contains 128K file I/O, but the program "
                                          "also uses UDGs \'T\' and/or \'U\'", BasicLineNo);
                             return (FALSE);
                   case  0 : break;
                 }
               }
               if (WhichChannel != '\0' && WhichChannel != 'm')                         /* Not tape nor microdrive/disk channel ? */
               {
                 if (**Index != ':' && **Index != 0x0D &&                                                   /* (End of statement) */
                     **Index != 0xAF &&                                                                                 /* (CODE) */
                     **Index != 0xE4 &&                                                                                 /* (DATA) */
                     **Index != 0xC4 &&                                                                                 /* (LINE) */
                     **Index != 0xAA)                                                                                /* (SCREEN$) */
                 {
                   ReportError ("{BERROR in line %d, statement %d}B\n- The \"%s\" channel does not use filenames",
                                BasicLineNo, StatementNo, BTokenMap[WhichChannel].Token);
                   return (FALSE);
                 }
               }
               else
               {
                 if (**Index == '\"')                                                                      /* Look for a filename */
                 {
                   if (*(*Index + 1) == '\"' &&                                                   /* Empty string (not allowed) ? */
                       *(*Index + 2) != '\"')                                    /* Concatenation - first char is a " (allowed) ? */
                   {
                     ReportError ("{BERROR in line %d, statement %d}B\n- Empty filename not allowed", BasicLineNo, StatementNo);
                     return (FALSE);
                   }
                   while (**Index == '\"')            /* Concatenated strings are ok, since they allow the use of the " character */
                   {
                     while (*(++ (*Index)) != '\"')                                                         /* Find closing quote */
                       if (**Index == 0x0D)                                                                      /* End of line ? */
                       {
                         ReportError ("{BERROR in line %d, statement %d}B\n- Unexpected end of line", BasicLineNo, StatementNo);
                         return (FALSE);
                       }
                     (*Index) ++;                                                                                 /* Step past it */
                   }
                 }
                 else if (**Index == ':' || **Index == 0x0D ||                                              /* (End of statement) */
                          **Index == 0xAF ||                                                                            /* (CODE) */
                          **Index == 0xE4 ||                                                                            /* (DATA) */
                          **Index == 0xC4 ||                                                                            /* (LINE) */
                          **Index == 0xAA)                                                                           /* (SCREEN$) */
                 {
                   BADTOKEN ("filename", BTokenMap[**Index].Token);
                   return (FALSE);
                 }
                 else if (!HandleClass10 (BasicLineNo, StatementNo, Keyword, Index))              /* Look for a string expression */
                   return (FALSE);
               }
               if (**Index != ':' && **Index != 0x0D)                                       /* (Continue unless end of statement) */
               {
                 if (**Index == 0xAF)                                                                                     /* CODE */
                 {
                   (*Index) ++;
                   if (!HandleClass06 (BasicLineNo, StatementNo, Keyword, Index))            /* Find address (numeric expression) */
                     return (FALSE);
                   if (**Index != ',')
                   {
                     ReportError ("{BERROR in line %d, statement %d}B\n- %s CODE requires both address and length",
                                  BasicLineNo, StatementNo, BTokenMap[Keyword].Token);
                     return (FALSE);
                   }
                   (*Index) ++;
                   if (!HandleClass06 (BasicLineNo, StatementNo, Keyword, Index))             /* Find length (numeric expression) */
                     return (FALSE);
                 }
                 else if (**Index == 0xE4)                                                                                /* DATA */
                 {
                   (*Index) ++;
                   if (CheckEnd (BasicLineNo, StatementNo, Index))
                     return (FALSE);
                   if (!ScanVariable (BasicLineNo, StatementNo, Keyword, Index, &Type, &VarNameLen, -1))
                   {
                     if (VarNameLen == 0)
                       BADTOKEN ("variable", BTokenMap[**Index].Token);
                     return (FALSE);
                   }
                   if (VarNameLen != 1)                                                                    /* Not single letter ? */
                   {
                     ReportError ("{BERROR in line %d, statement %d}B\n- Wrong variable type; must be single character",
                                  BasicLineNo, StatementNo);
                     return (FALSE);
                   }
                   if (**Index != '(')                                         /* The variable must be followed by an empty index */
                   {
                     ReportError ("{BERROR in line %d, statement %d}B\n- DATA requires an array", BasicLineNo, StatementNo);
                     return (FALSE);
                   }
                   (*Index) ++;
                   if (**Index != ')')
                   {
                     ReportError ("{BERROR in line %d, statement %d}B\n- DATA requires an empty array index",
                                  BasicLineNo, StatementNo);
                     return (FALSE);
                   }
                 }
                 else if (**Index == 0xAA)                                                                             /* SCREEN$ */
                   (*Index) ++;
                 else if (**Index == 0xCA)                                                                                /* LINE */
                 {
                   (*Index) ++;
                   if (!HandleClass06 (BasicLineNo, StatementNo, Keyword, Index))      /* Find starting line (numeric expression) */
                     return (FALSE);
                 }
                 else
                 {
                   ReportError ("{BERROR in line %d, statement %d}B\n- Unknown file-type \"%s\"",
                                BasicLineNo, StatementNo, BTokenMap[**Index].Token);
                   return (FALSE);
                 }
               }
               break;
    case 0xCF: if (!SignalInterface1 (BasicLineNo, StatementNo, 0))                                                      /* (CAT) */
                 return (FALSE);
               if (**Index == '#')                                                       /* A stream may precede the drive number */
               {
                 (*Index) ++;
                 if (!ScanStream (BasicLineNo, StatementNo, Keyword, Index))
                   return (FALSE);
                 if (**Index != ',')                                                                /* (Required separator token) */
                 {
                   BADTOKEN ("\",\"", BTokenMap[**Index].Token);
                   return (FALSE);
                 }
               }
               if (!HandleClass06 (BasicLineNo, StatementNo, Keyword, Index))           /* Find drive number (numeric expression) */
                 return (FALSE);
               break;
    case 0xD0: if (!ScanChannel (BasicLineNo, StatementNo, Keyword, Index, &WhichChannel))                            /* (FORMAT) */
                 return (FALSE);
               switch (WhichChannel)
               {
                 case 'm' : if (CheckEnd (BasicLineNo, StatementNo, Index))         /* "m" requires an additional new volume name */
                              return (FALSE);
                            if (**Index == '\"')                                                        /* Look for a volume name */
                            {
                              if (*(*Index + 1) == '\"' &&                                        /* Empty string (not allowed) ? */
                                  *(*Index + 2) != '\"')                         /* Concatenation - first char is a " (allowed) ? */
                              {
                                ReportError ("{BERROR in line %d, statement %d}B\n- Empty volume name not allowed",
                                             BasicLineNo, StatementNo);
                                return (FALSE);
                              }
                              while (**Index == '\"') /* Concatenated strings are ok, since they allow the use of the " character */
                              {
                                while (*(++ (*Index)) != '\"')                                              /* Find closing quote */
                                  if (**Index == 0x0D)                                                           /* End of line ? */
                                  {
                                    ReportError ("{BERROR in line %d, statement %d}B\n- Unexpected end of line",
                                                 BasicLineNo, StatementNo);
                                    return (FALSE);
                                  }
                                (*Index) ++;                                                                      /* Step past it */
                              }
                            }
                            else if (!HandleClass10 (BasicLineNo, StatementNo, Keyword, Index))   /* Look for a string expression */
                              return (FALSE);
                            break;
                 case 't' :                                                 /* The port channels requires an additional baud rate */
                 case 'b' :
                 case 'j' : if (**Index != ';')                   /* The joystick channel requires a operand to turn it on or off */
                            {
                              BADTOKEN ("\";\"", BTokenMap[**Index].Token);
                              return (FALSE);
                            }
                            (*Index) ++;
                            if (CheckEnd (BasicLineNo, StatementNo, Index))
                              return (FALSE);
                            if (!HandleClass06 (BasicLineNo, StatementNo, Keyword, Index))       /* Look for a numeric expression */
                              return (FALSE);
                            break;
                 default  : ReportError ("{BERROR in line %d, statement %d}B\n- You cannot FORMAT the \"%s\" channel",
                                         BasicLineNo, StatementNo, BTokenMap[WhichChannel].Token);
                            return (FALSE);
               }
               break;
    case 0xD1: for (MoveLoop = 0 ; MoveLoop < 2 ; MoveLoop ++)                                                          /* (MOVE) */
               {
                 if (**Index == '#')
                 {
                   (*Index) ++;                                                                       /* (Step past the '#' mark) */
                   if (!ScanStream (BasicLineNo, StatementNo, Keyword, Index))
                     return (FALSE);
                 }
                 else
                 {
                   if (!ScanChannel (BasicLineNo, StatementNo, Keyword, Index, &WhichChannel))
                     return (FALSE);
                   switch (WhichChannel)
                   {
                     case 'm' : if (CheckEnd (BasicLineNo, StatementNo, Index))            /* "m" requires an additional filename */
                                  return (FALSE);
                                if (**Index == '\"')                                                       /* Look for a filename */
                                {
                                  if (*(*Index + 1) == '\"' &&                                    /* Empty string (not allowed) ? */
                                      *(*Index + 2) != '\"')                     /* Concatenation - first char is a " (allowed) ? */
                                  {
                                    ReportError ("{BERROR in line %d, statement %d}B\n- Empty filename not allowed",
                                                 BasicLineNo, StatementNo);
                                    return (FALSE);
                                  }
                                  while (**Index == '\"')
                                  {
                                    while (*(++ (*Index)) != '\"')
                                      if (**Index == 0x0D)
                                      {
                                        ReportError ("{BERROR in line %d, statement %d}B\n- Unexpected end of line",
                                                     BasicLineNo, StatementNo);
                                        return (FALSE);
                                      }
                                    (*Index) ++;
                                  }
                                }
                                else if (!HandleClass10 (BasicLineNo, StatementNo, Keyword, Index))
                                  return (FALSE);
                                break;
                     case 't' :
                     case 'b' :
                     case 'n' :
                     case 'd' : break;                                       /* All these are okay and don't use extra parameters */
                     case 's' : if (MoveLoop == 0)                                               /* The "s" channel is write-only */
                                {
                                  ReportError ("{BERROR in line %d, statement %d}B\n- You cannot MOVE from the \"s\" channel",
                                               BasicLineNo, StatementNo);
                                  return (FALSE);
                                }
                                break;
                     case 'k' : if (MoveLoop == 1)                                                /* The "k" channel is read-only */
                                {
                                  ReportError ("{BERROR in line %d, statement %d}B\n- You cannot MOVE to the \"k\" channel",
                                               BasicLineNo, StatementNo);
                                  return (FALSE);
                                }
                                break;
                     default  : ReportError ("{BERROR in line %d, statement %d}B\n- You cannot MOVE from/to the \"%s\" channel",
                                             BasicLineNo, StatementNo, BTokenMap[WhichChannel].Token);
                                return (FALSE);
                   }
                 }
                 if (MoveLoop == 0)
                 {
                   if (**Index != 0xCC)                                                                    /* Required token 'TO' */
                   {
                     BADTOKEN ("\"TO\"", BTokenMap[**Index].Token);
                     return (FALSE);
                   }
                   (*Index) ++;
                 }
               }
               break;
    case 0xD2: if (!ScanChannel (BasicLineNo, StatementNo, Keyword, Index, &WhichChannel))                             /* (ERASE) */
                 return (FALSE);
               switch (WhichChannel)
               {
                 case 'm' : if (CheckEnd (BasicLineNo, StatementNo, Index))                /* "m" requires an additional filename */
                              return (FALSE);
                            if (**Index == '\"')                                                           /* Look for a filename */
                            {
                              if (*(*Index + 1) == '\"' &&                                        /* Empty string (not allowed) ? */
                                  *(*Index + 2) != '\"')                         /* Concatenation - first char is a " (allowed) ? */
                              {
                                ReportError ("{BERROR in line %d, statement %d}B\n- Empty filename not allowed",
                                             BasicLineNo, StatementNo);
                                return (FALSE);
                              }
                              while (**Index == '\"')
                              {
                                while (*(++ (*Index)) != '\"')
                                  if (**Index == 0x0D)
                                  {
                                    ReportError ("{BERROR in line %d, statement %d}B\n- Unexpected end of line",
                                                 BasicLineNo, StatementNo);
                                    return (FALSE);
                                  }
                                (*Index) ++;
                              }
                            }
                            else if (!HandleClass10 (BasicLineNo, StatementNo, Keyword, Index))
                              return (FALSE);
                            break;
                 default  : ReportError ("{BERROR in line %d, statement %d}B\n- You can only ERASE from the \"m\" channel",
                                         BasicLineNo, StatementNo);
                            return (FALSE);
               }
               break;
    case 0xD3: if (!ScanStream (BasicLineNo, StatementNo, Keyword, Index))                                            /* (OPEN #) */
                 return (FALSE);
               if (**Index != ';' && **Index != ',')                                                          /* (Required token) */
               {
                 BADTOKEN ("\";\"", BTokenMap[**Index].Token);
                 return (FALSE);
               }
               (*Index) ++;
               if (!ScanChannel (BasicLineNo, StatementNo, Keyword, Index, &WhichChannel))
                 return (FALSE);
               switch (WhichChannel)
               {
                 case 'm' : if (CheckEnd (BasicLineNo, StatementNo, Index))                /* "m" requires an additional filename */
                              return (FALSE);
                            if (**Index == '\"')                                                           /* Look for a filename */
                            {
                              if (*(*Index + 1) == '\"' &&                                        /* Empty string (not allowed) ? */
                                  *(*Index + 2) != '\"')                         /* Concatenation - first char is a " (allowed) ? */
                              {
                                ReportError ("{BERROR in line %d, statement %d}B\n- Empty filename not allowed",
                                             BasicLineNo, StatementNo);
                                return (FALSE);
                              }
                              while (**Index == '\"')
                              {
                                while (*(++ (*Index)) != '\"')
                                  if (**Index == 0x0D)
                                  {
                                    ReportError ("{BERROR in line %d, statement %d}B\n- Unexpected end of line",
                                                 BasicLineNo, StatementNo);
                                    return (FALSE);
                                  }
                                (*Index) ++;
                              }
                            }
                            else if (!HandleClass10 (BasicLineNo, StatementNo, Keyword, Index))
                              return (FALSE);
                            break;
                 case 's' :
                 case 'k' :
                 case 'p' :
                 case 't' :
                 case 'b' :
                 case 'n' :
                 case 0xAF:
                 case 0xCF:
                 case '#' : break;                                           /* All these are okay and don't use extra parameters */
                 default  : ReportError ("{BERROR in line %d, statement %d}B\n- You cannot attach a stream to the \"%s\" channel",
                                         BasicLineNo, StatementNo, BTokenMap[WhichChannel].Token);
                            return (FALSE);
               }
               if (**Index != ':' && **Index != 0x0D)                                       /* (Continue unless end of statement) */
               {
                 if (**Index == 0xBF)                                                                                       /* IN */
                 {
                   (*Index) ++;
                   if (!SignalInterface1 (BasicLineNo, StatementNo, 2))                                  /* This is Opus specific */
                     return (FALSE);
                 }
                 else if (**Index == 0xDF ||                                                                               /* OUT */
                          **Index == 0xB9)                                                                                 /* EXP */
                 {
                   (*Index) ++;
                   if (!SignalInterface1 (BasicLineNo, StatementNo, 2))                                  /* This is Opus specific */
                     return (FALSE);
                   if (!HandleClass06 (BasicLineNo, StatementNo, Keyword, Index))                      /* Find numeric expression */
                     return (FALSE);
                 }
                 else if (**Index == 0xA5)                                                                                 /* RND */
                 {
                   (*Index) ++;
                   if (!SignalInterface1 (BasicLineNo, StatementNo, 2))                                  /* This is Opus specific */
                     return (FALSE);
                   if (!HandleClass06 (BasicLineNo, StatementNo, Keyword, Index))                      /* Find numeric expression */
                     return (FALSE);
                   if (**Index == ',')                                                         /* RND may take a second parameter */
                   {
                     (*Index) ++;
                     if (!HandleClass06 (BasicLineNo, StatementNo, Keyword, Index))
                       return (FALSE);
                   }
                 }
               }
               break;
    case 0xD4: if (!ScanStream (BasicLineNo, StatementNo, Keyword, Index))                                           /* (CLOSE #) */
                 return (FALSE);
               break;
  }
  return (TRUE);
}

bool HandleClass12 (int BasicLineNo, int StatementNo, int Keyword, byte **Index)

/**********************************************************************************************************************************/
/* Class 12 = One or more string expressions, separated by commas, must follow.                                                   */
/**********************************************************************************************************************************/

{
  bool Type;
  bool More = TRUE;

  while (More)
  {
    if (!ScanExpression (BasicLineNo, StatementNo, Keyword, Index, &Type, 0))                               /* Find an expression */
      return (FALSE);
    if (Type)                                                                                                   /* Must be string */
    {
      ReportError ("{BERROR in line %d, statement %d}B\n- \"%s\" requires string parameters",
                   BasicLineNo, StatementNo, BTokenMap[Keyword].Token);
      return (FALSE);
    }
    if (**Index == ':' || **Index == 0x0D)                                                   /* End of statement or end of line ? */
      More = FALSE;
    else if (**Index == ',')                                                                                       /* Separator ? */
      (*Index) ++;
    else if (**Index != ')')
    {
      BADTOKEN ("\",\"", BTokenMap[**Index].Token);
      return (FALSE);
    }
  }
  return (TRUE);
}

bool HandleClass13 (int BasicLineNo, int StatementNo, int Keyword, byte **Index)

/**********************************************************************************************************************************/
/* Class 13 = One or more expressions, separated by commas, must follow (DATA, FN and DIM)                                        */
/**********************************************************************************************************************************/

{
  bool Type;
  bool More = TRUE;

  if (**Index == ')' && Keyword == 0xA8)                                                  /* FN requires zero or more expressions */
    return (TRUE);                               /* (The closing bracket is a required character and stepped over in CheckSyntax) */
  while (More)
  {
    if (!ScanExpression (BasicLineNo, StatementNo, Keyword, Index, &Type, 0))                               /* Find an expression */
      return (FALSE);                                                                              /* (Don't care about the type) */
    if (Keyword == 0xE9 && !Type)                                                              /* DIM requires numeric dimensions */
    {
      ReportError ("{BERROR in line %d, statement %d}B\n- \"DIM\" requires numeric dimensions", BasicLineNo, StatementNo);
      return (FALSE);
    }
    if (Keyword == 0xE9 || Keyword == 0xA8)                                              /* FN and DIM end with a closing bracket */
    {
      if (CheckEnd (BasicLineNo, StatementNo, Index))
        return (FALSE);
      if (**Index == ')')
        More = FALSE;
    }
    if (**Index == ':' || **Index == 0x0D)                                                   /* End of statement or end of line ? */
      More = FALSE;
    else if (**Index == ',')                                                                                       /* Separator ? */
      (*Index) ++;
    else if (**Index != ')')
    {
      BADTOKEN ("\",\"", BTokenMap[**Index].Token);
      return (FALSE);
    }
  }
  return (TRUE);
}

bool HandleClass14 (int BasicLineNo, int StatementNo, int Keyword, byte **Index)

/**********************************************************************************************************************************/
/* Class 14 = One or more variables, separated by commas, must follow (READ)                                                      */
/**********************************************************************************************************************************/

{
  bool Type;
  bool More = TRUE;
  int  VarNameLen;

  while (More)
  {
    if (!ScanVariable (BasicLineNo, StatementNo, Keyword, Index, &Type, &VarNameLen, 2))                    /* We need a variable */
    {
      if (VarNameLen == 0)                                                                                    /* (Not a variable) */
        BADTOKEN ("variable", BTokenMap[**Index].Token);
      return (FALSE);
    }
    if (**Index == ':' || **Index == 0x0D)                                                   /* End of statement or end of line ? */
      More = FALSE;
    else if (**Index == ',')                                                                                       /* Separator ? */
      (*Index) ++;
    else
    {
      BADTOKEN ("\",\"", BTokenMap[**Index].Token);
      return (FALSE);
    }
  }
  return (TRUE);
}

bool HandleClass15 (int BasicLineNo, int StatementNo, int Keyword, byte **Index)

/**********************************************************************************************************************************/
/* Class 15 = DEF FN                                                                                                              */
/**********************************************************************************************************************************/

{
  bool Type;
  int  VarNameLen;

  if (!ScanVariable (BasicLineNo, StatementNo, Keyword, Index, &Type, &VarNameLen, -1))
  {
    if (VarNameLen == 0)
      BADTOKEN ("variable", BTokenMap[**Index].Token);
    return (FALSE);
  }
  if (VarNameLen != 1)                                                                                     /* Not single letter ? */
  {
    ReportError ("{BERROR in line %d, statement %d}B\n- Wrong variable type; must be single character", BasicLineNo, StatementNo);
    return (FALSE);
  }
  if (**Index == '(')                                                 /* Arguments to be passed to the expression while running ? */
  {
    (*Index) ++;
    if (CheckEnd (BasicLineNo, StatementNo, Index))
      return (FALSE);
    if (**Index == ')')
    {
      ReportError ("{BERROR in line %d, statement %d}B\n- Empty parameter array not allowed", BasicLineNo, StatementNo);
      return (FALSE);
    }
    while (**Index != ')')
    {
      if (!ScanVariable (BasicLineNo, StatementNo, Keyword, Index, &Type, &VarNameLen, -1))
      {
        if (VarNameLen == 0)
          BADTOKEN ("variable", BTokenMap[**Index].Token);
        return (FALSE);
      }
      if (VarNameLen != 1)                                                                                 /* Not single letter ? */
      {
        ReportError ("{BERROR in line %d, statement %d}B\n- Wrong variable type; must be single character",
                     BasicLineNo, StatementNo);
        return (FALSE);
      }
      if (CheckEnd (BasicLineNo, StatementNo, Index))
        return (FALSE);
      if (**Index != ')')
      {
        if (**Index == ',')
          (*Index) ++;
        else
        {
          BADTOKEN ("\",\"", BTokenMap[**Index].Token);
          return (FALSE);
        }
      }
    }
    (*Index) ++;
  }
  if (CheckEnd (BasicLineNo, StatementNo, Index))
    return (FALSE);
  if (**Index != '=')
  {
    BADTOKEN ("\"=\"", BTokenMap[**Index].Token);
    return (FALSE);
  }
  (*Index) ++;
  if (CheckEnd (BasicLineNo, StatementNo, Index))
    return (FALSE);
  return (ScanExpression (BasicLineNo, StatementNo, Keyword, Index, &Type, 0));                             /* Find an expression */
}

bool CheckSyntax (int BasicLineNo, byte *Line)

/**********************************************************************************************************************************/
/* Pre   : `Line' points to the converted BASIC line. An initial syntax check has been done already -                             */
/*         - The line number makes sense;                                                                                         */
/*         - Keywords are at the beginning of each statement and not within a statement;                                          */
/*         - There are less than 128 statements in the line;                                                                      */
/*         - Brackets match on a per-line basis (but not necessarily on a per-statement basis!)                                   */
/*         - Quotes match;                                                                                                        */
/* Post  : The line has been checked against 'normal' Spectrum BASIC syntax. Extended devices that change the normal syntax       */
/*         (such as Interface 1 or disk interfaces) are not understood and will generate error messages.                          */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  byte  StrippedLine[MAXLINELENGTH + 1];
  byte *StrippedIndex;
  byte  Keyword;
  bool  AllOk       = TRUE;
  bool  VarType;
  int   StatementNo = 0;
  int   ClassIndex  = -1;

  StrippedIndex = &(StrippedLine[0]);
  while (*Line != 0x0D)                                          /* First clean up the line, dropping number expansions and trash */
  {
    switch (*Line)
    {
      case  0 :
      case  1 :
      case  2 :
      case  3 :
      case  4 :
      case  5 :
      case  6 :
      case  7 :
      case  8 :
      case  9 :
      case 10 :
      case 11 :
      case 12 :
      case 13 : break;
      case 14 : *(StrippedIndex ++) = *Line; Line += 5; break;                 /* EXCEPTION: keep the marker, but drop the number */
      case 15 : break;
      case 16 :
      case 17 :
      case 18 :
      case 19 :
      case 20 :
      case 21 : Line ++; break;
      case 22 :
      case 23 : Line += 2; break;
      case 24 :
      case 25 :
      case 26 :
      case 27 :
      case 28 :
      case 29 :
      case 30 :
      case 31 :
      case 32 : break;                                                                      /* (We don't care for spaces either!) */
      default : *(StrippedIndex ++) = *Line; break;                                                   /* Pass on only 'good' bits */
    }
    Line ++;
  }
  *(StrippedIndex ++) = 0x0D;
  *StrippedIndex = '\0';
  StrippedIndex = &(StrippedLine[0]);                                                                         /* Ok, here goes... */
  while (AllOk && *StrippedIndex != 0x0D)                                                                /* Handle each statement */
  {
    StatementNo ++;
    Keyword = *(StrippedIndex ++);
    if (Keyword == 0xEA)                                                                                               /* 'REM' ? */
      return (TRUE);                                                                        /* Then we're done checking this line */
    if (BTokenMap[Keyword].TokenType != 0 && BTokenMap[Keyword].TokenType != 1 && BTokenMap[Keyword].TokenType != 2)  /* (Sanity) */
    {
      if (Keyword == 0xA9)                                                             /* EXCEPTION: POINT may be used as command */
      {
        if (*StrippedIndex != '#')                                                /* It must be followed by a stream in that case */
        {
          ReportError ("{BERROR}B - Keyword (\"%s\") error in line %d, statement %d",
                       BTokenMap[Keyword].Token, BasicLineNo, StatementNo);
          return (FALSE);
        }
        StrippedIndex ++;
        if (!ScanStream (BasicLineNo, StatementNo, Keyword, &StrippedIndex))       /* (Also signals Interface1/Opus specificness) */
          return (FALSE);
        if (*StrippedIndex != ';')
        {
          BADTOKEN ("\";\"", BTokenMap[*StrippedIndex].Token);
          return (FALSE);
        }
        StrippedIndex ++;
        if (!HandleClass06 (BasicLineNo, StatementNo, Keyword, &StrippedIndex))
          return (FALSE);
      }
      else
      {
        ReportError ("{BERROR}B - Keyword (\"%s\") error in line %d, statement %d",
                     BTokenMap[Keyword].Token, BasicLineNo, StatementNo);
        return (FALSE);
      }
    }
    else
    {
      ClassIndex = -1;
      if ((Keyword == 0xE1 || Keyword == 0xF0) && *StrippedIndex == '#')           /* EXCEPTION: LIST and LLIST may take a stream */
      {
        StrippedIndex ++;
        if (!ScanStream (BasicLineNo, StatementNo, Keyword, &StrippedIndex))       /* (Also signals Interface1/Opus specificness) */
          return (FALSE);
        if (*StrippedIndex != ':' && *StrippedIndex != 0x0D)                                       /* Line number is not required */
        {
          if (*StrippedIndex != ',')
          {
            BADTOKEN ("\",\"", BTokenMap[*StrippedIndex].Token);
            return (FALSE);
          }
          StrippedIndex ++;
        }
      }
      while (AllOk && BTokenMap[Keyword].KeywordClass[++ ClassIndex])                              /* Handle all class parameters */
      {
        if (*StrippedIndex == 0x0D)
        {
          if (BTokenMap[Keyword].KeywordClass[ClassIndex] != 3 &&                       /* Class 5 and 3 need 0 or more arguments */
              BTokenMap[Keyword].KeywordClass[ClassIndex] != 5)
          {
            if ((Keyword == 0xEB && BTokenMap[Keyword].KeywordClass[ClassIndex] == 0xCD) ||/* 'FOR' doesn't need 'STEP' parameter */
                (Keyword == 0xFC && BTokenMap[Keyword].KeywordClass[ClassIndex] == ',')) /* 'DRAW' doesn't need a third parameter */
              ClassIndex ++;
            else
            {
              ReportError ("{BERROR in line %d, statement %d}B\n- Unexpected end of line", BasicLineNo, StatementNo);
              AllOk = FALSE;
            }
          }
        }
        else if (BTokenMap[Keyword].KeywordClass[ClassIndex] >= 32)                                  /* Required token or class ? */
        {
          if (*StrippedIndex != BTokenMap[Keyword].KeywordClass[ClassIndex])                                  /* (Required token) */
          {
            if ((Keyword == 0xEB && BTokenMap[Keyword].KeywordClass[ClassIndex] == 0xCD && *StrippedIndex == ':') ||
                (Keyword == 0xFC && BTokenMap[Keyword].KeywordClass[ClassIndex] == ',' && *StrippedIndex == ':'))
              ClassIndex ++;                                            /* EXCEPTION: 'FOR' does not require the 'STEP' parameter */
                                                                        /* EXCEPTION: 'DRAW' does not require the third parameter */
            else
            {                                                                                                /* (Token not there) */
              ReportError ("{BERROR in line %d, statement %d}B\n- Expected \"%s\", but got \"%s\"",
                           BasicLineNo, StatementNo, BTokenMap[BTokenMap[Keyword].KeywordClass[ClassIndex]].Token,
                           BTokenMap[*StrippedIndex].Token);
              AllOk = FALSE;
            }
          }
          else
            StrippedIndex ++;
        }
        else                                                                                                   /* (Command class) */
          switch (BTokenMap[Keyword].KeywordClass[ClassIndex])
          {
            case  1 : AllOk = HandleClass01 (BasicLineNo, StatementNo, Keyword, &StrippedIndex, &VarType); break;
            case  2 : AllOk = HandleClass02 (BasicLineNo, StatementNo, Keyword, &StrippedIndex, VarType); break;
            case  3 : AllOk = HandleClass03 (BasicLineNo, StatementNo, Keyword, &StrippedIndex); break;
            case  4 : AllOk = HandleClass04 (BasicLineNo, StatementNo, Keyword, &StrippedIndex); break;
            case  5 : AllOk = HandleClass05 (BasicLineNo, StatementNo, Keyword, &StrippedIndex); break;
            case  6 : AllOk = HandleClass06 (BasicLineNo, StatementNo, Keyword, &StrippedIndex); break;
            case  7 : AllOk = HandleClass07 (BasicLineNo, StatementNo, Keyword, &StrippedIndex); break;
            case  8 : AllOk = HandleClass08 (BasicLineNo, StatementNo, Keyword, &StrippedIndex); break;
            case  9 : AllOk = HandleClass09 (BasicLineNo, StatementNo, Keyword, &StrippedIndex); break;
            case 10 : AllOk = HandleClass10 (BasicLineNo, StatementNo, Keyword, &StrippedIndex); break;
            case 11 : AllOk = HandleClass11 (BasicLineNo, StatementNo, Keyword, &StrippedIndex); break;
            case 13 : AllOk = HandleClass13 (BasicLineNo, StatementNo, Keyword, &StrippedIndex); break;
            case 14 : AllOk = HandleClass14 (BasicLineNo, StatementNo, Keyword, &StrippedIndex); break;
            case 15 : AllOk = HandleClass15 (BasicLineNo, StatementNo, Keyword, &StrippedIndex); break;
          }
      }
    }
    if (AllOk && Keyword != 0xFA)                                        /* Handling 'IF' and AllOk (i.e. just read the "THEN" ?) */
    {                                                                                        /* (Nope, go check end of statement) */
      if (*StrippedIndex != ':' && *StrippedIndex != 0x0D)
      {
        if (Keyword == 0xFB && *StrippedIndex == '#')                                            /* EXCEPTION: 'CLS #' is allowed */
        {
          StrippedIndex ++;
          if (!SignalInterface1 (BasicLineNo, StatementNo, 0))                              /* (But is Opus/Interface 1 specific) */
            return (FALSE);
          if (*StrippedIndex != ':' && *StrippedIndex != 0x0D)
          {
            ReportError ("{BERROR in line %d, statement %d}B\n- Expected end of statement, but got \"%s\"",
                         BasicLineNo, StatementNo, BTokenMap[*StrippedIndex].Token);
            AllOk = FALSE;
          }
        }
        else
        {
          ReportError ("{BERROR in line %d, statement %d}B\n- Expected end of statement, but got \"%s\"",
                       BasicLineNo, StatementNo, BTokenMap[*StrippedIndex].Token);
          AllOk = FALSE;
        }
      }
    }
    if (AllOk && *StrippedIndex == ':')               /* (Placing this check here allows weird (but legal) construction "THEN :") */
    {
      StrippedIndex ++;
      while (*StrippedIndex == ':')                                              /* (More consecutive ':' separators are allowed) */
      {
        StrippedIndex ++;
        StatementNo ++;
      }
    }
  }
  return (AllOk);
}

int _ConvertBASIC (char *FileName, word *ObjectSize)

/**********************************************************************************************************************************/
/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> MAIN PROGRAM <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
/* Import: MatchToken, HandleNumbers, ExpandSequences, PrepareLine, CheckSyntax.                                                  */
/* Returns:  1 = No errors, no warnings.                                                                                          */
/*           0 = No errors, some warnings.                                                                                        */
/*          -1 = Some errors.                                                                                                     */
/**********************************************************************************************************************************/

{
  struct MouseShape *OldShape;
  FILE *FpIn;
  char *BasicIndex;                                                        /* Current scan position in the (converted) ASCII line */
  byte *ResultIndex;                                                          /* Current write index in to the binary result line */
  byte  Token;
  int   LineCount        = 0;                                                                     /* Line count in the ASCII file */
  int   BasicLineNo;                                                                                 /* Current BASIC line number */
  int   SubLineCount;                                                                                 /* Current statement number */
  bool  ExpectKeyword;                                                       /* If TRUE, the next scanned token must be a keyword */
  bool  InString;                                                                                     /* TRUE while inside quotes */
  int   BracketCount     = 0;                                                               /* Match opening and closing brackets */
  int   ObjectLength;                                                                      /* Binary length of one converted line */
  word  BlockSize        = 0;                                                                      /* Total size of the TAP block */
  bool  AllOk            = TRUE;
  bool  EndOfFile        = FALSE;

  Is48KProgram   = -1;
  UsesInterface1 = -1;
  PreviousBasicLineNo = -1;
  NumErrors = NumWarnings = 0;
  ReInitComboWindow (&ConvertWindow, 0);
  if ((FpIn = fopen (FileName, "rt")) == NULL)
  {
    ReportFatal ("{BFATAL}B - Cannot open source file");
    return (-1);
  }
  BlockSize = 0;
  ReportInfo ("Converting BASIC...");
  OldShape = CurrentMouseShape;
  SetMouseShape (DefaultWaitingMouseShape);
  while (NumFatals == 0 && !EndOfFile)
  {
    if (fgets (BASICLineIn, MAXLINELENGTH + 1, FpIn) != NULL)
    {
      LineCount ++;
      if (strlen (BASICLineIn) >= MAXLINELENGTH)
      {                                                                                 /* We don't require an end-of-line marker */
        ReportFatal ("{BFATAL}B - Line %d too long", LineCount);
        AllOk = FALSE;
      }
      else if ((BasicLineNo = PrepareLine (BASICLineIn, LineCount, &BasicIndex)) < 0)
      {
        if (BasicLineNo == -1)                                                                                         /* (Error) */
          AllOk = FALSE;
        else                                                                                   /* (Line should simply be skipped) */
          ;
      }
      else if (BasicLineNo >= 10000)
      {
        ReportFatal ("{BFATAL}B - Line number %d is larger than the maximum allowed", BasicLineNo);
        AllOk = FALSE;
      }
      else
      {
        InString = FALSE;
        ExpectKeyword = TRUE;
        SubLineCount = 1;
        ResultIndex = ResultingLine + 4;                                              /* Reserve space for line number and length */
        while (*BasicIndex && AllOk)
        {
          if (InString)
          {
            if (*BasicIndex == '\"')
            {
              InString = FALSE;
              *(ResultIndex ++) = *(BasicIndex ++);
              while (*BasicIndex == ' ')                                                                  /* Skip trailing spaces */
                BasicIndex ++;
            }
            else
              switch (ExpandSequences (BasicLineNo, &BasicIndex, &ResultIndex, FALSE))
              {
                case -1 : AllOk = FALSE; break;                                                     /* (Error - already reported) */
                case  0 : *(ResultIndex ++) = *(BasicIndex ++); break;                                     /* (No expansion made) */
                case  1 : break;
              }
          }
          else if (*BasicIndex == '\"')
          {
            if (ExpectKeyword)
            {
              ReportError ("{BERROR in line %d, statement %d}B\n- Expected keyword but got quote", BasicLineNo, SubLineCount);
              AllOk = FALSE;
            }
            else
            {
              InString = TRUE;
              *(ResultIndex ++) = *(BasicIndex ++);
            }
          }
          else if (ExpectKeyword)
            switch (MatchToken (BasicLineNo, TRUE, &BasicIndex, &Token))
            {
              case -2 : AllOk = FALSE; break;                                                       /* (Error - already reported) */
              case -1 : ReportError ("{BERROR in line %d, statement %d}B\n- Expected keyword but got token \"%s\"",
                                     BasicLineNo, SubLineCount, BTokenMap[Token].Token);                         /* (Not keyword) */
                        AllOk = FALSE;
                        break;
              case  0 : ReportError ("{BERROR in line %d, statement %d}B\n- Expected keyword but got \"%s\"",       /* (No match) */
                                     BasicLineNo, SubLineCount, BTokenMap[(byte)(*BasicIndex)].Token);
                        AllOk = FALSE;
                        break;
              case  1 : *(ResultIndex ++) = Token;                                                             /* (Found keyword) */
                        if (Token != ':')                                                   /* Special exception; empty statement */
                          ExpectKeyword = FALSE;
                        if (Token == 0xEA)                                                              /* Special exception; REM */
                          while (*BasicIndex)                                 /* Simply copy over the remaining part of the line, */
                                                                                       /* disregarding token or number expansions */
                                                                  /* As brackets aren't tested for, the match counting stops here */
                                                              /* (a closing bracket in a REM statement will not be seen by BASIC) */
                            switch (ExpandSequences (BasicLineNo, &BasicIndex, &ResultIndex, FALSE))
                            {
                              case -1 : AllOk = FALSE; break;
                              case  0 : *(ResultIndex ++) = *(BasicIndex ++); break;
                              case  1 : break;
                            }
                        break;
            }
          else if (*BasicIndex == '(')                                                                         /* Opening bracket */
          {
            BracketCount ++;
            *(ResultIndex ++) = *(BasicIndex ++);
          }
          else if (*BasicIndex == ')')                                                                         /* Closing bracket */
          {
            if (-- BracketCount < 0)                                                        /* More closing than opening brackets */
            {
              ReportError ("{BERROR in line %d, statement %d}B\n- Too many closing brackets", BasicLineNo, SubLineCount);
              AllOk = FALSE;
            }
            else
              *(ResultIndex ++) = *(BasicIndex ++);
          }
          else
            switch (MatchToken (BasicLineNo, FALSE, &BasicIndex, &Token))
            {
              case -2 : AllOk = FALSE; break;                                                       /* (Error - already reported) */
              case -1 : ReportError ("{BERROR in line %d, statement %d}B\n- Unexpected keyword \"%s\"",    /* (Match but keyword) */
                                     BasicLineNo, SubLineCount, BTokenMap[Token].Token);
                        AllOk = FALSE;
                        break;
              case  0 : switch (HandleNumbers (BasicLineNo, &BasicIndex, &ResultIndex))                             /* (No token) */
                        {
                          case 0 :  switch (ExpandSequences (BasicLineNo, &BasicIndex, &ResultIndex, TRUE))        /* (No number) */
                                    {
                                      case -1 : AllOk = FALSE; break;                               /* (Error - already reported) */
                                      case  0 : if (isalpha (*BasicIndex))                                 /* (No expansion made) */
                                                  while (isalnum (*BasicIndex))                    /* Skip full strings in one go */
                                                    *(ResultIndex ++) = *(BasicIndex ++);
                                                else
                                                  *(ResultIndex ++) = *(BasicIndex ++);
                                                break;
                                      case  1 : break;
                                    }
                                    break;
                          case -1 : AllOk = FALSE; break;
                        }
                        break;
              case  1 : *(ResultIndex ++) = Token;                                                   /* (Found token, no keyword) */
                        if (Token == ':' || Token == 0xCB)
                        {
                          ExpectKeyword = TRUE;
                          if (BracketCount != 0)                                                          /* All brackets match ? */
                          {
                            ReportError ("{BERROR in line %d, statement %d}B\n- Too few closing brackets",
                                         BasicLineNo, SubLineCount);
                            AllOk = FALSE;
                          }
                          if (++ SubLineCount > 127)
                          {
                            ReportError ("{BERROR}B - Line %d has too many statements", BasicLineNo);
                            AllOk = FALSE;
                          }
                        }
                        break;
            }
        }
        *(ResultIndex ++) = 0x0D;
        if (AllOk && BracketCount != 0)                                                                   /* All brackets match ? */
        {
          ReportError ("{BERROR in line %d, statement %d}B\n- Too few closing brackets", BasicLineNo, SubLineCount);
          AllOk = FALSE;
        }
        if (AllOk && DoCheckSyntax)
          AllOk = CheckSyntax (BasicLineNo, ResultingLine + 4);                           /* Check the syntax of the decoded line */
        if (AllOk)
        {
          ObjectLength = (int)(ResultIndex - ResultingLine);
          ResultingLine[0] = (byte)(BasicLineNo >> 8);                                             /* Line number is put reversed */
          ResultingLine[1] = (byte)(BasicLineNo & 0xFF);
          ResultingLine[2] = (byte)((ObjectLength - 4) & 0xFF);                                 /* Make sure this runs on any CPU */
          ResultingLine[3] = (byte)((ObjectLength - 4) >> 8);
          if (BlockSize + ObjectLength > 41500)                                        /* (= 65368-23755-<some work/stack space>) */
          {
            ReportFatal ("{BFATAL}B - Object file too large at line %d!", BasicLineNo);
            AllOk = FALSE;
          }
          else
            memcpy (_ObjectData + BlockSize, ResultingLine, ObjectLength);
          BlockSize += ObjectLength;
        }
      }
    }
    else
      EndOfFile = TRUE;
  }
  if (AllOk)
  {
    ReportInfo ("\nDone! (%d %s)", NumWarnings, NumWarnings == 1 ? "warning" : "warnings");
    ReportInfo ("INFO - Listing contains %d %s", LineCount, LineCount == 1 ? "line" : "lines");
    ReportInfo ("INFO - BASIC size is %d bytes", BlockSize);
  }
  else
  {
    ReportInfo ("\n(%d %s, %d %s)", NumWarnings, NumWarnings == 1 ? "warning" : "warnings",
                                    NumErrors, NumErrors == 1 ? "error" : "errors");
    ReportInfo ("Listing as far as done contains %d %s", LineCount - 1, LineCount == 2 ? "line" : "lines");
  }
  if (Is48KProgram >= 0)
    ReportInfo ("INFO - Note: this program can only be used in %dK mode", Is48KProgram ? 48 : 128);
  switch (UsesInterface1)
  {
    case -1 : break;                                                                                           /* Neither of them */
    case 0  : ReportInfo ("INFO - Note: this program requires Interface 1 or Opus Discovery"); break;
    case 1  : ReportInfo ("INFO - Note: this program requires Interface 1"); break;
    case 2  : ReportInfo ("INFO - Note: this program requires an Opus Discovery"); break;
  }
  fclose (FpIn);
  *ObjectSize = BlockSize;
  SetMouseShape (OldShape);
  return (AllOk ? (NumWarnings == 0 ? 1 : 0) : -1);
}

bool EditBASICBlock (short WhichWindow, struct TapeEntry_s *Entry, char *FileName, word *BlockLength, word BASICLength,
                     bool Inserting)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points the the tape entry, `OldSize' is 0 if the entry has just     */
/*         been inserted, otherwise holds the length of the block.                                                                */
/* Post  : The user has edited this BASIC block. If the block was now being inserted, he has been presented a file requester to   */
/*         import a .BAS file. If that was escaped, an empty editor has been started so the user can supply his own text.         */
/*         If it was an existing piece of text, this text has been written out in a temporary text file, the editor has started   */
/*         and if the file had been changed, it is re-inserted into the XMS block.                                                */
/*         If anything has been changed (either now inserted or edited), the tape is adjusted already when leaving and TRUE is    */
/*         returned. Any error is reported here, in which case FALSE is returned. If the text block has not been changed, FALSE   */
/*         is returned as well. If it should be created now and no text has actually been inserted, the tape block has been       */
/*         removed already.                                                                                                       */
/* Import: _ConvertBASIC.                                                                                                         */
/**********************************************************************************************************************************/

{
  struct WindowInfo CWindow = {50, 100, 510, 315, 510, 315, 510, 315, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                               WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL, "BASIC Converter"};
  struct ButtonInfo PCancel = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 203, 290, 8, DBLACK, DWHITE, NULL, 0, "@Cancel", NULL};
  struct ButtonInfo POk     = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 282, 290, 8, DBLACK, DWHITE, NULL, 1, "@Ok", NULL};
  struct ButtonInfo BCase   = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 400, 30, 0, DBLACK, DWHITE, NULL, 10, NULL, NULL};
  struct ButtonInfo BCheck  = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 400, 50, 0, DBLACK, DWHITE, NULL, 11, NULL, NULL};
  struct ButtonInfo PConvrt = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 425, 80, 8, DBLACK, DWHITE, NULL, 3, "Con@vert", NULL};
  struct ButtonInfo PEdit   = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 425, 120, 8, DBLACK, DWHITE, NULL, 2, "@Edit", NULL};
  struct ButtonInfo *ButtonP;
  struct TapeEntry_s *PrevEntry;
  struct TapeEntry_s *DirP;
  struct TapeEntry_s *NewHeaderBlock;
  struct TapeRecord_s *Header;
  comboent          Dummy;
  bool              Done      = FALSE;
  bool              Cancelled = TRUE;
  bool              FileIsTmp = FALSE;
  FILE             *Fp;
  char              NoOptions = '\0';
  word              NewBlockLength;
  word              VARSLength;
  dword             XMSIndex;
  dword             PrevXMSIndex;
  word              Cnt;
  int               ConvertResult;

  BCase.Contents = &CaseSensitive;
  BCheck.Contents = &DoCheckSyntax;
  POk.Flags |= BF_GHOSTED;
  AddButton (&CWindow, &PConvrt, FALSE);
  AddButton (&CWindow, &PCancel, TRUE);
  AddButton (&CWindow, &POk, FALSE);
  AddButton (&CWindow, &PEdit, FALSE);
  AddButton (&CWindow, &BCase, FALSE);
  AddButton (&CWindow, &BCheck, FALSE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  MakeComboWindow (&ConvertWindow, &CWindow, COMT_VIEW, COMF_NOBUFFER | COMF_VSLIDER | COMF_SYNCSLIDER,
                   CWindow.StartX + 10, CWindow.StartY + 25, 0, 0, 70, 60, 25, NULL, FALSE, 0, NULL, DBLACK, DWHITE, 0, 0);
  Wprintf (&CWindow, 1, 1, "{+1Output Window}N");
  Wprintf (&CWindow, 3, 69, "{+1Case sensitive}N");
  Wprintf (&CWindow, 5, 69, "{+1Check syntax}N");

  FileIsTmp = FALSE;
  if (Inserting)                                                                                           /* Inserting a block ? */
  {
    Done = FALSE;
    strcpy (_Mask, "*.BAS");
    while (!Done)
    {
      if (!FileRequester (180, 50, 48, 29, 20, _Path, _File, _Mask, "Select BASIC File", TRUE, FALSE))          /* Ask for a file */
      {                                                           /* User doesn't want to insert a file, but rather something new */
        if (TPConfig.EditPath[0] == '\0')                                                                  /* No editor defined ? */
        {
          ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - No editor has been defined");
          DeleteTapeBlock (WhichWindow, Entry, FALSE, FALSE);
          goto __FILESETUPERROR;
        }
        MakeFullPath (_TempFileName, TmpDirectory, mktemp (NULL));
        strcpy (_TempFileName + strlen (_TempFileName) - 3, "BAS");
        FileIsTmp = TRUE;
        SpawnApplication (TPConfig.EditPath, &NoOptions, _TempFileName);                                      /* Start the editor */
        if (SwapReturn != SWAP_OK)                                                                      /* Something went wrong ? */
        {
          DeleteTapeBlock (WhichWindow, Entry, FALSE, FALSE);
          goto __FILESETUPERROR;
        }
        if ((Fp = fopen (_TempFileName, "rb")) == NULL)                                               /* Open the typed text file */
        {
          DeleteTapeBlock (WhichWindow, Entry, FALSE, FALSE);
          if (errno != ENOENT)                                                /* User simply aborted his editor without writing ? */
          {
            ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Cannot open %s\n      - %s", _TempFileName, StrError (errno));
            goto __FILESETUPERROR;
          }
        }
        Done = TRUE;
      }
      else
      {
        MakeFullPath (_TempFileName, _Path, _File);
        FileIsTmp = FALSE;
        if ((Fp = fopen (_TempFileName, "rb")) == NULL)
        {
          ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Cannot open %s\n      - %s", _File, StrError (errno));
          goto __FILESETUPERROR;
        }
        Done = TRUE;
      }
    }
  }
  else                                                                                        /* (Editing an existing text block) */
  {
    strcpy (_TempFileName, FileName);
    SpawnApplication (TPConfig.EditPath, &NoOptions, _TempFileName);                                          /* Start the editor */
    if (SwapReturn != SWAP_OK)                                                                          /* Something went wrong ? */
      goto __FILESETUPERROR;
    if ((Fp = fopen (_TempFileName, "rb")) == NULL)                                                /* Re-open the typed text file */
    {
      ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Cannot open %s\n      - %s", _TempFileName, StrError (errno));
      goto __FILESETUPERROR;
    }
  }
  fclose (Fp);

  Done = FALSE;
  Cancelled = FALSE;
  while (!Done)
  {
    switch (HandleWindow (&CWindow, &ButtonP))
    {
      case WINH_CLOSE : Done = TRUE; Cancelled = TRUE; break;
      case WINH_RNOHAND : switch (ButtonP->ReactNum)
                          {
                            case 0 : Done = TRUE; Cancelled = TRUE; break;                                            /* (Cancel) */
                            case 1 : Done = TRUE; Cancelled = FALSE; break;                                               /* (Ok) */
                            case 2 : SpawnApplication (TPConfig.EditPath, &NoOptions, _TempFileName);                   /* (Edit) */
                                     if (SwapReturn != SWAP_OK)                                         /* Something went wrong ? */
                                       goto __FILESETUPERROR;
                                     if ((Fp = fopen (_TempFileName, "rb")) == NULL)               /* Re-open the typed text file */
                                     {
                                       ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Cannot open %s\n      - %s",
                                                       _TempFileName, StrError (errno));
                                       goto __FILESETUPERROR;
                                     }
                                     fclose (Fp);
                                     break;
                            case 3 : switch ((ConvertResult = _ConvertBASIC (_TempFileName, &NewBlockLength)))       /* (Convert) */
                                     {
                                       case -1 : POk.Flags |= BF_GHOSTED; break;
                                       case  0 : 
                                       case  1 : POk.Flags &= ~BF_GHOSTED; break;
                                     }
                                     if (!Inserting && ConvertResult != -1)
                                     {
                                       VARSLength = Entry->BlockLength - BASICLength;
                                       if (VARSLength > 0)                           /* Did the previous block have a VARS area ? */
                                       {
                                         memcpy (_ObjectData + NewBlockLength, WorkBuffer + BASICLength, VARSLength);
                                         BASICLength = NewBlockLength;
                                         NewBlockLength += VARSLength;
                                         ReportInfo ("INFO - Appended old variables area of %d bytes", VARSLength);
                                       }
                                       else
                                       {
                                         ReportInfo ("INFO - Old variables area was empty");
                                         BASICLength = NewBlockLength;
                                       }
                                       ReportInfo ("INFO - LOAD size is %d bytes", NewBlockLength);
                                     }
                                     else
                                       BASICLength = NewBlockLength;
                                     UpdateButtonText (&CWindow, &POk, 0);
                                     break;
                          }
    }
    HandleComboWindow (&ConvertWindow, &Dummy);
    if (GlobalKey == K_ESC) { Done = TRUE; Cancelled = TRUE; }
    MouseStatus ();
  }
  __FILESETUPERROR:
  if (FileIsTmp)
    unlink (_TempFileName);
  if (!Cancelled)
  {
    if (Inserting)
    {
      XMSIndex = ResizeTapeBlock (WhichWindow, Entry, NewBlockLength + 20, 0);                   /* Create space for a header too */
      Header = (struct TapeRecord_s *)WorkBuffer;
      Header->HType = 0x00;                                                                                      /* (BASIC block) */
      memset (Header->HName, ' ', 10);
      Header->HLength = NewBlockLength;
      Header->HStart = 0;                                                                                    /* (Auto-start line) */
      Header->HExtra = BASICLength;
      if ((NewHeaderBlock = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
        FatalError (-1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
      NewHeaderBlock->BlockLength = 17;
      NewHeaderBlock->AllocLength = 20;
      NewHeaderBlock->BlockTypeID = 0x10;
      NewHeaderBlock->FlagByte = 0x00;                                                              /* Assume normal header block */
      NewHeaderBlock->Flags.FlagByteUsed = TRUE;
      NewHeaderBlock->Flags.ParityByteUsed = TRUE;
      NewHeaderBlock->Timing.LengthPilotPulse = NS_LENPILOT;                                              /* Make it normal speed */
      NewHeaderBlock->Timing.NumPilotPulses = NS_HDPILOT;
      NewHeaderBlock->Timing.LengthSyncPulse[0] = NS_LENSYNC1;
      NewHeaderBlock->Timing.LengthSyncPulse[1] = NS_LENSYNC2;
      NewHeaderBlock->Timing.LengthZeroBitPulse = NS_LEN0BIT;
      NewHeaderBlock->Timing.LengthOneBitPulse = NS_LEN1BIT;
      NewHeaderBlock->Timing.NumBitsInFirstByte = 8;
      NewHeaderBlock->Timing.NumBitsInLastByte = 8;
      NewHeaderBlock->Flags.IsNormalSpeed = TRUE;
      NewHeaderBlock->Flags.IsHeader = TRUE;
      GetBlockTypeInfo (NewHeaderBlock->BlockTypeID, &(NewHeaderBlock->IsTapeData), &(NewHeaderBlock->IsIdentifiedData),
                        NULL, NULL, NULL);
      NewHeaderBlock->CalculatedParity = NewHeaderBlock->FlagByte;                                             /* Init the parity */
      for (Cnt = 0 ; Cnt < 17 ; Cnt ++)
        NewHeaderBlock->CalculatedParity ^= *(WorkBuffer + Cnt);                                          /* Calculate the parity */
      NewHeaderBlock->ParityByte = NewHeaderBlock->CalculatedParity;                                 /* Make the parity 'correct' */
      if (TapeWindow[WhichWindow].Directory == Entry)                                      /* Inserted at the start of the tape ? */
      {
        NewHeaderBlock->Next = Entry;
        TapeWindow[WhichWindow].Directory = NewHeaderBlock;
      }
      else
      {
        PrevEntry = NULL;
        DirP = TapeWindow[WhichWindow].Directory;                                                          /* Find previous entry */
        while (DirP != Entry)
        {
          PrevEntry = DirP;
          DirP = DirP->Next;
        }
        NewHeaderBlock->Next = Entry;                                                                  /* Insert the header block */
        PrevEntry->Next = NewHeaderBlock;
      }
      Entry->BlockLength -= 20;                                                              /* (Correct for the inserted header) */
      Entry->AllocLength -= 20;
      Entry->BlockTypeID = 0x10;
      Entry->FlagByte = 0xFF;                                                                         /* Assume normal data block */
      Entry->Flags.FlagByteUsed = TRUE;
      Entry->Flags.ParityByteUsed = TRUE;
      Entry->Timing.LengthPilotPulse = NS_LENPILOT;                                                       /* Make it normal speed */
      Entry->Timing.NumPilotPulses = NS_DTPILOT;
      Entry->Timing.LengthSyncPulse[0] = NS_LENSYNC1;
      Entry->Timing.LengthSyncPulse[1] = NS_LENSYNC2;
      Entry->Timing.LengthZeroBitPulse = NS_LEN0BIT;
      Entry->Timing.LengthOneBitPulse = NS_LEN1BIT;
      Entry->Timing.NumBitsInFirstByte = 8;
      Entry->Timing.NumBitsInLastByte = 8;
      Entry->Flags.IsNormalSpeed = TRUE;
      GetBlockTypeInfo (Entry->BlockTypeID, &(Entry->IsTapeData), &(Entry->IsIdentifiedData), NULL, NULL, NULL);
      Entry->CalculatedParity = Entry->FlagByte;
      for (Cnt = 0 ; Cnt < NewBlockLength ; Cnt ++)
        Entry->CalculatedParity ^= *(_ObjectData + Cnt);
      Entry->ParityByte = Entry->CalculatedParity;
      CopyToXMS (20, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, XMSIndex);                     /* Insert the header data */
      CopyToXMS (NewBlockLength, TapeWindow[WhichWindow].Tape.XMSHandle, _ObjectData, XMSIndex + 20);    /* Insert the BASIC data */
      EditHeaderBlock (WhichWindow, NewHeaderBlock);                                             /* Let the user adjust all items */
    }
    else                                                                                     /* (Editing an existing BASIC block) */
    {
      XMSIndex = ResizeTapeBlock (WhichWindow, Entry, NewBlockLength, 0);             /* Adjust the data area for this tape entry */
      CopyToXMS (NewBlockLength, TapeWindow[WhichWindow].Tape.XMSHandle, _ObjectData, XMSIndex);           /* Substitute the data */
      PrevEntry = NULL;
      if (TapeWindow[WhichWindow].Directory != Entry)                                /* This is not the first entry in the tape ? */
      {
        DirP = TapeWindow[WhichWindow].Directory;                                                          /* Find previous entry */
        while (DirP != Entry)
        {
          PrevEntry = DirP;
          DirP = DirP->Next;
        }
        if (!PrevEntry->Flags.IsHeader)                                                 /* Previous entry is not a header block ? */
          PrevEntry = NULL;
      }
      if (PrevEntry != NULL)
      {
        if (!QuestionRequester (-1, -1, FALSE, FALSE, NULL, NULL, NULL, NULL, NULL, NULL,
                                "Do you want to adjust the header block for this BASIC block as well?"))
        {
          PrevXMSIndex = XMSIndex - PrevEntry->AllocLength;
          CopyFromXMS (20, TapeWindow[WhichWindow].Tape.XMSHandle, PrevXMSIndex, WorkBuffer);
          Header = (struct TapeRecord_s *)WorkBuffer;
          Header->HLength = NewBlockLength;
          Header->HExtra = BASICLength;
          CopyToXMS (20, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, PrevXMSIndex);
          PrevEntry->CalculatedParity = PrevEntry->FlagByte;                                              /* (Correct its parity) */
          for (Cnt = 0 ; Cnt < 17 ; Cnt ++)
            PrevEntry->CalculatedParity ^= *(WorkBuffer + Cnt);
          PrevEntry->ParityByte = PrevEntry->CalculatedParity;
        }
      }
      else
        ErrorRequester (-1, -1, NULL, WBPics[2], "Notice", "Don't forget to insert/adjust the header block with the following "
                        "information:\n- Block length = %d bytes\n- BASIC length = %d", NewBlockLength, BASICLength);
    }
    *BlockLength = NewBlockLength;
    if (!Entry->Flags.FlagByteUsed)                                                              /* Make sure we have a flag byte */
    {
      Entry->Flags.FlagByteUsed = TRUE;
      Entry->FlagByte = 0xFF;
    }
    Entry->Flags.ParityByteUsed = TRUE;                                                 /* Create/update the parity matching byte */
    Entry->CalculatedParity = Entry->FlagByte;
    for (Cnt = 0 ; Cnt < NewBlockLength ; Cnt ++)
      Entry->CalculatedParity ^= *(_ObjectData + Cnt);
    Entry->ParityByte = Entry->CalculatedParity;
  }
  DestroyComboWindow (&ConvertWindow, 0);
  DestroyWindow (&CWindow);
  GlobalKey = 0x0000;
  return (!Cancelled);
}
