/**********************************************************************************************************************************/
/* Project      : TAPER                                                                                                           */
/* Module       : TPMAIN.H                                                                                                        */
/* Description  : Main program parts                                                                                              */
/* Version type : Main program interface                                                                                          */
/* Last changed : 27-12-1998  14:30                                                                                               */
/* Update count : 11                                                                                                              */
/* OS type      : PC (DOS)                                                                                                        */
/* Compiler     : Watcom C32 Optimizing Compiler Version 10.x                                                                     */
/* Linker       : Watcom C Linker                                                                                                 */
/*                                                                                                                                */
/*                Copyleft (C) 1997-2001 ThunderWare Research Center, written by Martijn van der Heide.                           */
/*                                                                                                                                */
/*                This program is free software; you can redistribute it and/or                                                   */
/*                modify it under the terms of the GNU General Public License                                                     */
/*                as published by the Free Software Foundation; either version 2                                                  */
/*                of the License, or (at your option) any later version.                                                          */
/*                                                                                                                                */
/*                This program is distributed in the hope that it will be useful,                                                 */
/*                but WITHOUT ANY WARRANTY; without even the implied warranty of                                                  */
/*                MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                                                   */
/*                GNU General Public License for more details.                                                                    */
/*                                                                                                                                */
/*                You should have received a copy of the GNU General Public License                                               */
/*                along with this program; if not, write to the Free Software                                                     */
/*                Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                                     */
/*                                                                                                                                */
/**********************************************************************************************************************************/

#include "tptape.h"

#ifndef TPMAIN_H_INC
#define TPMAIN_H_INC

#ifdef EXTERN
#undef EXTERN
#endif
#ifdef __TPMAIN_MAIN__
#define EXTERN
#else
#define EXTERN extern
#endif

#define SELECTBIT       0x02
#define EXPANDBIT       0x01
#define HEADERBLOCK     0x10                                                  /* Used in TAP indexes, TRUE = HEADER, FALSE = DATA */

#define TWLEFT          0
#define TWRIGHT         1

/**********************************************************************************************************************************/
/* Define the global variables                                                                                                    */
/**********************************************************************************************************************************/

struct TapeWindow_s
{
  struct ComboInfo        IndexWindow;
  comboent                SelectedLine;
  comboent                NumEntries;
  word                    TapeVersion;                                                                      /* 0x0000 if not used */
  int                     FileType;
  bool                    NoTape;
  bool                    TapeChanged;
  bool                    NewTape;
  bool                    EmptyTape;
  bool                    NonexistingTape;
  bool                    TapeFloppyId;
  bool                    BackupTape;
  bool                    ModeIs128K;
  bool                    AutoDecryptBlocks;
  bool                    DecryptAnything;
  char                    TapeName[MAXDIR];
  char                    BackupName[MAXDIR];
  char                    TempName[MAXDIR];
  char                    ConvTapeName[MAXDIR + 6];
  word                    NextJumpCode;
  struct Tape_s           Tape;                                                                     /* Intermediate format in XMS */
  struct TapeEntry_s     *Directory;
};

EXTERN struct TapeWindow_s   TapeWindow[2];                                                      /* (Indexed with TWLEFT/TWRIGHT) */

EXTERN struct FontInfo      *SpectrumFont;
EXTERN struct FontInfo      *SpectrumSmallFont;
EXTERN bool                  _HaveMixer;

EXTERN short                 GlobalDecodingScheme;                             /* Used to decrypt PURE DATA blocks for SpeedLocks */

/**********************************************************************************************************************************/
/* Define the function prototypes                                                                                                 */
/**********************************************************************************************************************************/

void  SurroundScreen       (struct WindowInfo *Window, short StartX, short StartY);
void  LowMemHandler        (int MemoryType, char *ErrorMessage);
void  AutoInsertNewTape    (short WhichWindow);
bool  TestFileOverwrite    (char *Path);
void  SelectTapeEntry      (short WhichWindow, comboent WhichEntry);
void  SelectNewTapeEntry   (short WhichWindow, comboent WhichEntry);
short GetFileType          (short WhichWindow);
void  SignalTapeMode       (short WhichWindow, bool ModeIs128K);
void  SignalTapeChanged    (short WhichWindow, bool Changed);
void  GhostEmptyTape       (short WhichWindow);
void  UnghostTape          (short WhichWindow);
void  SetTapeName          (short WhichWindow, char *FileName, bool UpdateButton);
void  DeselectAll          (short WhichWindow);
void  UpdateTapeSize       (short WhichWindow, long NewSize);

#endif
