/**********************************************************************************************************************************/
/* Project      : TAPER                                                                                                           */
/* Module       : TPMENU.C                                                                                                        */
/* Description  : Menu stuff                                                                                                      */
/* Version type : Program module                                                                                                  */
/* Last changed : 14-07-1999  7:05                                                                                                */
/* Update count : 63                                                                                                              */
/* OS type      : PC (DOS)                                                                                                        */
/* Compiler     : Watcom C32 Optimizing Compiler Version 10.x                                                                     */
/* Linker       : Watcom C Linker                                                                                                 */
/* Notes        : The BlockCRC routine was converted from 16-bit assembler and is copyright (C) Claus Jahn.                       */
/*                                                                                                                                */
/*                Copyleft (C) 1997-2001 ThunderWare Research Center, written by Martijn van der Heide.                           */
/*                                                                                                                                */
/*                This program is free software; you can redistribute it and/or                                                   */
/*                modify it under the terms of the GNU General Public License                                                     */
/*                as published by the Free Software Foundation; either version 2                                                  */
/*                of the License, or (at your option) any later version.                                                          */
/*                                                                                                                                */
/*                This program is distributed in the hope that it will be useful,                                                 */
/*                but WITHOUT ANY WARRANTY; without even the implied warranty of                                                  */
/*                MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                                                   */
/*                GNU General Public License for more details.                                                                    */
/*                                                                                                                                */
/*                You should have received a copy of the GNU General Public License                                               */
/*                along with this program; if not, write to the Free Software                                                     */
/*                Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                                     */
/*                                                                                                                                */
/**********************************************************************************************************************************/

#define __TPMENU_MAIN__

#include "wbench32/wbench32.h"
#include <dos.h>
#include <fcntl.h>
#ifdef __WATCOMC__
#include <io.h>
#endif
#include "tpconfig.h"
#include "tpscreen.h"
#include "tpmain.h"
#include "tpfile.h"
#include "tpaudio.h"
#include "tptape.h"
#include "tpedit.h"
#include "tpdisass.h"
#include "tpmenu.h"

#pragma pack          ()

/**********************************************************************************************************************************/
/* Define the things we need even if the menu strip is not linked                                                                 */
/**********************************************************************************************************************************/

static struct _BlockCRC_s
{
  word FloodIndex;
  word Sum;
  word FloodSum;
  byte XORMask;
  word XORSum;
};

static char            _Path[MAXDIR + 1];
static char            _File[13];
static char            _Mask[13]                = "*.*";
static short           _ActiveWindowMenu        = -1;

static char            _PreviousTapePath[2][MAXDIR + 1] = {{"\0"}, {"\0"}};

static struct WindowInfo _CalcWindow = {230, 200, 180, 54, 180, 54, 180, 54, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                        WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER | WINF_NOMOVE, NULL,
                                        "Calculating Play Time"};

/**********************************************************************************************************************************/
/* Define all static routines - the handlers for the menu strip entries                                                           */
/**********************************************************************************************************************************/

static void _GetBlockCRC         (short WhichWindow, struct TapeEntry_s *Entry, struct _BlockCRC_s *CRC);
static char _MatchBlockType      (struct TapeEntry_s *Entry1, struct TapeEntry_s *Entry2);
static bool _TestSelected        (short WhichWindow);
static void _SetupTapePath       (short WhichWindow);
static int  _HandleMENew         (struct MenuEntry *Entry);
static int  _HandleMEOpen        (struct MenuEntry *Entry);
static int  _HandleMESaveAs      (struct MenuEntry *Entry);
static int  _HandleMESave        (struct MenuEntry *Entry);
static int  _HandleMESampler     (struct MenuEntry *Entry);
static int  _HandleMEImport      (struct MenuEntry *Entry);
static int  _HandleMETapeCRC     (struct MenuEntry *Entry);
static int  _HandleMEPlayTime    (struct MenuEntry *Entry);
static int  _HandleMEMtchTapes   (struct MenuEntry *Entry);
static int  _HandleMESelNone     (struct MenuEntry *Entry);
static int  _HandleMEInvSel      (struct MenuEntry *Entry);
static int  _HandleMEDelEntry    (struct MenuEntry *Entry);
static int  _HandleMEDelSel      (struct MenuEntry *Entry);
// static int  _HandleMEResizeUp    (struct MenuEntry *Entry);
// static int  _HandleMEResizeDn    (struct MenuEntry *Entry);
static int  _HandleMEDupEntry    (struct MenuEntry *Entry);
static int  _HandleMEPlayEntry   (struct MenuEntry *Entry);
static int  _HandleMEPlaySel     (struct MenuEntry *Entry);
static int  _HandleMEPlayFrom    (struct MenuEntry *Entry);
static int  _HandleMEPlayTape    (struct MenuEntry *Entry);
static int  _HandleMESaveRaw     (struct MenuEntry *Entry);
static int  _HandleMEDecDirRec   (struct MenuEntry *Entry);
static int  _HandleMEMakeGroup   (struct MenuEntry *Entry);
static int  _HandleMESelGetPlTim (struct MenuEntry *Entry);
static int  _HandleMESelCheckCRC (struct MenuEntry *Entry);
static int  _HandleMESetDecrypt  (struct MenuEntry *Entry);
static int  _HandleMEMatchThis   (struct MenuEntry *Entry);
static int  _HandleMEMatchOther  (struct MenuEntry *Entry);
static int  _HandleMEBlockCRC    (struct MenuEntry *Entry);
static int  _HandleMESoundVol    (struct MenuEntry *Entry);

/**********************************************************************************************************************************/
/* Define the menu strip entries                                                                                                  */
/**********************************************************************************************************************************/

struct MenuEntry MELNew       = {'n', TWLEFT, 0, 0, 0, _HandleMENew,        "New"};
struct MenuEntry MERNew       = {'n', TWRIGHT, 0, 0, 0, _HandleMENew,       "New"};
struct MenuEntry MELOpen      = {'j', TWLEFT, 0, 0, 0, _HandleMEOpen,       "Open ..."};
struct MenuEntry MEROpen      = {'j', TWRIGHT, 0, 0, 0, _HandleMEOpen,      "Open ..."};
struct MenuEntry MELSaveAs    = {'w', TWLEFT, 0, 0, 0, _HandleMESaveAs,     "Save as ..."};
struct MenuEntry MERSaveAs    = {'w', TWRIGHT, 0, 0, 0, _HandleMESaveAs,    "Save as ..."};
struct MenuEntry MELSave      = {'s', TWLEFT, 0, 0, 0, _HandleMESave,       "Save"};
struct MenuEntry MERSave      = {'s', TWRIGHT, 0, 0, 0, _HandleMESave,      "Save"};
struct MenuEntry MELSampler   = {'`', TWLEFT, 0, 0, 0, _HandleMESampler,    "Sampler"};
struct MenuEntry MERSampler   = {'`', TWRIGHT, 0, 0, 0, _HandleMESampler,   "Sampler"};
struct MenuEntry MELImport    = {'m', TWLEFT, 0, 0, 0, _HandleMEImport,     "Import tape"};
struct MenuEntry MERImport    = {'m', TWRIGHT, 0, 0, 0, _HandleMEImport,    "Import tape"};
struct MenuEntry MELTapeCRC   = {0, TWLEFT, 0, 0, 0, _HandleMETapeCRC,      "Get tape CRC"};
struct MenuEntry MERTapeCRC   = {0, TWRIGHT, 0, 0, 0, _HandleMETapeCRC,     "Get tape CRC"};
struct MenuEntry MELPlayTime  = {0, TWLEFT, 0, 0, 0, _HandleMEPlayTime,     "Get tape play time"};
struct MenuEntry MERPlayTime  = {0, TWRIGHT, 0, 0, 0, _HandleMEPlayTime,    "Get tape play time"};
struct MenuEntry MESeperate   = {0, 0, MEF_SEPERATOR, 0, 0, NULL, ""};
struct MenuEntry MELAutoDecr  = {0, TWLEFT, MEF_BOOL | MEF_BOOLCLEAR,
                       &(TapeWindow[TWLEFT].AutoDecryptBlocks), 0, NULL,    "Auto-decrypt"};
struct MenuEntry MERAutoDecr  = {0, TWRIGHT, MEF_BOOL | MEF_BOOLCLEAR,
                      &(TapeWindow[TWRIGHT].AutoDecryptBlocks), 0, NULL,    "Auto-decrypt"};
struct MenuEntry MELDecrAny   = {0, TWLEFT, MEF_BOOL | MEF_BOOLCLEAR,
                       &(TapeWindow[TWLEFT].DecryptAnything), 0, NULL,      "Decrypt all"};
struct MenuEntry MERDecrAny   = {0, TWRIGHT, MEF_BOOL | MEF_BOOLCLEAR,
                      &(TapeWindow[TWRIGHT].DecryptAnything), 0, NULL,      "Decrypt all"};
struct MenuEntry MELSetDecr   = {0, TWLEFT, 0, 0, 0, _HandleMESetDecrypt,   "Set values ..."};
struct MenuEntry MERSetDecr   = {0, TWRIGHT, 0, 0, 0, _HandleMESetDecrypt,  "Set values ..."};
struct MenuEntry MELBackup    = {0, TWLEFT, MEF_BOOL | MEF_BOOLCLEAR,
                              &(TapeWindow[TWLEFT].BackupTape), 0, NULL,    "Backup tape"};
struct MenuEntry MERBackup    = {0, TWRIGHT, MEF_BOOL | MEF_BOOLCLEAR,
                             &(TapeWindow[TWRIGHT].BackupTape), 0, NULL,    "Backup tape"};
struct MenuEntry MEMtchTapes  = {0, 0, 0, 0, 0, _HandleMEMtchTapes,         "Match tapes"};
struct MenuEntry MEQuit       = {'q', 0xFF, MEF_KEYSELECT, 0, 0, NULL,      "Quit"};

struct MenuEntry MELSelNone   = {'a', TWLEFT, 0, 0, 0, _HandleMESelNone,    "Deselect all"};
struct MenuEntry MERSelNone   = {'a', TWRIGHT, 0, 0, 0, _HandleMESelNone,   "Deselect all"};
struct MenuEntry MELInvSel    = {'v', TWLEFT, 0, 0, 0, _HandleMEInvSel,     "Invert selected"};
struct MenuEntry MERInvSel    = {'v', TWRIGHT, 0, 0, 0, _HandleMEInvSel,    "Invert selected"};
struct MenuEntry MELInsert    = {'i', TWLEFT | 0xF0, 0, 0, 0, NULL,         "Insert ..."};
struct MenuEntry MERInsert    = {'i', TWRIGHT | 0xF0, 0, 0, 0, NULL,        "Insert ..."};
struct MenuEntry MELSaveRaw   = {0, TWLEFT, 0, 0, 0, _HandleMESaveRaw,      "Save raw data"};
struct MenuEntry MERSaveRaw   = {0, TWRIGHT, 0, 0, 0, _HandleMESaveRaw,     "Save raw data"};
struct MenuEntry MELDelete    = {0, TWLEFT, 0, 0, 0, NULL,                  "Delete"};
struct MenuEntry MERDelete    = {0, TWRIGHT, 0, 0, 0, NULL,                 "Delete"};
struct MenuEntry MELDelEntry  = {'d', TWLEFT, 0, 0, 0, _HandleMEDelEntry,     "Entry"};
struct MenuEntry MERDelEntry  = {'d', TWRIGHT, 0, 0, 0, _HandleMEDelEntry,    "Entry"};
struct MenuEntry MELDelSel    = {0, TWLEFT, 0, 0, 0, _HandleMEDelSel,         "Selected"};
struct MenuEntry MERDelSel    = {0, TWRIGHT, 0, 0, 0, _HandleMEDelSel,        "Selected"};
// struct MenuEntry MELResize    = {0, TWLEFT, 0, 0, 0, NULL,                  "Resize"};
// struct MenuEntry MERResize    = {0, TWRIGHT, 0, 0, 0, NULL,                 "Resize"};
// struct MenuEntry MELResizeUp  = {0, TWLEFT, 0, 0, 0, _HandleMEResizeUp,       "Expand"};
// struct MenuEntry MERResizeUp  = {0, TWRIGHT, 0, 0, 0, _HandleMEResizeUp,      "Expand"};
// struct MenuEntry MELResizeDn  = {0, TWLEFT, 0, 0, 0, _HandleMEResizeDn,       "Shrink"};
// struct MenuEntry MERResizeDn  = {0, TWRIGHT, 0, 0, 0, _HandleMEResizeDn,      "Shrink"};
struct MenuEntry MELDupEntry  = {'u', TWLEFT, 0, 0, 0, _HandleMEDupEntry,   "Duplicate entry"};
struct MenuEntry MERDupEntry  = {'u', TWRIGHT, 0, 0, 0, _HandleMEDupEntry,  "Duplicate entry"};
struct MenuEntry MELDecDirRec = {0, TWLEFT, 0, 0, 0, _HandleMEDecDirRec,    "Decode entry"};
struct MenuEntry MERDecDirRec = {0, TWRIGHT, 0, 0, 0, _HandleMEDecDirRec,   "Decode entry"};
struct MenuEntry MELEdHeader  = {'t', TWLEFT | 0xA0, 0, 0, 0, NULL,         "Edit header"};
struct MenuEntry MEREdHeader  = {'t', TWRIGHT | 0xA0, 0, 0, 0, NULL,        "Edit header"};
struct MenuEntry MELMtchBlock = {0, TWLEFT, 0, 0, 0, NULL,                  "Find match"};
struct MenuEntry MERMtchBlock = {0, TWRIGHT, 0, 0, 0, NULL,                 "Find match"};
struct MenuEntry MELMtchThis  = {'[', TWLEFT, 0, 0, 0, _HandleMEMatchThis,       "Same tape"};
struct MenuEntry MERMtchThis  = {'[', TWRIGHT, 0, 0, 0, _HandleMEMatchThis,      "Same tape"};
struct MenuEntry MELMtchOther = {']', TWLEFT, 0, 0, 0, _HandleMEMatchOther,      "Other tape"};
struct MenuEntry MERMtchOther = {']', TWRIGHT, 0, 0, 0, _HandleMEMatchOther,     "Other tape"};
struct MenuEntry MELBlockCRC  = {'\\', TWLEFT, 0, 0, 0, _HandleMEBlockCRC,       "Get CRC"};
struct MenuEntry MERBlockCRC  = {'\\', TWRIGHT, 0, 0, 0, _HandleMEBlockCRC,      "Get CRC"};
struct MenuEntry MELBlocks    = {0, TWLEFT, 0, 0, 0, NULL,                  "Selected block"};
struct MenuEntry MERBlocks    = {0, TWRIGHT, 0, 0, 0, NULL,                 "Selected block"};
struct MenuEntry MELMakeGroup = {'z', TWLEFT, 0, 0, 0, _HandleMEMakeGroup,       "Make group"};
struct MenuEntry MERMakeGroup = {'z', TWRIGHT, 0, 0, 0, _HandleMEMakeGroup,      "Make group"};
struct MenuEntry MELMakeLoop  = {0, TWLEFT | 0x02, 0, 0, 0, _HandleMEMakeGroup,  "Make loop"};
struct MenuEntry MERMakeLoop  = {0, TWRIGHT | 0x02, 0, 0, 0, _HandleMEMakeGroup, "Make loop"};
struct MenuEntry MELEdAllTim  = {0, TWLEFT | 0xC0, 0, 0, 0, NULL,                "Edit timing"};
struct MenuEntry MEREdAllTim  = {0, TWRIGHT | 0xC0, 0, 0, 0, NULL,               "Edit timing"};
struct MenuEntry MELSGetPlTim = {0, TWLEFT, 0, 0, 0, _HandleMESelGetPlTim,       "Get play time"};
struct MenuEntry MERSGetPlTim = {0, TWRIGHT, 0, 0, 0, _HandleMESelGetPlTim,      "Get play time"};
struct MenuEntry MELSCheckCRC = {0, TWLEFT, 0, 0, 0, _HandleMESelCheckCRC,       "Group parity"};
struct MenuEntry MERSCheckCRC = {0, TWRIGHT, 0, 0, 0, _HandleMESelCheckCRC,      "Group parity"};
struct MenuEntry MELPlay      = {0, TWLEFT, 0, 0, 0, NULL,                  "Play"};
struct MenuEntry MERPlay      = {0, TWRIGHT, 0, 0, 0, NULL,                 "Play"};
struct MenuEntry MELPlayEntry = {0, TWLEFT, 0, 0, 0, _HandleMEPlayEntry,      "Entry"};
struct MenuEntry MERPlayEntry = {0, TWRIGHT, 0, 0, 0, _HandleMEPlayEntry,     "Entry"};
struct MenuEntry MELPlaySel   = {0, TWLEFT, 0, 0, 0, _HandleMEPlaySel,        "Selected"};
struct MenuEntry MERPlaySel   = {0, TWRIGHT, 0, 0, 0, _HandleMEPlaySel,       "Selected"};
struct MenuEntry MELPlayFrom  = {'f', TWLEFT, 0, 0, 0, _HandleMEPlayFrom,     "From entry"};
struct MenuEntry MERPlayFrom  = {'f', TWRIGHT, 0, 0, 0, _HandleMEPlayFrom,    "From entry"};
struct MenuEntry MELPlayTape  = {'p', TWLEFT, 0, 0, 0, _HandleMEPlayTape,     "All tape"};
struct MenuEntry MERPlayTape  = {'p', TWRIGHT, 0, 0, 0, _HandleMEPlayTape,    "All tape"};
struct MenuEntry MELViewAs    = {0, TWLEFT, 0, 0, 0, NULL,                  "View as"};
struct MenuEntry MERViewAs    = {0, TWRIGHT, 0, 0, 0, NULL,                 "View as"};
struct MenuEntry MELViewDump  = {'e', TWLEFT | 0x10, 0, 0, 0, NULL,           "Hex editor"};
struct MenuEntry MERViewDump  = {'e', TWRIGHT | 0x10, 0, 0, 0, NULL,          "Hex editor"};
struct MenuEntry MELViewHead  = {'g', TWLEFT | 0x20, 0, 0, 0, NULL,           "Header"};
struct MenuEntry MERViewHead  = {'g', TWRIGHT | 0x20, 0, 0, 0, NULL,          "Header"};
struct MenuEntry MELViewBasic = {'l', TWLEFT | 0x30, 0, 0, 0, NULL,           "BASIC listing"};
struct MenuEntry MERViewBasic = {'l', TWRIGHT | 0x30, 0, 0, 0, NULL,          "BASIC listing"};
struct MenuEntry MELViewScrn  = {'r', TWLEFT | 0x40, 0, 0, 0, NULL,           "Screen image"};
struct MenuEntry MERViewScrn  = {'r', TWRIGHT | 0x40, 0, 0, 0, NULL,          "Screen image"};
struct MenuEntry MELViewDiss  = {'/', TWLEFT | 0x50, 0, 0, 0, NULL,           "Disassembly"};
struct MenuEntry MERViewDiss  = {'/', TWRIGHT | 0x50, 0, 0, 0, NULL,          "Disassembly"};
struct MenuEntry MELViewInfo  = {0, TWLEFT | 0x08, 0, 0, 0, NULL,           "View entry info"};
struct MenuEntry MERViewInfo  = {0, TWRIGHT | 0x08, 0, 0, 0, NULL,          "View entry info"};

struct MenuEntry MEQuiet      = {0, 0, MEF_BOOL | MEF_BOOLCLEAR,
                                         &SystemDefaults.Quiet, 0, NULL,   "Disable sounds"};
struct MenuEntry MESoundVol   = {0, 0, 0, 0, 0, _HandleMESoundVol,         "Sound volume"};
struct MenuEntry MEEmulFlsh   = {0, 0, MEF_BOOL | MEF_BOOLCLEAR,
                                        &TPConfig.EmulateFlash, 0, NULL,   "Emulate flash"};
struct MenuEntry MESlowFlsh   = {0, 0, MEF_BOOL | MEF_BOOLCLEAR,
                                           &TPConfig.SlowFlash, 0, NULL,   "Slow flash"};
struct MenuEntry METraceSnd   = {0, 0, MEF_BOOL | MEF_BOOLCLEAR,
                                    &TPConfig.TraceSoundBlocks, 0, NULL,   "Trace SB blocks"};
struct MenuEntry MEShowTime   = {0, 0, MEF_BOOL | MEF_BOOLCLEAR,
                                 &TPConfig.ShowRemainingPlayTime, 0, NULL, "Show play time"};

/**********************************************************************************************************************************/
/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> START OF THE NON-EXPORTED LIBRARY ROUTINES <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
/**********************************************************************************************************************************/

static void _GetBlockCRC (short WhichWindow, struct TapeEntry_s *Entry, struct _BlockCRC_s *CRC)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape entry.                                           */
/* Post  : The data CRC value is calculated and returned in `CRC'.                                                                */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  dword         XMSIndex;
  dword         XMSSubIndex;
  dword         LengthToGo;
  word          ReadLength;
  dword         DReadLength;
  word register Cnt;
  byte register NextByte;

  XMSIndex = GetTapeBlockXMSIndex (WhichWindow, Entry);
  XMSSubIndex = 0;
  if (CRC->FloodIndex == 0)                                                                          /* Check bounds to make sure */
    CRC->FloodIndex = 4096;
  LengthToGo = Entry->BlockLength;
  while (LengthToGo > 0)
  {
    if (LengthToGo > 49152)
      ReadLength = 49152;
    else
      ReadLength = (word)LengthToGo;
    DReadLength = (dword)(ReadLength & 0xFFFC) + (ReadLength % 4 ? 4 : 0);
    CopyFromXMS (DReadLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex + XMSSubIndex, WorkBuffer);
    for (Cnt = 0 ; Cnt < ReadLength ; Cnt ++)
    {
      NextByte = *(WorkBuffer + Cnt);
      CRC->Sum = (word)(CRC->Sum + NextByte);
      CRC->FloodSum = (word)(CRC->FloodSum + NextByte * CRC->FloodIndex);
      CRC->XORMask ^= NextByte;
      CRC->XORSum = (word)(CRC->XORSum + CRC->XORMask);
      if (-- CRC->FloodIndex == 0)
        CRC->FloodIndex = 4096;
    }
    XMSSubIndex += ReadLength;
    LengthToGo -= ReadLength;
  }
}

static char _MatchBlockType (struct TapeEntry_s *Entry1, struct TapeEntry_s *Entry2)

/**********************************************************************************************************************************/
/* Pre   : `Entry1' and `Entry2' should be compared.                                                                              */
/*         A check must have been done that they at least have the same block type ID (and the same block length)                 */
/* Post  : If all fields related to this specific block type match, `C' (light-green) is returned, otherwise `4' (green).         */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
#define MATCHYES  'A';
#define MATCHNO   '2';
  char ReturnColor = MATCHYES;                                                                          /* Prepare for full match */

  switch (Entry1->BlockTypeID)
  {
    case 0x10 : if (Entry1->PauseLength               != Entry2->PauseLength ||                    /* (Standard Speed Data Block) */
                    Entry1->Flags.FlagByteUsed ^ Entry2->Flags.FlagByteUsed ||                /* (Only 1 of them has flag/parity) */
                    Entry1->Flags.ParityByteUsed ^ Entry2->Flags.ParityByteUsed ||
                    (Entry1->Flags.FlagByteUsed && Entry1->FlagByte != Entry2->FlagByte) ||
                    (Entry1->Flags.ParityByteUsed && Entry1->ParityByte != Entry2->ParityByte))
                  ReturnColor = MATCHNO;
                break;
    case 0x11 : if (Entry1->Timing.LengthPilotPulse   != Entry2->Timing.LengthPilotPulse ||         /* (Turbo Loading Data Block) */
                    Entry1->Timing.LengthSyncPulse[0] != Entry2->Timing.LengthSyncPulse[0] ||
                    Entry1->Timing.LengthSyncPulse[1] != Entry2->Timing.LengthSyncPulse[1] ||
                    Entry1->Timing.LengthZeroBitPulse != Entry2->Timing.LengthZeroBitPulse ||
                    Entry1->Timing.LengthOneBitPulse  != Entry2->Timing.LengthOneBitPulse ||
                    Entry1->Timing.NumPilotPulses     != Entry2->Timing.NumPilotPulses ||
                    Entry1->Timing.NumBitsInLastByte  != Entry2->Timing.NumBitsInLastByte ||
                    Entry1->PauseLength               != Entry2->PauseLength ||
                    Entry1->Flags.FlagByteUsed ^ Entry2->Flags.FlagByteUsed ||                /* (Only 1 of them has flag/parity) */
                    Entry1->Flags.ParityByteUsed ^ Entry2->Flags.ParityByteUsed ||
                    (Entry1->Flags.FlagByteUsed && Entry1->FlagByte != Entry2->FlagByte) ||
                    (Entry1->Flags.ParityByteUsed && Entry1->ParityByte != Entry2->ParityByte))
                  ReturnColor = MATCHNO;
                break;
    case 0x12 : if (Entry1->Timing.LengthPilotPulse   != Entry2->Timing.LengthPilotPulse ||                        /* (Pure Tone) */
                    Entry1->Timing.NumPilotPulses     != Entry2->Timing.NumPilotPulses)
                  ReturnColor = MATCHNO;
                break;
    case 0x13 : break;                                                                  /* (Row of Pulses with Different Timings) */
    case 0x14 : if (Entry1->Timing.LengthZeroBitPulse != Entry2->Timing.LengthZeroBitPulse ||                /* (Pure Data Block) */
                    Entry1->Timing.LengthOneBitPulse  != Entry2->Timing.LengthOneBitPulse ||
                    Entry1->Timing.NumBitsInLastByte  != Entry2->Timing.NumBitsInLastByte ||
                    Entry1->PauseLength               != Entry2->PauseLength)
                  ReturnColor = MATCHNO;
                break;
    case 0x15 : if (Entry1->PauseLength               != Entry2->PauseLength ||                             /* (Direct Recording) */
                    Entry1->Timing.NumBitsInLastByte  != Entry2->Timing.NumBitsInLastByte)          /* (Samplerate not important) */
                  ReturnColor = MATCHNO;
                break;
    case 0x20 : if (Entry1->PauseLength               != Entry2->PauseLength)       /* (Pause (Silence) or Stop the Tape Command) */
                  ReturnColor = MATCHNO;
                break;
    case 0x21 : break;                                                                                           /* (Group Start) */
    case 0x22 : break;                                                                                             /* (Group End) */
    case 0x23 : if (Entry1->PauseLength               != Entry2->PauseLength)                                  /* (Jump To Block) */
                  ReturnColor = MATCHNO;
                break;
    case 0x24 : if (Entry1->PauseLength               != Entry2->PauseLength)                                     /* (Loop Start) */
                  ReturnColor = MATCHNO;
                break;
    case 0x25 : break;                                                                                              /* (Loop End) */
    case 0x26 : break;                                                                                         /* (Call Sequence) */
    case 0x27 : break;                                                                                  /* (Return from Sequence) */
    case 0x28 : break;                                                                                          /* (Select Block) */
    case 0x2A : break;                                                                              /* (Stop Tape if in 48K Mode) */
    case 0x30 : break;                                                                                      /* (Text Description) */
    case 0x31 : if (Entry1->PauseLength               != Entry2->PauseLength)                                  /* (Message Block) */
                  ReturnColor = MATCHNO;
                break;
    case 0x32 : break;                                                                                          /* (Archive info) */
    case 0x33 : break;                                                                                         /* (Hardware Type) */
    case 0x34 : break;                                                                                        /* (Emulation Info) */
    case 0x35 : if (Entry1->FlagByte                  != Entry2->FlagByte)                                 /* (Custom Info Block) */
                  ReturnColor = MATCHNO;
                break;
    case 0x40 : break;                                                                                        /* (Snapshot Block) */
    case 0x5A : break;                                                                          /* ('Z' - Merging of 2 TZX files) */
    default   : break;                                                                        /* (Unknown block for this version) */
  }
  return (ReturnColor);
}

static bool _TestSelected (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : The return value is TRUE only if at least one entry in the tape has been selected.                                     */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirP;
  bool                AtLeastOne = FALSE;

  DirP = TapeWindow[WhichWindow].Directory;
  while (!AtLeastOne && DirP != NULL)
    if (DirP->ComboSelected & SELECTBIT)
      AtLeastOne = TRUE;
    else
      DirP = DirP->Next;
  return (AtLeastOne);
}

static void _SetupTapePath (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : The name of the current tape has been moved to the Path and File variables.                                            */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  short register Cnt;

  //strcpy (_Mask, "*.*");
  if (TapeWindow[WhichWindow].NoTape ||
      !strcmp (TapeWindow[WhichWindow].TapeName, TAPENONE))                                     /* New tape - no valid name yet ? */
  {
    if (_PreviousTapePath[WhichWindow][0] == '\0')                                                      /* No previous path yet ? */
      strcpy (_Path, CurPath);
    else
      strcpy (_Path, _PreviousTapePath[WhichWindow]);
    _File[0] = '\0';
    return;
  }
  strcpy (_Path, TapeWindow[WhichWindow].TapeName);
  Cnt = strlen (_Path);
  while (-- Cnt > 0 && _Path[Cnt] != '\\')                                                                       /* Find the path */
    ;
  if (Cnt == 0)                                                                                    /* No path - only a filename ? */
  {
    strcpy (_File, TapeWindow[WhichWindow].TapeName);
    strcpy (_Path, CurPath);
  }
  else
  {
    strcpy (_File, _Path + Cnt + 1);
    if (_Path[1] == ':' && _Path[2] == '\\' && Cnt == 2)                                                  /* Leave root directory */
      _Path[3] = '\0';
    else
      _Path[Cnt] = '\0';
  }
}

static int _HandleMENew (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the activated menu entry.                                                                            */
/* Post  : The current tape has been released from memory, possibly after saving. The tape name has been set to a 'NONAME' value. */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  bool Go = TRUE;

  if (TapeWindow[Entry->ReactNum].TapeChanged)
    switch (QuestionRequester (-1, -1, TRUE, FALSE, NULL, NULL, NULL, NULL, NULL, NULL, "Tape has been changed! Write back?"))
    {
      case 0 : if (TapeWindow[Entry->ReactNum].NewTape)                                                                  /* (Yes) */
                 Go = ForceNewTapeName (Entry->ReactNum);
               else
                 Go = TestFileType (Entry->ReactNum);
               if (Go)
                 WriteTape (Entry->ReactNum, FALSE);
               break;
      case 1 : break;                                                                                                     /* (No) */
      case 2 : return (WINH_NONE);                                                                                    /* (Cancel) */
    }
  ReleaseTape (Entry->ReactNum);
  TapeWindow[Entry->ReactNum].NonexistingTape = TRUE;
  TapeWindow[Entry->ReactNum].NewTape = TRUE;
  TapeWindow[Entry->ReactNum].EmptyTape = TRUE;
  TapeWindow[Entry->ReactNum].TapeFloppyId = 0;
  SetTapeName (Entry->ReactNum, TAPENONE, TRUE);
  SignalTapeChanged (Entry->ReactNum, FALSE);
  ReInitComboWindow (&(TapeWindow[Entry->ReactNum].IndexWindow), 0);
  PrintComboEntries (&(TapeWindow[Entry->ReactNum].IndexWindow));
  UpdateTapeSize (Entry->ReactNum, 0);
  GhostEmptyTape (Entry->ReactNum);
  GhostMenuBackup (Entry->ReactNum);
  return (WINH_NONE);
}

static int _HandleMEOpen (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the activated menu entry.                                                                            */
/* Post  : The current tape has been saved if changed and a new user-selected tape has been read.                                 */
/* Import: _SetupTapePath.                                                                                                        */
/**********************************************************************************************************************************/

{
  char SPath[MAXDIR + 1];
  bool Go = TRUE;
  bool Accepted = FALSE;

  _SetupTapePath (Entry->ReactNum);
  while (!Accepted)
    if (FileRequester (180, 50, 48, 29, 20, _Path, _File, _Mask, "Select New Tape", TRUE, FALSE))
    {
      MakeFullPath (SPath, _Path, _File);
      if (TapeWindow[Entry->ReactNum].TapeChanged)
        switch (QuestionRequester (-1, -1, TRUE, FALSE, NULL, NULL, NULL, NULL, NULL, NULL, "Tape has been changed! Write back?"))
        {
          case 0 : if (TapeWindow[Entry->ReactNum].NewTape)                                                              /* (Yes) */
                     Go = ForceNewTapeName (Entry->ReactNum);
                   else
                     Go = TestFileType (Entry->ReactNum);
                   if (Go)
                     WriteTape (Entry->ReactNum, FALSE);
                   break;
          case 1 : break;                                                                                                 /* (No) */
          case 2 : if (MaxFloppyIdChanged)                                                                            /* (Cancel) */
                     WriteConfig ();
                   return (WINH_NONE);
        }
      SetTapeName (Entry->ReactNum, SPath, TRUE);
      strcpy (_PreviousTapePath[Entry->ReactNum], _Path);
      if (TapeWindow[Entry->ReactNum].FileType == FT_UNKNOWN)
        ErrorRequester (-1, -1, "@Close", NULL, NULL, "{BERROR:}B\n  %s\n  not a Spectrum tape file",
                        TapeWindow[Entry->ReactNum].TapeName);
      else if (TapeWindow[Entry->ReactNum].FileType == FT_BADSOUND)
        ErrorRequester (-1, -1, "@Close", NULL, NULL, "{BERROR:}B\n  %s\n  is not understood.\n"
                        "  Please convert to (8-bit mono) .VOC format", TapeWindow[Entry->ReactNum].TapeName);
      else
      {
        ReadTape (Entry->ReactNum, FALSE);                                                                       /* Read the tape */
        UnghostMenu (Entry->ReactNum);
        SetupNewData (Entry->ReactNum);
        if (TapeWindow[Entry->ReactNum].IndexWindow.NumEntries >= 0)
          SelectNewTapeEntry (Entry->ReactNum, 0);
        Accepted = TRUE;
        if (_TAPNoise)
        {
          _TAPNoise = FALSE;
          SignalTapeChanged (Entry->ReactNum, TRUE);
          ErrorRequester (-1, -1, NULL, NULL, "Skipped Noise", "Skipped noise blocks in %s tape",
                          Entry->ReactNum == TWLEFT ? "source" : "destination");
        }
      }
    }
    else
      Accepted = TRUE;
  return (WINH_NONE);
}

static int _HandleMESave (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the activated menu entry.                                                                            */
/* Post  : The entire tape has been saved to disk.                                                                                */
/* Import: ForceNewTapeName.                                                                                                      */
/**********************************************************************************************************************************/

{
  bool Go = TRUE;

  if (!TapeWindow[Entry->ReactNum].TapeChanged)                          /* Ask for comfirmation if the tape has not been changed */
  {
    if (QuestionRequester (-1, -1, FALSE, FALSE, NULL, "@Cancel", NULL, NULL, NULL, "Confirm Unneeded Save", "Tape has not "
                           "been changed! Save anyway?"))
      return (WINH_NONE);
  }
  if (TapeWindow[Entry->ReactNum].NewTape)
    Go = ForceNewTapeName (Entry->ReactNum);
  else
    Go = TestFileType (Entry->ReactNum);
  if (Go)
    WriteTape (Entry->ReactNum, FALSE);
  return (WINH_NONE);
}

static int _HandleMESaveAs (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the activated menu entry.                                                                            */
/* Post  : The entire tape has been saved to disk, but under a user-defined (new) name.                                           */
/* Import: _SetupTapePath.                                                                                                        */
/**********************************************************************************************************************************/

{
  char TmpPath[MAXDIR + 1];
  bool Done = FALSE;

  _SetupTapePath (Entry->ReactNum);
  while (!Done)
    if (FileRequester (180, 50, 48, 29, 20, _Path, _File, _Mask, "Choose New Tape Name", FALSE, FALSE))
    {
      MakeFullPath (TmpPath, _Path, _File);
      if (!TestFileOverwrite (TmpPath))
      {
        SetTapeName (Entry->ReactNum, TmpPath, TRUE);                                               /* Make permanent if new tape */
        strcpy (_PreviousTapePath[Entry->ReactNum], _Path);
        if (TapeWindow[Entry->ReactNum].FileType == FT_UNKNOWN)
          ErrorRequester (-1, -1, "@Close", NULL, NULL, "{BERROR:}B\n  %s\n  unknown Spectrum tape file format", _File);
        else if (TapeWindow[Entry->ReactNum].FileType == FT_BADSOUND)
          ErrorRequester (-1, -1, "@Close", NULL, NULL, "{BERROR:}B\n  %s is not understood.\n"
                          "  Please use the .VOC format", _File);
        else if (TapeWindow[Entry->ReactNum].FileType == FT_ZXS)
          ErrorRequester (-1, -1, "@Close", NULL, NULL, "Sorry, the ZXS file format is read-only");
        else if (IsSnapshot (Entry->ReactNum))
          ErrorRequester (-1, -1, "@Close", NULL, NULL, "Sorry, snapshot formats are read-only");
        else
        {
          TapeWindow[Entry->ReactNum].BackupTape = FALSE;                                            /* No need for a tape backup */
          TapeWindow[Entry->ReactNum].NewTape = FALSE;
          WriteTape (Entry->ReactNum, FALSE);                                                                   /* Write the tape */
          Done = TRUE;
        }
      }
    }
    else
      Done = TRUE;
  return (WINH_NONE);
}

static int _HandleMESampler (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the activated menu entry.                                                                            */
/* Post  : The sampler has been called.                                                                                           */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if (Sampler (Entry->ReactNum))
    UnghostMenu (Entry->ReactNum);
  return (WINH_NONE);
}

static int _HandleMEImport (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the activated menu entry.                                                                            */
/* Post  : A (new) tape has been appended in memory.                                                                              */
/* Import: _SetupTapePath.                                                                                                        */
/**********************************************************************************************************************************/

{
  char     OldTapeName[MAXDIR];
  comboent OldNum;
  comboent NewNum;

  if (TapeWindow[Entry->ReactNum].NonexistingTape)                                                               /* No tape yet ? */
    return (_HandleMEOpen (Entry));                                                                    /* Then just open the file */
  AutoInsertNewTape (Entry->ReactNum);                                                                          /* (Sanity check) */
  OldNum = TapeWindow[Entry->ReactNum].IndexWindow.NumEntries;
  strcpy (OldTapeName, TapeWindow[Entry->ReactNum].TapeName);
  _SetupTapePath (Entry->ReactNum);
  if (FileRequester (180, 50, 48, 29, 20, _Path, _File, _Mask, "Select Tape to Import", TRUE, FALSE))
  {
    MakeFullPath (_Path, _Path, _File);
    SetTapeName (Entry->ReactNum, _Path, FALSE);
    strcpy (_PreviousTapePath[Entry->ReactNum], _Path);
    if (TapeWindow[Entry->ReactNum].FileType == FT_UNKNOWN)
      ErrorRequester (-1, -1, "@Close", NULL, NULL, "{BERROR:}B\n  %s\n  not a Spectrum tape file",
                      TapeWindow[Entry->ReactNum].TapeName);
    else if (TapeWindow[Entry->ReactNum].FileType == FT_BADSOUND)
      ErrorRequester (-1, -1, "@Close", NULL, NULL, "{BERROR:}B\n  %s\n  is not understood.\n"
                      "  Please convert to (8-bit mono) .VOC format", TapeWindow[Entry->ReactNum].TapeName);
    else
    {
      ReadTape (Entry->ReactNum, TRUE);                                                                         /* Merge the tape */
      SetTapeName (Entry->ReactNum, OldTapeName, FALSE);
      SetupNewData (Entry->ReactNum);
      NewNum = OldNum;
      if (TapeWindow[Entry->ReactNum].IndexWindow.NumEntries >= 0 && OldNum < 0)                       /* Tape was empty before ? */
        NewNum = 0;
      if (NewNum >= 0)
        SelectNewTapeEntry (Entry->ReactNum, NewNum);
      SignalTapeChanged (Entry->ReactNum, TRUE);
      UnghostMenu (Entry->ReactNum);
      if (_TAPNoise)
      {
        _TAPNoise = FALSE;
        ErrorRequester (-1, -1, NULL, NULL, "Skipped Noise", "Skipped noise blocks in %s tape",
                        Entry->ReactNum == TWLEFT ? "source" : "destination");
      }
    }
  }
  return (WINH_NONE);
}

static int _HandleMETapeCRC (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the activated menu entry.                                                                            */
/* Post  : The full CRC of the tape has been calculated and displayed.                                                            */
/* Import: _GetBlockCRC.                                                                                                          */
/**********************************************************************************************************************************/

{
  struct _BlockCRC_s  TapeCRC;
  struct TapeEntry_s *DirP;
  struct ComboEntry   CEntry;
  comboent            ComboLine;

  memset (&TapeCRC, 0, sizeof (struct _BlockCRC_s));
  for (ComboLine = 0 ; ComboLine <= TapeWindow[Entry->ReactNum].IndexWindow.NumEntries ; ComboLine ++)
  {
    GetComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, ComboLine);                           /* Fetch next entry */
    if (!(CEntry.Selected & ENT_NOSELECT) && CEntry.Identifier != 0)                                  /* Points to a data block ? */
    {
      DirP = (struct TapeEntry_s *)CEntry.Identifier;
      _GetBlockCRC (Entry->ReactNum, DirP, &TapeCRC);                                                /* Include CRC of that block */
    }
  }
  ErrorRequester (-1, -1, NULL, WBPics[2], "Tape CRC Value", "The CRC Value of this tape is "
                  "%04X%04X%02X%04X", TapeCRC.Sum, TapeCRC.FloodSum, TapeCRC.XORMask, TapeCRC.XORSum);
  return (WINH_NONE);
}

static int _HandleMEPlayTime (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the activated menu entry.                                                                            */
/* Post  : The play time of the entire tape has been calculated and displayed.                                                    */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirP;
  struct ComboEntry   CEntry;
  dword               XMSIndex = 0;
  dword               Dummy;
  comboent            Cnt;
  comboent            OldCnt;
  bool                StillOk  = TRUE;
  bool                IsTapeData;
  word                PlayTimeMinutes;
  byte                PlayTimeSeconds;
  dword               FullSeconds;

  ClearTapeLoopDetection (Entry->ReactNum);
  CountPlayTimeLength = TRUE;
  UsePlayTimeClock = FALSE;
  PlayTimeLength = 0;
  DrawWindow (&_CalcWindow, NULL, 0, FALSE);
  Wprintf (&_CalcWindow, 2, TEXT_CENTER, "Calculating play time ...");
  SetMouseShape (DefaultWaitingMouseShape);
  for (Cnt = 0 ; StillOk && Cnt <= TapeWindow[Entry->ReactNum].IndexWindow.NumEntries ; )
  {
    GetComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, Cnt);
    if (CEntry.Identifier > 0)                                                                                     /* A good line */
    {
      DirP = (struct TapeEntry_s *)CEntry.Identifier;
      if (!(CEntry.Selected & ENT_NOSELECT))
      {
        GetBlockTypeInfo (DirP->BlockTypeID, &IsTapeData, NULL, NULL, NULL, NULL);
        if (DirP->BlockTypeID == 0x31 || IsTapeData)                                                           /* (Message Block) */
        {
          OldCnt = Cnt;
          XMSIndex = GetTapeBlockXMSIndex (Entry->ReactNum, DirP);
          StillOk = WriteVOCBlock (Entry->ReactNum, -1, DirP, TapeWindow[Entry->ReactNum].Tape.XMSHandle, XMSIndex, &Dummy, &Cnt);
          if (Cnt != OldCnt + 1)
            Cnt = FindBlockNumber (Entry->ReactNum, Cnt);
        }
        else
          Cnt ++;
      }
      else
        Cnt ++;
    }
    else
      Cnt ++;
  }
  DestroyWindow (&_CalcWindow);
  SetMouseShape (DefaultNormalMouseShape);
  CountPlayTimeLength = FALSE;
  FullSeconds = (dword)((PlayTimeLength / 3500000.0) + 0.5);                                             /* (T-states to seconds) */
  PlayTimeMinutes = (word)(FullSeconds / 60);
  PlayTimeSeconds = (byte)(FullSeconds % 60);
  ErrorRequester (-1, -1, NULL, WBPics[2], "Tape Play Time", "The play time of this tape is "
                  "%d:%02d", PlayTimeMinutes, PlayTimeSeconds);
  return (WINH_NONE);
}

static int _HandleMESelGetPlTim (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the activated menu entry.                                                                            */
/* Post  : The play time of all selected entries has been calculated and displayed.                                               */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirP;
  struct ComboEntry   CEntry;
  dword               XMSIndex = 0;
  dword               Dummy;
  comboent            Cnt;
  comboent            OldCnt;
  bool                StillOk = TRUE;
  bool                IsTapeData;
  word                PlayTimeMinutes;
  byte                PlayTimeSeconds;
  dword               FullSeconds;

  if (!_TestSelected (Entry->ReactNum))
    return (WINH_NONE);
  ClearTapeLoopDetection (Entry->ReactNum);
  CountPlayTimeLength = TRUE;
  UsePlayTimeClock = FALSE;
  PlayTimeLength = 0;
  DrawWindow (&_CalcWindow, NULL, 0, FALSE);
  Wprintf (&_CalcWindow, 2, TEXT_CENTER, "Calculating play time ...");
  SetMouseShape (DefaultWaitingMouseShape);
  for (Cnt = 0 ; StillOk && Cnt <= TapeWindow[Entry->ReactNum].IndexWindow.NumEntries ; )
  {
    GetComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, Cnt);
    if (CEntry.Identifier > 0)                                                                                   /* A good line */
    {
      DirP = (struct TapeEntry_s *)CEntry.Identifier;
      if (CEntry.Selected & SELECTBIT)                                                                      /* Write this one ? */
      {
        GetBlockTypeInfo (DirP->BlockTypeID, &IsTapeData, NULL, NULL, NULL, NULL);
        if (DirP->BlockTypeID == 0x31 || IsTapeData)                                                         /* (Message Block) */
        {
          OldCnt = Cnt;
          XMSIndex = GetTapeBlockXMSIndex (Entry->ReactNum, DirP);
          StillOk = WriteVOCBlock (Entry->ReactNum, -1, DirP, TapeWindow[Entry->ReactNum].Tape.XMSHandle, XMSIndex, &Dummy, &Cnt);
          if (Cnt != OldCnt + 1)
            Cnt = FindBlockNumber (Entry->ReactNum, Cnt);
        }
        else
          Cnt ++;
      }
      else
        Cnt ++;
    }
    else
      Cnt ++;
  }
  DestroyWindow (&_CalcWindow);
  SetMouseShape (DefaultNormalMouseShape);
  CountPlayTimeLength = FALSE;
  FullSeconds = (dword)((PlayTimeLength / 3500000.0) + 0.5);                                             /* (T-states to seconds) */
  PlayTimeMinutes = (word)(FullSeconds / 60);
  PlayTimeSeconds = (byte)(FullSeconds % 60);
  ErrorRequester (-1, -1, NULL, WBPics[2], "Selected Entries Play Time", "The play time of these entries is "
                  "%d:%02d", PlayTimeMinutes, PlayTimeSeconds);
  return (WINH_NONE);
}

static int _HandleMESelCheckCRC (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the activated menu entry.                                                                            */
/* Post  : The play time of all selected entries has been calculated and displayed.                                               */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirP;
  struct ComboEntry   CEntry;
  dword               XMSIndex = 0;
  dword               XMSSubIndex;
  dword               LengthToGo;
  word                ReadLength;
  dword               DReadLength;
  comboent            Cnt;
  word      register  Cnt2;
  byte                BlockXOR = 0;
  byte                ParityByte = 0;

  if (!_TestSelected (Entry->ReactNum))
    return (WINH_NONE);
  for (Cnt = 0 ; Cnt <= TapeWindow[Entry->ReactNum].IndexWindow.NumEntries ; Cnt ++)
  {
    GetComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, Cnt);
    if (CEntry.Identifier > 0)                                                                                   /* A good line */
    {
      DirP = (struct TapeEntry_s *)CEntry.Identifier;
      if ((CEntry.Selected & SELECTBIT) && DirP->BlockTypeID == 0x14)                                        /* Check this one ? */
      {
        XMSIndex = GetTapeBlockXMSIndex (Entry->ReactNum, DirP);
        XMSSubIndex = 0;
        LengthToGo = DirP->BlockLength;
        while (LengthToGo > 0)
        {
          if (LengthToGo > 49152)
            ReadLength = 49152;
          else
            ReadLength = (word)LengthToGo;
          DReadLength = (dword)(ReadLength & 0xFFFC) + (ReadLength % 4 ? 4 : 0);
          CopyFromXMS (DReadLength, TapeWindow[Entry->ReactNum].Tape.XMSHandle, XMSIndex + XMSSubIndex, WorkBuffer);
          for (Cnt2 = 0 ; Cnt2 < ReadLength ; Cnt2 ++)
            BlockXOR ^= *(WorkBuffer + Cnt2);
          ParityByte = *(WorkBuffer + Cnt2 - 1);                                               /* Remember last byte of the block */
          XMSSubIndex += ReadLength;
          LengthToGo -= ReadLength;
        }
      }
    }
  }
  if (!BlockXOR)
    ErrorRequester (-1, -1, NULL, WBPics[2], "Group Parity", "The group parity is correct", BlockXOR);
  else
    ErrorRequester (-1, -1, NULL, NULL, "Group Parity", "The group parity is incorrect\n(expected %02X, but found %02X)",
                    ParityByte, BlockXOR ^ ParityByte);
  return (WINH_NONE);
}

static int _HandleMEBlockCRC (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the activated menu entry.                                                                            */
/* Post  : The CRC of the appointed block has been calculated and displayed.                                                      */
/* Import: _GetBlockCRC.                                                                                                          */
/**********************************************************************************************************************************/

{
  struct _BlockCRC_s  BlockCRC;
  struct TapeEntry_s *DirP;
  struct ComboEntry   CEntry;

  memset (&BlockCRC, 0, sizeof (struct _BlockCRC_s));
  GetComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, TapeWindow[Entry->ReactNum].SelectedLine);
  if (!(CEntry.Selected & ENT_NOSELECT) && CEntry.Identifier != 0)                                              /* (Sanity check) */
  {
    DirP = (struct TapeEntry_s *)CEntry.Identifier;
    _GetBlockCRC (Entry->ReactNum, DirP, &BlockCRC);                                               /* Calculate CRC of this block */
    ErrorRequester (-1, -1, NULL, WBPics[2], "Block CRC Value", "The CRC Value of this block is "
                    "%04X%04X%02X%04X", BlockCRC.Sum, BlockCRC.FloodSum, BlockCRC.XORMask, BlockCRC.XORSum);
  }
  return (WINH_NONE);
}

static int _HandleMESelNone (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the activated menu entry.                                                                            */
/* Post  : All selected blocks of the tape have been deselected.                                                                  */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  DeselectAll (Entry->ReactNum);
  return (WINH_NONE);
}

static int _HandleMEInvSel (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the activated menu entry.                                                                            */
/* Post  : All selected blocks have been deselected and vice versa.                                                               */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct ComboEntry CEntry;
  comboent register Cnt;

  for (Cnt = 0 ; Cnt <= TapeWindow[Entry->ReactNum].IndexWindow.NumEntries ; Cnt ++)
  {
    GetComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, Cnt);
    if (CEntry.Identifier != 0)                                                                  /* It's a tape directory entry ? */
    {
      CEntry.Selected ^= SELECTBIT;                                                                       /* Toggle select status */
      SetTapeBlockSelectFlag ((struct TapeEntry_s *)CEntry.Identifier, CEntry.Selected);                           /* Then update */
      PutComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, Cnt, TRUE);
    }
  }
  return (WINH_NONE);
}

static int _HandleMEDelEntry (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the activated menu entry.                                                                            */
/* Post  : The currently selected tape block has been removed. The user has confirmed the action.                                 */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirP;
  struct TapeEntry_s *NextDirP;
  struct ComboEntry   CEntry;
  comboent            SAdjust;
  bool                GroupDone     = TRUE;
  bool                LoopDone      = TRUE;
  char                EntryText[80];

  if (TapeWindow[Entry->ReactNum].EmptyTape)                                                                    /* (Sanity Check) */
    return (WINH_NONE);
  if (TapeWindow[Entry->ReactNum].SelectedLine < 0)                                                             /* (Sanity check) */
    return (WINH_NONE);
  GetComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, TapeWindow[Entry->ReactNum].SelectedLine);
  strcpy (EntryText, CEntry.Text);
  EntryText[4] = ' ';                                                                                   /* (expand/collapse char) */
  if (TPConfig.ConfirmDelete)                                                                 /* Confirmation needed: handle it ? */
    if (QuestionRequester (-1, -1, FALSE, FALSE, NULL, "@Cancel", NULL, NULL, NULL, "Confirm Delete",
                           "Are you sure you want to delete the %s tape entry\n\"%s}N\"?",
                           Entry->ReactNum == TWLEFT ? "source" : "destination", EntryText))
      return (WINH_NONE);
  if (CEntry.Identifier != 0)                                                                                   /* (Sanity check) */
  {
    DirP = (struct TapeEntry_s *)CEntry.Identifier;
    SAdjust = TapeWindow[Entry->ReactNum].SelectedLine;
    if (SAdjust > 0)                                                                    /* Poiting to at least the second entry ? */
    {
      GetComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, SAdjust - 1);             /* Peek at the previous entry */
      if ((CEntry.Selected & ENT_NOSELECT) || CEntry.Identifier == 0)                  /* The previous entry was non-selectable ? */
        SAdjust --;                                /* Then we're now at the second line of a multi-line entry. Point to the first */
    }
    if ((DirP->BlockTypeID == 0x21 || DirP->BlockTypeID == 0x24) && (DirP->ComboSelected & EXPANDBIT))
    {                                                                                          /* Entry is collapsed Group/Loop ? */
      if (TPConfig.ConfirmDelete)
        if (DirP->BlockTypeID == 0x21)
        {
          if (QuestionRequester (-1, -1, FALSE, FALSE, NULL, "@Cancel", NULL, NULL, NULL, "Confirm Delete Entire Group",
                                 "Are you sure you want to delete the entire group\n%s}N?", EntryText + 13))
            return (WINH_NONE);
          GroupDone = FALSE;
        }
        else
        {
          if (QuestionRequester (-1, -1, FALSE, FALSE, NULL, "@Cancel", NULL, NULL, NULL, "Confirm Delete Entire Loop",
                                 "Are you sure you want to delete the entire loop?"))
            return (WINH_NONE);
          LoopDone = FALSE;
        }
      while ((!GroupDone || !LoopDone) && DirP != NULL)                                               /* Delete entire Group/Loop */
      {
        NextDirP = DirP->Next;
        if (DirP->BlockTypeID == 0x21)                                                                           /* Group Start ? */
          GroupDone = FALSE;                                                                           /* Then we need to do more */
        else if (DirP->BlockTypeID == 0x22)                                                                        /* Group End ? */
          GroupDone = TRUE;                                                                     /* Then we're done after this one */
        else if (DirP->BlockTypeID == 0x24)                                                                       /* Loop Start ? */
          LoopDone = FALSE;                                                                            /* Then we need to do more */
        else if (DirP->BlockTypeID == 0x25)                                                                         /* Loop End ? */
          LoopDone = TRUE;                                                                      /* Then we're done after this one */
        DeleteTapeBlock (Entry->ReactNum, DirP, FALSE, TRUE);
        DirP = NextDirP;
      }
      SetupNewData (Entry->ReactNum);
    }
    else
      DeleteTapeBlock (Entry->ReactNum, DirP, TRUE, TRUE);
  }
  SelectNewTapeEntry (Entry->ReactNum, SAdjust);
  return (WINH_NONE);
}

static int _HandleMEDelSel (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the activated menu entry.                                                                            */
/* Post  : All selected tape block have been removed. The user has confirmed the action.                                          */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirP;
  struct TapeEntry_s *NextDirP;

  if (!_TestSelected (Entry->ReactNum))
    return (WINH_NONE);
  if (TPConfig.ConfirmDelete)                                                                 /* Confirmation needed: handle it ? */
    if (QuestionRequester (-1, -1, FALSE, FALSE, NULL, "@Cancel", NULL, NULL, NULL, "Confirm Delete",
                           "Are you sure you want to delete all selected entries in the %s tape?",
                           Entry->ReactNum == TWLEFT ? "source" : "destination"))
      return (WINH_NONE);
  DirP = TapeWindow[Entry->ReactNum].Directory;
  while (DirP != NULL)
  {
    NextDirP = DirP->Next;
    if (DirP->ComboSelected & SELECTBIT)
      DeleteTapeBlock (Entry->ReactNum, DirP, FALSE, TRUE);
    DirP = NextDirP;
  }
  TapeWindow[Entry->ReactNum].SelectedLine = 0;
  SetupNewData (Entry->ReactNum);
  return (WINH_NONE);
}

// static int _HandleMEResizeUp (struct MenuEntry *Entry)
// 
/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the activated menu entry.                                                                            */
/* Post  : The currently selected tape block has been expanded.                                                                   */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/
// 
// {
// Entry=Entry;
// return(WINH_NONE);
// }
// 
// static int _HandleMEResizeDn (struct MenuEntry *Entry)
// 
/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the activated menu entry.                                                                            */
/* Post  : The currently selected tape block has been shrunk.                                                                     */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/
// 
// {
// Entry=Entry;//8888
// return(WINH_NONE);
// }

static int _HandleMEDupEntry (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the activated menu entry.                                                                            */
/* Post  : The currently selected tape block has been duplicated and selected.                                                    */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct ComboEntry   TEntry;
  struct TapeEntry_s *DirP;

  if (TapeWindow[Entry->ReactNum].SelectedLine >= 0)                                                            /* (Sanity check) */
  {
    GetComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &TEntry, TapeWindow[Entry->ReactNum].SelectedLine);
    DirP = (struct TapeEntry_s *)TEntry.Identifier;
    CopyTapeBlock (Entry->ReactNum, DirP, Entry->ReactNum, DirP, FALSE, TRUE);                                       /* Duplicate */
    SetupNewData (Entry->ReactNum);
    SelectNewTapeEntry (Entry->ReactNum, TapeWindow[Entry->ReactNum].SelectedLine);
  }
  return (WINH_NONE);
}

static int _HandleMEPlayEntry (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the activated menu entry.                                                                            */
/* Post  : The currently appointed tape block has been played through the SoundBlaster.                                           */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirP;
  struct ComboEntry   CEntry;
  dword               XMSIndex = 0;
  dword               Dummy;
  word                SampleRate;
  bool                IsTapeData;
  bool                GroupDone = TRUE;
  bool                LoopDone  = TRUE;
  bool                StillOk  = TRUE;
  short               PlayLoopCount;
  comboent            BlockNo;
  comboent            OldBlockNo;

  if (TapeWindow[Entry->ReactNum].SelectedLine < 0)                                                             /* (Sanity check) */
    return (WINH_NONE);
  for (PlayLoopCount = TPConfig.ShowRemainingPlayTime ? 0 : 1 ; PlayLoopCount < 2 ; PlayLoopCount ++)
  {
    ClearTapeLoopDetection (Entry->ReactNum);
    if (PlayLoopCount == 0)
    {
      CountPlayTimeLength = TRUE;
      UsePlayTimeClock = FALSE;
      PlayTimeLength = 0;
      DrawWindow (&_CalcWindow, NULL, 0, FALSE);
      Wprintf (&_CalcWindow, 2, TEXT_CENTER, "Calculating play time ...");
      SetMouseShape (DefaultWaitingMouseShape);
    }
    else
    {
      CountPlayTimeLength = FALSE;
      UsePlayTimeClock = TPConfig.ShowRemainingPlayTime;
      if (!SetupVOCWriteTimers (Entry->ReactNum, &SampleRate))
      {
        UsePlayTimeClock = FALSE;
        return (WINH_NONE);
      }
      if (TPConfig.PlayCriticalSection)
        EnterCriticalSection ();
    }
    GetComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, TapeWindow[Entry->ReactNum].SelectedLine);
    BlockNo = GetBlockNumber (CEntry.Text);                                                           /* Get current block number */
    if (CEntry.Identifier > 0)                                                                             /* Must be a good line */
    {
      DirP = (struct TapeEntry_s *)CEntry.Identifier;
      if ((DirP->BlockTypeID == 0x21 || DirP->BlockTypeID == 0x24) && (DirP->ComboSelected & EXPANDBIT))
      {                                                                                        /* Entry is collapsed Group/Loop ? */
        if (DirP->BlockTypeID == 21)
          GroupDone = FALSE;
        else
          LoopDone = FALSE;
        while (StillOk && (!GroupDone || !LoopDone) && DirP != NULL)
        {
          OldBlockNo = BlockNo;
          GetBlockTypeInfo (DirP->BlockTypeID, &IsTapeData, NULL, NULL, NULL, NULL);
          if (DirP->BlockTypeID == 0x21)                                                                         /* Group Start ? */
            GroupDone = FALSE;                                                                         /* Then we need to do more */
          else if (DirP->BlockTypeID == 0x22)                                                                      /* Group End ? */
            GroupDone = TRUE;                                                                   /* Then we're done after this one */
          else if (DirP->BlockTypeID == 0x24)                                                                     /* Loop Start ? */
            LoopDone = FALSE;                                                                          /* Then we need to do more */
          else if (DirP->BlockTypeID == 0x25)                                                                       /* Loop End ? */
            LoopDone = TRUE;                                                                    /* Then we're done after this one */
          else if (DirP->BlockTypeID == 0x31 || IsTapeData)                                                    /* (Message Block) */
          {
            XMSIndex = GetTapeBlockXMSIndex (Entry->ReactNum, DirP);
            StillOk = WriteVOCBlock (Entry->ReactNum, -1, DirP, TapeWindow[Entry->ReactNum].Tape.XMSHandle, XMSIndex, &Dummy,
                                     &BlockNo);
            if (BlockNo == OldBlockNo + 1)
              DirP = DirP->Next;
            else
            {
              DirP = TapeWindow[Entry->ReactNum].Directory;
              OldBlockNo = BlockNo;
              while (OldBlockNo > 0 && DirP)
              {
                DirP = DirP->Next;
                OldBlockNo --;
              }
            }
          }
          else 
          {
            DirP = DirP->Next;
            BlockNo ++;
          }
        }
      }
      else
      {
        GetBlockTypeInfo (DirP->BlockTypeID, &IsTapeData, NULL, NULL, NULL, NULL);
        if (IsTapeData)
        {
          XMSIndex = GetTapeBlockXMSIndex (Entry->ReactNum, DirP);
          StillOk = WriteVOCBlock (Entry->ReactNum, -1, DirP, TapeWindow[Entry->ReactNum].Tape.XMSHandle, XMSIndex,
                                   &Dummy, &BlockNo);
        }
        else
          StillOk = FALSE;
      }
    }
    else
      StillOk = FALSE;
    if (PlayLoopCount == 0)
    {
      DestroyWindow (&_CalcWindow);
      SetMouseShape (DefaultNormalMouseShape);
      StillOk = TRUE;
    }
  }
  FlushWriteBuffer ((bool)!StillOk);
  UsePlayTimeClock = FALSE;
  if (TPConfig.PlayCriticalSection)
    LeaveCriticalSection ();
  if (_InfiniteTapeLoop)                                                                   /* Stopped due to infinite tape loop ? */
    ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR: Infinite tape loop");
  else if (SB_DMANoIRQTimeOut)
    ErrorRequester (-1, -1, NULL, NULL, NULL, "{BERROR}B: %s", GetSB_DMANoIRQTimeOutMessage ());
  else if (*SB_DMAOverrun)                                                                        /* Stopped due to DMA overrun ? */
    if (TPConfig.PlaySoundBlaster)
      ErrorRequester (-1, -1, NULL, NULL, NULL, "{BERROR}B: DMA overrun\n       Choose a lower sample frequency");
    else
      ErrorRequester (-1, -1, NULL, NULL, NULL, "{BERROR}B: Sorry, your machine isn't fast enough for this feature");
  return (WINH_NONE);
}

static int _HandleMEPlaySel (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the activated menu entry.                                                                            */
/* Post  : All selected tape blocks have been played through the SoundBlaster.                                                    */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirP;
  struct ComboEntry   CEntry;
  dword               XMSIndex = 0;
  dword               Dummy;
  word                SampleRate;
  short               PlayLoopCount;
  comboent            Cnt;
  comboent            OldCnt;
  comboent            OldSelected;
  bool                StillOk = TRUE;
  bool                IsTapeData;
  bool                AtLeastOne = FALSE;

  if (!_TestSelected (Entry->ReactNum))
    return (WINH_NONE);
  OldSelected = TapeWindow[Entry->ReactNum].SelectedLine;
  for (PlayLoopCount = TPConfig.ShowRemainingPlayTime ? 0 : 1 ; PlayLoopCount < 2 ; PlayLoopCount ++)
  {
    ClearTapeLoopDetection (Entry->ReactNum);
    if (PlayLoopCount == 0)
    {
      CountPlayTimeLength = TRUE;
      UsePlayTimeClock = FALSE;
      PlayTimeLength = 0;
      DrawWindow (&_CalcWindow, NULL, 0, FALSE);
      Wprintf (&_CalcWindow, 2, TEXT_CENTER, "Calculating play time ...");
      SetMouseShape (DefaultWaitingMouseShape);
    }
    else
    {
      CountPlayTimeLength = FALSE;
      UsePlayTimeClock = TPConfig.ShowRemainingPlayTime;
      if (!SetupVOCWriteTimers (Entry->ReactNum, &SampleRate))
      {
        UsePlayTimeClock = FALSE;
        return (WINH_NONE);
      }
      TmpExpandTape (Entry->ReactNum);                                               /* Expand all collapsed blocks while playing */
      if (TPConfig.PlayCriticalSection)
        EnterCriticalSection ();
    }
    for (Cnt = 0 ; StillOk && Cnt <= TapeWindow[Entry->ReactNum].IndexWindow.NumEntries ; )
    {
      GetComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, Cnt);
      if (CEntry.Identifier > 0)                                                                                   /* A good line */
      {
        DirP = (struct TapeEntry_s *)CEntry.Identifier;
        if (CEntry.Selected & SELECTBIT)                                                                      /* Write this one ? */
        {
          if (PlayLoopCount != 0 && TPConfig.TraceSoundBlocks)
            SelectEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), Cnt);
          GetBlockTypeInfo (DirP->BlockTypeID, &IsTapeData, NULL, NULL, NULL, NULL);
          if (DirP->BlockTypeID == 0x31 || IsTapeData)                                                         /* (Message Block) */
          {
            OldCnt = Cnt;
            XMSIndex = GetTapeBlockXMSIndex (Entry->ReactNum, DirP);
            StillOk = WriteVOCBlock (Entry->ReactNum, -1, DirP, TapeWindow[Entry->ReactNum].Tape.XMSHandle, XMSIndex, &Dummy, &Cnt);
            if (StillOk)
              AtLeastOne = TRUE;
            if (Cnt != OldCnt + 1)
              Cnt = FindBlockNumber (Entry->ReactNum, Cnt);
          }
          else
            Cnt ++;
        }
        else
          Cnt ++;
      }
      else
        Cnt ++;
    }
    if (PlayLoopCount == 0)
    {
      DestroyWindow (&_CalcWindow);
      SetMouseShape (DefaultNormalMouseShape);
      StillOk = TRUE;
    }
  }
  FlushWriteBuffer ((bool)!StillOk);
  UsePlayTimeClock = FALSE;
  if (TPConfig.PlayCriticalSection)
    LeaveCriticalSection ();
  if (_InfiniteTapeLoop)                                                                   /* Stopped due to infinite tape loop ? */
    ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR: Infinite tape loop");
  else if (SB_DMANoIRQTimeOut)
    ErrorRequester (-1, -1, NULL, NULL, NULL, "{BERROR}B: %s", GetSB_DMANoIRQTimeOutMessage ());
  else if (*SB_DMAOverrun)                                                                        /* Stopped due to DMA overrun ? */
    if (TPConfig.PlaySoundBlaster)
      ErrorRequester (-1, -1, NULL, NULL, NULL, "{BERROR}B: DMA overrun\n       Choose a lower sample frequency");
    else if (AtLeastOne && TPConfig.TraceSoundBlocks)
      ErrorRequester (-1, -1, NULL, NULL, NULL, "{BERROR}B: Sorry, your machine isn't fast enough for this feature\n"
                                                 "It may work if you disable the 'Trace Sound Blocks' config option");
    else
      ErrorRequester (-1, -1, NULL, NULL, NULL, "{BERROR}B: Sorry, your machine isn't fast enough for this feature");
  TmpRestoreTape (Entry->ReactNum);                                                                   /* Restore collapsed status */
  if (TPConfig.TraceSoundBlocks)
    SelectEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), OldSelected);
  return (WINH_NONE);
}

static int _HandleMEPlayFrom (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the activated menu entry.                                                                            */
/* Post  : All tape blocks from the active entry onward have been played through the SoundBlaster.                                */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirP;
  struct ComboEntry   CEntry;
  dword               XMSIndex = 0;
  dword               Dummy;
  word                SampleRate;
  short               PlayLoopCount;
  comboent            Cnt;
  comboent            OldCnt;
  comboent            OldSelected;
  bool                StillOk  = TRUE;
  bool                IsTapeData;
  bool                AtLeastOne = FALSE;

  OldSelected = TapeWindow[Entry->ReactNum].SelectedLine;
  for (PlayLoopCount = TPConfig.ShowRemainingPlayTime ? 0 : 1 ; PlayLoopCount < 2 ; PlayLoopCount ++)
  {
    ClearTapeLoopDetection (Entry->ReactNum);
    if (PlayLoopCount == 0)
    {
      CountPlayTimeLength = TRUE;
      UsePlayTimeClock = FALSE;
      PlayTimeLength = 0;
      DrawWindow (&_CalcWindow, NULL, 0, FALSE);
      Wprintf (&_CalcWindow, 2, TEXT_CENTER, "Calculating play time ...");
      SetMouseShape (DefaultWaitingMouseShape);
    }
    else
    {
      CountPlayTimeLength = FALSE;
      UsePlayTimeClock = TPConfig.ShowRemainingPlayTime;
      if (!SetupVOCWriteTimers (Entry->ReactNum, &SampleRate))
      {
        UsePlayTimeClock = FALSE;
        return (WINH_NONE);
      }
      TmpExpandTape (Entry->ReactNum);                                               /* Expand all collapsed blocks while playing */
      if (TPConfig.PlayCriticalSection)
        EnterCriticalSection ();
    }
    for (Cnt = OldSelected ; StillOk && Cnt <= TapeWindow[Entry->ReactNum].IndexWindow.NumEntries ; )
    {
      GetComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, Cnt);
      if (CEntry.Identifier > 0)                                                                                   /* A good line */
      {
        DirP = (struct TapeEntry_s *)CEntry.Identifier;
        if (!(CEntry.Selected & ENT_NOSELECT))
        {
          if (PlayLoopCount != 0 && TPConfig.TraceSoundBlocks)
            SelectEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), Cnt);
          GetBlockTypeInfo (DirP->BlockTypeID, &IsTapeData, NULL, NULL, NULL, NULL);
          if (DirP->BlockTypeID == 0x31 || IsTapeData)                                                         /* (Message Block) */
          {
            OldCnt = Cnt;
            XMSIndex = GetTapeBlockXMSIndex (Entry->ReactNum, DirP);
            StillOk = WriteVOCBlock (Entry->ReactNum, -1, DirP, TapeWindow[Entry->ReactNum].Tape.XMSHandle, XMSIndex, &Dummy, &Cnt);
            if (StillOk)
              AtLeastOne = TRUE;
            if (Cnt != OldCnt + 1)
              Cnt = FindBlockNumber (Entry->ReactNum, Cnt);
          }
          else
            Cnt ++;
        }
        else
          Cnt ++;
      }
      else
        Cnt ++;
    }
    if (PlayLoopCount == 0)
    {
      DestroyWindow (&_CalcWindow);
      SetMouseShape (DefaultNormalMouseShape);
      StillOk = TRUE;
    }
  }
  FlushWriteBuffer ((bool)!StillOk);
  UsePlayTimeClock = FALSE;
  if (TPConfig.PlayCriticalSection)
    LeaveCriticalSection ();
  if (_InfiniteTapeLoop)                                                                   /* Stopped due to infinite tape loop ? */
    ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR: Infinite tape loop");
  else if (SB_DMANoIRQTimeOut)
    ErrorRequester (-1, -1, NULL, NULL, NULL, "{BERROR}B: %s", GetSB_DMANoIRQTimeOutMessage ());
  else if (*SB_DMAOverrun)                                                                        /* Stopped due to DMA overrun ? */
    if (TPConfig.PlaySoundBlaster)
      ErrorRequester (-1, -1, NULL, NULL, NULL, "{BERROR}B: DMA overrun\n       Choose a lower sample frequency");
    else if (AtLeastOne && TPConfig.TraceSoundBlocks)
      ErrorRequester (-1, -1, NULL, NULL, NULL, "{BERROR}B: Sorry, your machine isn't fast enough for this feature\n"
                                                 "It may work if you disable the 'Trace Sound Blocks' config option");
    else
      ErrorRequester (-1, -1, NULL, NULL, NULL, "{BERROR}B: Sorry, your machine isn't fast enough for this feature");
  TmpRestoreTape (Entry->ReactNum);                                                                   /* Restore collapsed status */
  if (TPConfig.TraceSoundBlocks)
    SelectEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), OldSelected);
  return (WINH_NONE);
}

static int _HandleMEPlayTape (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the activated menu entry.                                                                            */
/* Post  : All tape blocks have been played through the SoundBlaster.                                                             */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirP;
  struct ComboEntry   CEntry;
  dword               XMSIndex = 0;
  dword               Dummy;
  word                SampleRate;
  short               PlayLoopCount;
  comboent            Cnt;
  comboent            OldCnt;
  comboent            OldSelected;
  bool                StillOk  = TRUE;
  bool                IsTapeData;
  bool                AtLeastOne = FALSE;

  OldSelected = TapeWindow[Entry->ReactNum].SelectedLine;
  for (PlayLoopCount = TPConfig.ShowRemainingPlayTime ? 0 : 1 ; PlayLoopCount < 2 ; PlayLoopCount ++)
  {
    ClearTapeLoopDetection (Entry->ReactNum);
    if (PlayLoopCount == 0)
    {
      CountPlayTimeLength = TRUE;
      UsePlayTimeClock = FALSE;
      PlayTimeLength = 0;
      DrawWindow (&_CalcWindow, NULL, 0, FALSE);
      Wprintf (&_CalcWindow, 2, TEXT_CENTER, "Calculating play time ...");
      SetMouseShape (DefaultWaitingMouseShape);
    }
    else
    {
      CountPlayTimeLength = FALSE;
      UsePlayTimeClock = TPConfig.ShowRemainingPlayTime;
      if (!SetupVOCWriteTimers (Entry->ReactNum, &SampleRate))
      {
        UsePlayTimeClock = FALSE;
        return (WINH_NONE);
      }
      TmpExpandTape (Entry->ReactNum);                                               /* Expand all collapsed blocks while playing */
      if (TPConfig.PlayCriticalSection)
        EnterCriticalSection ();
    }
    for (Cnt = 0 ; StillOk && Cnt <= TapeWindow[Entry->ReactNum].IndexWindow.NumEntries ; )
    {
      GetComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, Cnt);
      if (CEntry.Identifier > 0)                                                                                   /* A good line */
      {
        DirP = (struct TapeEntry_s *)CEntry.Identifier;
        if (!(CEntry.Selected & ENT_NOSELECT))
        {
          if (PlayLoopCount != 0 && TPConfig.TraceSoundBlocks)
            SelectEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), Cnt);
          GetBlockTypeInfo (DirP->BlockTypeID, &IsTapeData, NULL, NULL, NULL, NULL);
          if (DirP->BlockTypeID == 0x31 || IsTapeData)                                                         /* (Message Block) */
          {
            OldCnt = Cnt;
            XMSIndex = GetTapeBlockXMSIndex (Entry->ReactNum, DirP);
            StillOk = WriteVOCBlock (Entry->ReactNum, -1, DirP, TapeWindow[Entry->ReactNum].Tape.XMSHandle, XMSIndex, &Dummy, &Cnt);
            if (StillOk)
              AtLeastOne = TRUE;
            if (Cnt != OldCnt + 1)
              Cnt = FindBlockNumber (Entry->ReactNum, Cnt);
          }
          else
            Cnt ++;
        }
        else
          Cnt ++;
      }
      else
        Cnt ++;
    }
    if (PlayLoopCount == 0)
    {
      DestroyWindow (&_CalcWindow);
      SetMouseShape (DefaultNormalMouseShape);
      StillOk = TRUE;
    }
  }
  FlushWriteBuffer ((bool)!StillOk);
  UsePlayTimeClock = FALSE;
  if (TPConfig.PlayCriticalSection)
    LeaveCriticalSection ();
  if (_InfiniteTapeLoop)                                                                   /* Stopped due to infinite tape loop ? */
    ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR: Infinite tape loop");
  else if (SB_DMANoIRQTimeOut)
    ErrorRequester (-1, -1, NULL, NULL, NULL, "{BERROR}B: %s", GetSB_DMANoIRQTimeOutMessage ());
  else if (*SB_DMAOverrun)                                                                        /* Stopped due to DMA overrun ? */
    if (TPConfig.PlaySoundBlaster)
      ErrorRequester (-1, -1, NULL, NULL, NULL, "{BERROR}B: DMA overrun\n       Choose a lower sample frequency");
    else if (AtLeastOne && TPConfig.TraceSoundBlocks)
      ErrorRequester (-1, -1, NULL, NULL, NULL, "{BERROR}B: Sorry, your machine isn't fast enough for this feature\n"
                                                 "It may work if you disable the 'Trace Sound Blocks' config option");
    else
      ErrorRequester (-1, -1, NULL, NULL, NULL, "{BERROR}B: Sorry, your machine isn't fast enough for this feature");
  TmpRestoreTape (Entry->ReactNum);                                                                   /* Restore collapsed status */
  if (TPConfig.TraceSoundBlocks)
    SelectEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), OldSelected);
  return (WINH_NONE);
}

static int _HandleMESaveRaw (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the activated menu entry.                                                                            */
/* Post  : The currently appointed block's XMS data has been written to a raw file.                                               */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct ComboEntry CEntry;
  char              RPath[MAXDIR + 1] = "\0";
  char              RFile[13]         = "\0";
  char              RMask[13]         = "*.*";
  dword             BlockIndex        = 0;
  bool              More              = TRUE;
  word              Size;
  int               Fh;

  if (TapeWindow[Entry->ReactNum].SelectedLine < 0)                                                             /* (Sanity check) */
    return (WINH_NONE);
  GetComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, TapeWindow[Entry->ReactNum].SelectedLine);
  if (CEntry.Identifier > 0)                                                                               /* Must be a good line */
  {
    while (More)
      if (FileRequester (180, 50, 48, 29, 20, RPath, RFile, RMask, "Choose Raw File Name", FALSE, FALSE))
      {
        MakeFullPath (RPath, RPath, RFile);
        if (!TestFileOverwrite (RPath))
          More = FALSE;
      }
      else
        return (WINH_NONE);
    if ((Fh = open (RPath, O_BINARY | O_RDWR | O_CREAT | O_TRUNC, 0777)) == -1)
    { ErrorRequester (-1, -1, "@Stop", NULL, NULL, "Error creating file %s", RPath); return (WINH_NONE); }
    More = TRUE;
    while (More)
    {
      Size = GetTapeBlock (Entry->ReactNum, &BlockIndex, &More);
      if (Size != 0)
      {
        if (write (Fh, WorkBuffer, Size) == -1)                                                                 /* Write the file */
        { close (Fh); ErrorRequester (-1, -1, "@Close", NULL, NULL, "Error writing to file %s", RPath); return (WINH_NONE); }
      }
    }
    close (Fh);
  }
  return (WINH_NONE);
}

static int _HandleMEDecDirRec (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the activated menu entry.                                                                            */
/* Post  : The currently appointed Direct Recording block has been decoded.                                                       */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirP;
  struct ComboEntry   CEntry;

  if (TapeWindow[Entry->ReactNum].SelectedLine < 0)                                                             /* (Sanity check) */
    return (WINH_NONE);
  GetComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, TapeWindow[Entry->ReactNum].SelectedLine);
  if (CEntry.Identifier != 0)                                                                                   /* (Sanity check) */
  {
    DirP = (struct TapeEntry_s *)CEntry.Identifier;
    DecodeDRB (Entry->ReactNum, DirP);
  }
  return (WINH_NONE);
}

static int _HandleMEMakeGroup (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : All selected (consequetive) entries have been joined into a Group block.                                               */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *NewEntry;
  struct TapeEntry_s *DirP;
  struct TapeEntry_s *StartOfGroup;
  struct TapeEntry_s *EndOfGroup;
  short               WhichWindow;
  bool                FoundStart = FALSE;
  bool                FoundEnd = FALSE;
  byte                StartID;
  byte                EndID;
  short               CountEntries;
  short               AdjustSelected = 0;                                                      /* 0 = select bar before the group */
                                                                                               /* 1 = select bar within the group */
                                                                                               /* 2 = select bar after the group  */

  WhichWindow = (Entry->ReactNum & 0x0001);
  switch (Entry->ReactNum & 0x0002)
  {
    case 0x0000 : StartID = 0x21 ; EndID = 0x22; break;                                                             /* Make Group */
    case 0x0001 : StartID = 0x24 ; EndID = 0x25; break;                                                              /* Make Loop */
  }
  DirP = TapeWindow[WhichWindow].Directory;
  CountEntries = 0;
  while (DirP != NULL && !FoundEnd)
  {
    if (DirP->ComboSelected & SELECTBIT)
    {
      if (!FoundStart)
      {
        FoundStart = TRUE;
        StartOfGroup = DirP;
      }
    }
    else if (FoundStart)                                                                                       /* First one after */
    {
      FoundEnd = TRUE;
      EndOfGroup = DirP;
    }
    if (TapeWindow[WhichWindow].SelectedLine == CountEntries && FoundStart)          /* Selection bar is inside the group range ? */
      AdjustSelected = 1;
    CountEntries ++;
    DirP = DirP->Next;
  }
  if (TapeWindow[WhichWindow].SelectedLine >= CountEntries)                                 /* Selection bar is after the group ? */
    AdjustSelected = 2;
  if ((NewEntry = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
    FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
  memset (NewEntry, 0, sizeof (struct TapeEntry_s));
  NewEntry->BlockTypeID = StartID;                                                                          /* (Group/Loop Start) */
  NewEntry->ComboSelected = 0x00;
  NewEntry->Flags.IsNormalSpeed = TRUE;
  GetBlockTypeInfo (StartID, &(NewEntry->IsTapeData), &(NewEntry->IsIdentifiedData), NULL, NULL, NULL);
  if (!FoundStart || TapeWindow[WhichWindow].Directory == NULL || TapeWindow[WhichWindow].SelectedLine < 0)       /* Empty tape ? */
  { free (NewEntry); return (WINH_NONE); }                                                        /* Sanity check - cannot happen */
  else
  {
    if (TapeWindow[WhichWindow].Directory == StartOfGroup)                                     /* Insert before the first entry ? */
    {
      NewEntry->Next = TapeWindow[WhichWindow].Directory;
      TapeWindow[WhichWindow].Directory = NewEntry;
    }
    else
    {
      DirP = TapeWindow[WhichWindow].Directory;
      while (DirP->Next != NULL && DirP->Next != StartOfGroup)
        DirP = DirP->Next;
      NewEntry->Next = StartOfGroup;
      DirP->Next = NewEntry;                                                                              /* The actual insertion */
    }
  }
  TapeWindow[WhichWindow].NumEntries ++;
  _MakeGroupCancelled = FALSE;
  switch (StartID)
  {
    case 0x21 : EditGroupStrtBlock (WhichWindow, NewEntry); break;                                              /* Get group name */
    case 0x24 : EditLoopStrtBlock (WhichWindow, NewEntry); break;                                    /* Get number of repetitions */
  }
  if (_MakeGroupCancelled)                                                                                         /* Cancelled ? */
  {
    DeleteTapeBlock (WhichWindow, NewEntry, FALSE, FALSE);                                       /* Then we don't want it anymore */
    return (WINH_NONE);
  }
  if (FoundEnd)                                                          /* Does the end of the block happen before end of tape ? */
  {                                                                                        /* Then insert a Group End block there */
    if ((NewEntry = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
      FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
    memset (NewEntry, 0, sizeof (struct TapeEntry_s));
    NewEntry->BlockTypeID = EndID;                                                                            /* (Group/Loop End) */
    NewEntry->ComboSelected = 0x00;
    NewEntry->Flags.IsNormalSpeed = TRUE;
    GetBlockTypeInfo (EndID, &(NewEntry->IsTapeData), &(NewEntry->IsIdentifiedData), NULL, NULL, NULL);
    NewEntry->Next = NULL;
    DirP = StartOfGroup;
    while (DirP->Next != NULL && DirP->Next != EndOfGroup)
      DirP = DirP->Next;
    NewEntry->Next = EndOfGroup;
    DirP->Next = NewEntry;
    TapeWindow[WhichWindow].NumEntries ++;
  }
  DeselectAll (WhichWindow);
  SignalTapeChanged (WhichWindow, TRUE);
  TapeWindow[WhichWindow].SelectedLine += AdjustSelected;                       /* Adjust the select bar for the inserted entries */
  SetupNewData (WhichWindow);                                                            /* Make new listing and fix subgroupings */
  return (WINH_NONE);
}

static int _HandleMESetDecrypt (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : The user has set the decryption values for special decoding schemes.                                                   */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct WindowInfo CWindow   = {193, 225, 292, 140, 292, 140, 292, 140, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                 WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL, "Set Auto-Decrypt Values"};
  struct ButtonInfo PCancel   = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 81, 110, 8, DBLACK, DWHITE, NULL, 1, "@Cancel", NULL};
  struct ButtonInfo POk       = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 167, 110, 8, DBLACK, DWHITE, NULL, 0, "@Ok", NULL};
  struct ButtonInfo BDefault  = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 208, 10, 1, DBLACK, DWHITE, NULL, 4, NULL, NULL};
  struct ButtonInfo BValXOR   = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 172, 60, 3, DBLACK, DWHITE, NULL, 2,
                                 NULL, NULL};
  struct ButtonInfo BValADD   = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 172, 80, 3, DBLACK, DWHITE, NULL, 3,
                                 NULL, NULL};
  struct ButtonInfo *ButtonP;
  struct MenuEntry  *ThisEntry;
  bool              UseDefault;
  char              ValXOR[6];
  char              ValADD[6];
  byte              EValXOR;
  byte              EValADD;
  bool              Done      = FALSE;
  bool              Cancelled = FALSE;
  bool              Changed   = FALSE;

  ThisEntry = Entry; /* keep compiler happy */
  BDefault.Contents = &UseDefault;
  BValXOR.Contents = ValXOR;
  BValADD.Contents = ValADD;
  EValXOR = DecryptXOR;
  EValADD = DecryptADD;
  if ((UseDefault = !UseDecryptUserValues))
  { BValXOR.Flags |= BF_GHOSTED; BValADD.Flags |= BF_GHOSTED; }
  sprintf (ValXOR, "%u", EValXOR);
  sprintf (ValADD, "%u", EValADD);
  AddButton (&CWindow, &POk, TRUE);
  AddButton (&CWindow, &BDefault, FALSE);
  AddButton (&CWindow, &BValXOR, FALSE);
  AddButton (&CWindow, &BValADD, FALSE);
  AddButton (&CWindow, &PCancel, FALSE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  Wprintf (&CWindow, 1, 14, "{+1Use Default Values}N");
  Wprintf (&CWindow, 3,  2, "SpeedLock 2-3: XOR 152 (0x98), ADD 11 (0x0B)");
  Wprintf (&CWindow, 4,  2, "SpeedLock 4-7: XOR 193 (0xC1), ADD 17 (0x11)");
  Wprintf (&CWindow, 6, 17, "{+1XOR Value}N");
  Wprintf (&CWindow, 8, 17, "{+1ADD Value}N");
  while (!Done)
  {
    switch (HandleWindow (&CWindow, &ButtonP))
    {
      case WINH_CLOSE   : Done = TRUE; Cancelled = TRUE; break;
      case WINH_RNOHAND : switch (ButtonP->ReactNum)
                          {
                            case 0  : Done = TRUE; Cancelled = FALSE; break;                                              /* (Ok) */
                            case 1  : Done = TRUE; Cancelled = TRUE; break;                                           /* (Cancel) */
                          }
                          break;
      case WINH_BUTTBOOL: switch (ButtonP->ReactNum)
                          {
                            case 4  : if (UseDefault)                                                     /* Use default values ? */
                                      {
                                        BValXOR.Flags |= BF_GHOSTED;
                                        BValADD.Flags |= BF_GHOSTED;
                                      }
                                      else
                                      {
                                        BValXOR.Flags &= ~BF_GHOSTED;
                                        BValADD.Flags &= ~BF_GHOSTED;
                                      }
                                      UpdateButtonText (&CWindow, &BValXOR, BF_FITTXT);
                                      UpdateButtonText (&CWindow, &BValADD, BF_FITTXT);
                                      break;
                          }
                          break;
      case WINH_EDITEND : if (ButtonP->Flags & BF_FITINPUT)                                            /* Contents were changed ? */
                            switch (ButtonP->ReactNum)
                            {
                              case 2  : EValXOR = atol (ValXOR);
                                        sprintf (ValXOR, "%u", EValXOR);
                                        UpdateButtonText (&CWindow, &BValXOR, BF_FITTXT);
                                        break;
                              case 3  : EValADD = atol (ValADD);
                                        sprintf (ValADD, "%u", EValADD);
                                        UpdateButtonText (&CWindow, &BValADD, BF_FITTXT);
                                        break;
                            }
                          break;
    }
    if (GlobalKey == K_ESC)  { Done = TRUE; Cancelled = TRUE; }
    MouseStatus ();
  }
  DestroyWindow (&CWindow);
  GlobalKey = 0x0000;
  if (Cancelled)
    return (WINH_NONE);
  Changed = (UseDefault == UseDecryptUserValues || EValXOR != DecryptXOR || EValADD != DecryptADD);
  if (Changed)
  {
    UseDecryptUserValues = !UseDefault;
    DecryptXOR = EValXOR;
    DecryptADD = EValADD;
  }
  return (WINH_NONE);
}

static int _HandleMEMatchThis (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : The currently appointed entry has been checked against the rest of the tape.                                           */
/*         Any block that matches both the data and block-specific info received a green block number. Blocks that only match the */
/*         data received a light-green block number. If no matches were found, the errorbeeper has been sound.                    */
/*         Any colouring from a previous match has been negated.                                                                  */
/* Import: _GetBlockCRC, _MatchBlockType.                                                                                         */
/**********************************************************************************************************************************/

{
  struct _BlockCRC_s  ThisCRC;
  struct _BlockCRC_s  TestCRC;
  struct TapeEntry_s *DirP;
  struct TapeEntry_s *TestDirP;
  struct ComboEntry   CEntry;
  comboent            ComboLine;
  comboent            ComboLine2;
  char                NewText[256];
  char                ColorByte;
  bool                AnyChange = FALSE;
  bool                AnyMatch = FALSE;

  if (TapeWindow[Entry->ReactNum].SelectedLine < 0)                                                             /* (Sanity check) */
    return (WINH_NONE);
  if (TapeWindow[Entry->ReactNum ^ 0x0001].IndexWindow.NumEntries >= 0)                       /* Cleanup the other tape if needed */
    for (ComboLine = 0 ; ComboLine <= TapeWindow[Entry->ReactNum ^ 0x0001].IndexWindow.NumEntries ; ComboLine ++)
    {
      GetComboEntry (&(TapeWindow[Entry->ReactNum ^ 0x0001].IndexWindow), &CEntry, ComboLine);
      if (!strncmp (CEntry.Text, "{+", 2))
      {
        strcpy (CEntry.Text, CEntry.Text + 3);
        strcpy (CEntry.Text + 5, CEntry.Text + 8);
        PutComboEntry (&(TapeWindow[Entry->ReactNum ^ 0x0001].IndexWindow), &CEntry, ComboLine, TRUE);
        AnyChange = TRUE;
      }
    }
  GetComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, TapeWindow[Entry->ReactNum].SelectedLine);
  if (CEntry.Identifier != 0)                                                                                   /* (Sanity check) */
  {
    DirP = (struct TapeEntry_s *)CEntry.Identifier;
    memset (&ThisCRC, 0, sizeof (struct _BlockCRC_s));
    _GetBlockCRC (Entry->ReactNum, DirP, &ThisCRC);                                                      /* Get CRC of this block */
    for (ComboLine = 0 ; ComboLine <= TapeWindow[Entry->ReactNum].IndexWindow.NumEntries ; ComboLine ++)
    {
      GetComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, ComboLine);                         /* Fetch next entry */
      if (!strncmp (CEntry.Text, "{+", 2))                                             /* This entry matched any previous check ? */
      {
        strcpy (CEntry.Text, CEntry.Text + 3);                                                           /* Remove the color info */
        strcpy (CEntry.Text + 5, CEntry.Text + 8);
        PutComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, ComboLine, FALSE);
        AnyChange = TRUE;
      }
      if (!(CEntry.Selected & ENT_NOSELECT) && CEntry.Identifier != 0)                                /* Points to a data block ? */
      {
        TestDirP = (struct TapeEntry_s *)CEntry.Identifier;
        if (TestDirP != DirP && TestDirP->BlockTypeID == DirP->BlockTypeID && TestDirP->BlockLength == DirP->BlockLength)
        {                                                                        /* Same type and size (but not the same entry) ? */
          memset (&TestCRC, 0, sizeof (struct _BlockCRC_s));
          _GetBlockCRC (Entry->ReactNum, TestDirP, &TestCRC);                                            /* Get CRC of that block */
          if (!memcmp (&ThisCRC, &TestCRC, sizeof (struct _BlockCRC_s)))                             /* Data parts are the same ? */
          {
            ColorByte = _MatchBlockType (DirP, TestDirP);                 /* See if the block info (timing, etc) are the same too */
            GetComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, ComboLine);                    /* Fetch the entry */
            if (!strncmp (CEntry.Text, "    ", 4) && ComboLine != 0)                                      /* Empty block number ? */
            {
              ComboLine2 = ComboLine - 1;                                             /* The previous line holds the block number */
              GetComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, ComboLine2);        /* Fetch that entry instead */
            }
            else
              ComboLine2 = ComboLine;
            sprintf (NewText, "{+%c", ColorByte);                                                           /* Insert color start */
            strncpy (NewText + 3, CEntry.Text, 5);                                                   /* Add original block number */
            NewText[8] = '\0';
            strcat (NewText, "{++");                                                                          /* Insert color end */
            strcat (NewText, CEntry.Text + 5);                                                        /* Append entry description */
            strcpy (CEntry.Text, NewText);
            PutComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, ComboLine2, FALSE);           /* Store the change */
            AnyChange = TRUE;
            AnyMatch = TRUE;
          }
        }
      }
    }
  }
  if (AnyMatch)                                                                                          /* Did we have a match ? */
  {                                                                                       /* Then color the current entry as well */
    ComboLine = TapeWindow[Entry->ReactNum].SelectedLine;
    GetComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, ComboLine);
    if (!strncmp (CEntry.Text, "    ", 4) && ComboLine != 0)
    {
      ComboLine2 = ComboLine - 1;
      GetComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, ComboLine2);
    }
    else
      ComboLine2 = ComboLine;
    sprintf (NewText, "{+A");
    strncpy (NewText + 3, CEntry.Text, 5);
    NewText[8] = '\0';
    strcat (NewText, "{++");
    strcat (NewText, CEntry.Text + 5);
    strcpy (CEntry.Text, NewText);
    PutComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, ComboLine2, FALSE);
  }
  else
    SoundErrorBeeper (BEEP_SHORT);
  if (AnyChange)
    PrintComboEntries (&(TapeWindow[Entry->ReactNum].IndexWindow));
  return (WINH_NONE);
}

static int _HandleMEMatchOther (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : The currently appointed entry has been checked against the rest of the tape.                                           */
/*         Any block that matches both the data and block-specific info received a green block number. Block that only match the  */
/*         data received a light-green block number. If no matches were found, the errorbeeper has been sound.                    */
/*         Any coloring from a previous match has been negated.                                                                   */
/* Import: _GetBlockCRC, _MatchBlockType.                                                                                         */
/**********************************************************************************************************************************/

{
  struct _BlockCRC_s  ThisCRC;
  struct _BlockCRC_s  TestCRC;
  struct TapeEntry_s *DirP;
  struct TapeEntry_s *TestDirP;
  struct ComboEntry   CEntry;
  comboent            ComboLine;
  comboent            ComboLine2;
  char                NewText[256];
  char                ColorByte;
  bool                AnyChange = FALSE;
  bool                AnyMatch = FALSE;

  if (TapeWindow[Entry->ReactNum].SelectedLine < 0)                                                             /* (Sanity check) */
    return (WINH_NONE);
  if (TapeWindow[Entry->ReactNum ^ 0x0001].SelectedLine < 0)                                                    /* (Sanity check) */
    return (WINH_NONE);
  for (ComboLine = 0 ; ComboLine <= TapeWindow[Entry->ReactNum].IndexWindow.NumEntries ; ComboLine ++)
  {
    GetComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, ComboLine);
    if (!strncmp (CEntry.Text, "{+", 2))
    {
      strcpy (CEntry.Text, CEntry.Text + 3);
      strcpy (CEntry.Text + 5, CEntry.Text + 8);
      PutComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, ComboLine, TRUE);
      AnyChange = TRUE;
    }
  }
  GetComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, TapeWindow[Entry->ReactNum].SelectedLine);
  if (CEntry.Identifier != 0)                                                                                   /* (Sanity check) */
  {
    DirP = (struct TapeEntry_s *)CEntry.Identifier;
    memset (&ThisCRC, 0, sizeof (struct _BlockCRC_s));
    _GetBlockCRC (Entry->ReactNum, DirP, &ThisCRC);                                                      /* Get CRC of this block */
    for (ComboLine = 0 ; ComboLine <= TapeWindow[Entry->ReactNum ^ 0x0001].IndexWindow.NumEntries ; ComboLine ++)
    {
      GetComboEntry (&(TapeWindow[Entry->ReactNum ^ 0x0001].IndexWindow), &CEntry, ComboLine);                /* Fetch next entry */
      if (!strncmp (CEntry.Text, "{+", 2))                                             /* This entry matched any previous check ? */
      {
        strcpy (CEntry.Text, CEntry.Text + 3);                                                           /* Remove the color info */
        strcpy (CEntry.Text + 5, CEntry.Text + 8);
        PutComboEntry (&(TapeWindow[Entry->ReactNum ^ 0x0001].IndexWindow), &CEntry, ComboLine, FALSE);
        AnyChange = TRUE;
      }
      if (!(CEntry.Selected & ENT_NOSELECT) && CEntry.Identifier != 0)                                /* Points to a data block ? */
      {
        TestDirP = (struct TapeEntry_s *)CEntry.Identifier;
        if (TestDirP->BlockTypeID == DirP->BlockTypeID && TestDirP->BlockLength == DirP->BlockLength)
        {                                                                        /* Same type and size (but not the same entry) ? */
          memset (&TestCRC, 0, sizeof (struct _BlockCRC_s));
          _GetBlockCRC (Entry->ReactNum ^ 0x0001, TestDirP, &TestCRC);                                   /* Get CRC of that block */
          if (!memcmp (&ThisCRC, &TestCRC, sizeof (struct _BlockCRC_s)))                             /* Data parts are the same ? */
          {
            ColorByte = _MatchBlockType (DirP, TestDirP);                 /* See if the block info (timing, etc) are the same too */
            GetComboEntry (&(TapeWindow[Entry->ReactNum ^ 0x0001].IndexWindow), &CEntry, ComboLine);           /* Fetch the entry */
            if (!strncmp (CEntry.Text, "    ", 4) && ComboLine != 0)                                      /* Empty block number ? */
            {
              ComboLine2 = ComboLine - 1;                                             /* The previous line holds the block number */
              GetComboEntry (&(TapeWindow[Entry->ReactNum ^ 0x0001].IndexWindow), &CEntry, ComboLine2);     /* Fetch that instead */
            }
            else
              ComboLine2 = ComboLine;
            sprintf (NewText, "{+%c", ColorByte);                                                           /* Insert color start */
            strncpy (NewText + 3, CEntry.Text, 5);                                                   /* Add original block number */
            NewText[8] = '\0';
            strcat (NewText, "{++");                                                                          /* Insert color end */
            strcat (NewText, CEntry.Text + 5);                                                        /* Append entry description */
            strcpy (CEntry.Text, NewText);
            PutComboEntry (&(TapeWindow[Entry->ReactNum ^ 0x0001].IndexWindow), &CEntry, ComboLine2, FALSE);  /* Store the change */
            AnyChange = TRUE;
            AnyMatch = TRUE;
          }
        }
      }
    }
  }
  if (AnyChange)
    PrintComboEntries (&(TapeWindow[Entry->ReactNum ^ 0x0001].IndexWindow));
  if (AnyMatch)                                                                                          /* Did we have a match ? */
  {                                                                                       /* Then color the current entry as well */
    ComboLine = TapeWindow[Entry->ReactNum].SelectedLine;
    GetComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, ComboLine);
    if (!strncmp (CEntry.Text, "    ", 4) && ComboLine != 0)
    {
      ComboLine2 = ComboLine - 1;
      GetComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, ComboLine2);
    }
    else
      ComboLine2 = ComboLine;
    sprintf (NewText, "{+A");
    strncpy (NewText + 3, CEntry.Text, 5);
    NewText[8] = '\0';
    strcat (NewText, "{++");
    strcat (NewText, CEntry.Text + 5);
    strcpy (CEntry.Text, NewText);
    PutComboEntry (&(TapeWindow[Entry->ReactNum].IndexWindow), &CEntry, ComboLine2, TRUE);
  }
  else
    SoundErrorBeeper (BEEP_SHORT);
  return (WINH_NONE);
}

static int _HandleMEMtchTapes (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : The CRC is generated of both tapes. Reported is whether these CRC values are equal.                                    */
/* Import: _GetBlockCRC.                                                                                                          */
/**********************************************************************************************************************************/

{
  struct _BlockCRC_s  ThisCRC;
  struct _BlockCRC_s  OthCRC;
  struct TapeEntry_s *DirP;
  struct MenuEntry   *ThisEntry;
  struct ComboEntry   CEntry;
  comboent            ComboLine;

  ThisEntry = Entry; /* keep compiler happy */
  if (TapeWindow[TWLEFT].SelectedLine < 0)                                                                      /* (Sanity check) */
    return (WINH_NONE);
  if (TapeWindow[TWRIGHT].SelectedLine < 0)                                                                     /* (Sanity check) */
    return (WINH_NONE);
  memset (&ThisCRC, 0, sizeof (struct _BlockCRC_s));
  memset (&OthCRC, 0, sizeof (struct _BlockCRC_s));
  for (ComboLine = 0 ; ComboLine <= TapeWindow[TWLEFT].IndexWindow.NumEntries ; ComboLine ++)
  {
    GetComboEntry (&(TapeWindow[TWLEFT].IndexWindow), &CEntry, ComboLine);                                    /* Fetch next entry */
    if (!(CEntry.Selected & ENT_NOSELECT) && CEntry.Identifier != 0)                                  /* Points to a data block ? */
    {
      DirP = (struct TapeEntry_s *)CEntry.Identifier;
      _GetBlockCRC (TWLEFT, DirP, &ThisCRC);                                                         /* Include CRC of that block */
    }
  }
  for (ComboLine = 0 ; ComboLine <= TapeWindow[TWRIGHT].IndexWindow.NumEntries ; ComboLine ++)
  {
    GetComboEntry (&(TapeWindow[TWRIGHT].IndexWindow), &CEntry, ComboLine);                                   /* Fetch next entry */
    if (!(CEntry.Selected & ENT_NOSELECT) && CEntry.Identifier != 0)                                  /* Points to a data block ? */
    {
      DirP = (struct TapeEntry_s *)CEntry.Identifier;
      _GetBlockCRC (TWRIGHT, DirP, &OthCRC);                                                         /* Include CRC of that block */
    }
  }
  if (!memcmp (&ThisCRC, &OthCRC, sizeof (struct _BlockCRC_s)))                                   /* The check - tapes same CRC ? */
    ErrorRequester (-1, -1, NULL, WBPics[2], "Match Tapes", "The tapes are equal");
  else                                                                   /* Not fully equal. See if the Spectrum data parts match */
  {
    memset (&ThisCRC, 0, sizeof (struct _BlockCRC_s));
    memset (&OthCRC, 0, sizeof (struct _BlockCRC_s));
    for (ComboLine = 0 ; ComboLine <= TapeWindow[TWLEFT].IndexWindow.NumEntries ; ComboLine ++)
    {
      GetComboEntry (&(TapeWindow[TWLEFT].IndexWindow), &CEntry, ComboLine);
      if (!(CEntry.Selected & ENT_NOSELECT) && CEntry.Identifier != 0)
      {                                                                           /* Consider only Spectrum data blocks this time */
        DirP = (struct TapeEntry_s *)CEntry.Identifier;
        if (DirP->BlockTypeID == 0x10 || DirP->BlockTypeID == 0x11 || DirP->BlockTypeID == 0x14 || DirP->BlockTypeID == 0x15)
          _GetBlockCRC (TWLEFT, DirP, &ThisCRC);
      }
    }
    for (ComboLine = 0 ; ComboLine <= TapeWindow[TWRIGHT].IndexWindow.NumEntries ; ComboLine ++)
    {
      GetComboEntry (&(TapeWindow[TWRIGHT].IndexWindow), &CEntry, ComboLine);
      if (!(CEntry.Selected & ENT_NOSELECT) && CEntry.Identifier != 0)
      {
        DirP = (struct TapeEntry_s *)CEntry.Identifier;
        if (DirP->BlockTypeID == 0x10 || DirP->BlockTypeID == 0x11 || DirP->BlockTypeID == 0x14 || DirP->BlockTypeID == 0x15)
          _GetBlockCRC (TWRIGHT, DirP, &OthCRC);
      }
    }
    if (!memcmp (&ThisCRC, &OthCRC, sizeof (struct _BlockCRC_s)))                                 /* The check - tapes same CRC ? */
      ErrorRequester (-1, -1, NULL, WBPics[2], "Match Tapes", "The data of these tapes are equal");
    else
      ErrorRequester (-1, -1, NULL, WBPics[2], "Match Tapes", "The tapes do not match");
  }
  return (WINH_NONE);
}

static int _HandleMESoundVol (struct MenuEntry *Entry)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : The settings of the SoundBlaster software mixer have been adjusted (routine in TPAUDIO.C)                              */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  Entry = Entry; /* keep compiler happy */
  SetSoftMixerOutput ();
  return (WINH_NONE);
}

/**********************************************************************************************************************************/
/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> START OF THE EXPORTED LIBRARY ROUTINES <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
/**********************************************************************************************************************************/

void SetActiveWindowMenu (struct WindowInfo *MenuWindow, short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `MenuWindow' points to the start of the menu strip.                         */
/* Post  : The active window has shortcut keys. The menu widths have been adjusted.                                               */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if (WhichWindow == TWLEFT && _ActiveWindowMenu != TWLEFT)
  {
    _ActiveWindowMenu = TWLEFT;
    MELNew.Flags       |= MEF_KEYSELECT;   MERNew.Flags       &= ~MEF_KEYSELECT;
    MELOpen.Flags      |= MEF_KEYSELECT;   MEROpen.Flags      &= ~MEF_KEYSELECT;
    MELSaveAs.Flags    |= MEF_KEYSELECT;   MERSaveAs.Flags    &= ~MEF_KEYSELECT;
    MELSave.Flags      |= MEF_KEYSELECT;   MERSave.Flags      &= ~MEF_KEYSELECT;
    MELSampler.Flags   |= MEF_KEYSELECT;   MERSampler.Flags   &= ~MEF_KEYSELECT;
    MELImport.Flags    |= MEF_KEYSELECT;   MERImport.Flags    &= ~MEF_KEYSELECT;
    MELSelNone.Flags   |= MEF_KEYSELECT;   MERSelNone.Flags   &= ~MEF_KEYSELECT;
    MELInvSel.Flags    |= MEF_KEYSELECT;   MERInvSel.Flags    &= ~MEF_KEYSELECT;
    MELInsert.Flags    |= MEF_KEYSELECT;   MERInsert.Flags    &= ~MEF_KEYSELECT;
    MELDelEntry.Flags  |= MEF_KEYSELECT;   MERDelEntry.Flags  &= ~MEF_KEYSELECT;
    MELDupEntry.Flags  |= MEF_KEYSELECT;   MERDupEntry.Flags  &= ~MEF_KEYSELECT;
    MELPlayFrom.Flags  |= MEF_KEYSELECT;   MERPlayFrom.Flags  &= ~MEF_KEYSELECT;
    MELPlayTape.Flags  |= MEF_KEYSELECT;   MERPlayTape.Flags  &= ~MEF_KEYSELECT;
    MELViewDump.Flags  |= MEF_KEYSELECT;   MERViewDump.Flags  &= ~MEF_KEYSELECT;
    MELViewHead.Flags  |= MEF_KEYSELECT;   MERViewHead.Flags  &= ~MEF_KEYSELECT;
    MELViewBasic.Flags |= MEF_KEYSELECT;   MERViewBasic.Flags &= ~MEF_KEYSELECT;
    MELViewScrn.Flags  |= MEF_KEYSELECT;   MERViewScrn.Flags  &= ~MEF_KEYSELECT;
    MELViewDiss.Flags  |= MEF_KEYSELECT;   MERViewDiss.Flags  &= ~MEF_KEYSELECT;
    MELEdHeader.Flags  |= MEF_KEYSELECT;   MEREdHeader.Flags  &= ~MEF_KEYSELECT;
    MELMakeGroup.Flags |= MEF_KEYSELECT;   MERMakeGroup.Flags &= ~MEF_KEYSELECT;
    MELMtchThis.Flags |= MEF_KEYSELECT;    MERMtchThis.Flags &= ~MEF_KEYSELECT;
    MELMtchOther.Flags |= MEF_KEYSELECT;   MERMtchOther.Flags &= ~MEF_KEYSELECT;
    MELBlockCRC.Flags |= MEF_KEYSELECT;    MERBlockCRC.Flags &= ~MEF_KEYSELECT;
  }
  else if (WhichWindow == TWRIGHT && _ActiveWindowMenu != TWRIGHT)
  {
    _ActiveWindowMenu = TWRIGHT;
    MERNew.Flags       |= MEF_KEYSELECT;   MELNew.Flags       &= ~MEF_KEYSELECT;
    MEROpen.Flags      |= MEF_KEYSELECT;   MELOpen.Flags      &= ~MEF_KEYSELECT;
    MERSaveAs.Flags    |= MEF_KEYSELECT;   MELSaveAs.Flags    &= ~MEF_KEYSELECT;
    MERSave.Flags      |= MEF_KEYSELECT;   MELSave.Flags      &= ~MEF_KEYSELECT;
    MERSampler.Flags   |= MEF_KEYSELECT;   MELSampler.Flags   &= ~MEF_KEYSELECT;
    MERImport.Flags    |= MEF_KEYSELECT;   MELImport.Flags    &= ~MEF_KEYSELECT;
    MERSelNone.Flags   |= MEF_KEYSELECT;   MELSelNone.Flags   &= ~MEF_KEYSELECT;
    MERInvSel.Flags    |= MEF_KEYSELECT;   MELInvSel.Flags    &= ~MEF_KEYSELECT;
    MERInsert.Flags    |= MEF_KEYSELECT;   MELInsert.Flags    &= ~MEF_KEYSELECT;
    MERDelEntry.Flags  |= MEF_KEYSELECT;   MELDelEntry.Flags  &= ~MEF_KEYSELECT;
    MERDupEntry.Flags  |= MEF_KEYSELECT;   MELDupEntry.Flags  &= ~MEF_KEYSELECT;
    MERPlayFrom.Flags  |= MEF_KEYSELECT;   MELPlayFrom.Flags  &= ~MEF_KEYSELECT;
    MERPlayTape.Flags  |= MEF_KEYSELECT;   MELPlayTape.Flags  &= ~MEF_KEYSELECT;
    MERViewDump.Flags  |= MEF_KEYSELECT;   MELViewDump.Flags  &= ~MEF_KEYSELECT;
    MERViewHead.Flags  |= MEF_KEYSELECT;   MELViewHead.Flags  &= ~MEF_KEYSELECT;
    MERViewBasic.Flags |= MEF_KEYSELECT;   MELViewBasic.Flags &= ~MEF_KEYSELECT;
    MERViewScrn.Flags  |= MEF_KEYSELECT;   MELViewScrn.Flags  &= ~MEF_KEYSELECT;
    MERViewDiss.Flags  |= MEF_KEYSELECT;   MELViewDiss.Flags  &= ~MEF_KEYSELECT;
    MEREdHeader.Flags  |= MEF_KEYSELECT;   MELEdHeader.Flags  &= ~MEF_KEYSELECT;
    MERMakeGroup.Flags |= MEF_KEYSELECT;   MELMakeGroup.Flags &= ~MEF_KEYSELECT;
    MERMtchThis.Flags |= MEF_KEYSELECT;    MELMtchThis.Flags &= ~MEF_KEYSELECT;
    MERMtchOther.Flags |= MEF_KEYSELECT;   MELMtchOther.Flags &= ~MEF_KEYSELECT;
    MERBlockCRC.Flags |= MEF_KEYSELECT;    MELBlockCRC.Flags &= ~MEF_KEYSELECT;
  }
  UpdateMenuWidths (MenuWindow->FirstMenu);
}

void GhostMenuNoSoundBlaster (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : None. (No SoundBlaster compatible device has been detected).                                                           */
/* Post  : All appropriate menu entries have been ghosted.                                                                        */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if (SoundBlasterInfo.MixerType == SBMTYPE_NONE)                                                /* Soundcard has no soft mixer ? */
    MESoundVol.Flags |= MEF_GHOSTED;
  if (!SoundBlasterInfo.Detected || SoundBlasterInfo.DSPMajorVersion < 2)
  {
    MELSampler.Flags |= MEF_GHOSTED;
    MERSampler.Flags |= MEF_GHOSTED;
    if (TPConfig.PlaySoundBlaster)
      if (WhichWindow == TWLEFT)
        MELPlay.Flags |= MEF_GHOSTED;
      else
        MERPlay.Flags |= MEF_GHOSTED;
    else
      if (WhichWindow == TWLEFT)
        MELPlay.Flags &= ~MEF_GHOSTED;
      else
        MERPlay.Flags &= ~MEF_GHOSTED;
  }
}

void GhostMenuEmptyTape (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' holds either TWLEFT or TWRIGHT, and the tape is empty (or no tape at all).                               */
/* Post  : All appropriate menu entries have been ghosted.                                                                        */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if (WhichWindow == TWLEFT)
  {
    MELSaveAs.Flags |= MEF_GHOSTED;
    MELSave.Flags |= MEF_GHOSTED;
    MELBackup.Flags |= MEF_GHOSTED;
    MELSelNone.Flags |= MEF_GHOSTED;
    MELInvSel.Flags |= MEF_GHOSTED;
    MELDelete.Flags |= MEF_GHOSTED;
    // MELResize.Flags |= MEF_GHOSTED;
    MELDupEntry.Flags |= MEF_GHOSTED;
    MELPlay.Flags |= MEF_GHOSTED;
    MELSaveRaw.Flags |= MEF_GHOSTED;
    MELDecDirRec.Flags |= MEF_GHOSTED;
    MELEdHeader.Flags |= MEF_GHOSTED;
    MELMtchBlock.Flags |= MEF_GHOSTED;
    MELBlocks.Flags |= MEF_GHOSTED;
    MELViewInfo.Flags |= MEF_GHOSTED;
    MELViewAs.Flags |= MEF_GHOSTED;
  }
  else
  {
    MERSaveAs.Flags |= MEF_GHOSTED;
    MERSave.Flags |= MEF_GHOSTED;
    MERBackup.Flags |= MEF_GHOSTED;
    MERSelNone.Flags |= MEF_GHOSTED;
    MERInvSel.Flags |= MEF_GHOSTED;
    MERDelete.Flags |= MEF_GHOSTED;
    // MERResize.Flags |= MEF_GHOSTED;
    MERDupEntry.Flags |= MEF_GHOSTED;
    MERPlay.Flags |= MEF_GHOSTED;
    MERSaveRaw.Flags |= MEF_GHOSTED;
    MERDecDirRec.Flags |= MEF_GHOSTED;
    MEREdHeader.Flags |= MEF_GHOSTED;
    MERMtchBlock.Flags |= MEF_GHOSTED;
    MERBlocks.Flags |= MEF_GHOSTED;
    MERViewInfo.Flags |= MEF_GHOSTED;
    MERViewAs.Flags |= MEF_GHOSTED;
  }
}

void UnghostMenu (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT and the tape is no longer empty.                                             */
/* Post  : All appropriate menu entries have been unghosted.                                                                      */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if (WhichWindow == TWLEFT)
  {
    MELSaveAs.Flags &= ~MEF_GHOSTED;
    MELSave.Flags &= ~MEF_GHOSTED;
    MELBackup.Flags &= ~MEF_GHOSTED;
    MELSelNone.Flags &= ~MEF_GHOSTED;
    MELInvSel.Flags &= ~MEF_GHOSTED;
    MELDelete.Flags &= ~MEF_GHOSTED;
    // MELResize.Flags &= ~MEF_GHOSTED;
    MELDupEntry.Flags &= ~MEF_GHOSTED;
    MELMtchBlock.Flags &= ~MEF_GHOSTED;
    if ((TPConfig.PlaySoundBlaster && SoundBlasterInfo.Detected) || !TPConfig.PlaySoundBlaster)
      MELPlay.Flags &= ~MEF_GHOSTED;
    MELSaveRaw.Flags &= ~MEF_GHOSTED;
    MELDecDirRec.Flags &= ~MEF_GHOSTED;
    MELViewInfo.Flags &= ~MEF_GHOSTED;
    MELViewAs.Flags &= ~MEF_GHOSTED;
  }
  else
  {
    MERSaveAs.Flags &= ~MEF_GHOSTED;
    MERSave.Flags &= ~MEF_GHOSTED;
    MERBackup.Flags &= ~MEF_GHOSTED;
    MERSelNone.Flags &= ~MEF_GHOSTED;
    MERInvSel.Flags &= ~MEF_GHOSTED;
    MERDelete.Flags &= ~MEF_GHOSTED;
    // MERResize.Flags &= ~MEF_GHOSTED;
    MERDupEntry.Flags &= ~MEF_GHOSTED;
    MERMtchBlock.Flags &= ~MEF_GHOSTED;
    if ((TPConfig.PlaySoundBlaster && SoundBlasterInfo.Detected) || !TPConfig.PlaySoundBlaster)
      MERPlay.Flags &= ~MEF_GHOSTED;
    MERSaveRaw.Flags &= ~MEF_GHOSTED;
    MERDecDirRec.Flags &= ~MEF_GHOSTED;
    MERViewInfo.Flags &= ~MEF_GHOSTED;
    MERViewAs.Flags &= ~MEF_GHOSTED;
  }
}

void GhostMenuBackup (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : The 'Backup' menu entry has been ghosted.                                                                              */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if (WhichWindow == TWLEFT)
    MELBackup.Flags |= MEF_GHOSTED;
  else
    MERBackup.Flags |= MEF_GHOSTED;
}

void UnghostMenuBackup (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : The 'Backup' menu entry has been unghosted.                                                                            */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if (WhichWindow == TWLEFT)
    MELBackup.Flags &= ~MEF_GHOSTED;
  else
    MERBackup.Flags &= ~MEF_GHOSTED;
}

void GhostMenuDecDirRec (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : The 'Decode entry' menu entry has been ghosted.                                                                        */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if (WhichWindow == TWLEFT)
    MELDecDirRec.Flags |= MEF_GHOSTED;
  else
    MERDecDirRec.Flags |= MEF_GHOSTED;
}

void UnghostMenuDecDirRec (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : The 'Decode entry' menu entry has been unghosted.                                                                      */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if (WhichWindow == TWLEFT)
    MELDecDirRec.Flags &= ~MEF_GHOSTED;
  else
    MERDecDirRec.Flags &= ~MEF_GHOSTED;
}

void GhostMenuEdHeader (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : The 'Decode block' menu entry has been ghosted.                                                                        */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if (WhichWindow == TWLEFT)
    MELEdHeader.Flags |= MEF_GHOSTED;
  else
    MEREdHeader.Flags |= MEF_GHOSTED;
}

void UnghostMenuEdHeader (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : The 'Decode block' menu entry has been unghosted.                                                                      */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if (WhichWindow == TWLEFT)
    MELEdHeader.Flags &= ~MEF_GHOSTED;
  else
    MEREdHeader.Flags &= ~MEF_GHOSTED;
}

void GhostMenuViewAs (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : The 'View as' and 'Save raw data' submenus have been ghosted.                                                          */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if (WhichWindow == TWLEFT)
  {
    MELViewAs.Flags |= MEF_GHOSTED;
    MELSaveRaw.Flags |= MEF_GHOSTED;
  }
  else
  {
    MERViewAs.Flags |= MEF_GHOSTED;
    MERSaveRaw.Flags |= MEF_GHOSTED;
  }
}

void UnghostMenuViewAs (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : The 'View as' and 'Save raw data' submenus have been unghosted.                                                        */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if (WhichWindow == TWLEFT)
  {
    MELViewAs.Flags &= ~MEF_GHOSTED;
    MELSaveRaw.Flags &= ~MEF_GHOSTED;
  }
  else
  {
    MERViewAs.Flags &= ~MEF_GHOSTED;
    MERSaveRaw.Flags &= ~MEF_GHOSTED;
  }
}

void InstallMenu (struct WindowInfo *Window)

/**********************************************************************************************************************************/
/* Pre   : `Window' points to the window to which the menu strip should be attached.                                              */
/* Post  : The menu strip has been attached to this window.                                                                       */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct MenuInfo *MenuP;

  Window->MenuFont = _UsedMenuFont;
  MenuP = AddMenu (Window, "\x11 Window", 0);
  AddMenuEntry (MenuP, &MELNew);
  AddMenuEntry (MenuP, &MELOpen);
  AddMenuEntry (MenuP, &MELSaveAs);
  AddMenuEntry (MenuP, &MELSave);
  AddMenuEntry (MenuP, &MELSampler);
  AddMenuEntry (MenuP, &MELImport);
  AddMenuEntry (MenuP, &MELTapeCRC);
  AddMenuEntry (MenuP, &MELPlayTime);
  AddMenuEntry (MenuP, &MESeperate);
  AddMenuEntry (MenuP, &MELBackup);
  AddMenuEntry (MenuP, &MELAutoDecr);
  AddMenuEntry (MenuP, &MELDecrAny);
  AddMenuEntry (MenuP, &MELSetDecr);
  AddMenuEntry (MenuP, &MESeperate);
  AddMenuEntry (MenuP, &MEMtchTapes);
  AddMenuEntry (MenuP, &MEQuit);

  MenuP = AddMenu (Window, "\x11 Entry", 0);
  AddMenuEntry (MenuP, &MELSelNone);
  AddMenuEntry (MenuP, &MELInvSel);
  AddMenuEntry (MenuP, &MELInsert);
  AddMenuEntry (MenuP, &MELSaveRaw);
  AddMenuEntry (MenuP, &MELDelete);
  ConvertToSubMenu (MenuP, &MELDelete);
  AddMenuEntry (MELDelete.SubMenu, &MELDelEntry);
  AddMenuEntry (MELDelete.SubMenu, &MELDelSel);
  // AddMenuEntry (MenuP, &MELResize);
  // ConvertToSubMenu (MenuP, &MELResize);
  // AddMenuEntry (MELResize.SubMenu, &MELResizeUp);
  // AddMenuEntry (MELResize.SubMenu, &MELResizeDn);
  AddMenuEntry (MenuP, &MELDupEntry);
  AddMenuEntry (MenuP, &MELDecDirRec);
  AddMenuEntry (MenuP, &MELEdHeader);
  AddMenuEntry (MenuP, &MELBlocks);
  AddMenuEntry (MenuP, &MELMtchBlock);
  ConvertToSubMenu (MenuP, &MELMtchBlock);
  AddMenuEntry (MELMtchBlock.SubMenu, &MELMtchThis);
  AddMenuEntry (MELMtchBlock.SubMenu, &MELMtchOther);
  AddMenuEntry (MELMtchBlock.SubMenu, &MELBlockCRC);
  ConvertToSubMenu (MenuP, &MELBlocks);
  AddMenuEntry (MELBlocks.SubMenu, &MELMakeGroup);
  AddMenuEntry (MELBlocks.SubMenu, &MELMakeLoop);
  AddMenuEntry (MELBlocks.SubMenu, &MELEdAllTim);
  AddMenuEntry (MELBlocks.SubMenu, &MELSGetPlTim);
  AddMenuEntry (MELBlocks.SubMenu, &MELSCheckCRC);
  AddMenuEntry (MenuP, &MESeperate);
  AddMenuEntry (MenuP, &MELPlay);
  ConvertToSubMenu (MenuP, &MELPlay);
  AddMenuEntry (MELPlay.SubMenu, &MELPlayEntry);
  AddMenuEntry (MELPlay.SubMenu, &MELPlaySel);
  AddMenuEntry (MELPlay.SubMenu, &MELPlayFrom);
  AddMenuEntry (MELPlay.SubMenu, &MELPlayTape);
  AddMenuEntry (MenuP, &MESeperate);
  AddMenuEntry (MenuP, &MELViewAs);
  ConvertToSubMenu (MenuP, &MELViewAs);
  AddMenuEntry (MELViewAs.SubMenu, &MELViewDump);
  AddMenuEntry (MELViewAs.SubMenu, &MELViewHead);
  AddMenuEntry (MELViewAs.SubMenu, &MELViewBasic);
  AddMenuEntry (MELViewAs.SubMenu, &MELViewScrn);
  AddMenuEntry (MELViewAs.SubMenu, &MELViewDiss);
  AddMenuEntry (MenuP, &MELViewInfo);

  MenuP = AddMenu (Window, "\x10 Window", 0);
  AddMenuEntry (MenuP, &MERNew);
  AddMenuEntry (MenuP, &MEROpen);
  AddMenuEntry (MenuP, &MERSaveAs);
  AddMenuEntry (MenuP, &MERSave);
  AddMenuEntry (MenuP, &MERSampler);
  AddMenuEntry (MenuP, &MERImport);
  AddMenuEntry (MenuP, &MERTapeCRC);
  AddMenuEntry (MenuP, &MERPlayTime);
  AddMenuEntry (MenuP, &MESeperate);
  AddMenuEntry (MenuP, &MERBackup);
  AddMenuEntry (MenuP, &MERAutoDecr);
  AddMenuEntry (MenuP, &MERDecrAny);
  AddMenuEntry (MenuP, &MERSetDecr);

  MenuP = AddMenu (Window, "\x10 Entry", 0);
  AddMenuEntry (MenuP, &MERSelNone);
  AddMenuEntry (MenuP, &MERInvSel);
  AddMenuEntry (MenuP, &MERInsert);
  AddMenuEntry (MenuP, &MERSaveRaw);
  AddMenuEntry (MenuP, &MERDelete);
  ConvertToSubMenu (MenuP, &MERDelete);
  AddMenuEntry (MERDelete.SubMenu, &MERDelEntry);
  AddMenuEntry (MERDelete.SubMenu, &MERDelSel);
  // AddMenuEntry (MenuP, &MERResize);
  // ConvertToSubMenu (MenuP, &MERResize);
  // AddMenuEntry (MERResize.SubMenu, &MERResizeUp);
  // AddMenuEntry (MERResize.SubMenu, &MERResizeDn);
  AddMenuEntry (MenuP, &MERDupEntry);
  AddMenuEntry (MenuP, &MERDecDirRec);
  AddMenuEntry (MenuP, &MEREdHeader);
  AddMenuEntry (MenuP, &MERMtchBlock);
  ConvertToSubMenu (MenuP, &MERMtchBlock);
  AddMenuEntry (MERMtchBlock.SubMenu, &MERMtchThis);
  AddMenuEntry (MERMtchBlock.SubMenu, &MERMtchOther);
  AddMenuEntry (MERMtchBlock.SubMenu, &MERBlockCRC);
  AddMenuEntry (MenuP, &MERBlocks);
  ConvertToSubMenu (MenuP, &MERBlocks);
  AddMenuEntry (MERBlocks.SubMenu, &MERMakeGroup);
  AddMenuEntry (MERBlocks.SubMenu, &MERMakeLoop);
  AddMenuEntry (MERBlocks.SubMenu, &MEREdAllTim);
  AddMenuEntry (MERBlocks.SubMenu, &MERSGetPlTim);
  AddMenuEntry (MERBlocks.SubMenu, &MERSCheckCRC);
  AddMenuEntry (MenuP, &MESeperate);
  AddMenuEntry (MenuP, &MERPlay);
  ConvertToSubMenu (MenuP, &MERPlay);
  AddMenuEntry (MERPlay.SubMenu, &MERPlayEntry);
  AddMenuEntry (MERPlay.SubMenu, &MERPlaySel);
  AddMenuEntry (MERPlay.SubMenu, &MERPlayFrom);
  AddMenuEntry (MERPlay.SubMenu, &MERPlayTape);
  AddMenuEntry (MenuP, &MESeperate);
  AddMenuEntry (MenuP, &MERViewAs);
  ConvertToSubMenu (MenuP, &MERViewAs);
  AddMenuEntry (MERViewAs.SubMenu, &MERViewDump);
  AddMenuEntry (MERViewAs.SubMenu, &MERViewHead);
  AddMenuEntry (MERViewAs.SubMenu, &MERViewBasic);
  AddMenuEntry (MERViewAs.SubMenu, &MERViewScrn);
  AddMenuEntry (MERViewAs.SubMenu, &MERViewDiss);
  AddMenuEntry (MenuP, &MERViewInfo);

  MenuP = AddMenu (Window, "System", 0);
  AddMenuEntry (MenuP, &MEQuiet);
  AddMenuEntry (MenuP, &MESoundVol);
  AddMenuEntry (MenuP, &MESeperate);
  AddMenuEntry (MenuP, &MEEmulFlsh);
  AddMenuEntry (MenuP, &MESlowFlsh);
  AddMenuEntry (MenuP, &METraceSnd);
  AddMenuEntry (MenuP, &MEShowTime);
}

bool ForceNewTapeName (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : The user has been forced a 'Save As...' requester and the tape has been saved with this name.                          */
/* Import: _SetupTapePath.                                                                                                        */
/**********************************************************************************************************************************/

{
  bool Done = FALSE;
  char NewPath[MAXDIR + 1];

  _SetupTapePath (WhichWindow);
  while (!Done)
    if (FileRequester (180, 50, 48, 29, 20, _Path, _File, _Mask, "Choose New Tape Name", FALSE, FALSE))
    {
      if (MaxFloppyIdChanged)
        WriteConfig ();
      MakeFullPath (NewPath, _Path, _File);
      if (!TestFileOverwrite (NewPath))
      {
        SetTapeName (WhichWindow, NewPath, TRUE);
        strcpy (_PreviousTapePath[WhichWindow], _Path);
        if (TapeWindow[WhichWindow].FileType == FT_UNKNOWN)
          ErrorRequester (-1, -1, "@Close", NULL, NULL, "{BERROR:}B\n  %s\n  unknown Spectrum tape file format", _File);
        else if (TapeWindow[WhichWindow].FileType == FT_ZXS)
          ErrorRequester (-1, -1, "@Close", NULL, NULL, "Sorry, the ZXS file format is read-only");
        else if (IsSnapshot (WhichWindow))
          ErrorRequester (-1, -1, "@Close", NULL, NULL, "Sorry, snapshot formats are read-only");
        else if (TapeWindow[WhichWindow].FileType == FT_BADSOUND)
          ErrorRequester (-1, -1, "@Close", NULL, NULL, "{BERROR:}B\n  %s is not understood.\n"
                          "  Please use the .VOC format", _File);
        else
        {
          TapeWindow[WhichWindow].BackupTape = FALSE;                                                /* No need for a tape backup */
          TapeWindow[WhichWindow].NewTape = FALSE;
          return (TRUE);
        }
      }
      else
        _SetupTapePath (WhichWindow);
    }
    else
    {
      if (MaxFloppyIdChanged)
        WriteConfig ();
      return (FALSE);
    }
  return (FALSE);
}

bool TestFileType (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : The file type is tested. If it was FT_ZXS, the user has been asked for a new output file.                              */
/*         TRUE is returned if the type is not FT_ZXS or the user chose a new (valid) output file.                                */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  bool Done = FALSE;
  bool Cancelled = FALSE;

  while (!Done)
    if (IsSnapshot (WhichWindow) || TapeWindow[WhichWindow].FileType == FT_ZXS)
      switch (QuestionRequester (-1, -1, FALSE, FALSE, "@Cancel", "@Reselect", NULL, NULL, NULL, NULL,
                                 "%s read-only.\nSelect new outputfile?",
                                 IsSnapshot (WhichWindow) ? "Snapshot file formats are" : "The ZXS file format is"))
      {
        case 0 : Done = ForceNewTapeName (WhichWindow);                                                                  /* (Yes) */
                 break;
        case 1 : Done = TRUE; Cancelled = TRUE; break;                                                                    /* (No) */
      }
    else
      Done = TRUE;
  return (!Cancelled);
}

void TestForSelectedEntries (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : A test if done to see whether there are selected tape entries. Depending on these, the submenu `Selected block'        */
/*         and its entries are (un)ghosted.                                                                                       */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirP;
  bool                AtLeastOne = FALSE;
  bool                SelectGap  = FALSE;
  bool                NotSel     = TRUE;

  DirP = TapeWindow[WhichWindow].Directory;
  while (DirP != NULL)
  {
    if (DirP->ComboSelected & SELECTBIT)
    {
      if (!AtLeastOne)                                                                      /* First encountered selected entry ? */
        AtLeastOne = TRUE;
      else if (NotSel)                                               /* Selected entries, then nonselected, then selected again ? */
        SelectGap = TRUE;                                                                                       /* Signal the gap */
      NotSel = FALSE;
    }
    else
      NotSel = TRUE;
    DirP = DirP->Next;
  }
  if (AtLeastOne)
  {
    if (WhichWindow == TWLEFT)
      MELBlocks.Flags &= ~MEF_GHOSTED;
    else
      MERBlocks.Flags &= ~MEF_GHOSTED;
    if (SelectGap)
      if (WhichWindow == TWLEFT)
      {
        MELMakeGroup.Flags |= MEF_GHOSTED;
        MELMakeLoop.Flags |= MEF_GHOSTED;
      }
      else
      {
        MERMakeGroup.Flags |= MEF_GHOSTED;
        MERMakeLoop.Flags |= MEF_GHOSTED;
      }
    else
      if (WhichWindow == TWLEFT)
      {
        MELMakeGroup.Flags &= ~MEF_GHOSTED;
        MELMakeLoop.Flags &= ~MEF_GHOSTED;
      }
      else
      {
        MERMakeGroup.Flags &= ~MEF_GHOSTED;
        MERMakeLoop.Flags &= ~MEF_GHOSTED;
      }
  }
  else
    if (WhichWindow == TWLEFT)
      MELBlocks.Flags |= MEF_GHOSTED;
    else
      MERBlocks.Flags |= MEF_GHOSTED;
}

void DeleteEntryByHand (void)

/**********************************************************************************************************************************/
/* Pre   : None ([DEL] was pressed in the main handler).                                                                          */
/* Post  : The same action is taken as when [ALT][D] was pressed.                                                                 */
/* Import: _HandleMEDelEntry.                                                                                                     */
/**********************************************************************************************************************************/

{
  if (!TapeWindow[_ActiveWindowMenu].EmptyTape)
    if (_ActiveWindowMenu == TWLEFT)                                             /* Delete entry from the currently active window */
      _HandleMEDelEntry (&MELDelEntry);
    else
      _HandleMEDelEntry (&MERDelEntry);
}
