/**********************************************************************************************************************************/
/* Project      : TAPER                                                                                                           */
/* Module       : TPSNAP.C                                                                                                        */
/* Description  : Snapshot to tape conversion                                                                                     */
/* Version type : Program module                                                                                                  */
/* Last changed : 09-07-1998  19:30                                                                                               */
/* Update count : 7                                                                                                               */
/* OS type      : PC (DOS)                                                                                                        */
/* Compiler     : Watcom C32 Optimizing Compiler Version 10.x                                                                     */
/* Linker       : Watcom C Linker                                                                                                 */
/*                                                                                                                                */
/*                Copyleft (C) 1997-2001 ThunderWare Research Center, written by Martijn van der Heide.                           */
/*                                                                                                                                */
/*                This program is free software; you can redistribute it and/or                                                   */
/*                modify it under the terms of the GNU General Public License                                                     */
/*                as published by the Free Software Foundation; either version 2                                                  */
/*                of the License, or (at your option) any later version.                                                          */
/*                                                                                                                                */
/*                This program is distributed in the hope that it will be useful,                                                 */
/*                but WITHOUT ANY WARRANTY; without even the implied warranty of                                                  */
/*                MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                                                   */
/*                GNU General Public License for more details.                                                                    */
/*                                                                                                                                */
/*                You should have received a copy of the GNU General Public License                                               */
/*                along with this program; if not, write to the Free Software                                                     */
/*                Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                                     */
/*                                                                                                                                */
/**********************************************************************************************************************************/

#define __TPSNAP_MAIN__

#include <io.h>
#include <fcntl.h>
#include <direct.h>
#include <dos.h>
#include "wbench32/wbench32.h"
#include "tpscreen.h"
#include "tpfile.h"
#include "tpaudio.h"
#include "tpconfig.h"
#include "tpmain.h"
#include "tpdisass.h"
#include "tpsnap.h"

#pragma pack          ()

/**********************************************************************************************************************************/
/* Define all info needed to work with snapshot files                                                                             */
/**********************************************************************************************************************************/

static char *_InvalidZ80File = "Invalid Z80 file %s";
static char *_InvalidSNAFile = "Invalid SNA file %s";

#define MAXZ80PAGES 8

static struct _Z80DataMap_s                                                              /* Full description/layout of a snapshot */
{
  word                          _Version;                                                             /* Normally 145, 201 or 300 */
  bool                          _VersionValid;
  bool                          _128KSnapshot;                                              /* TRUE when dealing with a 128K snap */
  struct Z80145SnapshotHeader_s _145SnapshotHeader;                                                     /* Snapshot header chunks */
  struct Z80201SnapshotHeader_s _201SnapshotHeader;
  struct Z80300SnapshotHeader_s _300SnapshotHeader;
  struct _MemoryStructure_s                                                /* Layout of the memory pages inside the snapshot file */
  {
    dword _Offset;                                            /* File offset pointing to the start of the (compressed) page image */
    word  _Length;                                                                            /* Length of the (compressed) image */
    bool  _InUse;                                                             /* TRUE if this page was found in the snapshot file */
  }                             _MemoryStructure[MAXZ80PAGES];
} _Z80DataMap;                                                                              /* (Filled after calling _GetZ80Info) */

/**********************************************************************************************************************************/
/* Here goes:                                                                                                                     */
/* The BASIC header                                                                                                               */
/**********************************************************************************************************************************/

static struct _SpectrumBASICHeader_s
{
  byte      __Code1[1];
  char      _FileName[10];
  byte      __Code2[6];
} _SpectrumBASICHeader = {{ "\x00" },
                          { "\x11\x05\x08\x08\x08\x08\x08\x08\x08\x08" },
                          { "\xE0\x02\x00\x00\xCB\x00" }};

/**********************************************************************************************************************************/
/* The BASIC body                                                                                                                 */
/**********************************************************************************************************************************/

#define LOADERPREPIECE  224                                                    /* Length of the BASIC part before the loader code */
static byte _SpectrumBASICData[LOADERPREPIECE] = {

   /* This first line is the only loaded one if it is MERGEd! (next line has length 65535 :-) */
   /* 0 REM {AT 2,5}{PAPER 0}{INK 7}TAPER Snapshot Loader{AT 4,7}Copyright (C) 1998{AT 6,3}ThunderWare Research Center{INK 255} */
   "\x00\x00\x53\x00\xEA"
   "\x16\x02\x05\x11\x00\x10\x07""TAPER Snapshot Loader"
   "\x16\x04\x07""Copyright (C) 1998"
   "\x16\x06\x03""ThunderWare Research Center\x10\xFF\x0D"

   /* 0 {INK 255}{PAPER 255}BORDER PI-PI : PAPER PI-PI : INK PI-PI : CLS
        PRINT "{AT 10,6}{INK 5}Loading snapshot ...
        {AT 12,9}{INK 6}{PAPER 2}{FLASH 1} PLEASE WAIT {AT 0,0}{FLASH 0}{PAPER 0}{INK 0}":
        RANDOMIZE USR (PEEK VAL "23627"+VAL "256"*PEEK VAL "23628") */
   "\x00\x00\xFF\xFF\x10\xFF\x11\xFF\xE7\xA7\x2D\xA7\x3A\xDA\xA7\x2D\xA7\x3A\xD9\xA7\x2D\xA7\x3A\xFB\x3A"
   "\xF5\"\x16\x0A\x06\x10\x05""Loading snapshot ..."
   "\x16\x0C\x09\x10\x06\x11\x02\x12\x01 PLEASE WAIT \x16\x00\x00\x12\x00\x11\x00\x10\x00\""
   "\x3A\xF9\xC0(\xBE\xB0\"23627\"+\xB0\"256\"*\xBE\xB0\"23628\")\x0D"

   /* Variables area - the 512 byte loader block is appended after this piece */
   "\xCD\x52\x00"         /* VARSTART  CALL 0052,RET_INSTR   ; Determine on which address we are now             */
   "\x3B"                 /* RETRET    DEC  SP                                                                   */
   "\x3B"                 /*           DEC  SP                                                                   */
   "\xE1"                 /*           POP  HL                                                                   */
   "\x01\x12\x00"         /*           LD   BC,0012                                                              */
   "\x09"                 /*           ADD  HL,BC            ; Find start of the 512 byte loader                 */
   "\x11\x00\xBE"         /*           LD   DE,+BE00                                                             */
   "\x01\x00\x02"         /*           LD   BC,+0200                                                             */
   "\xED\xB0"             /*           LDIR                  ; Move it into place                                */
   "\xC3\x00\xBE" };      /*           JP   BE00,LDSTART     ; And run it!                                       */

/**********************************************************************************************************************************/
/* The load/decrunch table handler in Z80 code - equal for both 48K and 128K programs                                             */
/**********************************************************************************************************************************/

static struct _SpectrumLoader_s
{
  byte      __Z80Code1[6];
  word      _LoadTable;
  byte      __Z80Code2[2];
  word      _IY;
  byte      __Z80Code3[1];
  word      _HL_a;
  byte      __Z80Code4[1];
  word      _DE_a;
  byte      __Z80Code5[1];
  word      _BC_a;
  byte      __Z80Code6[45];
  word      _LoadBytesRoutine;
  byte      __Z80Code7[6];
  word      _DecrunchAddress;
  byte      __Z80Code8[4];
} _SpectrumLoader   = {
  { "\xF3"                /* LDSTART   DI                                                                        */
    "\x31\x00\xC0"        /*           LD   SP,+C000         ; Corrupt only the 512 bytes space                  */
    "\xDD\x21" },         /*           LD   IX,TABLE         ; Point to the start of the page table              */
    0x0000,
  { "\xFD\x21" },         /*           LD   IY,SNAP_IY       ; These don't change anymore                        */
    0x0000,
  { "\x21" },             /*           LD   HL,SNAP_HL'                                                          */
    0x0000,
  { "\x11" },             /*           LD   DE,SNAP_DE'                                                          */
    0x0000,
  { "\x01" },             /*           LD   BC,SNAP_BC'                                                          */
    0x0000,
  { "\xD9"                /*           EXX                                                                       */
    "\xDD\x7E\x00"        /* LOADNEXT  LD   A,(IX+00)        ; Consider each page in the table                   */
    "\xA7"                /*           AND  A                ; Endmarker?                                        */
    "\x28\x34"            /*           JR   Z,LOADEND        ; Step out if so                                    */
    "\xDD\x66\x01"        /*           LD   H,(IX+01)        ; Determine start address                           */
    "\x2E\xFF"            /*           LD   L,+FF                                                                */
    "\xDD\x5E\x02"        /*           LD   E,(IX+02)        ; Get length                                        */
    "\xDD\x56\x03"        /*           LD   D,(IX+03)                                                            */
    "\x01\xFD\x7F"        /*           LD   BC,+7FFD         ; Set page register                                 */
    "\xED\x79"            /*           OUT  (C),A                                                                */
    "\x01\x04\x00"        /*           LD   BC,+0004         ; Prepare for next page                             */
    "\xDD\x09"            /*           ADD  IX,BC                                                                */
    "\xDD\xE5"            /*           PUSH IX                                                                   */
    "\xD5"                /*           PUSH DE               ; Keep a copy of the (unmodified) length            */
    "\x7A"                /*           LD   A,D              ; Block length 0?                                   */
    "\xB3"                /*           OR   E                                                                    */
    "\x20\x03"            /*           JR   NZ,GOLOAD                                                            */
    "\x11\x00\x40"        /*           LD   DE,+4000         ; Then it's not crunched and 16384 bytes are loaded */
    "\xE5"                /* GOLOAD    PUSH HL               ; Pass start to IX                                  */
    "\xDD\xE1"            /*           POP  IX                                                                   */
    "\x3E\xAA"            /*           LD   A,+AA            ; Flagbyte 0xAA                                     */
    "\x37"                /*           SCF                                                                       */
    "\xCD" },             /*           CALL LD_BLOCK         ; Load the block (custom loader)                    */
    0x0000,
  { "\xC1"                /*           POP  BC               ; BC = length                                       */
    "\xDD\xE5"            /*           PUSH IX                                                                   */
    "\xE1"                /*           POP  HL               ; HL = start - 1                                    */
    "\x23"                /*           INC  HL               ; HL = start                                        */
    "\xCD" },             /*           CALL DECRUNCH         ; Decrunch (if crunched)                            */
    0x0000,
  { "\xDD\xE1"            /*           POP  IX                                                                   */
    "\x18\xC6" }};        /*           JR   LOADNEXT                                                             */
#define _SpectrumLoaderLength  (sizeof (struct _SpectrumLoader_s))

/**********************************************************************************************************************************/
/* The decruncher in Z80 code                                                                                                     */
/* Requires: HL = start of crunched code                                                                                          */
/*           BC = length of crunched code                                                                                         */
/* Decrunches a 16K block that is loaded at the end of that block from the start of that same block.                              */
/**********************************************************************************************************************************/

static byte _SpectrumDecruncher[40] = {
    "\x7C"                /* DECSTART  LD   A,H              ; DE := HL % +C000                                  */
    "\xE6\xC0"            /*           AND  +C0              ; = start of that page                              */
    "\x57"                /*           LD   D,A                                                                  */
    "\x1E\x00"            /*           LD   E,0                                                                  */
    "\x78"                /* CHECKEND  LD   A,B              ; This nicely stops non-crunched blocks             */
    "\xB1"                /*           OR   C                                                                    */
    "\xC8"                /*           RET  Z                                                                    */
    "\x7E"                /* NEXTBYTE  LD   A,(HL)           ; Consider next byte                                */
    "\xFE\xED"            /*           CP   +ED              ; Crunch start?                                     */
    "\x20\x05"            /*           JR   NZ,NOEXPAND      ; Nope - go on                                      */
    "\x23"                /*           INC  HL               ; Next byte is the crunch confirm?                  */
    "\xBE"                /*           CP   (HL)                                                                 */
    "\x28\x05"            /*           JR   Z,DOEXPAND       ; Yes, handle it                                    */
    "\x2B"                /*           DEC  HL                                                                   */
    "\xED\xA0"            /* NOEXPAND  LDI                   ; Pass the byte straight on                         */
    "\x18\xEF"            /*           JR   CHECKEND         ; Next!                                             */
    "\x23"                /* DOEXPAND  INC  HL                                                                   */
    "\xC5"                /*           PUSH BC               ; Remember length left to go                        */
    "\x46"                /*           LD   B,(HL)           ; Fetch running length                              */
    "\x23"                /*           INC  HL                                                                   */
    "\x7E"                /*           LD   A,(HL)           ; Fetch repeated code                               */
    "\x23"                /*           INC  HL                                                                   */
    "\x12"                /* EXPNEXT   LD   (DE),A           ; Expand                                            */
    "\x13"                /*           INC  DE                                                                   */
    "\x10\xFC"            /*           DJNZ EXPNEXT                                                              */
    "\xC1"                /*           POP  BC               ; Crunch header is 4 bytes                          */
    "\x0B"                /*           DEC  BC                                                                   */
    "\x0B"                /*           DEC  BC                                                                   */
    "\x0B"                /*           DEC  BC                                                                   */
    "\x0B"                /*           DEC  BC                                                                   */
    "\x18\xDE" };         /*           JR   CHECKEND         ; Next!                                             */
#define _SpectrumDecruncherLength  40

/**********************************************************************************************************************************/
/* 48K programs continue here after all parts are loaded                                                                          */
/**********************************************************************************************************************************/

static struct _Spectrum48Load_s
{
  byte      __Z80Code1[13];
  byte      _AttrColour;
  byte      __Z80Code2[5];
  word      _FinalCodeAddress;
  byte      __Z80Code3[4];
  byte      _FinalCodeLength;
  byte      __Z80Code4[6];
  byte      _FinalStackLength;
  byte      __Z80Code5[5];
} _Spectrum48Load   = {
  { "\x01\xFD\x7F"        /* LDEND     LD   BC,+7FFD         ; Force +7FFD register into sane state              */
    "\x3E\x10"            /*           LD   A,+10            ; as the final block is loaded into '48K' memory    */
    "\xED\x79"            /*           OUT  (C),A                                                                */
    "\x21\xE0\x5A"        /*           LD   HL,+5AE0         ; Last attributes line                              */
    "\x06\x20"            /*           LD   B,+20            ; Fill the entire line                              */
    "\x3E" },             /*           LD   A,COLOR          ; with the chosen attribute colour                  */
    0x00,
  { "\x77"                /* FILLATTR  LD   (HL),A                                                               */
    "\x23"                /*           INC  HL                                                                   */
    "\x10\xFC"            /*           DJNZ FILLATTR         ; (B will nicely be 0 after this)                   */
    "\x21" },             /*           LD   HL,FINCODE       ; _Final48Loader                                    */
    0x0000,
  { "\x11\xE0\x56"        /*           LD   DE,CODEDEST      ; 1-last display line                               */
    "\x0E" },             /*           LD   C,CODELEN                                                            */
    0x00,
  { "\xED\xB0"            /*           LDIR                                                                      */
    "\x11\xEA\x57"        /*           LD   DE,STKDEST       ; _FinalStackFrame goes to last display line        */
    "\x0E" },             /*           LD   C,STKLEN                                                             */
    0x00,
  { "\xED\xB0"            /*           LDIR                                                                      */
    "\x31\xEA\x57" }};    /*           LD   SP,STKDEST+10    ; Point SP to the pre-stacked return value (=+56E0) */
#define _Spectrum48LoadLength  (sizeof (struct _Spectrum48Load_s))

/**********************************************************************************************************************************/
/* 48K programs are checked for all zeroes at the loader space - which is at least always true with 16K programs.                 */
/* If not, this routine is used, which loads the next block from the tape and overwrites the loader space with it.                */
/* Otherwise, there's no need to load an all zeroes block and a routine to clear the loader space is put instead.                 */
/**********************************************************************************************************************************/

static struct _Spectrum48Load21_s
{
  byte      __Z80Code1[14];
} _Spectrum48Load21 = {
  { "\xDD\x21\x00\xBE"    /*           LD   IX,START         ; Overwrite this 512 bytes block                    */
    "\x11\x00\x02"        /*           LD   DE,LENGTH                                                            */
    "\x3E\x55"            /*           LD   A,+55            ; Flagbyte 0x55                                     */
    "\xBC"                /*           CP   H                ; Cf := 1, Zf := 0                                  */
    "\x08"                /*           EX   AF,AF'                                                               */
    "\xC3\x62\x05" }};    /*           JP   0562,LD_BLOCK+12 ; Effectively goes to _Final48Loader after this     */
#define _Spectrum48Load21Length  (sizeof (struct _Spectrum48Load21_s))

static struct _Spectrum48Load22_s
{
  byte      __Z80Code1[4];
  byte      _ExtraLength;
  byte      __Z80Code2[4];
} _Spectrum48Load22 = {
  { "\x11\xE0\x55"        /*           LD   DE,EXTRACODE     ; Stick the clear code in the 2-last display line   */
    "\x0E" },             /*           LD   C,EXTRALEN                                                           */
    0x00,
  { "\xD5"                /*           PUSH DE               ; Store start address                               */
    "\xED\xB0"            /*           LDIR                                                                      */
    "\xC9" }};            /*           RET                   ; Jump to the clear code                            */
#define _Spectrum48Load22Length  (sizeof (struct _Spectrum48Load22_s))
static struct _Spectrum48Load23_s
{
  byte      __Z80Code1[13];
} _Spectrum48Load23 = {
  { "\x21\x00\xBE"        /*           LD   HL,+BE00         ; Simply clear the 512 loader code                  */
    "\x11\x01\xBE"        /*           LD   DE,+BE01                                                             */
    "\x01\xFF\x01"        /*           LD   BC,+01FF                                                             */
    "\x75"                /*           LD   (HL),L                                                               */
    "\xED\xB0"            /*           LDIR                                                                      */
    "\xC9" }};            /*           RET                   ; Jump to _Final48Loader                            */
#define _Spectrum48Load23Length  (sizeof (struct _Spectrum48Load23_s))

/**********************************************************************************************************************************/
/* 128K programs continue here after all parts are loaded                                                                         */
/**********************************************************************************************************************************/

static struct _Spectrum128Load_s
{
  byte      __Z80Code1[13];
  byte      _AttrColour;
  byte      __Z80Code2[5];
  word      _FinalCode1Address;
  byte      __Z80Code3[4];
  byte      _FinalCode1Length;
  byte      __Z80Code4[6];
  byte      _FinalCode2Length;
  byte      __Z80Code5[6];
  byte      _FinalStack1Length;
  byte      __Z80Code6[6];
  byte      _FinalStack2Length;
  byte      __Z80Code7[19];
} _Spectrum128Load  = {
  { "\x01\xFD\x7F"        /* LDEND     LD   BC,+7FFD         ; Force +7FFD register into sane state              */
    "\x3E\x10"            /*           LD   A,+10            ; as the final block is loaded into '48K' memory    */
    "\xED\x79"            /*           OUT  (C),A                                                                */
    "\x21\xE0\x5A"        /*           LD   HL,+5AE0         ; Last attributes line                              */
    "\x06\x20"            /*           LD   B,+20            ; Fill the entire line                              */
    "\x3E" },             /*           LD   A,COLOR          ; with the chosen attribute colour                  */
    0x00,
  { "\x77"                /* FILLATTR  LD   (HL),A                                                               */
    "\x23"                /*           INC  HL                                                                   */
    "\x10\xFC"            /*           DJNZ FILLATTR         ; (B will nicely be 0 after this)                   */
    "\x21" },             /*           LD   HL,FINCOD1       ; _Final48Loader                                    */
    0x0000,
  { "\x11\xE0\x56"        /*           LD   DE,COD1DEST      ; 1-last display line                               */
    "\x0E" },             /*           LD   C,COD1LEN                                                            */
    0x00,
  { "\xED\xB0"            /*           LDIR                                                                      */
    "\x11\xE0\x55"        /*           LD   DE,COD2DEST      ; _Final128Loader goes to 2-last display line       */
    "\x0E" },             /*           LD   C,COD2LEN                                                            */
    0x00,
  { "\xED\xB0"            /*           LDIR                                                                      */
    "\x11\xEA\x57"        /*           LD   DE,STK1DEST      ; _FinalStackFrame goes to last display line        */
    "\x0E" },             /*           LD   C,STK1LEN                                                            */
    0x00,
  { "\xED\xB0"            /*           LDIR                                                                      */
    "\x11\xE0\x54"        /*           LD   DE,STK2DEST      ; _Final128SoundStackFrame goes to 3-last line      */
    "\x0E" },             /*           LD   C,STK2LEN                                                            */
    0x00,
  { "\xED\xB0"            /*           LDIR                                                                      */
    "\x31\xEA\x57"        /*           LD   SP,STK1DEST+10   ; Point SP to the pre-stacked return value (=+55E0) */
    "\xDD\x21\x00\xBE"    /*           LD   IX,START         ; Overwrite this 512 bytes block                    */
    "\x11\x00\x02"        /*           LD   DE,LENGTH                                                            */
    "\x3E\x55"            /*           LD   A,+55            ; Flagbyte 0x55                                     */
    "\xBC"                /*           CP   H                ; Cf := 1, Zf := 0                                  */
    "\x08"                /*           EX   AF,AF'                                                               */
    "\xC3\x62\x05" }};    /*           JP   0562,LD_BLOCK+12 ; Effectively goes to _Final128Loader after this    */
#define _Spectrum128LoadLength  (sizeof (struct _Spectrum128Load_s))

/**********************************************************************************************************************************/
/* This code is executed by 128K programs only and sets the registers of the AY chip. It continues in the 48K final part below    */
/**********************************************************************************************************************************/

static struct _Final128Loader_s
{
  byte      __Z80Code1[26];
  byte      _StateFFFD;
  byte      __Z80Code2[5];
} _Final128Loader   = {
  { "\x31\xE0\x54"        /*           LD   SP,STK2DEST      ; Point SP to _Final128SoundStackFrame              */
    "\x16\x10"            /*           LD   D,+10            ; 16 registers to fill                              */
    "\x01\xFD\xFF"        /* NEXTSND   LD   BC,+FFFD                                                             */
    "\x15"                /*           DEC  D                                                                    */
    "\xED\x51"            /*           OUT  (C),D            ; Select register                                   */
    "\xF1"                /*           POP  AF               ; Fetch value to set                                */
    "\x06\xBF"            /*           LD   B,+BF                                                                */
    "\xED\x79"            /*           OUT  (C),A            ; Set register's value                              */
    "\x7A"                /*           LD   A,D                                                                  */
    "\xA7"                /*           AND  A                                                                    */
    "\x20\xF1"            /*           JR   NZ,NEXTSND       ; Next register                                     */
    "\x31\xEC\x57"        /*           LD   SP,STK1DEST+12   ; Set SP to continue to _Final48Loader              */
    "\x06\xFF"            /*           LD   B,+FF            ; Set AY register index from the snap               */
    "\x3E" },             /*           LD   A,LASTFFFD                                                           */
    0x00,
  { "\xED\x79"            /*           OUT  (C),A                                                                */
    "\xC3\xE0\x56" }};    /*           JP   FINCOD1          ; Continue in _Final48Loader                        */
#define _Final128LoaderLength  (sizeof (struct _Final128Loader_s))

/**********************************************************************************************************************************/
/* This code is executed by all programs. It sets the remaining registers and starts the program                                  */
/**********************************************************************************************************************************/

static struct _Final48Loader_s
{
  byte      __Z80Code1[23];
  word      _SP;
  byte      __Z80Code4[1];
  byte      _IMValue;
  byte      _DIorEI;
  byte      __Z80Code5[1];
  word      _PC;
} _Final48Loader    = {
  { "\xF1"                /*           POP  AF               ; Set border colour                                 */
    "\xD3\xFE"            /*           OUT  (+FE),A                                                              */
    "\xC1"                /*           POP  BC               ; BC := +7FFD                                       */
    "\xF1"                /*           POP  AF               ; A := value for this register                      */
    "\xED\x79"            /*           OUT  (C),A            ; Set page register                                 */
    "\xDD\xE1"            /*           POP  IX               ; Set IX                                            */
    "\xF1"                /*           POP  AF               ; Set AF'                                           */
    "\x08"                /*           EX   AF,AF'                                                               */
    "\xE1"                /*           POP  HL                                                                   */
    "\x7C"                /*           LD   A,H                                                                  */
    "\xED\x47"            /*           LD   I,A              ; Set I                                             */
    "\x7D"                /*           LD   A,L                                                                  */
    "\xED\x4F"            /*           LD   R,A              ; Set R (increases by 9 by the next instructions!)  */
    "\xE1"                /*           POP  HL               ; Set HL                                            */
    "\xD1"                /*           POP  DE               ; Set DE                                            */
    "\xC1"                /*           POP  BC               ; Set BC                                            */
    "\xF1"                /*           POP  AF               ; Set AF                                            */
    "\x31" },             /*           LD   SP,SNAP_SP       ; Set SP                                            */
    0x0000,
  { "\xED" },             /*           IM   x                ; Select interrupt mode                             */
    0x00,
    0x00,                 /*          (DI or EI)             ; Select IFF1                                       */
  { "\xC3" },             /*           JP   SNAP_PC          ; Run the snap!                                     */
    0x0000 };
#define _Final48LoaderLength  (sizeof (struct _Final48Loader_s))

/**********************************************************************************************************************************/
/* The stack used by the _Final128Loader. It contains the values for the AY registers                                             */
/**********************************************************************************************************************************/

static struct _Final128SoundStackFrame_s
{
  byte _Dummy;
  byte _State;
} _Final128SoundStackFrame[16];
#define _Final128SoundStackFrameLength  32

/**********************************************************************************************************************************/
/* The stack used by the _Final48Loader                                                                                           */
/**********************************************************************************************************************************/

static struct _FinalStackFrame_s
{
  word _ReturnAddress;
  byte _Dummy1;
  byte _BorderColour;
  word _Value7FFD;
  byte _Dummy2;
  byte _State7FFD;
  word _IX;
  byte _F_a;
  byte _A_a;
  byte _R;
  byte _I;
  word _HL_n;
  word _DE_n;
  word _BC_n;
  byte _F_n;
  byte _A_n;
} _FinalStackFrame;
#define _FinalStackFrameLength  (sizeof (_FinalStackFrame))

/**********************************************************************************************************************************/
/* The custom loader. Any speed can be handled by it.                                                                             */
/* Notice that the EAR bit value is NOT rotated as opposed to the ROM and is kept at 0x40!                                        */
/* ROM speed: a = 22, b = 25 (bd = 27) => delay = 32*22 = 704, sampler = 59*25 = 1475 => delay/sampler = 1/2                      */
/* Max speed: a = 1, b = 0 (bd = 2) => 279 + 32 = 311 (hb0 = 156 T) => 468 average bit => 7479 bps                                */
/* (see below) The pilot pulse must be between 1453 + 16a T and 3130 + 16a T (a=known)                                            */
/* (see below) The first sync pulse must be 840 + 16a T maximum (a=known)                                                         */
/* Average bit length = 3,500,000 / speed : (a=x, b=2x - so delay/sampler is kept at 1/2)                                         */
/*  1364 bps => 2566 T => hb0 = 855 T, hb1 = 1710 T, avg = 1283 (2565) T       <-- ROM speed                                      */
/*              279 + 32a + 43b = 2565 => 32x + 86x = 2286 => 118x = 2286 => x = 19.4                                             */
/*              279 + 32*20 + 43*39 = 2596 (2565 needed), so a = 20, b = 39, bd = 41                                              */
/*              PilotMin = 1453 + 16a = 1453 + 16*20 = 1773 T                                                                     */
/*              PilotMax = 3130 + 16a = 3130 + 16*20 = 3450 T => 2168 T                                                           */
/*              Sync0 = 840 + 16a = 840 + 16*20 = 1160 T => 667 T                                                                 */
/*  2250 bps => 1556 T => hb0 = 518 T, hb1 = 1036 T, avg = 777 (1554) T                                                           */
/*              279 + 32a + 43b = 1554 => 32x + 86x = 1275 => 118x = 1275 => x = 10.8                                             */
/*              279 + 32*11 + 43*22 = 1577 (1554 needed), so a = 11, b = 22, bd = 24                                              */
/*              PilotMin = 1453 + 16a = 1453 + 16*11 = 1629 T                                                                     */
/*              PilotMax = 3130 + 16a = 3130 + 16*11 = 3306 T => 2000 T                                                           */
/*              Sync0 = 840 + 16a = 840 + 16*11 = 1016 T => 600 T                                                                 */
/*  3000 bps => 1167 T => hb0 = 389 T, hb1 = 778 T, avg = 584 (1167) T                                                            */
/*              279 + 32a + 43b = 1167 => 32x + 86x = 888 => 118x = 888 => x = 7.5                                                */
/*              279 + 32*7 + 43*16 = 1191 (1167 needed), so a = 7, b = 16, bd = 18                                                */
/*              PilotMin = 1453 + 16a = 1453 + 16*7 = 1565 T                                                                      */
/*              PilotMax = 3130 + 16a = 3130 + 16*7 = 3242 T => 1900 T                                                            */
/*              Sync0 = 840 + 16a = 840 + 16*7 = 952 T => 550 T                                                                   */
/*  6000 bps => 584 T => hb0 = 195 T, hb1 = 390 T, avg = 293 (585) T                                                              */
/*              279 + 32a + 43b = 585 => 32x + 86x = 306 => 118x = 306 => x = 2.6                                                 */
/*              279 + 32*3 + 43*5 = 590 (585 needed), so a = 3, b = 5, bd = 7                                                     */
/*              PilotMin = 1453 + 16a = 1453 + 16*3 = 1501 T                                                                      */
/*              PilotMax = 3130 + 16a = 3130 + 16*3 = 3178 T => 1700 T                                                            */
/*              Sync0 = 840 + 16a = 840 + 16*3 = 888 T => 450 T                                                                   */
/**********************************************************************************************************************************/

static struct _TurboLoadVars_s
{
  /* Values as stored inside the TurboLoader code */
  byte      _Compare;    /* Variable 'bd' + starting value (+80) */
  byte      _Delay;      /* Variable 'a' */
  /* Timing values as stored in the TZX block header (unless ROM speed) */
  word      _LenPilot;   /* Number of pilot pulses is calculated to make the pilot take exactly 1 second */
                         /* This is just enough for the largest possible block to decompress and sync again */
  word      _LenSync0;   /* Both sync values are made equal */
  word      _Len0;       /* A '1' bit gets twice this value */
} _TurboLoadVars[4] = {{ 0x80 + 41, 20, 2168, 667, 855 },   /*  1364 bps - uses the normal ROM timing values! */
                       { 0x80 + 24, 11, 2000, 600, 518 },   /*  2250 bps */
                       { 0x80 + 18,  7, 1900, 550, 389 },   /*  3000 bps */
                       { 0x80 +  7,  3, 1700, 450, 195 }};  /*  6000 bps */

static struct _TurboLoader_s
{
  byte      __Z80Code1[94];
  byte      _Compare;                                                       /* Both _Delay and _Compare come from the above table */
  byte      __Z80Code2[23];
  byte      _Delay;
  byte      __Z80Code3[15];
  byte      _ColourXOR;                                                                    /* This value MUST be (0x40 | colour)! */
  byte      __Z80Code4[9];
} _TurboLoader   = {
  { "\xCD\x5B\xBF"        /* LD_BLOCK  CALL LD_BYTES         */
    "\xD8"                /*           RET  C                */
    "\xCF"                /*           RST  0008,ERROR_1     */
    "\x1A"                /*           DEFB +1A ; R Tape Loading Error */
    "\x14"                /* LD_BYTES  INC  D                */
    "\x08"                /*           EX   AF,AF'           */
    "\x15"                /*           DEC  D                */
    "\x3E\x08"            /*           LD   A,+08            (Border BLACK + MIC off) */
    "\xD3\xFE"            /*           OUT  (+FE),A          */
    "\xDB\xFE"            /*           IN   A,(+FE)          */
    "\xE6\x40"            /*           AND  +40              */
    "\x4F"                /*           LD   C,A              (BLACK/chosen colour) */
    "\xBF"                /*           CP   A                */
    "\xC0"                /* LD_BREAK  RET  NZ               */
    "\xCD\xCA\xBF"        /* LD_START  CALL LD_EDGE_1        */
    "\x30\xFA"            /*           JR   NC,LD_BREAK      */
    "\x26\x00"            /*           LD   H,+80            */

    /* Each subsequent 2 edges take:   65 + LD_EDGE_2 T = 283 + 32a + 43b T, bd = b + 2                      */
    /* Minimum:                                                                                              */
    /*   ROM => b = (+C6 - +9C - 2) + 1 = 41, a=22 => 427 + 32a + 59b T => 427 + 32*22 + 59*41 T = 3550 T    */
    /*          283 + 32a + 43b = 3432, a=20 => 283 + 32*20 + 43b = 3550 => 43b = 2627 => b = 61, bd = 63    */
    /*   b = 61, so each 2 leader pulses must fall within:                                                   */
    /*   283 + 32a + 43*61 T = 2906 + 32a T                                                                  */
    /*   One leader pulse must therefore be (2906 + 32a) / 2 = 1453 + 16a T minimum                          */
    /* Maximum, at timeout:                                                                                  */
    /*   ROM => b = +100 - +9C - 2 = 98, a=22 => 427 + 32a + 59b T => 427 + 32*22 + 59*98 T = 6913 T         */
    /*          283 + 32a + 43b = 6913, a=20 => 283 + 32*20 + 43b = 6913 => 43b = 6002 => b = 139, bd = 141  */
    /*   b = 139, so each 2 leader pulses must fall within:                                                  */
    /*   283 + 32a + 43*139 T = 6260 + 32a T                                                                 */
    /*   One leader pulse must therefore be (6260 + 32a) / 2 = 3130 + 16a T maximum                          */
    /* Notice that, as opposed to the ROM, we don't need the leader to be over a second. Only 256 pulses are */
    /* required (128 * 2 edges, counted in H).                                                               */

    "\x06\x75"            /* LD_LEADER LD   B,+73            7       */
    "\xCD\xC6\xBF"        /*           CALL LD_EDGE_2        17      */
    "\x30\xF1"            /*           JR   NC,LD_BREAK      7 (/12) */
    "\x3E\xB0"            /*           LD   A,+B2            7       */
    "\xB8"                /*           CP   B                4       */
    "\x30\xED"            /*           JR   NC,LD_START      7 (/12) */
    "\x24"                /*           INC  H                4       */
    "\x20\xF1"            /*           JR   NZ,LD_LEADER     7/12    */

    /* Each subsequent edge takes:   54 + LD_EDGE_1 T = 152 + 16a + 43b T, bd = b + 1                        */
    /* Sync0 maximum:                                                                                        */
    /*   ROM => b = +D4 - +C9 - 1 = 10, a=22 => 224 + 16a + 59b T => 224 + 16*22 + 59*10 T = 1166 T          */
    /*          152 + 16a + 43b = 1166, a=20 => 152 + 16*20 + 43b = 1166 => 43b = 694 => b = 16, bd = 17     */
    /*   b = 16, so the (first) sync pulse must be less than:                                                */
    /*   152 + 16a + 43*16 T = 840 + 16a T                                                                   */
    /* Leader maximum (at timeout):                                                                          */
    /*   ROM => b = +100 - +C9 - 1 = 56, a=22 => 224 + 16a + 59b T => 224 + 16*22 + 59*56 T = 3880 T         */
    /*          152 + 16a + 43b = 3880, a=20 => 152 + 16*20 + 43b = 3880 => 43b = 3408 => b = 79, bd = 80    */
    /*   Notice that this leader maximum is higher than above, so is not needed to be considered             */

    "\x06\xB0"            /* LD_SYNC   LD   B,+B0            7       */
    "\xCD\xCA\xBF"        /*           CALL LD_EDGE_1        17      */
    "\x30\xE2"            /*           JR   NC,LD_BREAK      7 (/12) */
    "\x78"                /*           LD   A,B              4       */
    "\xFE\xC1"            /*           CP   +C1              7       */
    "\x30\xF4"            /*           JR   NC,LD_SYNC       7/12    */

    "\xCD\xCA\xBF"        /*           CALL LD_EDGE_1        17      */
    "\xD0"                /*           RET  NC               5 (/11) */
    "\x26\x00"            /*           LD   H,+00            7       */
    "\x06\x80"            /*           LD   B,+80            7       */
    "\x18\x17"            /*           JR   LD_MARKER        12   = 48S */

    "\x08"                /* LD_LOOP   EX   AF,AF'           4       */
    "\x20\x05"            /*           JR   NZ,LD_FLAG       7/12 = 16F */
    "\xDD\x75\x00"        /*           LD   (IX+00),L        19      */
    "\x18\x09"            /*           JR   LD_NEXT          12   = 42D */

    "\xCB\x11"            /* LD_FLAG   RL   C                4       */
    "\xAD"                /*           XOR  L                4       */
    "\xC0"                /*           RET  NZ               5 (/11) */
    "\x79"                /*           LD   A,C              4       */
    "\x1F"                /*           RRA                   4       */
    "\x4F"                /*           LD   C,A              4       */
    "\x18\x03"            /*           JR   LD_FLNEXT        12   = 37F */

    "\xDD\x2B"            /* LD_NEXT   DEC  IX               10      (We're loading BACKWARD!) */
    "\x1B"                /*           DEC  DE               6       */
    "\x08"                /* LD_FLNEXT EX   AF,AF'           4       */
    "\x06\x82"            /*           LD   B,+82            7       */
    "\x2E\x01"            /* LD_MARKER LD   L,+01            7    = 34D/18F/7S */

    /* The very first (flag) bit takes (S) = 48 + 7 = 55 + LD_EDGE_2 T = 273 + 32a + 43b T                       */
    /* (D) 55 + 32 + 42 + 34 = 163, so                                                                           */
    /* Each first (data) bit takes: 163 + LD_EDGE_2 - 2b T = 295 + 32a + 43b T (ROM = 405 + 32a + 59b T)         */
    /* Each subsequent bit takes:   60 + LD_EDGE_2       T = 278 + 32a + 43b T (ROM = 422 + 32a + 59b T)         */

    "\xCD\xC6\xBF"        /* LD_8_BITS CALL LD_EDGE_2        17      */
    "\xD0"                /*           RET  NC               5 (/11) */
    "\x3E" },             /*           LD   A,COMPARE        7    (variable 'bd') */
    0x00,
  { "\xB8"                /*           CP   B                4       */
    "\xCB\x15"            /*           RL   L                8       */
    "\x06\x80"            /*           LD   B,+80            7       */
    "\x30\xF3"            /*           JR   NC,LD_8_BITS     7/12 = 55D/60N */

    "\x7C"                /*           LD   A,H              4       */
    "\xAD"                /*           XOR  L                4       */
    "\x67"                /*           LD   H,A              4       */
    "\x7A"                /*           LD   A,D              4       */
    "\xB3"                /*           OR   E                4       */
    "\x20\xD3"            /*           JR   NZ,LD_LOOP       (7/) 12 = 32D */
    "\x7C"                /*           LD   A,H                      */
    "\xFE\x01"            /*           CP   +01                      */
    "\xC9"                /*           RET                           */

    /* The LD_EDGE_2 routine takes:  22 + 2 * (98 + 16a) + 43b T = 218 + 32a + 43b T (ROM = 362 + 32a + 59b T), bd = b + 2 */

    "\xCD\xCA\xBF"        /* LD_EDGE_2 CALL LD_EDGE_1        17      */
    "\xD0"                /*           RET  NC               5 (/11) */

    /* The LD_EDGE_1 routine takes:  98 + 16a + 43b T (ROM = 170 + 16a + 59b T, a=22), bd = b + 1 */

    "\x3E" },             /* LD_EDGE_1 LD   A,DELAY          7    (variable 'a') */
    0x00,
  { "\x3D"                /* LD_DELAY  DEC  A                4       */
    "\x20\xFD"            /*           JR   NZ,LD_DELAY      7/12    */
    "\xA7"                /*           AND  A                4       */

    "\x04"                /* LD_SAMPLE INC  B                4    (variable 'b') */
    "\xC8"                /*           RET  Z                5 (/11) */
    "\xDB\xFE"            /*           IN   A,(+FE)          11      */
    "\xA9"                /*           XOR  C                4       */
    "\xE6\x40"            /*           AND  +40              7       */
    "\x28\xF7"            /*           JR   Z,LD_SAMPLE      7/12 = 43 */

    "\x79"                /*           LD   A,C              4       */
    "\xEE" },             /*           XOR  COLOUR           7       */
    0x00,
  { "\x4F"                /*           LD   C,A              4       */
    "\xE6\x07"            /*           AND  +07              7       */
    "\xF6\x08"            /*           OR   +08              7       */
    "\xD3\xFE"            /*           OUT  (+FE),A          11      */
    "\x37"                /*           SCF                   4       */
    "\xC9" }};            /*           RET                   10      */
#define _TurboLoaderLength  (sizeof (_TurboLoader))

static byte _UsedLoadingSpeed = 2;

/**********************************************************************************************************************************/
/* Define the static routines                                                                                                     */
/**********************************************************************************************************************************/

static bool _GetZ80Info              (int Fh, dword FileLength, char *FullPath, int FileType);
static bool _DecrunchZ80Block        (byte *BufferIn, word BlLength, byte *BufferOut, word OutLength);
static void _CrunchZ80Block          (byte *BufferIn, word BlLength, byte *BufferOut, word *CrunchedLength);
static void _ReverseBlock            (byte *BufferOut, byte *BufferIn);
static void _AddAnyBlockToTape       (short WhichWindow, struct TapeEntry_s *Entry);
static void _AddDataBlockToTape      (short WhichWindow, struct TapeEntry_s *Entry);
static void _TBTimedFlash            (struct WindowInfo *Window);
static void _ViewScreenBlock         (byte *Screen);
static bool _SetupConversionUserInfo (byte *BorderColour, bool *CompressAll);

/**********************************************************************************************************************************/
/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> START OF THE INTERNAL ROUTINES <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
/**********************************************************************************************************************************/

static bool _GetZ80Info (int Fh, dword FileLength, char *FullPath, int FileType)

/**********************************************************************************************************************************/
/* Pre   : `FullPath' holds the complete path to the file.                                                                        */
/* Post  : The version of the .Z80 file has been determined. If it is a valid file, TRUE is returned.                             */
/*         .SNA files are converted to (noncompressed) v1.45 .Z80 files.                                                          */
/*         The block 4000-7FFF is loaded and decompressed at _ObjectData, the block 8000-BFFF is loaded and decompressed at       */
/*         (_ObjectData + 16384). v1.45 .Z80 (and .SNA) files are fully loaded and decompressed at _ObjectData.                   */
/* Import: _DecrunchZ80Block.                                                                                                     */
/**********************************************************************************************************************************/

{
  struct MMSnapshotHeader_s SNASnapshotHeader;
  struct BlockInfo_s
  {
    word     Length;
    byte     Number;
  }          BlockInfo;
  dword      Offset         = sizeof (struct Z80145SnapshotHeader_s);
  word       NRead;
  word       CrunchedLength;

  memset (&_Z80DataMap, 0, sizeof (struct _Z80DataMap_s));
  if (FileType == FT_SNA)
  {
    if (FileLength != ((dword)49152 + sizeof (struct MMSnapshotHeader_s)))
    { ErrorRequester (-1, -1, "@Close", NULL, NULL, _InvalidSNAFile, FullPath); return (FALSE); }
    if (read (Fh, &SNASnapshotHeader, sizeof (struct MMSnapshotHeader_s)) == -1)
    { ErrorRequester (-1, -1, "@Close", NULL, NULL, _FileReadError, FullPath, strerror (errno)); return (FALSE); }
    if (read (Fh, _ObjectData, 49152) < 49152)
    { ErrorRequester (-1, -1, "@Close", NULL, NULL, _FileReadError, FullPath, strerror (errno));
      return (FALSE); }
    _Z80DataMap._Version = 145;                                               /* Make equivalent to non-compressed Z80 v1.45 snap */
    _Z80DataMap._VersionValid = TRUE;
    _Z80DataMap._145SnapshotHeader.A_n   = (SNASnapshotHeader.AF_n & 0xFF00) >> 8;
    _Z80DataMap._145SnapshotHeader.F_n   = SNASnapshotHeader.AF_n & 0x00FF;
    _Z80DataMap._145SnapshotHeader.BC_n  = SNASnapshotHeader.BC_n;
    _Z80DataMap._145SnapshotHeader.DE_n  = SNASnapshotHeader.DE_n;
    _Z80DataMap._145SnapshotHeader.HL_n  = SNASnapshotHeader.HL_n;
    _Z80DataMap._145SnapshotHeader.A_a   = (SNASnapshotHeader.AF_a & 0xFF00) >> 8;
    _Z80DataMap._145SnapshotHeader.F_a   = SNASnapshotHeader.AF_a & 0x00FF;
    _Z80DataMap._145SnapshotHeader.BC_a  = SNASnapshotHeader.BC_a;
    _Z80DataMap._145SnapshotHeader.DE_a  = SNASnapshotHeader.DE_a;
    _Z80DataMap._145SnapshotHeader.HL_a  = SNASnapshotHeader.HL_a;
    _Z80DataMap._145SnapshotHeader.IX    = SNASnapshotHeader.IX;
    _Z80DataMap._145SnapshotHeader.IY    = SNASnapshotHeader.IY;
    _Z80DataMap._145SnapshotHeader.SP    = SNASnapshotHeader.SP;
    _Z80DataMap._145SnapshotHeader.I     = SNASnapshotHeader.I;
    _Z80DataMap._145SnapshotHeader.R     = SNASnapshotHeader.R & 0x7F;
    _Z80DataMap._145SnapshotHeader.IFF1  = _Z80DataMap._145SnapshotHeader.IFF2 = SNASnapshotHeader.IFF1;
    _Z80DataMap._145SnapshotHeader.Stat1 = ((SNASnapshotHeader.R & 0x80) >> 7) |
                                           ((*(_ObjectData + 0x1C48) & 38) >> 2);             /* (Border colour = 0x5C48 bit 3-5) */
                                                                                                         /* (Leave bits 4-7 at 0) */
    _Z80DataMap._145SnapshotHeader.Stat2 = SNASnapshotHeader.IM;                                         /* (Leave bits 2-7 at 0) */
    _Z80DataMap._145SnapshotHeader.PC = *(_ObjectData + SNASnapshotHeader.SP - 0x4000) |            /* Fetch PC off the Z80 stack */
                                        (*(_ObjectData + SNASnapshotHeader.SP - 0x3FFF) << 8);
    _Z80DataMap._145SnapshotHeader.SP    = SNASnapshotHeader.SP + 2;
    return (TRUE);
  }
  if (FileLength < sizeof (struct Z80145SnapshotHeader_s))
  { ErrorRequester (-1, -1, "@Close", NULL, NULL, _InvalidZ80File, FullPath); return (FALSE); }
  if (read (Fh, &(_Z80DataMap._145SnapshotHeader), sizeof (struct Z80145SnapshotHeader_s)) == -1)
  { ErrorRequester (-1, -1, "@Close", NULL, NULL, _FileReadError, FullPath, strerror (errno)); return (FALSE); }
  if (_Z80DataMap._145SnapshotHeader.Stat1 == 0xFF)
    _Z80DataMap._145SnapshotHeader.Stat1 = 0x01;
  if (_Z80DataMap._145SnapshotHeader.PC != 0)
  {
    _Z80DataMap._Version = 145;
    _Z80DataMap._VersionValid = TRUE;
    if (_Z80DataMap._145SnapshotHeader.Stat1 & 0x20)                                                              /* Compressed ? */
    {                                                                      /* Load and decrunch the entire block into _ObjectData */
      CrunchedLength = FileLength - sizeof (struct Z80145SnapshotHeader_s);
      if (read (Fh, WorkBuffer, CrunchedLength) < CrunchedLength)
      { ErrorRequester (-1, -1, "@Close", NULL, NULL, _FileReadError, FullPath, strerror (errno)); return (FALSE); }
      if (!_DecrunchZ80Block (WorkBuffer, CrunchedLength, _ObjectData, 49152))
      { ErrorRequester (-1, -1, "@Close", NULL, NULL, _InvalidZ80File, FullPath); return (FALSE); }
    }
    else
    {
      if (read (Fh, _ObjectData, 49152) < 49152)
      { ErrorRequester (-1, -1, "@Close", NULL, NULL, _FileReadError, FullPath, strerror (errno));
        return (FALSE); }
    }
    return (TRUE);
  }
  if (FileLength < (sizeof (struct Z80145SnapshotHeader_s) + sizeof (struct Z80201SnapshotHeader_s)))
  { ErrorRequester (-1, -1, "@Close", NULL, NULL, _InvalidZ80File, FullPath); return (FALSE); }
  if (read (Fh, &(_Z80DataMap._201SnapshotHeader), sizeof (struct Z80201SnapshotHeader_s)) == -1)
  { ErrorRequester (-1, -1, "@Close", NULL, NULL, _FileReadError, FullPath, strerror (errno)); return (FALSE); }
  _Z80DataMap._Version = 201;
  _Z80DataMap._128KSnapshot = _Z80DataMap._201SnapshotHeader.HWMode >= 3 ? TRUE : FALSE;
  Offset += sizeof (struct Z80201SnapshotHeader_s);
  if (_Z80DataMap._201SnapshotHeader.AdditionalLength == 54)
  {
    if (FileLength < (sizeof (struct Z80145SnapshotHeader_s) + sizeof (struct Z80201SnapshotHeader_s) +
                      sizeof (struct Z80300SnapshotHeader_s)))
    { ErrorRequester (-1, -1, "@Close", NULL, NULL, _InvalidZ80File, FullPath); return (FALSE); }
    if (read (Fh, &(_Z80DataMap._300SnapshotHeader), sizeof (struct Z80300SnapshotHeader_s)) == -1)
    { ErrorRequester (-1, -1, "@Close", NULL, NULL, _FileReadError, FullPath, strerror (errno)); return (FALSE); }
    _Z80DataMap._Version = 300;
    _Z80DataMap._128KSnapshot = _Z80DataMap._201SnapshotHeader.HWMode >= 4 ? TRUE : FALSE;
    Offset += sizeof (struct Z80300SnapshotHeader_s);
  }
  _Z80DataMap._VersionValid = TRUE;
  if ((NRead = read (Fh, &BlockInfo, 3)) == -1)
  { ErrorRequester (-1, -1, "@Close", NULL, NULL, _FileReadError, FullPath, strerror (errno)); return (FALSE); }
  while (NRead == 3)                                                                                       /* More block follow ? */
  {
    if (BlockInfo.Number >= 12)
    { ErrorRequester (-1, -1, "@Close", NULL, NULL, _InvalidZ80File, FullPath); return (FALSE); }
    Offset += sizeof (struct BlockInfo_s);
    _Z80DataMap._MemoryStructure[BlockInfo.Number - 3]._Offset = Offset;
    _Z80DataMap._MemoryStructure[BlockInfo.Number - 3]._Length = BlockInfo.Length;
    _Z80DataMap._MemoryStructure[BlockInfo.Number - 3]._InUse  = TRUE;
    if (BlockInfo.Length == 0xFFFF)                                                       /* Z80 v3.05: This block not crunched ? */
      Offset += 16384;                                                                          /* Then count exactly 16384 bytes */
    else
      Offset += BlockInfo.Length;
    if (Offset > FileLength)
    { ErrorRequester (-1, -1, "@Close", NULL, NULL, _InvalidZ80File, FullPath); return (FALSE); }
    if (BlockInfo.Number == 8 ||                                                                             /* Page 5 (screen) ? */
       (_Z80DataMap._128KSnapshot && BlockInfo.Number == 5) ||                                                /* Page 8000-BFFF ? */
       (!_Z80DataMap._128KSnapshot && BlockInfo.Number == 4))
    {                                                                        /* Then read and decrunch the block into _ObjectData */
      if (BlockInfo.Length == 0xFFFF)
      {
        if (read (Fh, _ObjectData + (BlockInfo.Number == 8 ? 0 : 16384), 16384) < 16384)
        { ErrorRequester (-1, -1, "@Close", NULL, NULL, _FileReadError, FullPath, strerror (errno)); return (FALSE); }
      }
      else
      {
        if (read (Fh, WorkBuffer, BlockInfo.Length) < BlockInfo.Length)
        { ErrorRequester (-1, -1, "@Close", NULL, NULL, _FileReadError, strerror (errno)); return (FALSE); }
        _DecrunchZ80Block (WorkBuffer, BlockInfo.Length, _ObjectData + (BlockInfo.Number == 8 ? 0 : 16384), 16384);
      }
    }
    else                                                                                                  /* (Skip if not page 5) */
    {
      if (lseek (Fh, Offset, SEEK_SET) == -1)
      { ErrorRequester (-1, -1, "@Close", NULL, NULL, _FileSearchError, FullPath, strerror (errno)); return (FALSE); }
    }
    if ((NRead = read (Fh, &BlockInfo, 3)) == -1)
    { ErrorRequester (-1, -1, "@Close", NULL, NULL, _FileReadError, FullPath, strerror (errno)); return (FALSE); }
  }
  return (TRUE);
}

static bool _DecrunchZ80Block (byte *BufferIn, word BlLength, byte *BufferOut, word OutLength)

/**********************************************************************************************************************************/
/* Pre   : `BufferIn' points to the crunched spectrum block, `BlLength' holds the length, `BufferOut' points to the result.       */
/*         `OutLength' is the expected length after decrunching.                                                                  */
/*         The version of Z80 file must have been determined.                                                                     */
/* Post  : The block has been decrunched from `BufferIn' to `BufferOut'. The return value is TRUE if all went well.               */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  word register IndexIn;
  word register IndexOut       = 0;
  word register RunTimeLength;

  if (_Z80DataMap._Version == 145)
    BlLength -= 4;                                                                           /* Do not `decrunch' the end marker! */
  for (IndexIn = 0 ; IndexIn < BlLength ; IndexIn ++)                                       /* Decrunch it into the second buffer */
  {
    if (IndexOut >= OutLength)                                                                           /* Past maximum length ? */
      return (FALSE);                                                                           /* Then there's an error for sure */
    if (*(BufferIn + IndexIn) != 0xED)                                                                  /* Start of crunch code ? */
      *(BufferOut + (IndexOut ++)) = *(BufferIn + IndexIn);                                                      /* No, just copy */
    else
      if (*(BufferIn + IndexIn + 1) != 0xED)                                                                  /* Crunch confirm ? */
        *(BufferOut + (IndexOut ++)) = *(BufferIn + IndexIn);                                                    /* No, just copy */
      else                                                                                                      /* Yes, decrunch! */
      {
        RunTimeLength = *(BufferIn + IndexIn + 2);                                                             /* Run time length */
        while (RunTimeLength -- > 0 && IndexOut < OutLength)
          *(BufferOut + (IndexOut ++)) = *(BufferIn + IndexIn + 3);
        IndexIn += 3;
      }
  }
  if (_Z80DataMap._Version == 145)
  {
    if (IndexOut != OutLength)
      return (FALSE);
    if (*(BufferIn + IndexIn ++) != 0x00 || *(BufferIn + IndexIn ++) != 0xED ||
        *(BufferIn + IndexIn ++) != 0xED || *(BufferIn + IndexIn ++) != 0x00)
      return (FALSE);
    return (TRUE);
  }
  else
    return (IndexOut == OutLength);
}

static void _CrunchZ80Block (byte *BufferIn, word BlLength, byte *BufferOut, word *CrunchedLength)

/**********************************************************************************************************************************/
/* Pre   : `BufferIn' points to the uncrunched spectrum block, `BlLength' holds the length, `BufferOut' points to the result.     */
/*         `CrunchedLength' is the length after crunching.                                                                        */
/*         The version of Z80 file must have been determined.                                                                     */
/* Post  : The block has been crunched from `BufferIn' to `BufferOut'.                                                            */
/*         If the crunched size is larger than (or equal to) the non-crunched size, the block is returned directly with           */
/*         `CrunchedLength' set to 0x0000.                                                                                        */
/*         The crunched block is returned reversed (for backward loading).                                                        */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  word register IndexIn;
  word register IndexOut       = 0;
  word          LengthOut;
  word register RunTimeLength;
  bool          CrunchIt       = FALSE;
  byte          RepeatedCode;
  byte          PrevByte       = 0x00;
  byte         *BufferTemp;

  BufferTemp = WorkBuffer + (word)32768;                                                        /* Use the top 16Kb of Workbuffer */
  for (IndexIn = 0 ; IndexIn < BlLength - 4 ; IndexIn ++)                                     /* Crunch it into the second buffer */
  {
    if (*(BufferIn + IndexIn) == 0xED &&                                                                     /* Exceptional cases */
        *(BufferIn + IndexIn + 1) == 0xED)
      CrunchIt = TRUE;
    else if (*(BufferIn + IndexIn) == *(BufferIn + IndexIn + 1) &&
             *(BufferIn + IndexIn) == *(BufferIn + IndexIn + 2) &&
             *(BufferIn + IndexIn) == *(BufferIn + IndexIn + 3) &&
             *(BufferIn + IndexIn) == *(BufferIn + IndexIn + 4) &&                                        /* At least 5 repeats ? */
             PrevByte != 0xED)                                                                         /* Not right after a 0xED! */
      CrunchIt = TRUE;
    if (CrunchIt)                                                                                             /* Crunch a block ? */
    {
      CrunchIt = FALSE;
      RunTimeLength = 1;
      RepeatedCode = *(BufferIn + IndexIn);
      while (RunTimeLength < 255 && RunTimeLength + IndexIn < BlLength && *(BufferIn + RunTimeLength + IndexIn) == RepeatedCode)
        RunTimeLength ++;
      *(BufferTemp + (IndexOut ++)) = 0xED;
      *(BufferTemp + (IndexOut ++)) = 0xED;
      *(BufferTemp + (IndexOut ++)) = RunTimeLength;
      *(BufferTemp + (IndexOut ++)) = RepeatedCode;
      IndexIn += RunTimeLength - 1;
      PrevByte = 0x00;
    }
    else
      PrevByte = *(BufferTemp + (IndexOut ++)) = *(BufferIn + IndexIn);                                          /* No, just copy */
  }
  while (IndexIn < BlLength)
    *(BufferTemp + (IndexOut ++)) = *(BufferIn + (IndexIn ++));                                                 /* Copy last ones */
  if (IndexOut >= BlLength)                                                              /* Compressed larger than uncompressed ? */
  {
    memcpy (BufferTemp, BufferIn, BlLength);                                                /* Return the inputblock uncompressed */
    LengthOut = BlLength;
    IndexOut = 0x0000;                                                                                  /* Signal: not compressed */
  }
  else
    LengthOut = IndexOut;
  BufferTemp = WorkBuffer + (word)32768 + LengthOut - 1;                                       /* Point to the last crunched byte */
  for (IndexIn = 0 ; IndexIn < LengthOut ; IndexIn ++)                             /* Now reverse the result for backward loading */
    *(BufferOut + IndexIn) = *(BufferTemp - IndexIn);
  *CrunchedLength = IndexOut;
}

static void _ReverseBlock (byte *BufferOut, byte *BufferIn)
{
  word register  Cnt;
  byte          *BIn;
  byte          *BOut;

  for (BIn = BufferIn, BOut = BufferOut + 16383, Cnt = 0 ; Cnt < 16384 ; Cnt ++)
    *(BOut --) = *(BIn ++);
}

static void _AddAnyBlockToTape (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the prepared tape entry to be appended.                   */
/*         The data of the block must be ready in `WorkBuffer' and the BlockLength field must hold its length.                    */
/* Post  : The block has been appended to the tape after filling in the AllocLength field.                                        */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirP;

  Entry->AllocLength = (Entry->BlockLength & 0xFFFFFFFC) + (Entry->BlockLength % 4 ? 4 : 0);
  if (TapeWindow[WhichWindow].Tape.XMSHandle == -1)                                                              /* First entry ? */
    XMSAlloc (&(TapeWindow[WhichWindow].Tape.XMSHandle), Entry->AllocLength);
  else
    XMSRealloc (TapeWindow[WhichWindow].Tape.XMSHandle, TapeWindow[WhichWindow].Tape.Size + Entry->AllocLength, TRUE);
  CopyToXMS (Entry->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, TapeWindow[WhichWindow].Tape.Size);
  TapeWindow[WhichWindow].Tape.Size += Entry->AllocLength;
  Entry->Next = NULL;
  if (TapeWindow[WhichWindow].Directory == NULL)
    TapeWindow[WhichWindow].Directory = Entry;
  else
  {
    DirP = TapeWindow[WhichWindow].Directory;
    while (DirP->Next != NULL)
      DirP = DirP->Next;
    DirP->Next = Entry;
  }
  TapeWindow[WhichWindow].NumEntries ++;
}

static void _AddDataBlockToTape (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the prepared tape entry to be appended.                   */
/*         The data of the block must be ready in `WorkBuffer' and the BlockLength field must hold its length.                    */
/*         The block type is determined from the IsNormalSpeed field, so this field must be filled in.                            */
/* Post  : The block has been appended to the tape after filling in the AllocLength field.                                        */
/*         The Parity and CalculatedParity fields are calculated here too.                                                        */
/* Import: _AddAnyBlockToTape.                                                                                                    */
/**********************************************************************************************************************************/

{
  dword Cnt;

  if (Entry->Flags.IsNormalSpeed || !_UsedLoadingSpeed)
  {
    Entry->BlockTypeID = 0x10;                                                                               /* Normal speed data */
    Entry->PauseLength = 0;
    Entry->Flags.IsNormalSpeed = TRUE;
  }
  else
  {
    Entry->BlockTypeID = 0x11;                                                                                /* Turbo speed data */
    Entry->Timing.LengthPilotPulse = _TurboLoadVars[_UsedLoadingSpeed]._LenPilot;
    Entry->Timing.NumPilotPulses = (word)((dword)3500000 / _TurboLoadVars[_UsedLoadingSpeed]._LenPilot);   /* Make pilot 1 second */
    Entry->Timing.LengthSyncPulse[0] = Entry->Timing.LengthSyncPulse[1] = _TurboLoadVars[_UsedLoadingSpeed]._LenSync0;
    Entry->Timing.LengthZeroBitPulse = _TurboLoadVars[_UsedLoadingSpeed]._Len0;
    Entry->Timing.LengthOneBitPulse = _TurboLoadVars[_UsedLoadingSpeed]._Len0 * 2;
    Entry->PauseLength = 0;
  }
  Entry->ComboSelected = 0x00;
  Entry->IsTapeData = TRUE;
  Entry->IsIdentifiedData = TRUE;
  Entry->Flags.FlagByteUsed = Entry->Flags.ParityByteUsed = TRUE;
  Entry->CalculatedParity = Entry->FlagByte;
  for (Cnt = 0 ; Cnt < Entry->BlockLength ; Cnt ++)                                                           /* Calculate parity */
    Entry->CalculatedParity ^= *(WorkBuffer + Cnt);
  Entry->ParityByte = Entry->CalculatedParity;
  _AddAnyBlockToTape (WhichWindow, Entry);
}

static void _TBTimedFlash (struct WindowInfo *Window)

/**********************************************************************************************************************************/
/* Pre   : A spectrum screen has to be set up and drawn already.                                                                  */
/* Post  : This is the timed flash handler. Each run toggles the flashing parts.                                                  */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  FlashState = !FlashState;
  PutSpectrumScreen (WorkBuffer, Window->StartX + 4, Window->StartY + 4);
}

static void _ViewScreenBlock (byte *Screen)

/**********************************************************************************************************************************/
/* Pre   : `Screen' points to the 6912 bytes of screen data in memory.                                                            */
/* Post  : The screen has been shown to the user.                                                                                 */
/*         The user must explicitely use the close gadget or <ESC>.                                                               */
/* Import: _TBTimedFlash (indirectly).                                                                                            */
/**********************************************************************************************************************************/

{
  struct WindowInfo TBWindow  = {188, 140, 264, 200, 264, 200, 264, 200, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                 WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL, "Screen Image"};
  struct ButtonInfo *ButtonP;
  bool              Done;

  DrawWindow (&TBWindow, NULL, WING_CLOSE, TRUE);
  SurroundScreen (&TBWindow, 2, 2);
  memcpy (WorkBuffer, Screen, 6912);
  PutSpectrumScreen (WorkBuffer, TBWindow.StartX + 4, TBWindow.StartY + 4);
  if (ScreenContainsFlash && TPConfig.EmulateFlash)
    RegisterTimedHandler (TPConfig.SlowFlash ? 960 : 320, _TBTimedFlash, &TBWindow);
  Done = FALSE;
  while (!Done)
  {
    if (HandleWindow (&TBWindow, &ButtonP) == WINH_CLOSE || GlobalKey == K_ESC)
      Done = TRUE;
    MouseStatus ();
  }
  if (ScreenContainsFlash && TPConfig.EmulateFlash)
    ReleaseTimedHandler (_TBTimedFlash, &TBWindow);
  DestroyWindow (&TBWindow);
  GlobalKey = 0x0000;
}

static bool _SetupConversionUserInfo (byte *BorderColour, bool *CompressAll)

/**********************************************************************************************************************************/
/* Pre   : `_GetZ80Info' must have been called already, so the internal loading screen is available for viewing.                  */
/*         `BorderColour' is the default border colour (taken from the snapshot header).                                          */
/* Post  : The user selected the desired replay speed and possibly changed the border colour or selected an alternate screen.     */
/*         The speed is set in (global) `_UsedLoadingSpeed' and return value tells whether the user selected a different screen.  */
/* Import: _ViewScreenBlock.                                                                                                      */
/**********************************************************************************************************************************/

{
#define SP_MAX  4
  char *SpeedNames[SP_MAX]  = { "Normal speed    (1500 bps)",
                                "High speed      (2250 bps)",
                                "Turbo speed     (3000 bps)",
                                "Ludicrous speed (6000 bps)" };
  char *Methods[2] = { "Fastest",
                       "Cleanest" };
#define MethodIndex(_CompAll)   (_CompAll ? 0 : 1)
  static byte IBorderColours[8] = { 0, 1, 4, 5, 2, 3, 6, 7 };

  struct WindowInfo CWindow   = {147, 250, 343, 165, 343, 165, 343, 165, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                 WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL, "Setup Snapshot Conversion Info"};
  struct ButtonInfo POk       = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 147, 140, 8, DBLACK, DWHITE, NULL, 0, "@Ok", NULL};
  struct ButtonInfo BBCol     = {BT_TEXT, BC_TEXT, BF_FWIDTH | BF_FITTXT, 112, 20, 1, DBLACK, DWHITE, NULL, 0, "\0", NULL};
  struct ButtonInfo BBColSel  = {BT_PUSH, BC_GADGET, BF_FWIDTH | BF_KEYSELECT, 125, 20, 1, DBLACK, DWHITE, NULL, 4, "\012", NULL};
  struct ButtonInfo BDScreen  = {BT_TEXT, BC_TEXT, BF_FWIDTH | BF_FITTXT, 112, 50, 26, DBLACK, DWHITE, NULL, 0, NULL, NULL};
  struct ButtonInfo BDScrSel  = {BT_PUSH, BC_GADGET, BF_FWIDTH | BF_KEYSELECT, 275, 50, 1, DBLACK, DWHITE, NULL, 2, "\012", NULL};
  struct ButtonInfo BDScrView = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 293, 50, 6, DBLACK, DWHITE, NULL, 3, "@View", NULL};
  struct ButtonInfo BDSpeed   = {BT_TEXT, BC_TEXT, BF_FWIDTH | BF_FITTXT, 112, 80, 26, DBLACK, DWHITE, NULL, 0, NULL, NULL};
  struct ButtonInfo BDSpdSel  = {BT_PUSH, BC_GADGET, BF_FWIDTH | BF_KEYSELECT, 275, 80, 1, DBLACK, DWHITE, NULL, 1, "\012", NULL};
  struct ButtonInfo BDMethod  = {BT_TEXT, BC_TEXT, BF_FWIDTH | BF_FITTXT, 112, 110, 8, DBLACK, DWHITE, NULL, 0, NULL, NULL};
  struct ButtonInfo BDMethSel = {BT_PUSH, BC_GADGET, BF_FWIDTH | BF_KEYSELECT, 167, 110, 1, DBLACK, DWHITE, NULL, 5, "\007", NULL};
  struct ComboInfo  SWindow;
  struct ComboEntry SEntry;
  struct ButtonInfo *ButtonP;
  struct MouseShape *OldShape;
  char              ValLoadingSpeed[32];
  char              ValLoadingScreen[MAXDIR + 1];
  char              ValMethod[16];
  bool              Finished = FALSE;
  comboent          SelectedLine;
  short   register  Cnt;
  bool              DifferentLoadingScreen = FALSE;
  char              LoadingScreenPath[MAXDIR + 1];
  char              Path[MAXDIR + 1];
  char              File[MAXDIR + 1] = "\0";
  char              Mask[13] = "*.SCR";
  int               FhLoadingScreen;
  bool              Done = FALSE;

  OldShape = CurrentMouseShape;
  SetMouseShape (DefaultNormalMouseShape);
  _UsedLoadingSpeed = 2;                                                                              /* (Turbo speed - 3000 bps) */
  BBCol.BPen = IBorderColours[*BorderColour];
  BDSpeed.Contents = ValLoadingSpeed;
  BDScreen.Contents = ValLoadingScreen;
  BDMethod.Contents = ValMethod;
  strcpy (ValLoadingSpeed, SpeedNames[_UsedLoadingSpeed]);
  strcpy (ValLoadingScreen, "<internal>");
  strcpy (ValMethod, Methods[MethodIndex (*CompressAll)]);
  strcpy (Path, CurPath);
  AddButton (&CWindow, &POk, TRUE);
  AddButton (&CWindow, &BBCol, FALSE);
  AddButton (&CWindow, &BBColSel, FALSE);
  AddButton (&CWindow, &BDScreen, FALSE);
  AddButton (&CWindow, &BDScrSel, FALSE);
  AddButton (&CWindow, &BDScrView, FALSE);
  AddButton (&CWindow, &BDSpeed, FALSE);
  AddButton (&CWindow, &BDSpdSel, FALSE);
  AddButton (&CWindow, &BDMethod, FALSE);
  AddButton (&CWindow, &BDMethSel, FALSE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  Wprintf (&CWindow, 2,  1, "{+1Border Colour}N");
  Wprintf (&CWindow, 5,  1, "{+1Loading Screen}N");
  Wprintf (&CWindow, 8,  1, "{+1Loading Speed}N");
  Wprintf (&CWindow, 11, 1, "{+1Method}N");
  while (!Finished)
  {
    switch (HandleWindow (&CWindow, &ButtonP))
    {
      case WINH_CLOSE   : Finished = TRUE;
                          break;
      case WINH_RNOHAND : switch (ButtonP->ReactNum)
                          {
                            case 0 : Finished = TRUE; break;                                                              /* (Ok) */
                            case 1 : MakeComboWindow (&SWindow, &CWindow, COMT_SELECT, COMF_VSLIDER | COMF_SYNCSLIDER |
                                                      COMF_KEYSELECT | COMF_EXCLUSIVE, CWindow.StartX + BDSpeed.StartX + 1,
                                                      CWindow.StartY + BDSpeed.StartY + 16, 0, 1024, BDSpeed.FWidth + 2,
                                                      BDSpeed.FWidth, SP_MAX, NULL, FALSE, 0, NULL, DBLACK, DWHITE, 0, 0);
                                     for (Cnt = 0 ; Cnt < SP_MAX ; Cnt ++)                              /* (Loading speed select) */
                                     {
                                       SEntry.Selected = 0;
                                       SEntry.Identifier = Cnt;
                                       strcpy (SEntry.Text, SpeedNames[Cnt]);
                                       if (!InsertComboEntry (&SWindow, &SEntry, NULL, -1))
                                         FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                                     }
                                     SWindow.LastSelected = _UsedLoadingSpeed;                   /* Auto-select the current speed */
                                     PrintComboEntries (&SWindow);
                                     Finished = FALSE;
                                     while (!Finished)
                                     {
                                       MouseStatus ();
                                       if (HandleComboWindow (&SWindow, &SelectedLine) == WINH_LINEFIRST)
                                       {
                                         Finished = TRUE;
                                         _UsedLoadingSpeed = SWindow.LastSelected;
                                       }
                                       else
                                       {
                                         HandleWindow (&(SWindow.ComboWindow), &ButtonP);             /* (Just read the keyboard) */
                                         if (Mouse.Xx < SWindow.OverlayWindow.A_StartX || Mouse.Xx > SWindow.OverlayWindow.A_EndX ||
                                             Mouse.Yy < SWindow.OverlayWindow.A_StartY || Mouse.Yy > SWindow.OverlayWindow.A_EndY)
                                           if (Mouse.Left)                                 /* Left click outside the select box ? */
                                           {
                                             Finished = TRUE;
                                             DelayLeftMouseButton ();
                                           }
                                       }
                                       if (GlobalKey == K_ESC)
                                         Finished = TRUE;
                                       else if (GlobalKey == K_ENTER)
                                       { _UsedLoadingSpeed = SWindow.LastSelected; Finished = TRUE; }
                                     }
                                     GlobalKey = 0x0000;
                                     Finished = FALSE;
                                     DestroyComboWindow (&SWindow, 0);
                                     strcpy (ValLoadingSpeed, SpeedNames[_UsedLoadingSpeed]);
                                     UpdateButtonText (&CWindow, &BDSpeed, BF_FITTXT);
                                     break;
                            case 2 : Done = FALSE;                                                     /* (Loading screen select) */
                                     DifferentLoadingScreen = FALSE;
                                     while (!Done)
                                     {
                                       if (!FileRequester (180, 50, 48, 29, 20, Path, File, Mask, "Select Screen File",
                                                           TRUE, FALSE))
                                         Done = TRUE;
                                       else
                                       {
                                         MakeFullPath (LoadingScreenPath, Path, File);
                                         if ((FhLoadingScreen = open (LoadingScreenPath, O_BINARY | O_RDONLY)) == -1)
                                           ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Cannot open %s\n"
                                                           "      - %s", File, StrError (errno));
                                         else if (filelength (FhLoadingScreen) != 6912)
                                         {
                                           ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - %s is not a Spectrum screen file!",
                                                           File);
                                           close (FhLoadingScreen);
                                         }
                                         else if (read (FhLoadingScreen, WorkBuffer + 16384, 6912) != 6912)
                                         {
                                           ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Cannot read %s\n"
                                                           "      - %s", File, StrError (errno));
                                           close (FhLoadingScreen);
                                         }
                                         else
                                         {
                                           Done = TRUE;
                                           close (FhLoadingScreen);
                                           DifferentLoadingScreen = TRUE;
                                           BDScreen.Flags &= ~BF_FITTXT;
                                           BDScreen.Flags |= BF_FITDIR;
                                           strcpy (ValLoadingScreen, LoadingScreenPath);
                                           UpdateButtonText (&CWindow, &BDScreen, BF_FITDIR);
                                         }
                                       }
                                     }
                                     break;
                            case 3 : _ViewScreenBlock (DifferentLoadingScreen ? WorkBuffer + 16384 : _ObjectData);
                                     break;
                            case 4 : MakeComboWindow (&SWindow, &CWindow, COMT_SELECT, COMF_KEYSELECT | COMF_EXCLUSIVE|COMF_VSLIDER,
                                                      CWindow.StartX + BBCol.StartX + 1, CWindow.StartY + BBCol.StartY + 16,
                                                      0, 1024, 8, 1, 8, NULL, FALSE, 0, NULL, DBLACK, DWHITE, 0, 0);
                                     for (Cnt = 0 ; Cnt < 8 ; Cnt ++)                                   /* (Border colour select) */
                                     {
                                       SEntry.Selected = 0;
                                       SEntry.Identifier = Cnt;
                                       sprintf (SEntry.Text, "{+%d\xDB}N", IBorderColours[Cnt]);
                                       if (!InsertComboEntry (&SWindow, &SEntry, NULL, -1))
                                         FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                                     }
                                     SWindow.LastSelected = *BorderColour;                      /* Auto-select the current colour */
                                     PrintComboEntries (&SWindow);
                                     Finished = FALSE;
                                     while (!Finished)
                                     {
                                       MouseStatus ();
                                       if (HandleComboWindow (&SWindow, &SelectedLine) == WINH_LINEFIRST)
                                       {
                                         Finished = TRUE;
                                         *BorderColour = SWindow.LastSelected;
                                       }
                                       else
                                       {
                                         HandleWindow (&(SWindow.ComboWindow), &ButtonP);             /* (Just read the keyboard) */
                                         if (Mouse.Xx < SWindow.OverlayWindow.A_StartX || Mouse.Xx > SWindow.OverlayWindow.A_EndX ||
                                             Mouse.Yy < SWindow.OverlayWindow.A_StartY || Mouse.Yy > SWindow.OverlayWindow.A_EndY)
                                           if (Mouse.Left)                                 /* Left click outside the select box ? */
                                           {
                                             Finished = TRUE;
                                             DelayLeftMouseButton ();
                                           }
                                       }
                                       if (GlobalKey == K_ESC)
                                         Finished = TRUE;
                                       else if (GlobalKey == K_ENTER)
                                       { *BorderColour = SWindow.LastSelected; Finished = TRUE; }
                                     }
                                     GlobalKey = 0x0000;
                                     Finished = FALSE;
                                     DestroyComboWindow (&SWindow, 0);
                                     BBCol.BPen = IBorderColours[*BorderColour];
                                     DrawButton (&CWindow, &BBCol);
                                     break;
                            case 5 : *CompressAll = !*CompressAll;                                                    /* (Method) */
                                     strcpy (ValMethod, Methods[MethodIndex (*CompressAll)]);
                                     UpdateButtonText (&CWindow, &BDMethod, BF_FITTXT);
                                     break;
                          }
                          break;
    }
    if (GlobalKey == K_ESC)
      Finished = TRUE;
    MouseStatus ();
  }
  DestroyWindow (&CWindow);
  GlobalKey = 0x0000;
  SetMouseShape (OldShape);
  return (DifferentLoadingScreen);
}

/**********************************************************************************************************************************/
/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> START OF THE EXPORTED ROUTINES <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
/**********************************************************************************************************************************/

bool ConvertSnapshot (dword FileLength, short WhichWindow, int Fh)

/**********************************************************************************************************************************/
/* Pre   : `FileLength' holds the length of the snapshot file, `WhichWindow' is either TWLEFT or TWRIGHT, `Fh' is the file handle */
/*         of the open snapshot file.                                                                                             */
/* Post  : The snapshot has been converted to a compressed, turbo-loading TZX file.                                               */
/*         The user selected the speed and may have selected a loading screen file different from the one that the snapshot has.  */
/* Import: _GetZ80Info, _DecrunchZ80Block, CrunchZ80Block, _AddDataBlockToTape, _AddAnyBlockToTape, _SetupConversionUserInfo.     */
/**********************************************************************************************************************************/

{
#define PRESTAGES  4                                                       /* Number of stages (blocks) before the snapshot pages */
                                                                                               /* 1 = Text Description (memtype)  */
                                                                                               /* 2 = Loader header               */
                                                                                               /* 3 = Loader body                 */
                                                                                               /* 4 = (Compressed) loading screen */
#define LOADERADD  ((word)49152 - 512)                                                      /* ORG address of the Loader Z80 code */

  struct PageOrder_s                                                                          /* Each page as found in a Z80 file */
  {
    byte PageNumber;                                                                     /* Spectrum value, not Z80 value! (= +3) */
    word PageStart;                                                        /* Normal (non-paged) start address of this 16Kb chunk */
  };                     /* Notice that the 3 blocks for an 48K machine have different block numbers for a 48K and 128K snapshot! */
  struct PageOrder_s      PageOrder48S[3]            = {{ 1, 0x8000 },  // + With alternate loading screen
                                                        { 2, 0xC000 },  // +
                                                        { 5, 0x4000 }}; // +
  struct PageOrder_s      PageOrder128S[MAXZ80PAGES] = {{ 2, 0x8000 },  // +
                                                        { 1, 0xC000 },
                                                        { 3, 0xC000 },
                                                        { 4, 0xC000 },
                                                        { 6, 0xC000 },
                                                        { 7, 0xC000 },
                                                        { 0, 0xC000 },  // +
                                                        { 5, 0x4000 }}; // +
  struct PageOrder_s      PageOrder48N[3]            = {{ 5, 0x4000 },  // + With internal screen
                                                        { 1, 0x8000 },  // +
                                                        { 2, 0xC000 }}; // +
  struct PageOrder_s      PageOrder128N[MAXZ80PAGES] = {{ 5, 0x4000 },  // +
                                                        { 2, 0x8000 },  // +
                                                        { 1, 0xC000 },
                                                        { 3, 0xC000 },
                                                        { 4, 0xC000 },
                                                        { 6, 0xC000 },
                                                        { 7, 0xC000 },
                                                        { 0, 0xC000 }}; // +
  struct PageOrder_s     *UsedPageOrderTable;                  /* (Either PageOrder48 or PageOrder128, depending on the snapshot) */
  struct                  LoadTable_s                                      /* The loading table as stored within the Loader block */
  {
    byte                  PageNumber;                                      /* Either ((page number) | 0x10) or 0x10 for 48K snaps */
    byte                  LoadStartHi;                                  /* Where does the compressed block load in the Spectrum ? */
    word                  LoadLength;                                                             /* Size of the compressed block */
  }                       LoadTable[MAXZ80PAGES + 2];  /* First one = loading screen, last one has PageNumber = 0 to mark the end */
  struct TapeEntry_s     *Entry;
  struct Timing_s         NormalTiming;
  struct TapeEntry_s     *LoaderEntry;
  dword                   LoaderXMSIndex;
  byte                    NumStages = PRESTAGES + 2;                                                            /* 2 Final stage: */
                                                                                          /* 1 = Overwrite the loader             */
                                                                                          /* 2 = Insert LoadTable into the loader */
  byte                    Stage;
  byte                    NumPages;
  byte                    NumUsedPages;
  byte                    WhichPage;
  word                    CrunchedLength;
  word                    DecrunchedLength;
  word                    PageStart;
  word                    PageLength;
  word                    Cnt;
  dword                   PReqDivider;
  bool                    AddThisPage;
  bool                    DifferentScreen = FALSE;
  bool                    CompressAll = FALSE;
  bool                    LoaderSpaceIsEmpty = FALSE;
  byte                    BorderColour;

  /* Stage 0: Analyse the snapshot */
  UpdatePReq (0, "Status: Analysing snapshot ...");
  if (!_GetZ80Info (Fh, FileLength, TapeWindow[WhichWindow].TapeName, TapeWindow[WhichWindow].FileType))
    return (FALSE);
  NumPages = _Z80DataMap._128KSnapshot ? 8 : 3;
  NumStages += NumPages;
  PReqDivider = FileLength / NumStages;
  NormalTiming.LengthPilotPulse = NS_LENPILOT;                                                   /* Prepare for normal speed data */
  NormalTiming.LengthSyncPulse[0] = NS_LENSYNC1;
  NormalTiming.LengthSyncPulse[1] = NS_LENSYNC2;
  NormalTiming.LengthZeroBitPulse = NS_LEN0BIT;
  NormalTiming.LengthOneBitPulse = NS_LEN1BIT;
  NormalTiming.NumPilotPulses = NS_DTPILOT;                                                                      /* (Short pilot) */
  NormalTiming.NumSyncPulses = 2;
  NormalTiming.SampleRate = NS_SAMPRATE;
  NormalTiming.NumBitsInFirstByte = NS_LENFIRST;
  NormalTiming.NumBitsInLastByte = NS_LENLAST;

  BorderColour = (_Z80DataMap._145SnapshotHeader.Stat1 & 0x0E) >> 1;                           /* Get border colour from the snap */
  DifferentScreen = _SetupConversionUserInfo (&BorderColour, &CompressAll);
  if (DifferentScreen)
    UsedPageOrderTable = _Z80DataMap._128KSnapshot ? &PageOrder128S : &PageOrder48S;
  else
  {
    UsedPageOrderTable = _Z80DataMap._128KSnapshot ? &PageOrder128N : &PageOrder48N;
    NumStages --;
  }

  LoaderSpaceIsEmpty = TRUE;
  for (Cnt = 0 ; LoaderSpaceIsEmpty && Cnt < 512 ; Cnt ++)                     /* Check if there's data in 512 bytes loader space */
    if (*(_ObjectData + (word)32767 - Cnt) != 0x00)
      LoaderSpaceIsEmpty = FALSE;
  if (LoaderSpaceIsEmpty)
    NumStages --;

  /* Stage 1: Stick Text Description in front, saying whether it's a 48K or 128K snapshot */
  Stage = 1;
  UpdatePReq (PReqDivider * (Stage ++), "Status: Inserting Memory Description ...");
  if ((Entry = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
    FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
  Entry->BlockTypeID = 0x30;                                                                                  /* Text Description */
  Entry->ComboSelected = 0x00;
  Entry->IsTapeData = FALSE;
  Entry->IsIdentifiedData = FALSE;
  sprintf (WorkBuffer, "%dK Snapshot - ", _Z80DataMap._128KSnapshot ? 128 : 48);
  switch (_UsedLoadingSpeed)
  {
    case 0 : strcat (WorkBuffer, "1500 bps"); break;
    case 1 : strcat (WorkBuffer, "2250 bps"); break;
    case 2 : strcat (WorkBuffer, "3000 bps"); break;
    case 3 : strcat (WorkBuffer, "6000 bps"); break;
  }
  Entry->BlockLength = strlen (WorkBuffer);
  _AddAnyBlockToTape (WhichWindow, Entry);

  /* Stage 2: Write Loader Header */
  UpdatePReq (PReqDivider * (Stage ++), "Status: Building loader BASIC ...");
  Cnt = strlen (TapeWindow[WhichWindow].TapeName);                                      /* Add PC filename into Speccy block name */
  while (TapeWindow[WhichWindow].TapeName[-- Cnt] != '\\')
    ;
  PageLength = 2;
  memset (_SpectrumBASICHeader._FileName + 2, 0x08, 8);                                                 /* (Clear filename first) */
  while (TapeWindow[WhichWindow].TapeName[++ Cnt] != '.' && PageLength < 10)
    _SpectrumBASICHeader._FileName[PageLength ++] = TapeWindow[WhichWindow].TapeName[Cnt];
  if ((Entry = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
    FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
  memcpy (&(Entry->Timing), &NormalTiming, sizeof (struct Timing_s));
  Entry->BlockLength = 17;
  memcpy (WorkBuffer, &_SpectrumBASICHeader, 17);
  Entry->Flags.IsHeader = TRUE;
  Entry->Flags.IsNormalSpeed = TRUE;
  Entry->FlagByte = 0x00;
  _AddDataBlockToTape (WhichWindow, Entry);

  /* Stage 3: Write Loader Body */
  UpdatePReq (PReqDivider * (Stage ++), NULL);
  memset (&LoadTable, 0, sizeof (struct LoadTable_s) * (MAXZ80PAGES + 2));                                    /* Clear load table */
  memcpy (WorkBuffer, &_SpectrumBASICData, LOADERPREPIECE);                            /* Start with the BASIC part and relocator */
  PageLength = LOADERPREPIECE;
  LoaderXMSIndex = TapeWindow[WhichWindow].Tape.Size;                /* (Keep a copy - the load table must be inserted afterward) */
  _SpectrumLoader._IY   = _Z80DataMap._145SnapshotHeader.IY;
  _SpectrumLoader._HL_a = _Z80DataMap._145SnapshotHeader.HL_a;
  _SpectrumLoader._DE_a = _Z80DataMap._145SnapshotHeader.DE_a;
  _SpectrumLoader._BC_a = _Z80DataMap._145SnapshotHeader.BC_a;
  _SpectrumLoader._LoadBytesRoutine = 0xBF55;                                                             /* Install turbo loader */
  switch (_Z80DataMap._145SnapshotHeader.Stat2 & 0x03)
  {
    case 0x00 : _Final48Loader._IMValue = 0x46; break;
    case 0x02 : _Final48Loader._IMValue = 0x5E; break;
    default   : _Final48Loader._IMValue = 0x56; break;
  }
  _Final48Loader._SP     = _Z80DataMap._145SnapshotHeader.SP;
  _Final48Loader._DIorEI = (_Z80DataMap._145SnapshotHeader.IFF2 ? 0xFB : 0xF3);
  _Final48Loader._PC     = (_Z80DataMap._Version == 145 ? _Z80DataMap._145SnapshotHeader.PC : _Z80DataMap._201SnapshotHeader.NewPC);
  _FinalStackFrame._ReturnAddress = 0x56E0;
  _FinalStackFrame._BorderColour = BorderColour;
  _FinalStackFrame._Value7FFD = 0x7FFD;
  _FinalStackFrame._State7FFD = _Z80DataMap._201SnapshotHeader.State7FFD;
  _FinalStackFrame._IX   = _Z80DataMap._145SnapshotHeader.IX;
  _FinalStackFrame._A_a  = _Z80DataMap._145SnapshotHeader.A_a;
  _FinalStackFrame._F_a  = _Z80DataMap._145SnapshotHeader.F_a;
  _FinalStackFrame._R    = ((_Z80DataMap._145SnapshotHeader.R - 9) & 0x7F) |         /* R will be incremented by 9 once finished! */
                            (_Z80DataMap._145SnapshotHeader.Stat1 & 0x01 ? 0x80 : 0x00);
  _FinalStackFrame._I    = _Z80DataMap._145SnapshotHeader.I;
  _FinalStackFrame._HL_n = _Z80DataMap._145SnapshotHeader.HL_n;
  _FinalStackFrame._DE_n = _Z80DataMap._145SnapshotHeader.DE_n;
  _FinalStackFrame._BC_n = _Z80DataMap._145SnapshotHeader.BC_n;
  _FinalStackFrame._A_n  = _Z80DataMap._145SnapshotHeader.A_n;
  _FinalStackFrame._F_n  = _Z80DataMap._145SnapshotHeader.F_n;
  _Final128Loader._StateFFFD = _Z80DataMap._201SnapshotHeader.StateFFFD;
  _TurboLoader._Compare = _TurboLoadVars[_UsedLoadingSpeed]._Compare;
  _TurboLoader._Delay   = _TurboLoadVars[_UsedLoadingSpeed]._Delay;
  if (BorderColour == 0x00)                                                                      /* Border is going to be black ? */
    _TurboLoader._ColourXOR = 0x41;                           /* Then use blue as counter colour (black/black cannot be seen....) */
  else
    _TurboLoader._ColourXOR = 0x40 | BorderColour;           /* Use the ultimate colour as counter colour for the loading stripes */
  if (_Z80DataMap._128KSnapshot && _Z80DataMap._Version >= 200)                   /* Pre-v2.00 128K versions use the '48K' loader */
  {
    _SpectrumLoader._DecrunchAddress = LOADERADD + _SpectrumLoaderLength + _Spectrum128LoadLength;
    _SpectrumLoader._LoadTable = _SpectrumLoader._DecrunchAddress + _SpectrumDecruncherLength;
    memcpy (WorkBuffer + PageLength, &_SpectrumLoader, _SpectrumLoaderLength);
    PageLength += _SpectrumLoaderLength;
    _Spectrum128Load._AttrColour = BorderColour | (BorderColour << 3);                /* Set attribute colour for corruption part */
    _Spectrum128Load._FinalCode1Address = _SpectrumLoader._LoadTable + sizeof (struct LoadTable_s) * (NumPages + 1) + 1;
    _Spectrum128Load._FinalCode1Length = _Final48LoaderLength;
    _Spectrum128Load._FinalCode2Length = _Final128LoaderLength;
    _Spectrum128Load._FinalStack1Length = _FinalStackFrameLength;
    _Spectrum128Load._FinalStack2Length = _Final128SoundStackFrameLength;
    _FinalStackFrame._ReturnAddress = 0x55E0;
    memcpy (WorkBuffer + PageLength, &_Spectrum128Load, _Spectrum128LoadLength);
    PageLength += _Spectrum128LoadLength;
    memcpy (WorkBuffer + PageLength, &_SpectrumDecruncher, _SpectrumDecruncherLength);
    PageLength += _SpectrumDecruncherLength;
    PageLength += sizeof (struct LoadTable_s) * (NumPages + 1) + 1;
    memcpy (WorkBuffer + PageLength, &_Final48Loader, _Final48LoaderLength);
    PageLength += _Final48LoaderLength;
    memcpy (WorkBuffer + PageLength, &_Final128Loader, _Final128LoaderLength);
    PageLength += _Final128LoaderLength;
    memcpy (WorkBuffer + PageLength, &_FinalStackFrame, _FinalStackFrameLength);
    PageLength += _FinalStackFrameLength;
    for (Cnt = 0 ; Cnt < 16 ; Cnt ++)
      _Final128SoundStackFrame[15 - Cnt]._State = _Z80DataMap._201SnapshotHeader.SoundRegisters[Cnt];
    memcpy (WorkBuffer + PageLength, &_Final128SoundStackFrame, _Final128SoundStackFrameLength);
    PageLength += _Final128SoundStackFrameLength;
  }
  else
  {
    _FinalStackFrame._State7FFD = 0x10;                                                                       /* Force sane state */
    _SpectrumLoader._DecrunchAddress = LOADERADD + _SpectrumLoaderLength + _Spectrum48LoadLength +
                                       (LoaderSpaceIsEmpty ? _Spectrum48Load22Length : _Spectrum48Load21Length);
    _SpectrumLoader._LoadTable = _SpectrumLoader._DecrunchAddress + _SpectrumDecruncherLength;
    memcpy (WorkBuffer + PageLength, &_SpectrumLoader, _SpectrumLoaderLength);
    PageLength += _SpectrumLoaderLength;
    _Spectrum48Load._AttrColour = BorderColour | (BorderColour << 3);
    _Spectrum48Load._FinalCodeAddress = _SpectrumLoader._LoadTable + sizeof (struct LoadTable_s) * (NumPages + 1) + 1;
    _Spectrum48Load._FinalCodeLength = _Final48LoaderLength;
    _Spectrum48Load._FinalStackLength = _FinalStackFrameLength;
    memcpy (WorkBuffer + PageLength, &_Spectrum48Load, _Spectrum48LoadLength);
    PageLength += _Spectrum48LoadLength;
    if (LoaderSpaceIsEmpty)
    {
      _Spectrum48Load22._ExtraLength = _Spectrum48Load23Length;
      memcpy (WorkBuffer + PageLength, &_Spectrum48Load22, _Spectrum48Load22Length);
      PageLength += _Spectrum48Load22Length;
    }
    else
    {
      memcpy (WorkBuffer + PageLength, &_Spectrum48Load21, _Spectrum48Load21Length);
      PageLength += _Spectrum48Load21Length;
    }
    memcpy (WorkBuffer + PageLength, &_SpectrumDecruncher, _SpectrumDecruncherLength);
    PageLength += _SpectrumDecruncherLength;
    PageLength += sizeof (struct LoadTable_s) * (NumPages + 1) + 1;
    memcpy (WorkBuffer + PageLength, &_Final48Loader, _Final48LoaderLength);
    PageLength += _Final48LoaderLength;
    memcpy (WorkBuffer + PageLength, &_FinalStackFrame, _FinalStackFrameLength);
    PageLength += _FinalStackFrameLength;
    if (LoaderSpaceIsEmpty)
    {
      memcpy (WorkBuffer + PageLength, &_Spectrum48Load23, _Spectrum48Load23Length);
      PageLength += _Spectrum48Load23Length;
    }
  }
  if ((Entry = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
    FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
  memcpy (&(Entry->Timing), &NormalTiming, sizeof (struct Timing_s));
  if (PageLength < LOADERPREPIECE + 512)
    memset (WorkBuffer + PageLength, 0, LOADERPREPIECE + 512 - PageLength);
  memcpy (WorkBuffer + LOADERPREPIECE + 0x0155, &_TurboLoader, _TurboLoaderLength);           /* Stick the turbo loader at 0xBF55 */
  Entry->BlockLength = LOADERPREPIECE + 512;
  Entry->Flags.IsHeader = FALSE;
  Entry->Flags.IsNormalSpeed = TRUE;
  Entry->FlagByte = 0xFF;
  _AddDataBlockToTape (WhichWindow, Entry);
  LoaderEntry = Entry;

  /* Stage 4: Write Compressed, Turbo-Loading Loading Screen if wanted */
  if (DifferentScreen)
  {
    UpdatePReq (PReqDivider * (Stage ++), "Status: Adding compressed loading screen ...");
    if ((Entry = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
      FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
    memcpy (&(Entry->Timing), &NormalTiming, sizeof (struct Timing_s));
    _CrunchZ80Block (DifferentScreen ? WorkBuffer + 16384 : _ObjectData, 6912, WorkBuffer, &(LoadTable[0].LoadLength));
    Entry->BlockLength = (dword)LoadTable[0].LoadLength;
    if (Entry->BlockLength == 0x0000)
      Entry->BlockLength = (dword)6912;
    LoadTable[0].LoadStartHi = 0x7F;                                                              /* (Loading from 7FFF downward) */
    LoadTable[0].PageNumber = 0x10;
    Entry->Flags.IsHeader = FALSE;
    Entry->Flags.IsNormalSpeed = FALSE;
    Entry->FlagByte = 0xAA;
    _AddDataBlockToTape (WhichWindow, Entry);
    NumUsedPages = 1;
  }
  else
    NumUsedPages = 0;

  /* Stage 5 and up: Write Compressed, Turbo-Loading 16Kb Pages */
  for (WhichPage = 0 ; WhichPage < NumPages ; WhichPage ++)
  {
    AddThisPage = FALSE;
    UpdatePReq (PReqDivider * (Stage ++), "Status: Adding compressed snapshot pages ...");
    if ((Entry = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
      FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
    memcpy (&(Entry->Timing), &NormalTiming, sizeof (struct Timing_s));
    if (_Z80DataMap._Version == 145)                                          /* (SNA files are effectively handled here as well) */
    {
      if (DifferentScreen)
        switch (WhichPage)
        {
          case 0 : PageStart = (word)32768; DecrunchedLength = 16384 - 512; break;
          case 1 : PageStart = (word)49152; DecrunchedLength = 16384; break;
          case 2 : PageStart = (word)16384; DecrunchedLength = 16384; break;
        }
      else
        switch (WhichPage)
        {
          case 0 : PageStart = (word)16384; DecrunchedLength = 16384; break;
          case 1 : PageStart = (word)32768; DecrunchedLength = 16384 - 512; break;
          case 2 : PageStart = (word)49152; DecrunchedLength = 16384; break;
        }
      _CrunchZ80Block (_ObjectData + PageStart - 16384, DecrunchedLength, WorkBuffer, &(LoadTable[NumUsedPages].LoadLength));
      if (!CompressAll && PageStart == 16384)                                                         /* Create cleanest loader ? */
        if (LoadTable[NumUsedPages].LoadLength > (16384 - 6912))                           /* Will overwrite part of the screen ? */
        {
          _ReverseBlock (WorkBuffer, _ObjectData + PageStart - 16384);                              /* Use the block uncompressed */
          LoadTable[NumUsedPages].LoadLength = 0x0000;
        }
      LoadTable[NumUsedPages].PageNumber = 0x10;                 /* Force sane state so that 48K snaps can be loaded in 128K mode */
      AddThisPage = TRUE;
    }
    else if (_Z80DataMap._MemoryStructure[UsedPageOrderTable[WhichPage].PageNumber]._InUse)
    {
      if (lseek (Fh, _Z80DataMap._MemoryStructure[UsedPageOrderTable[WhichPage].PageNumber]._Offset, SEEK_SET) == -1)
      { ErrorRequester (-1, -1, "@Close", NULL, NULL, _FileSearchError, TapeWindow[WhichWindow].TapeName, strerror (errno));
        return (FALSE); }
      CrunchedLength = _Z80DataMap._MemoryStructure[UsedPageOrderTable[WhichPage].PageNumber]._Length;
      if (CrunchedLength == 0xFFFF)
      {
        if (read (Fh, _ObjectData, 16384) < 16384)
        { ErrorRequester (-1, -1, "@Close", NULL, NULL, _FileReadError, TapeWindow[WhichWindow].TapeName, strerror (errno));
          return (FALSE); }
      }
      else
      {
        if (read (Fh, WorkBuffer, CrunchedLength) < CrunchedLength)
        { ErrorRequester (-1, -1, "@Close", NULL, NULL, _FileReadError, TapeWindow[WhichWindow].TapeName, strerror (errno));
          return (FALSE); }
        _DecrunchZ80Block (WorkBuffer, CrunchedLength, _ObjectData, 16384);
      }
      for (Cnt = 0 ; !AddThisPage && Cnt < 16384 ; Cnt ++)                                  /* Check if there's data in this page */
        if (*(_ObjectData + Cnt) != 0x00)                                                            /* (Empty pages are skipped) */
          AddThisPage = TRUE;
      if (AddThisPage)
      {
        if (DifferentScreen)
          DecrunchedLength = (WhichPage == 0 ? 16384 - 512 : 16384);
        else
          DecrunchedLength = (WhichPage == 1 ? 16384 - 512 : 16384);
        _CrunchZ80Block (_ObjectData, DecrunchedLength, WorkBuffer, &(LoadTable[NumUsedPages].LoadLength));
        PageStart = UsedPageOrderTable[WhichPage].PageStart;
        if (_Z80DataMap._128KSnapshot)
          LoadTable[NumUsedPages].PageNumber = UsedPageOrderTable[WhichPage].PageNumber | 0x10;
        else
          LoadTable[NumUsedPages].PageNumber = 0x10;             /* Force sane state so that 48K snaps can be loaded in 128K mode */
        if (!CompressAll && PageStart == 16384)                                                       /* Create cleanest loader ? */
          if (LoadTable[NumUsedPages].LoadLength > (16384 - 6912))                         /* Will overwrite part of the screen ? */
          {
            _ReverseBlock (WorkBuffer, _ObjectData + PageStart - 16384);                            /* Use the block uncompressed */
            LoadTable[NumUsedPages].LoadLength = 0x0000;
          }
        AddThisPage = TRUE;
      }
    }
    if (AddThisPage)
    {
      Entry->BlockLength = (dword)LoadTable[NumUsedPages].LoadLength;
      if (Entry->BlockLength == 0x0000)
        Entry->BlockLength = (dword)DecrunchedLength;
      LoadTable[NumUsedPages].LoadStartHi = (byte)((PageStart + DecrunchedLength - 1) >> 8);
      Entry->Flags.IsHeader = FALSE;
      Entry->Flags.IsNormalSpeed = FALSE;
      Entry->FlagByte = 0xAA;
      _AddDataBlockToTape (WhichWindow, Entry);
      NumUsedPages ++;
    }
    else
      free (Entry);
  }

  /* Stage last: Overwrite The Loader (Non-Compressed, Non-Turbo) */
  if (!LoaderSpaceIsEmpty)
  {
    UpdatePReq (PReqDivider * (Stage ++), "Status: Adding final snapshot block ...");
    if ((Entry = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
      FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
    memcpy (&(Entry->Timing), &NormalTiming, sizeof (struct Timing_s));
    Entry->BlockLength = 512;
    if (_Z80DataMap._Version == 145)
      memcpy (WorkBuffer, _ObjectData + LOADERADD - 16384, 512);
    else
    {
      WhichPage = DifferentScreen ? 0 : 1;
      if (lseek (Fh, _Z80DataMap._MemoryStructure[UsedPageOrderTable[WhichPage].PageNumber]._Offset, SEEK_SET) == -1)
      { ErrorRequester (-1, -1, "@Close", NULL, NULL, _FileSearchError, TapeWindow[WhichWindow].TapeName, strerror (errno));
        return (FALSE); }
      CrunchedLength = _Z80DataMap._MemoryStructure[UsedPageOrderTable[WhichPage].PageNumber]._Length;
      if (CrunchedLength == 0xFFFF)
      {
        if (read (Fh, WorkBuffer, 16384) < 16384)
        { ErrorRequester (-1, -1, "@Close", NULL, NULL, _FileReadError, TapeWindow[WhichWindow].TapeName, strerror (errno));
          return (FALSE); }
      }
      else
      {
        if (read (Fh, _ObjectData, CrunchedLength) < CrunchedLength)
        { ErrorRequester (-1, -1, "@Close", NULL, NULL, _FileReadError, TapeWindow[WhichWindow].TapeName, strerror (errno));
          return (FALSE); }
        _DecrunchZ80Block (_ObjectData, CrunchedLength, WorkBuffer, 16384);
      }
      memcpy (WorkBuffer, WorkBuffer + 16384 - 512, 512);
    }
    Entry->Flags.IsHeader = FALSE;
    Entry->Flags.IsNormalSpeed = TRUE;
    Entry->FlagByte = 0x55;
    _AddDataBlockToTape (WhichWindow, Entry);
  }

  UpdatePReq (PReqDivider * (Stage ++), "Status: Finishing loader ...");
  CopyFromXMS (LOADERPREPIECE + 512, TapeWindow[WhichWindow].Tape.XMSHandle, LoaderXMSIndex, WorkBuffer);
  memcpy (WorkBuffer + _SpectrumLoader._LoadTable - LOADERADD + LOADERPREPIECE,                        /* Insert the loader table */
          &LoadTable, sizeof (struct LoadTable_s) * (NumPages + 1) + 1);
  CopyToXMS (LOADERPREPIECE + 512, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, LoaderXMSIndex);
  LoaderEntry->CalculatedParity = LoaderEntry->FlagByte;                                            /* (Re)calculate block parity */
  for (Cnt = 0 ; Cnt < LOADERPREPIECE + 512 ; Cnt ++)
    LoaderEntry->CalculatedParity ^= *(WorkBuffer + Cnt);
  LoaderEntry->ParityByte = LoaderEntry->CalculatedParity;

  return(TRUE);
}
