/**********************************************************************************************************************************/
/* Project      : TAPER                                                                                                           */
/* Module       : TPSNAP.H                                                                                                        */
/* Description  : Snapshot to tape conversion                                                                                     */
/* Version type : Module interface                                                                                                */
/* Last changed : 29-06-1998  19:00                                                                                               */
/* Update count : 3                                                                                                               */
/* OS type      : PC (DOS)                                                                                                        */
/* Compiler     : Watcom C32 Optimizing Compiler Version 10.x                                                                     */
/* Linker       : Watcom C Linker                                                                                                 */
/*                                                                                                                                */
/*                Copyleft (C) 1997-2001 ThunderWare Research Center, written by Martijn van der Heide.                           */
/*                                                                                                                                */
/*                This program is free software; you can redistribute it and/or                                                   */
/*                modify it under the terms of the GNU General Public License                                                     */
/*                as published by the Free Software Foundation; either version 2                                                  */
/*                of the License, or (at your option) any later version.                                                          */
/*                                                                                                                                */
/*                This program is distributed in the hope that it will be useful,                                                 */
/*                but WITHOUT ANY WARRANTY; without even the implied warranty of                                                  */
/*                MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                                                   */
/*                GNU General Public License for more details.                                                                    */
/*                                                                                                                                */
/*                You should have received a copy of the GNU General Public License                                               */
/*                along with this program; if not, write to the Free Software                                                     */
/*                Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                                     */
/*                                                                                                                                */
/**********************************************************************************************************************************/

#ifndef TPSNAP_H_INC
#define TPSNAP_H_INC

#include "tpconfig.h"
#include "tptape.h"

#ifdef EXTERN
#undef EXTERN
#endif
#ifdef __TPSNAP_MAIN__
#define EXTERN
#else
#define EXTERN extern
#endif

struct Z80145SnapshotHeader_s                                              /* Snapshot header for Z80 snap v1.45 (Little Endian!) */
{
  byte                      A_n, F_n;                                      /* AF                                  */
  word                      BC_n;                                          /* BC                                  */
  word                      HL_n;                                          /* HL                                  */
  word                      PC;                                            /* PC                                  */
  word                      SP;                                            /* SP                                  */
  byte                      I;                                             /* I                                   */
  byte                      R;                                             /* R (bit 0-6)                         */
  byte                      Stat1;                                         /* Bit   0: Bit 7 of R                 */
                                                                           /* Bit 1-3: Border Color               */
                                                                           /* Bit   4: Basic SamRam switched in ? */
                                                                           /* Bit   5: Compressed datablocks ?    */
  word                      DE_n;                                          /* DE                                  */
  word                      BC_a;                                          /* BC'                                 */
  word                      DE_a;                                          /* DE'                                 */
  word                      HL_a;                                          /* HL'                                 */
  byte                      A_a, F_a;                                      /* AF'                                 */
  word                      IY;                                            /* IY                                  */
  word                      IX;                                            /* IX                                  */
  byte                      IFF1;                                          /* in bit 0                            */
  byte                      IFF2;
  byte                      Stat2;                                         /* Bit 0-1: IM 0/1/2                   */
                                                                           /* Bit   2: Issue 2 emulation ?        */
                                                                           /* Bit   3: Double int frequency ?     */
                                                                           /* Bit 4-5: 01 = High Video Sync       */
                                                                           /*          11 = Low                   */
                                                                           /*          else Normal Video Sync     */
                                                                           /* Bit 6-7: 00 = Cursor Joystick,      */
                                                                           /*          01 = Kempston,             */
                                                                           /*          10 = Interface 2 #1,       */
                                                                           /*          11 = Interface 2 #2        */
};

struct Z80201SnapshotHeader_s                                              /* Additional snapshot header for 2.01 */
{
  word                      AdditionalLength;                              /* Length of additional header         */
  word                      NewPC;                                         /* PC for v2.01 and up                 */
  byte                      HWMode;                                        /* Hardware mode                       */
  byte                      State7FFD;                                     /* Last OUT to 7FFD                    */
  byte                      I1Paged;                                       /* 0xFF if Interface I ROM paged in    */
  byte                      Emulation;                                     /* Bit 0 : R register emulation on ?   */
                                                                           /* Bit 1 : LDIR emulation on ?         */
  byte                      StateFFFD;                                     /* Last OUT to FFFD (soundchip regnum) */
  byte                      SoundRegisters[16];                            /* Values of the sound chip registers  */
};

struct Z80300SnapshotHeader_s                                              /* Additional snapshot header for 3.00 */
{
  word                      TStateLo;                                      /* Low T-state counter                 */
  byte                      TStateHi;                                      /* Hi T-state counter                  */
  byte                      xxxxx1;                                        /* (QL emul)                           */
  byte                      MGTPaged;                                      /* 0xFF if MGT ROM paged in            */
  byte                      MFPaged;                                       /* 0xFF if MF ROM paged in             */
  byte                      RAMLo;                                         /* 0xFF if 0x0000-0x1FFF is RAM        */
  byte                      RAMHi;                                         /* 0xFF if 0x2000-0x3FFF is RAM        */
  word                      UKey0, UKey1, UKey2, UKey3, UKey4;             /* ASCII words of user defined keys    */
  word                      MKey0, MKey1, MKey2, MKey3, MKey4;             /* Mappings of these keys              */
  byte                      MGTType;                                       /* MGT type                            */
  byte                      DiscInhBS;                                     /* Disciple inhibit button status:     */
                                                                           /*     0x00 = out                      */
                                                                           /*     0xFF = in                       */
  byte                      DiscInhFlag;                                   /* Disciple inhibit flag:              */
                                                                           /*     0x00 = ROM pageable             */
                                                                           /*     0xFF = ROM not pageable         */
};

struct MMSnapshotHeader_s                                                  /* Snapshot format for MM snapshot (Little-Endian!) */
{
  byte                      I;                                             /* I                                   */
  word                      HL_a;                                          /* HL'                                 */
  word                      DE_a;                                          /* DE'                                 */
  word                      BC_a;                                          /* BC'                                 */
  word                      AF_a;                                          /* AF'                                 */
  word                      HL_n;                                          /* HL                                  */
  word                      DE_n;                                          /* DE                                  */
  word                      BC_n;                                          /* BC                                  */
  word                      IY;                                            /* IY                                  */
  word                      IX;                                            /* IX                                  */
  byte                      IFF1;                                          /* Bit 2: IFF1                         */
  byte                      R;                                             /* R                                   */
  word                      AF_n;                                          /* AF                                  */
  word                      SP;                                            /* SP                                  */
  byte                      IM;                                            /* IM 0/1/2                            */
  byte                      xxxxx1;                                        /* Unused (padding) byte               */
};

/**********************************************************************************************************************************/
/* Define the global variables                                                                                                    */
/**********************************************************************************************************************************/

/**********************************************************************************************************************************/
/* Define the interface functions                                                                                                 */
/**********************************************************************************************************************************/

bool ConvertSnapshot (dword FileLength, short WhichWindow, int Fh);

#endif
