/**********************************************************************************************************************************/
/* Project      : TAPER                                                                                                           */
/* Module       : TPTAPE.H                                                                                                        */
/* Description  : Intermediate tapefile handler                                                                                   */
/* Version type : Module interface                                                                                                */
/* Last changed : 31-03-1998  6:00                                                                                                */
/* Update count : 19                                                                                                              */
/* OS type      : PC (DOS)                                                                                                        */
/* Compiler     : Watcom C32 Optimizing Compiler Version 10.x                                                                     */
/* Linker       : Watcom C Linker                                                                                                 */
/*                                                                                                                                */
/*                Copyleft (C) 1997-2001 ThunderWare Research Center, written by Martijn van der Heide.                           */
/*                                                                                                                                */
/*                This program is free software; you can redistribute it and/or                                                   */
/*                modify it under the terms of the GNU General Public License                                                     */
/*                as published by the Free Software Foundation; either version 2                                                  */
/*                of the License, or (at your option) any later version.                                                          */
/*                                                                                                                                */
/*                This program is distributed in the hope that it will be useful,                                                 */
/*                but WITHOUT ANY WARRANTY; without even the implied warranty of                                                  */
/*                MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                                                   */
/*                GNU General Public License for more details.                                                                    */
/*                                                                                                                                */
/*                You should have received a copy of the GNU General Public License                                               */
/*                along with this program; if not, write to the Free Software                                                     */
/*                Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                                     */
/*                                                                                                                                */
/**********************************************************************************************************************************/

#ifndef TPTAPE_H_INC
#define TPTAPE_H_INC

#ifdef EXTERN
#undef EXTERN
#endif
#ifdef __TPTAPE_MAIN__
#define EXTERN
#else
#define EXTERN extern
#endif

#define TAPETEMP      ".TMP"
#define TAPEBACKUP    ".BAK"
#define TAPELSAVE     "LEFTTAPE.TZX"
#define TAPERSAVE     "RGHTTAPE.TZX"
#define TAPENONE      "(new tape)"                                                            /* Tape name after 'New' menu entry */

#define BUEXPANDBIT   0x80                                                        /* Backup bit for temporary group expand (play) */

/**********************************************************************************************************************************/
/* Define the global variables                                                                                                    */
/**********************************************************************************************************************************/

struct Timing_s
{
  word            LengthPilotPulse;                                                                    /* Length of a PILOT pulse */
  word            LengthSyncPulse[2];                                                                          /* The SYNC pulses */
  word            LengthZeroBitPulse;                                                               /* Length of a ZERO bit pulse */
  word            LengthOneBitPulse;                                                                 /* Length of a ONE bit pulse */
  word            NumPilotPulses;                                                                   /* Num pulses in a PILOT tone */
  dword           SampleRate;                                                                                            /* In Hz */
  byte            NumSyncPulses;                                                                     /* The number of sync pulses */
  byte            NumBitsInFirstByte;                                                          /* Num used bits in the first byte */
  byte            CustomFirstByte;                                                         /* Not used if NumBitsInFirstByte is 8 */
  byte            NumBitsInLastByte;                                                            /* Num used bits in the last byte */
  byte            CustomLastByte;                                                           /* Not used if NumBitsInLastByte is 8 */
};

struct TapeEntry_s                                                                                             /* Directory entry */
{
  struct TapeEntry_s *Next;                                                                                      /* (Linked list) */
  dword             BlockLength;                                          /* Length of the datablock, without parity or flag byte */
  dword             AllocLength;                                                                     /* Length as a multiple of 4 */
  byte              BlockTypeID;                                                                   /* Uses values from TZX format */
  byte              FlagByte;
  byte              ParityByte;
  byte              CalculatedParity;
  byte              ComboSelected;                                                     /* Keeps a copy of the combo entry's value */
  bool              IsTapeData;                                           /* TRUE if this is some tape block (as opposed to info) */
  bool              IsIdentifiedData;                                    /* TRUE if all bytes can be shown ('View as' menu entry) */
  struct            TEFlags_s                                                      /* Pack all flags where possible to save bytes */
  {
    bool            FlagByteUsed : 1;
    bool            ParityByteUsed : 1;
    bool            IsHeader : 1;                                                          /* TRUE if this is a header tape block */
    bool            IsNormalSpeed : 1;                          /* TRUE for plain (.TAP) tape blocks (don't need the Timing part) */
    bool            IsWritten : 1;                               /* Set TRUE after the block is written (infinite loop detection) */
  }                 Flags;
  struct Timing_s   Timing;                                                                            /* Block sound timing info */
  word              PauseLength;                                                   /* Length of pause after this block (millisec) */
  word              SrcJumpCode;                                                   /* Jump index number - if this is a jump block */
  word              DstJumpCode;                                                /* Jump index number - if this is the destination */
};

struct Tape_s
{
  dword             Size;                                                                 /* Total size of the intermediate block */
  short             XMSHandle;
};

EXTERN bool _MustInitializeJumpCodes;
EXTERN bool _PreReadingTape;
EXTERN bool _TAPNoise;
EXTERN bool _InfiniteTapeLoop;

/**********************************************************************************************************************************/
/* Define the function prototypes                                                                                                 */
/**********************************************************************************************************************************/

bool     UpdatePReq             (dword FileIndex, char *StatusText);
bool     ReadTape               (short WhichWindow, bool Merge);
void     WriteTape              (short WhichWindow, bool FatalSave);
void     ReleaseTape            (short WhichWindow);
dword    GetTapeBlockXMSIndex   (short WhichWindow, struct TapeEntry_s *Entry);
word     GetTapeBlock           (short WhichWindow, dword *NextByte, bool *More);
void     DeleteTapeBlock        (short WhichWindow, struct TapeEntry_s *Entry, bool RebuildTape, bool Changed);
void     CopyTapeBlock          (short SourceWindow, struct TapeEntry_s *SDirP, short DestinationWindow,
                                 struct TapeEntry_s *DDirP, bool Move, bool CompleteCollapsedGroups);
short    DragTapeBlock          (void);
void     SetTapeBlockSelectFlag (struct TapeEntry_s *Entry, short NewValue);
void     SetTapeBlockExpandFlag (struct TapeEntry_s *Entry, short NewValue);
void     SetupNewData           (short WhichWindow);
void     ExpandCollapse         (short WhichWindow, struct TapeEntry_s *Entry);
void     DecodeDRB              (short WhichWindow, struct TapeEntry_s *Entry);
comboent GetBlockNumber         (char *ComboTextLine);
comboent FindBlockNumber        (short WhichWindow, comboent BlockNumber);
void     ViewTapeBlockInfo      (short WhichWindow);
short    GetBlockTypeIndex      (byte BlockTypeID);
bool     GetBlockTypeInfo       (byte BlockTypeID, bool *IsTapeData, bool *IsIdentifiedData, char **FullName, char **TapeWindowName,
                                 short *NumInfoLines);
void     TmpExpandTape          (short WhichWindow);
void     TmpRestoreTape         (short WhichWindow);
void     ClearTapeLoopDetection (short WhichWindow);

#endif
